Package: splitGraph
Title: Dataset Dependency Graphs for Leakage-Aware Evaluation
Version: 0.2.0
Authors@R: c(
    person("Selcuk", "Korkmaz", role = c("aut","cre"), email = "selcukorkmaz@gmail.com",
           comment = c(ORCID = "0000-0003-4632-6850"))
    )
Description: Represent biomedical dataset structure as typed dependency
    graphs so that sample provenance, repeated-measure structure, study
    design, batch effects, and temporal relationships are explicit and
    inspectable. Validates dataset structure, detects sample-level overlap,
    derives deterministic split constraints, and produces a tool-agnostic
    split specification for leakage-aware evaluation workflows.
License: MIT + file LICENSE
URL: https://github.com/selcukorkmaz/splitGraph
BugReports: https://github.com/selcukorkmaz/splitGraph/issues
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: graphics, igraph
Suggests: jsonlite, knitr, pkgload, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
RoxygenNote: 7.3.3
Packaged: 2026-04-29 20:11:03 UTC; selcuk
Author: Selcuk Korkmaz [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4632-6850>)
Maintainer: Selcuk Korkmaz <selcukorkmaz@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-29 20:30:02 UTC
Built: R 4.5.2; ; 2026-05-01 17:44:09 UTC; unix

Package: splitGraph
Title: Dataset Dependency Graphs for Leakage-Aware Evaluation
Version: 0.3.0
Authors@R: c(
    person("Selcuk", "Korkmaz", role = c("aut","cre"), email = "selcukorkmaz@gmail.com",
           comment = c(ORCID = "0000-0003-4632-6850"))
    )
Description: Represent biomedical dataset structure as typed dependency
    graphs so that sample provenance, repeated-measure structure, study
    design, batch effects, and temporal relationships are explicit and
    inspectable. Validates dataset structure, detects sample-level overlap,
    derives deterministic split constraints, and produces a tool-agnostic
    split specification for leakage-aware evaluation workflows.
License: MIT + file LICENSE
URL: https://github.com/selcukorkmaz/splitGraph
BugReports: https://github.com/selcukorkmaz/splitGraph/issues
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: graphics, igraph, stats, utils
Suggests: bioLeak, jsonlite, knitr, pkgload, rmarkdown, testthat (>=
        3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
RoxygenNote: 7.3.3
Packaged: 2026-07-03 11:48:54 UTC; selcuk
Author: Selcuk Korkmaz [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4632-6850>)
Maintainer: Selcuk Korkmaz <selcukorkmaz@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-03 12:40:02 UTC
Built: R 4.5.2; ; 2026-07-03 15:14:03 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,dependency_constraint)
S3method(as.data.frame,depgraph_validation_report)
S3method(as.data.frame,graph_edge_set)
S3method(as.data.frame,graph_node_set)
S3method(as.data.frame,graph_query_result)
S3method(as.data.frame,leakage_risk_summary)
S3method(as.data.frame,split_constraint)
S3method(as.data.frame,split_spec)
S3method(as.data.frame,split_spec_validation)
S3method(plot,dependency_graph)
S3method(print,dependency_constraint)
S3method(print,dependency_graph)
S3method(print,depgraph_validation_report)
S3method(print,graph_edge_set)
S3method(print,graph_node_set)
S3method(print,graph_query_result)
S3method(print,leakage_risk_summary)
S3method(print,split_constraint)
S3method(print,split_spec)
S3method(print,split_spec_validation)
S3method(print,splitgraph_json_report)
S3method(summary,dependency_constraint)
S3method(summary,dependency_graph)
S3method(summary,depgraph_validation_report)
S3method(summary,graph_edge_set)
S3method(summary,graph_node_set)
S3method(summary,graph_query_result)
S3method(summary,leakage_risk_summary)
S3method(summary,split_constraint)
S3method(summary,split_spec)
S3method(summary,split_spec_validation)
export(as_igraph)
export(as_split_spec)
export(build_dependency_graph)
export(build_depgraph)
export(create_edges)
export(create_nodes)
export(dependency_constraint)
export(dependency_graph)
export(depgraph_validation_report)
export(derive_split_constraints)
export(detect_dependency_components)
export(detect_shared_dependencies)
export(graph_edge_set)
export(graph_from_metadata)
export(graph_node_set)
export(graph_query_result)
export(grouping_vector)
export(ingest_metadata)
export(leakage_risk_summary)
export(migrate_dependency_graph_json)
export(migrate_split_spec_json)
export(new_depgraph)
export(new_depgraph_edges)
export(new_depgraph_nodes)
export(query_edge_type)
export(query_neighbors)
export(query_node_type)
export(query_paths)
export(query_shortest_paths)
export(read_dependency_graph)
export(read_split_spec)
export(relatedness_edges_from_kinship)
export(spatial_edges_from_coords)
export(split_constraint)
export(split_spec)
export(split_spec_validation)
export(summarize_leakage_risks)
export(validate_depgraph)
export(validate_graph)
export(validate_graph_json)
export(validate_split_spec)
export(validate_split_spec_json)
export(write_dependency_graph)
export(write_split_spec)
importFrom(graphics,plot)
importFrom(igraph,ecount)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,vcount)

# splitGraph: Dataset Dependency Graphs for Leakage-Aware Evaluation

`splitGraph` is an R package for representing biomedical dataset structure as a
typed dependency graph so that leakage-relevant relationships can be made
explicit, validated, queried, and converted into deterministic split
constraints.

It does not fit models, run preprocessing pipelines, or generate resamples by
itself. Its job is to encode dataset structure *before* evaluation so that
overlap, provenance, and time-ordering assumptions are inspectable instead of
implicit.

![Example dependency graph](man/figures/README-plot-1.png)

The plot above shows six samples (blue) that share three subjects (orange),
two batches (green), two timepoints (red), and two outcome classes (brown).
A plain `vfold_cv` on this dataset would violate subject, batch, *and* time
structure at the same time — and that is exactly what the graph is designed
to make visible.

## Why It Exists

In biomedical evaluation workflows, leakage often comes from dataset structure
rather than obvious coding mistakes. Samples may share:

- the same subject
- the same batch
- the same study
- the same collection timepoint
- the same assay provenance
- the same derived feature set
- the same outcome definition

If those relationships are not modeled explicitly, a train/test split can look
correct while still violating the intended scientific separation.

`splitGraph` makes those dependencies first-class objects.

## What It Does (and Does Not Do)

**Does:**

- metadata ingestion with canonical ID normalization
- one-shot graph construction from canonical metadata via
  `graph_from_metadata()`
- typed node and edge constructors backed by `igraph`
- structural, semantic, and leakage-relevant validation, with a documented
  `validation_overrides` mechanism for explicit exceptions
- typed query and traversal helpers (with a safety cap on `query_paths()`)
- projected sample-dependency detection
- split-constraint derivation for subject, batch, study, time, and composite
  modes
- translation of constraints into a stable, tool-agnostic `split_spec`
- split-spec preflight validation and leakage summary helpers
- JSON serialization for `dependency_graph` and `split_spec` so handoff
  objects are portable across sessions and languages (`write_*()` /
  `read_*()`, requires `jsonlite`)
- typed layered `plot()` method with per-type colors and a node-type legend
- `print()`, `summary()`, and `as.data.frame()` on all core S3 objects

**Does not:**

- fit models or run preprocessing pipelines
- generate resamples (`rsample` does that)
- implement leakage-aware *training* workflows
- provide a general-purpose graph analytics toolkit

The package is intentionally narrow: dataset dependency structure for
leakage-aware evaluation design.

## Installation

From GitHub:

```r
install.packages("remotes")
remotes::install_github("selcukorkmaz/splitGraph")
```

To use the JSON serialization API, also install `jsonlite`:

```r
install.packages("jsonlite")
```

## Quick Start

The fastest path is `graph_from_metadata()`, which auto-detects canonical
columns in a metadata frame and assembles a validated `dependency_graph`:

```r
library(splitGraph)

meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
  batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
  timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
  time_index   = c(0, 1, 0, 1, 0, 1),
  outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl")
)

g <- graph_from_metadata(meta, graph_name = "demo")
plot(g)

validation <- validate_graph(g)
subject_constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(subject_constraint, graph = g)
validate_split_spec(spec)
summarize_leakage_risks(g, constraint = subject_constraint, split_spec = spec)

# Persist the spec for a downstream consumer (R or non-R):
path <- tempfile(fileext = ".json")
write_split_spec(spec, path)
spec2 <- read_split_spec(path)
```

For full control over node labels, attribute columns, and the feature-set
provenance edges, use `create_nodes()` / `create_edges()` /
`build_dependency_graph()` directly. `graph_from_metadata()` auto-builds
the six sample-rooted canonical edges, `timepoint_precedes`, and the
appropriate outcome edge (`sample_has_outcome` by default, or
`subject_has_outcome` when `outcome_scope = "subject"`). The
`featureset_generated_from_study` and `featureset_generated_from_batch`
edges always require the explicit constructor path.

## Downstream Handoff

`split_spec` is the tool-agnostic handoff object produced by
`as_split_spec()`. `splitGraph` does not know about any particular
resampling package — downstream consumers are expected to provide their own
adapters so that `splitGraph` stays neutral and has no runtime dependency
on them.

The typical end-to-end flow is:

1. `graph_from_metadata(meta)` → typed `dependency_graph`
2. `derive_split_constraints(g, mode = ...)` → `split_constraint`
3. `as_split_spec(constraint, graph = g)` → `split_spec`
4. (optional) `write_split_spec(spec, path)` → JSON, for cross-session or
   cross-language handoff
5. adapter in the downstream package → native resamples

The `sample_data` frame carried by `split_spec` exposes exactly what an
adapter needs: `sample_id` for joining against the observation frame,
`group_id` for grouped resampling, `batch_group` / `study_group` for
blocking, and `order_rank` for ordered evaluation. An adapter can be built
on top of, for example, `rsample::group_vfold_cv()` (grouped CV keyed to
`group_id`) or `rsample::rolling_origin()` (ordered evaluation keyed to
`order_rank`).

For three small, self-contained adapter examples (a base-R LOGO adapter,
plus illustrative `rsample::group_vfold_cv()` and `rsample::rolling_origin()`
adapters), see the **Adapter cookbook** vignette:

```r
vignette("adapter-cookbook", package = "splitGraph")
```

## Core Concepts

### Node types

- `Sample`, `Subject`, `Batch`, `Study`, `Timepoint`, `Assay`, `FeatureSet`,
  `Outcome`

### Canonical edge types

- `sample_belongs_to_subject`
- `sample_processed_in_batch`
- `sample_from_study`
- `sample_collected_at_timepoint`
- `sample_measured_by_assay`
- `sample_uses_featureset`
- `sample_has_outcome`
- `subject_has_outcome`
- `timepoint_precedes`
- `featureset_generated_from_study`
- `featureset_generated_from_batch`

### Main S3 objects

`graph_node_set`, `graph_edge_set`, `dependency_graph`,
`depgraph_validation_report`, `graph_query_result`, `split_constraint`,
`split_spec`, `split_spec_validation`, `leakage_risk_summary`.

## Main Functions

| Layer | Functions |
|---|---|
| Ingestion and construction | `ingest_metadata()`, `graph_from_metadata()`, `create_nodes()`, `create_edges()`, `build_dependency_graph()`, `dependency_graph()`, `as_igraph()` |
| Validation | `validate_graph()` (with `validation_overrides`), `validate_split_spec()` |
| Queries | `query_node_type()`, `query_edge_type()`, `query_neighbors()`, `query_paths()` (capped by default), `query_shortest_paths()`, `detect_dependency_components()`, `detect_shared_dependencies()` |
| Constraint derivation | `derive_split_constraints()`, `grouping_vector()` |
| Split-spec translation | `as_split_spec()`, `summarize_leakage_risks()` |
| Serialization (JSON) | `write_dependency_graph()`, `read_dependency_graph()`, `write_split_spec()`, `read_split_spec()` |

## Example Queries

```r
query_node_type(g, "Subject")
query_edge_type(g, "sample_processed_in_batch")
query_neighbors(g, node_ids = "sample:S1", edge_types = "sample_belongs_to_subject")
detect_shared_dependencies(g, via = "Batch")
detect_dependency_components(g, via = c("Subject", "Batch"))
```

## Example Split Designs

```r
subject_constraint <- derive_split_constraints(g, mode = "subject")
batch_constraint   <- derive_split_constraints(g, mode = "batch")
study_constraint   <- derive_split_constraints(g, mode = "study")
time_constraint    <- derive_split_constraints(g, mode = "time")

strict_composite <- derive_split_constraints(
  g, mode = "composite", strategy = "strict",
  via = c("Subject", "Batch")
)

rule_based_composite <- derive_split_constraints(
  g, mode = "composite", strategy = "rule_based",
  priority = c("batch", "study", "subject", "time")
)
```

## Serialization (JSON)

Both core handoff objects can be written to a stable, schema-versioned
JSON format and read back, so a `dependency_graph` or `split_spec` is
portable across R sessions and across language boundaries.

```r
graph_path <- tempfile(fileext = ".json")
spec_path  <- tempfile(fileext = ".json")

write_dependency_graph(g, graph_path)
write_split_spec(spec, spec_path)

g2    <- read_dependency_graph(graph_path)
spec2 <- read_split_spec(spec_path)
```

The JSON schema is documented under `?write_dependency_graph` and
`?write_split_spec`. Each file carries a `schema_version` field; reading a
file written under a different schema version emits a warning but still
loads. `NA` values in `sample_data` round-trip as JSON `null`. The
`jsonlite` package (a `Suggests` dep) must be installed.

## Plot Method

`plot(g)` renders a typed, layered layout with per-type node colors and an
auto-generated node-type legend. Layers: Sample (top), peer dependencies
(Subject / Batch / Study / Timepoint) in the middle band,
Assay / FeatureSet next, Outcome (bottom).

```r
plot(g)                              # typed layered layout (default)
plot(g, layout = "sugiyama")         # alternative hierarchical layout
plot(g, show_labels = FALSE)         # hide node labels on dense graphs
plot(g, legend = FALSE)              # suppress the legend
plot(g, legend_position = "bottomright")
plot(g, node_colors = c(Sample = "#000000"))  # override type colors
```

## Citation

```r
citation("splitGraph")
```

produces:

> Korkmaz S (2026). *splitGraph: Dataset Dependency Graphs for
> Leakage-Aware Evaluation*. R package version 0.2.0.
> <https://github.com/selcukorkmaz/splitGraph>

## License

MIT. See `LICENSE`.

## Appendix: Design Guarantees

The package prefers explicit failure over silent guessing. In particular:

- unknown sample IDs, ambiguous direct assignments, and conflicting
  duplicate nodes or edges are rejected rather than silently resolved
- contradictory time-order metadata are rejected rather than reconciled
  arbitrarily
- validation truth is not changed by severity filters, and generated
  split specs are re-validated against the package's own preflight rules
- documented exceptions go through `validation_overrides` (e.g.
  `allow_multi_subject_samples`); the same override is honored by both
  `validate_graph()` and `derive_split_constraints(mode = "subject")`
- `query_paths()` defaults to a finite path-length cap so traversal cannot
  explode on dense graphs; pass `max_length = Inf` to opt out
- the on-disk JSON format is schema-versioned, and loading a file with a
  different `schema_version` warns rather than failing silently

# Splitgraph
![Build status](https://github.com/splitgraph/splitgraph/workflows/build_all/badge.svg)
[![Coverage Status](https://coveralls.io/repos/github/splitgraph/splitgraph/badge.svg?branch=master)](https://coveralls.io/github/splitgraph/splitgraph?branch=master)
[![PyPI version](https://badge.fury.io/py/splitgraph.svg)](https://badge.fury.io/py/splitgraph)
[![Discord chat room](https://img.shields.io/discord/718534846472912936.svg)](https://discord.gg/4Qe2fYA)
[![Follow](https://img.shields.io/badge/twitter-@Splitgraph-blue.svg)](https://twitter.com/Splitgraph)

## Overview

**Splitgraph** is a tool for building, versioning and querying reproducible datasets. It's inspired
by Docker and Git, so it feels familiar. And it's powered by [PostgreSQL](https://postgresql.org), so it [works seamlessly with existing tools](https://www.splitgraph.com/connect) in the Postgres ecosystem. Use Splitgraph to package your data into self-contained **data images** that you can [share with other Splitgraph instances](https://www.splitgraph.com/docs/getting-started/decentralized-demo).

[**Splitgraph.com**](https://www.splitgraph.com), or **Splitgraph Cloud**, is a public Splitgraph instance where you can share and discover data. It's a Splitgraph peer powered by the **Splitgraph Core** code in this repository, adding proprietary features like a data catalog, multitenancy, and a distributed SQL proxy.

You can explore [40k+ open datasets](https://www.splitgraph.com/explore) in the catalog. You can also connect directly to the [Data Delivery Network](https://www.splitgraph.com/connect) and query any of the datasets, without installing anything.

To install `sgr` (the command line client) or a local Splitgraph Engine, see the [Installation](#installation) section of this readme.

### Build and Query Versioned, Reproducible Datasets

[**Splitfiles**](https://www.splitgraph.com/docs/concepts/splitfiles) give you a declarative language, inspired by Dockerfiles, for expressing data transformations in ordinary SQL familiar to any researcher or business analyst. You can reference other images, or even other databases, with a simple JOIN.

![](pics/splitfile.png)

When you build data with Splitfiles, you get provenance tracking of the resulting data: it's possible to find out what sources went into every dataset and know when to rebuild it if the sources ever change. You can easily integrate Splitgraph into your existing CI pipelines, to keep your data up-to-date and stay on top of changes to upstream sources.

Splitgraph images are also version-controlled, and you can manipulate them with Git-like operations through a CLI. You can check out any image into a PostgreSQL schema and interact with it using any PostgreSQL client. Splitgraph will capture your changes to the data, and then you can commit them as delta-compressed changesets that you can package into new images.

Splitgraph supports PostgreSQL [foreign data wrappers](https://wiki.postgresql.org/wiki/Foreign_data_wrappers). We call this feature [mounting](https://www.splitgraph.com/docs/concepts/mounting). With mounting, you can query other databases (like PostgreSQL/MongoDB/MySQL) or open data providers (like [Socrata](https://www.splitgraph.com/docs/ingesting-data/socrata)) from your Splitgraph instance with plain SQL. You can even snapshot the results or use them in Splitfiles.

### Why Splitgraph?

Splitgraph isn't opinionated and doesn't break existing abstractions. To any existing PostgreSQL application, Splitgraph images are just another database. We have carefully designed Splitgraph to not break the abstraction of a PostgreSQL table and wire protocol, because doing otherwise would mean throwing away a vast existing ecosystem of applications, users, libraries and extensions. This means that a lot of tools that work with PostgreSQL work with Splitgraph out of the box.

![](pics/splitfiles.gif)

## Components

The code in this repository, known as **Splitgraph Core**, contains:

- **[`sgr` command line client](https://www.splitgraph.com/docs/architecture/sgr-client)**: `sgr` is the main command line tool used to work with Splitgraph "images" (data snapshots). Use it to ingest data, work with splitfiles, and push data to Splitgraph.com.
- **[Splitgraph Engine](engine/README.md)**: a [Docker image](https://hub.docker.com/r/splitgraph/engine) of the latest Postgres with Splitgraph and other required extensions pre-installed.
- **[Splitgraph Python library](https://www.splitgraph.com/docs/python-api/splitgraph.core)**: All Splitgraph functionality is available in the Python API, offering first-class support for data science workflows including Jupyter notebooks and Pandas dataframes.

## Docs

Documentation is available at https://www.splitgraph.com/docs, specifically:

- [Installation](https://www.splitgraph.com/docs/getting-started/installation)
- [FAQ](https://www.splitgraph.com/docs/getting-started/frequently-asked-questions)

We also recommend reading our Blog, including some of our favorite posts:

- [Supercharging `dbt` with Splitgraph: versioning, sharing, cross-DB joins](https://www.splitgraph.com/blog/dbt)
- [Querying 40,000+ datasets with SQL](https://www.splitgraph.com/blog/40k-sql-datasets)
- [Foreign data wrappers: PostgreSQL's secret weapon?](https://www.splitgraph.com/blog/foreign-data-wrappers)

## Installation

Pre-requisites:

- Docker is required to run the Splitgraph Engine. `sgr` must have access to Docker. You either need to [install Docker locally](https://docs.docker.com/install/) or have access to a remote Docker socket.

For Linux and OSX, once Docker is running, install Splitgraph with a single script:

```
$ bash -c "$(curl -sL https://github.com/splitgraph/splitgraph/releases/latest/download/install.sh)"
```

This will download the `sgr` binary and set up the Splitgraph Engine Docker container.

Alternatively, you can get the `sgr` single binary from [the releases page](https://github.com/splitgraph/splitgraph/releases) and run [`sgr engine add`](https://www.splitgraph.com/docs/sgr/engine-management/engine-add) to create an engine.

See the [installation guide](https://www.splitgraph.com/docs/getting-started/installation) for more installation methods.

## Quick start guide

You can follow the [quick start guide](https://www.splitgraph.com/docs/getting-started/five-minute-demo) that will guide you through the basics of using Splitgraph with public and private data.

Alternatively, Splitgraph comes with plenty of [examples](examples) to get you started.

If you're stuck or have any questions, check out the [documentation](https://www.splitgraph.com/docs/) or join our [Discord channel](https://discord.gg/4Qe2fYA)!

## Contributing

### Setting up a development environment

  * Splitgraph requires Python 3.6 or later.
  * Install [Poetry](https://github.com/python-poetry/poetry): `curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python` to manage dependencies
  * Install pre-commit hooks (we use [Black](https://github.com/psf/black) to format code)
  * `git clone --recurse-submodules https://github.com/splitgraph/splitgraph.git`
  * `poetry install`
  * To build the [engine](https://www.splitgraph.com/docs/architecture/splitgraph-engine) Docker image: `cd engine && make`

### Running tests

The test suite requires [docker-compose](https://github.com/docker/compose). You will also
need to add these lines to your `/etc/hosts` or equivalent:

```
127.0.0.1       local_engine
127.0.0.1       remote_engine
127.0.0.1       objectstorage
```

To run the core test suite, do

```
docker-compose -f test/architecture/docker-compose.core.yml up -d
poetry run pytest -m "not mounting and not example"
```

To run the test suite related to "mounting" and importing data from other databases
(PostgreSQL, MySQL, Mongo), do

```
docker-compose -f test/architecture/docker-compose.core.yml -f test/architecture/docker-compose.mounting.yml up -d
poetry run pytest -m mounting
```

Finally, to test the [example projects](https://github.com/splitgraph/splitgraph/tree/master/examples), do

```
# Example projects spin up their own engines
docker-compose -f test/architecture/docker-compose.core.yml -f test/architecture/docker-compose.core.yml down -v
poetry run pytest -m example
```

All of these tests run in [CI](https://github.com/splitgraph/splitgraph/actions).

# splitspec — Python reference consumer for splitGraph `split_spec`

A pure-Python reader for the tool-agnostic `split_spec` JSON interchange format
written by `splitGraph::write_split_spec()`. It recovers the sample-level
grouping, ordering, and stratum annotations needed to drive scikit-learn
resamplers, and is the reference implementation used by the cross-language
conformance check.

- The reader (`splitspec.load_split_spec`) depends only on the standard library.
- `SplitSpec.to_frame()` additionally needs `pandas`.
- `SplitSpec.group_kfold()` imports scikit-learn (and numpy) lazily.

## Usage

```python
import sys
sys.path.insert(0, "<path to this inst/python directory>")
from splitspec import load_split_spec

spec = load_split_spec("split_spec.json")
spec.grouping()      # {sample_id: group_id}, == R grouping_vector()
spec.groups()        # group_id per sample, for GroupKFold(groups=...)
spec.order_ranks()   # order_rank per sample, for TimeSeriesSplit
spec.to_frame()      # pandas DataFrame of sample_data
```

In R, locate this directory with:

```r
system.file("python", package = "splitGraph")
```

## Conformance

`conformance.py <split_spec.json> <out.json>` reads a spec and writes back the
grouping and ordering it recovered. The R test `test-python-conformance.R`
runs it and asserts the result matches `grouping_vector()` and
`sample_data$order_rank` exactly. It is skipped when `python3` is unavailable
and never runs on CRAN.

## Compatibility

The reader accepts any `split_spec` sharing schema **major** version `0`
(the compatibility boundary), matching the R side's policy.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}

## ----build-spec---------------------------------------------------------------
meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
  batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
  timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
  time_index   = c(0, 1, 0, 1, 0, 1),
  outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl"),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "cookbook")
subject_constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(subject_constraint, graph = g)
spec

## ----show-sample-data---------------------------------------------------------
as.data.frame(spec)[, c("sample_id", "group_id", "batch_group", "order_rank")]

## ----adapter-base-r-----------------------------------------------------------
logo_folds <- function(spec, observation_data, sample_id_col = "sample_id") {
  stopifnot(inherits(spec, "split_spec"))
  if (!sample_id_col %in% names(observation_data)) {
    stop("`observation_data` must contain a `", sample_id_col, "` column.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$group_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )
  joined$.row <- seq_len(nrow(joined))
  groups <- split(joined$.row, joined[[spec$group_var]])

  lapply(names(groups), function(g) {
    list(
      group   = g,
      train   = unlist(groups[setdiff(names(groups), g)], use.names = FALSE),
      assess  = groups[[g]]
    )
  })
}

# Pretend we have an observation frame keyed by sample_id.
obs <- data.frame(
  sample_id = meta$sample_id,
  x = rnorm(nrow(meta)),
  y = rbinom(nrow(meta), 1, 0.5)
)

folds <- logo_folds(spec, obs)
length(folds)
folds[[1]]

## ----adapter-rsample-group, eval=FALSE----------------------------------------
# spec_to_group_vfold <- function(spec, observation_data,
#                                 v = NULL,
#                                 sample_id_col = "sample_id") {
#   stopifnot(inherits(spec, "split_spec"))
#   if (!requireNamespace("rsample", quietly = TRUE)) {
#     stop("Install rsample to use this adapter.")
#   }
# 
#   joined <- merge(
#     observation_data,
#     spec$sample_data[, c("sample_id", spec$group_var)],
#     by.x = sample_id_col, by.y = "sample_id", sort = FALSE
#   )
# 
#   n_groups <- length(unique(joined[[spec$group_var]]))
#   if (is.null(v)) v <- n_groups
# 
#   rsample::group_vfold_cv(
#     data  = joined,
#     group = !!spec$group_var,
#     v     = v
#   )
# }

## ----adapter-rsample-rolling, eval=FALSE--------------------------------------
# spec_to_rolling_origin <- function(spec, observation_data,
#                                    sample_id_col = "sample_id",
#                                    initial = NULL,
#                                    assess = 1L) {
#   stopifnot(inherits(spec, "split_spec"))
#   if (is.null(spec$time_var)) {
#     stop("This split_spec has no `time_var`; ordered evaluation is not available.")
#   }
#   if (!requireNamespace("rsample", quietly = TRUE)) {
#     stop("Install rsample to use this adapter.")
#   }
# 
#   joined <- merge(
#     observation_data,
#     spec$sample_data[, c("sample_id", spec$time_var)],
#     by.x = sample_id_col, by.y = "sample_id", sort = FALSE
#   )
#   ordered <- joined[order(joined[[spec$time_var]]), , drop = FALSE]
# 
#   if (is.null(initial)) initial <- max(1L, floor(nrow(ordered) * 0.6))
#   rsample::rolling_origin(ordered, initial = initial, assess = assess)
# }

## ----serialize, eval = requireNamespace("jsonlite", quietly = TRUE)-----------
tmp <- tempfile(fileext = ".json")
write_split_spec(spec, tmp)

# Inspect the first ~30 lines so the on-disk format is visible.
cat(readLines(tmp, n = 30), sep = "\n")

# And read it back exactly.
spec2 <- read_split_spec(tmp)
identical(spec$sample_data$group_id, spec2$sample_data$group_id)

unlink(tmp)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}

or_empty <- function(x) {
  if (is.null(x)) character() else x
}

## ----metadata-----------------------------------------------------------------
meta <- data.frame(
  sample_id = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id = c("P1", "P1", "P2", "P3", "P4", "P2"),
  batch_id = c("B1", "B2", "B1", "B3", NA, "B1"),
  study_id = c("ST1", "ST1", "ST1", "ST2", "ST3", "ST2"),
  timepoint_id = c("T0", "T1", "T0", "T2", NA, "T1"),
  assay_id = c("RNAseq", "RNAseq", "RNAseq", "RNAseq", "Proteomics", "RNAseq"),
  featureset_id = c("FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_PROT", "FS_GLOBAL"),
  outcome_id = c("O_case", "O_case", "O_ctrl", "O_case", "O_ctrl", "O_ctrl"),
  stringsAsFactors = FALSE
)

meta

## ----fast-path----------------------------------------------------------------
quick_graph <- graph_from_metadata(
  data.frame(
    sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
    subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
    batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
    timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
    time_index   = c(0, 1, 0, 1, 0, 1),
    outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl")
  ),
  graph_name = "quick_demo"
)

quick_graph

## ----construction-------------------------------------------------------------
meta <- ingest_metadata(meta, dataset_name = "VignetteDemo")

sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")

time_nodes <- create_nodes(
  data.frame(
    timepoint_id = c("T0", "T1", "T2"),
    time_index = c(0L, 1L, 2L),
    visit_label = c("baseline", "follow_up", "late_follow_up"),
    stringsAsFactors = FALSE
  ),
  type = "Timepoint",
  id_col = "timepoint_id",
  attr_cols = c("time_index", "visit_label")
)

assay_nodes <- create_nodes(
  data.frame(
    assay_id = c("RNAseq", "Proteomics"),
    modality = c("transcriptomics", "proteomics"),
    platform = c("NovaSeq", "Orbitrap"),
    stringsAsFactors = FALSE
  ),
  type = "Assay",
  id_col = "assay_id",
  attr_cols = c("modality", "platform")
)

featureset_nodes <- create_nodes(
  data.frame(
    featureset_id = c("FS_GLOBAL", "FS_PROT"),
    featureset_name = c("global_rna_signature", "proteomics_panel"),
    derivation_scope = c("per_dataset", "external"),
    feature_count = c(500L, 80L),
    stringsAsFactors = FALSE
  ),
  type = "FeatureSet",
  id_col = "featureset_id",
  attr_cols = c("featureset_name", "derivation_scope", "feature_count")
)

outcome_nodes <- create_nodes(
  data.frame(
    outcome_id = c("O_case", "O_ctrl"),
    outcome_name = c("response", "response"),
    outcome_type = c("binary", "binary"),
    observation_level = c("subject", "subject"),
    stringsAsFactors = FALSE
  ),
  type = "Outcome",
  id_col = "outcome_id",
  attr_cols = c("outcome_name", "outcome_type", "observation_level")
)

subject_edges <- create_edges(
  meta, "sample_id", "subject_id",
  "Sample", "Subject", "sample_belongs_to_subject"
)

batch_edges <- create_edges(
  meta, "sample_id", "batch_id",
  "Sample", "Batch", "sample_processed_in_batch",
  allow_missing = TRUE
)

study_edges <- create_edges(
  meta, "sample_id", "study_id",
  "Sample", "Study", "sample_from_study"
)

time_edges <- create_edges(
  meta, "sample_id", "timepoint_id",
  "Sample", "Timepoint", "sample_collected_at_timepoint",
  allow_missing = TRUE
)

assay_edges <- create_edges(
  meta, "sample_id", "assay_id",
  "Sample", "Assay", "sample_measured_by_assay"
)

featureset_edges <- create_edges(
  meta, "sample_id", "featureset_id",
  "Sample", "FeatureSet", "sample_uses_featureset"
)

outcome_edges <- create_edges(
  data.frame(
    subject_id = c("P1", "P2", "P3", "P4"),
    outcome_id = c("O_case", "O_ctrl", "O_case", "O_ctrl"),
    stringsAsFactors = FALSE
  ),
  "subject_id", "outcome_id",
  "Subject", "Outcome", "subject_has_outcome"
)

precedence_edges <- create_edges(
  data.frame(
    from_timepoint = c("T0", "T1"),
    to_timepoint = c("T1", "T2"),
    stringsAsFactors = FALSE
  ),
  "from_timepoint", "to_timepoint",
  "Timepoint", "Timepoint", "timepoint_precedes"
)

featureset_from_study <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    study_id = "ST1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "study_id",
  "FeatureSet", "Study", "featureset_generated_from_study"
)

featureset_from_batch <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    batch_id = "B1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "batch_id",
  "FeatureSet", "Batch", "featureset_generated_from_batch"
)

## ----construction-output------------------------------------------------------
sample_nodes
as.data.frame(sample_nodes)[, c("node_id", "node_type", "node_key", "label")]

edge_preview <- do.call(rbind, lapply(
  list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  as.data.frame
))

edge_preview[, c("from", "to", "edge_type")]

## ----graph--------------------------------------------------------------------
graph <- build_dependency_graph(
  nodes = list(
    sample_nodes, subject_nodes, batch_nodes, study_nodes,
    time_nodes, assay_nodes, featureset_nodes, outcome_nodes
  ),
  edges = list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  graph_name = "vignette_graph",
  dataset_name = attr(meta, "dataset_name")
)

graph
summary(graph)

## ----plot, fig.width = 7, fig.height = 5--------------------------------------
plot(graph)

## ----plot-options, eval = FALSE-----------------------------------------------
# plot(graph, layout = "sugiyama")         # alternative hierarchical layout
# plot(graph, show_labels = FALSE)         # hide labels on dense graphs
# plot(graph, legend = FALSE)              # suppress the legend
# plot(graph, legend_position = "bottomright")
# plot(graph, node_colors = c(Sample = "#000000"))

## ----validation---------------------------------------------------------------
validation <- validate_graph(graph)

validation
as.data.frame(validation)[, c("level", "severity", "code", "message")]

## ----strictness---------------------------------------------------------------
tryCatch(
  derive_split_constraints(graph, mode = "subject", samples = c("S1", "BAD")),
  error = function(e) e$message
)

## ----overrides-fixture--------------------------------------------------------
multi_nodes <- graph_node_set(data.frame(
  node_id   = c("sample:S1", "subject:P1", "subject:P2"),
  node_type = c("Sample", "Subject", "Subject"),
  node_key  = c("S1", "P1", "P2"),
  label     = c("S1", "P1", "P2"),
  attrs     = I(list(list(), list(), list())),
  stringsAsFactors = FALSE
))

multi_edges <- graph_edge_set(data.frame(
  edge_id   = c("sample_belongs_to_subject:1", "sample_belongs_to_subject:2"),
  from      = c("sample:S1", "sample:S1"),
  to        = c("subject:P1", "subject:P2"),
  edge_type = c("sample_belongs_to_subject", "sample_belongs_to_subject"),
  attrs     = I(list(list(), list())),
  stringsAsFactors = FALSE
))

multi_graph <- dependency_graph(nodes = multi_nodes, edges = multi_edges)

## ----overrides-default--------------------------------------------------------
default_report <- validate_graph(multi_graph)
default_report$valid
default_report$issues[, c("severity", "code", "message")]

tryCatch(
  derive_split_constraints(multi_graph, mode = "subject"),
  error = function(e) e$message
)

## ----overrides-on-------------------------------------------------------------
# Validator: pass.
permissive_report <- validate_graph(
  multi_graph,
  validation_overrides = list(allow_multi_subject_samples = TRUE)
)
permissive_report$valid

# Constraint derivation: pick the first listed subject and record the
# ambiguity in metadata$warnings instead of erroring.
multi_graph$metadata$validation_overrides <-
  list(allow_multi_subject_samples = TRUE)
relaxed_constraint <- derive_split_constraints(multi_graph, mode = "subject")
relaxed_constraint$sample_map[, c("sample_id", "group_id")]
relaxed_constraint$metadata$warnings

## ----neighbors-and-paths------------------------------------------------------
neighbors_s1 <- query_neighbors(graph, node_ids = "sample:S1", direction = "out")
neighbors_s1
as.data.frame(neighbors_s1)[, c("seed_node_id", "node_id", "node_type", "edge_type")]

subject_outcome_path <- query_shortest_paths(
  graph,
  from = "sample:S1",
  to = "outcome:O_case",
  edge_types = c("sample_belongs_to_subject", "subject_has_outcome")
)

subject_outcome_path
as.data.frame(subject_outcome_path)

## ----projected-dependencies---------------------------------------------------
shared_dependencies <- detect_shared_dependencies(
  graph,
  via = c("Subject", "Batch", "FeatureSet")
)

as.data.frame(shared_dependencies)[, c(
  "sample_id_1", "sample_id_2", "shared_node_type", "shared_node_id", "edge_type"
)]

dependency_components <- detect_dependency_components(
  graph,
  via = c("Subject", "Batch")
)

as.data.frame(dependency_components)

## ----constraints--------------------------------------------------------------
subject_constraint <- derive_split_constraints(graph, mode = "subject")
batch_constraint <- derive_split_constraints(graph, mode = "batch")
study_constraint <- derive_split_constraints(graph, mode = "study")
time_constraint <- derive_split_constraints(graph, mode = "time")

strict_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "strict",
  via = c("Subject", "Batch")
)

rule_based_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "rule_based",
  priority = c("batch", "study", "subject", "time")
)

constraint_overview <- do.call(rbind, lapply(
  list(
    subject = subject_constraint,
    batch = batch_constraint,
    study = study_constraint,
    time = time_constraint,
    composite_strict = strict_constraint,
    composite_rule = rule_based_constraint
  ),
  function(x) {
    data.frame(
      strategy = x$strategy,
      groups = length(unique(x$sample_map$group_id)),
      warnings = if (is.null(x$metadata$warnings)) 0L else length(x$metadata$warnings),
      stringsAsFactors = FALSE
    )
  }
))

constraint_overview <- cbind(constraint = row.names(constraint_overview), constraint_overview)
row.names(constraint_overview) <- NULL

constraint_overview

## ----batch-constraint---------------------------------------------------------
batch_constraint
as.data.frame(batch_constraint)[, c("sample_id", "group_id", "group_label", "explanation")]

## ----time-constraint----------------------------------------------------------
time_constraint
as.data.frame(time_constraint)[, c("sample_id", "group_id", "timepoint_id", "order_rank")]

## ----composite-constraints----------------------------------------------------
strict_constraint
as.data.frame(strict_constraint)[, c("sample_id", "group_id", "constraint_type")]

rule_based_constraint
as.data.frame(rule_based_constraint)[, c("sample_id", "group_id", "constraint_type", "group_label")]

## ----precedence-only----------------------------------------------------------
precedence_meta <- data.frame(
  sample_id = c("S1", "S2", "S3"),
  subject_id = c("P1", "P1", "P2"),
  study_id = c("ST1", "ST1", "ST2"),
  timepoint_id = c("T0", "T1", "T2"),
  stringsAsFactors = FALSE
)

precedence_graph <- build_dependency_graph(
  nodes = list(
    create_nodes(precedence_meta, type = "Sample", id_col = "sample_id"),
    create_nodes(precedence_meta, type = "Subject", id_col = "subject_id"),
    create_nodes(precedence_meta, type = "Study", id_col = "study_id"),
    create_nodes(
      data.frame(timepoint_id = c("T0", "T1", "T2"), stringsAsFactors = FALSE),
      type = "Timepoint",
      id_col = "timepoint_id"
    )
  ),
  edges = list(
    create_edges(
      precedence_meta, "sample_id", "subject_id",
      "Sample", "Subject", "sample_belongs_to_subject"
    ),
    create_edges(
      precedence_meta, "sample_id", "study_id",
      "Sample", "Study", "sample_from_study"
    ),
    create_edges(
      precedence_meta, "sample_id", "timepoint_id",
      "Sample", "Timepoint", "sample_collected_at_timepoint"
    ),
    create_edges(
      data.frame(
        from_timepoint = c("T0", "T1"),
        to_timepoint = c("T1", "T2"),
        stringsAsFactors = FALSE
      ),
      "from_timepoint", "to_timepoint",
      "Timepoint", "Timepoint", "timepoint_precedes"
    )
  ),
  graph_name = "precedence_only_graph"
)

precedence_time_constraint <- derive_split_constraints(precedence_graph, mode = "time")

precedence_time_constraint$metadata$time_order_source
as.data.frame(precedence_time_constraint)[, c("sample_id", "timepoint_id", "time_index", "order_rank")]

## ----split-spec---------------------------------------------------------------
split_spec <- as_split_spec(strict_constraint, graph = graph)
split_spec

as.data.frame(split_spec)[, c(
  "sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank"
)]

split_spec_validation <- validate_split_spec(split_spec)
split_spec_validation
as.data.frame(split_spec_validation)

## ----risk-summary-------------------------------------------------------------
risk_summary <- summarize_leakage_risks(
  graph,
  constraint = strict_constraint,
  split_spec = split_spec
)

risk_summary
as.data.frame(risk_summary)[, c("source", "severity", "category", "message")]

## ----serialize, eval = requireNamespace("jsonlite", quietly = TRUE)-----------
spec_path <- tempfile(fileext = ".json")
write_split_spec(split_spec, spec_path)

# Round-trip it back into R unchanged.
spec_round_trip <- read_split_spec(spec_path)
identical(split_spec$sample_data$group_id, spec_round_trip$sample_data$group_id)

unlink(spec_path)

## ----cookbook-pointer, eval = FALSE-------------------------------------------
# vignette("adapter-cookbook", package = "splitGraph")

## ----case-study-1-------------------------------------------------------------
subject_groups <- grouping_vector(subject_constraint)
time_groups <- time_constraint$sample_map[, c("sample_id", "group_id", "timepoint_id", "order_rank")]

subject_groups
time_groups

## ----case-study-2-------------------------------------------------------------
cross_study_issues <- as.data.frame(validation)[
  as.data.frame(validation)$code == "subject_cross_study_overlap",
  c("severity", "code", "message")
]

p2_shared <- detect_shared_dependencies(
  graph,
  via = "Subject",
  samples = c("S3", "S6")
)

study_only_map <- study_constraint$sample_map[, c("sample_id", "group_id", "group_label")]
strict_map <- strict_constraint$sample_map[, c("sample_id", "group_id", "constraint_type")]

cross_study_issues
as.data.frame(p2_shared)
study_only_map[study_only_map$sample_id %in% c("S3", "S6"), ]
strict_map[strict_map$sample_id %in% c("S3", "S6"), ]

## ----case-study-3-------------------------------------------------------------
batch_missing <- batch_constraint$sample_map[
  batch_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "group_label", "explanation")
]

rule_based_missing <- rule_based_constraint$sample_map[
  rule_based_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "constraint_type", "group_label", "explanation")
]

split_spec_missing <- as.data.frame(split_spec)[
  as.data.frame(split_spec)$sample_id == "S5",
  c("sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank")
]

batch_missing
rule_based_missing
split_spec_missing

## ----case-study-4-------------------------------------------------------------
strategy_summary <- data.frame(
  constraint = c("subject", "batch", "study", "time", "composite_strict", "composite_rule"),
  groups = c(
    length(unique(subject_constraint$sample_map$group_id)),
    length(unique(batch_constraint$sample_map$group_id)),
    length(unique(study_constraint$sample_map$group_id)),
    length(unique(time_constraint$sample_map$group_id)),
    length(unique(strict_constraint$sample_map$group_id)),
    length(unique(rule_based_constraint$sample_map$group_id))
  ),
  warnings = c(
    length(or_empty(subject_constraint$metadata$warnings)),
    length(or_empty(batch_constraint$metadata$warnings)),
    length(or_empty(study_constraint$metadata$warnings)),
    length(or_empty(time_constraint$metadata$warnings)),
    length(or_empty(strict_constraint$metadata$warnings)),
    length(or_empty(rule_based_constraint$metadata$warnings))
  ),
  recommended_resampling = c(
    as_split_spec(subject_constraint, graph = graph)$recommended_resampling,
    as_split_spec(batch_constraint, graph = graph)$recommended_resampling,
    as_split_spec(study_constraint, graph = graph)$recommended_resampling,
    as_split_spec(time_constraint, graph = graph)$recommended_resampling,
    as_split_spec(strict_constraint, graph = graph)$recommended_resampling,
    as_split_spec(rule_based_constraint, graph = graph)$recommended_resampling
  ),
  stringsAsFactors = FALSE
)

strategy_summary

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}

or_empty <- function(x) {
  if (is.null(x)) character() else x
}

## ----metadata-----------------------------------------------------------------
meta <- data.frame(
  sample_id = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id = c("P1", "P1", "P2", "P3", "P4", "P2"),
  batch_id = c("B1", "B2", "B1", "B3", NA, "B1"),
  study_id = c("ST1", "ST1", "ST1", "ST2", "ST3", "ST2"),
  timepoint_id = c("T0", "T1", "T0", "T2", NA, "T1"),
  assay_id = c("RNAseq", "RNAseq", "RNAseq", "RNAseq", "Proteomics", "RNAseq"),
  featureset_id = c("FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_PROT", "FS_GLOBAL"),
  outcome_id = c("O_case", "O_case", "O_ctrl", "O_case", "O_ctrl", "O_ctrl"),
  stringsAsFactors = FALSE
)

meta

## ----fast-path----------------------------------------------------------------
quick_graph <- graph_from_metadata(
  data.frame(
    sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
    subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
    batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
    timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
    time_index   = c(0, 1, 0, 1, 0, 1),
    outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl")
  ),
  graph_name = "quick_demo"
)

quick_graph

## ----construction-------------------------------------------------------------
meta <- ingest_metadata(meta, dataset_name = "VignetteDemo")

sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")

time_nodes <- create_nodes(
  data.frame(
    timepoint_id = c("T0", "T1", "T2"),
    time_index = c(0L, 1L, 2L),
    visit_label = c("baseline", "follow_up", "late_follow_up"),
    stringsAsFactors = FALSE
  ),
  type = "Timepoint",
  id_col = "timepoint_id",
  attr_cols = c("time_index", "visit_label")
)

assay_nodes <- create_nodes(
  data.frame(
    assay_id = c("RNAseq", "Proteomics"),
    modality = c("transcriptomics", "proteomics"),
    platform = c("NovaSeq", "Orbitrap"),
    stringsAsFactors = FALSE
  ),
  type = "Assay",
  id_col = "assay_id",
  attr_cols = c("modality", "platform")
)

featureset_nodes <- create_nodes(
  data.frame(
    featureset_id = c("FS_GLOBAL", "FS_PROT"),
    featureset_name = c("global_rna_signature", "proteomics_panel"),
    derivation_scope = c("per_dataset", "external"),
    feature_count = c(500L, 80L),
    stringsAsFactors = FALSE
  ),
  type = "FeatureSet",
  id_col = "featureset_id",
  attr_cols = c("featureset_name", "derivation_scope", "feature_count")
)

outcome_nodes <- create_nodes(
  data.frame(
    outcome_id = c("O_case", "O_ctrl"),
    outcome_name = c("response", "response"),
    outcome_type = c("binary", "binary"),
    observation_level = c("subject", "subject"),
    stringsAsFactors = FALSE
  ),
  type = "Outcome",
  id_col = "outcome_id",
  attr_cols = c("outcome_name", "outcome_type", "observation_level")
)

subject_edges <- create_edges(
  meta, "sample_id", "subject_id",
  "Sample", "Subject", "sample_belongs_to_subject"
)

batch_edges <- create_edges(
  meta, "sample_id", "batch_id",
  "Sample", "Batch", "sample_processed_in_batch",
  allow_missing = TRUE
)

study_edges <- create_edges(
  meta, "sample_id", "study_id",
  "Sample", "Study", "sample_from_study"
)

time_edges <- create_edges(
  meta, "sample_id", "timepoint_id",
  "Sample", "Timepoint", "sample_collected_at_timepoint",
  allow_missing = TRUE
)

assay_edges <- create_edges(
  meta, "sample_id", "assay_id",
  "Sample", "Assay", "sample_measured_by_assay"
)

featureset_edges <- create_edges(
  meta, "sample_id", "featureset_id",
  "Sample", "FeatureSet", "sample_uses_featureset"
)

outcome_edges <- create_edges(
  data.frame(
    subject_id = c("P1", "P2", "P3", "P4"),
    outcome_id = c("O_case", "O_ctrl", "O_case", "O_ctrl"),
    stringsAsFactors = FALSE
  ),
  "subject_id", "outcome_id",
  "Subject", "Outcome", "subject_has_outcome"
)

precedence_edges <- create_edges(
  data.frame(
    from_timepoint = c("T0", "T1"),
    to_timepoint = c("T1", "T2"),
    stringsAsFactors = FALSE
  ),
  "from_timepoint", "to_timepoint",
  "Timepoint", "Timepoint", "timepoint_precedes"
)

featureset_from_study <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    study_id = "ST1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "study_id",
  "FeatureSet", "Study", "featureset_generated_from_study"
)

featureset_from_batch <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    batch_id = "B1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "batch_id",
  "FeatureSet", "Batch", "featureset_generated_from_batch"
)

## ----construction-output------------------------------------------------------
sample_nodes
as.data.frame(sample_nodes)[, c("node_id", "node_type", "node_key", "label")]

edge_preview <- do.call(rbind, lapply(
  list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  as.data.frame
))

edge_preview[, c("from", "to", "edge_type")]

## ----graph--------------------------------------------------------------------
graph <- build_dependency_graph(
  nodes = list(
    sample_nodes, subject_nodes, batch_nodes, study_nodes,
    time_nodes, assay_nodes, featureset_nodes, outcome_nodes
  ),
  edges = list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  graph_name = "vignette_graph",
  dataset_name = attr(meta, "dataset_name")
)

graph
summary(graph)

## ----plot, fig.width = 7, fig.height = 5--------------------------------------
plot(graph)

## ----plot-options, eval = FALSE-----------------------------------------------
# plot(graph, layout = "sugiyama")         # alternative hierarchical layout
# plot(graph, show_labels = FALSE)         # hide labels on dense graphs
# plot(graph, legend = FALSE)              # suppress the legend
# plot(graph, legend_position = "bottomright")
# plot(graph, node_colors = c(Sample = "#000000"))

## ----validation---------------------------------------------------------------
validation <- validate_graph(graph)

validation
as.data.frame(validation)[, c("level", "severity", "code", "message")]

## ----strictness---------------------------------------------------------------
tryCatch(
  derive_split_constraints(graph, mode = "subject", samples = c("S1", "BAD")),
  error = function(e) e$message
)

## ----overrides-fixture--------------------------------------------------------
multi_nodes <- graph_node_set(data.frame(
  node_id   = c("sample:S1", "subject:P1", "subject:P2"),
  node_type = c("Sample", "Subject", "Subject"),
  node_key  = c("S1", "P1", "P2"),
  label     = c("S1", "P1", "P2"),
  attrs     = I(list(list(), list(), list())),
  stringsAsFactors = FALSE
))

multi_edges <- graph_edge_set(data.frame(
  edge_id   = c("sample_belongs_to_subject:1", "sample_belongs_to_subject:2"),
  from      = c("sample:S1", "sample:S1"),
  to        = c("subject:P1", "subject:P2"),
  edge_type = c("sample_belongs_to_subject", "sample_belongs_to_subject"),
  attrs     = I(list(list(), list())),
  stringsAsFactors = FALSE
))

multi_graph <- dependency_graph(nodes = multi_nodes, edges = multi_edges)

## ----overrides-default--------------------------------------------------------
default_report <- validate_graph(multi_graph)
default_report$valid
default_report$issues[, c("severity", "code", "message")]

tryCatch(
  derive_split_constraints(multi_graph, mode = "subject"),
  error = function(e) e$message
)

## ----overrides-on-------------------------------------------------------------
# Validator: pass.
permissive_report <- validate_graph(
  multi_graph,
  validation_overrides = list(allow_multi_subject_samples = TRUE)
)
permissive_report$valid

# Constraint derivation: pick the first listed subject and record the
# ambiguity in metadata$warnings instead of erroring.
multi_graph$metadata$validation_overrides <-
  list(allow_multi_subject_samples = TRUE)
relaxed_constraint <- derive_split_constraints(multi_graph, mode = "subject")
relaxed_constraint$sample_map[, c("sample_id", "group_id")]
relaxed_constraint$metadata$warnings

## ----neighbors-and-paths------------------------------------------------------
neighbors_s1 <- query_neighbors(graph, node_ids = "sample:S1", direction = "out")
neighbors_s1
as.data.frame(neighbors_s1)[, c("seed_node_id", "node_id", "node_type", "edge_type")]

subject_outcome_path <- query_shortest_paths(
  graph,
  from = "sample:S1",
  to = "outcome:O_case",
  edge_types = c("sample_belongs_to_subject", "subject_has_outcome")
)

subject_outcome_path
as.data.frame(subject_outcome_path)

## ----projected-dependencies---------------------------------------------------
shared_dependencies <- detect_shared_dependencies(
  graph,
  via = c("Subject", "Batch", "FeatureSet")
)

as.data.frame(shared_dependencies)[, c(
  "sample_id_1", "sample_id_2", "shared_node_type", "shared_node_id", "edge_type"
)]

dependency_components <- detect_dependency_components(
  graph,
  via = c("Subject", "Batch")
)

as.data.frame(dependency_components)

## ----constraints--------------------------------------------------------------
subject_constraint <- derive_split_constraints(graph, mode = "subject")
batch_constraint <- derive_split_constraints(graph, mode = "batch")
study_constraint <- derive_split_constraints(graph, mode = "study")
time_constraint <- derive_split_constraints(graph, mode = "time")

strict_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "strict",
  via = c("Subject", "Batch")
)

rule_based_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "rule_based",
  priority = c("batch", "study", "subject", "time")
)

constraint_overview <- do.call(rbind, lapply(
  list(
    subject = subject_constraint,
    batch = batch_constraint,
    study = study_constraint,
    time = time_constraint,
    composite_strict = strict_constraint,
    composite_rule = rule_based_constraint
  ),
  function(x) {
    data.frame(
      strategy = x$strategy,
      groups = length(unique(x$sample_map$group_id)),
      warnings = if (is.null(x$metadata$warnings)) 0L else length(x$metadata$warnings),
      stringsAsFactors = FALSE
    )
  }
))

constraint_overview <- cbind(constraint = row.names(constraint_overview), constraint_overview)
row.names(constraint_overview) <- NULL

constraint_overview

## ----batch-constraint---------------------------------------------------------
batch_constraint
as.data.frame(batch_constraint)[, c("sample_id", "group_id", "group_label", "explanation")]

## ----time-constraint----------------------------------------------------------
time_constraint
as.data.frame(time_constraint)[, c("sample_id", "group_id", "timepoint_id", "order_rank")]

## ----composite-constraints----------------------------------------------------
strict_constraint
as.data.frame(strict_constraint)[, c("sample_id", "group_id", "constraint_type")]

rule_based_constraint
as.data.frame(rule_based_constraint)[, c("sample_id", "group_id", "constraint_type", "group_label")]

## ----site-structure-----------------------------------------------------------
site_meta <- data.frame(
  sample_id  = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id = c("P1", "P1", "P2", "P2", "P3", "P3"),
  site_id    = c("NYC", "NYC", "BOS", "BOS", "NYC", "BOS"),
  stringsAsFactors = FALSE
)

site_graph <- graph_from_metadata(site_meta, graph_name = "multi-site")

# Group samples so that no collection site straddles a train/test split.
site_constraint <- derive_split_constraints(site_graph, mode = "site")
as.data.frame(site_constraint)[, c("sample_id", "group_id", "group_label")]

## ----site-spec----------------------------------------------------------------
subject_then_block_by_site <- as_split_spec(
  derive_split_constraints(site_graph, mode = "subject"),
  graph = site_graph
)
subject_then_block_by_site$block_vars
head(subject_then_block_by_site$sample_data[, c("sample_id", "group_id", "site_group")])

## ----region-structure---------------------------------------------------------
region_meta <- data.frame(
  sample_id = c("S1", "S2", "S3", "S4"),
  region_id = c("cortex", "cortex", "hippocampus", "hippocampus"),
  stringsAsFactors = FALSE
)
region_graph <- graph_from_metadata(region_meta, graph_name = "regions")
as.data.frame(derive_split_constraints(region_graph, mode = "region"))[
  , c("sample_id", "group_id", "group_label")
]

## ----platform-assay-----------------------------------------------------------
tech_meta <- data.frame(
  sample_id   = c("S1", "S2", "S3", "S4"),
  platform_id = c("illumina", "illumina", "nanopore", "nanopore"),
  assay_id    = c("rnaseq", "rnaseq", "wgs", "wgs"),
  stringsAsFactors = FALSE
)
tech_graph <- graph_from_metadata(tech_meta, graph_name = "tech")

as.data.frame(derive_split_constraints(tech_graph, mode = "platform"))[
  , c("sample_id", "group_id", "group_label")
]
as.data.frame(derive_split_constraints(tech_graph, mode = "assay"))[
  , c("sample_id", "group_id", "group_label")
]

## ----cluster-plot, fig.width = 7, fig.height = 5------------------------------
mixed_meta <- data.frame(
  sample_id   = c("S1", "S2", "S3", "S4"),
  subject_id  = c("P1", "P1", "P2", "P2"),
  site_id     = c("NYC", "NYC", "BOS", "BOS"),
  platform_id = c("illumina", "illumina", "nanopore", "nanopore"),
  stringsAsFactors = FALSE
)
plot(graph_from_metadata(mixed_meta, graph_name = "mixed_structure"))

## ----relatedness-demo---------------------------------------------------------
kin <- data.frame(
  id1     = c("P1", "P2", "P1"),
  id2     = c("P2", "P3", "P4"),
  kinship = c(0.25, 0.20, 0.02),   # P1-P4 is below the 0.1 threshold
  stringsAsFactors = FALSE
)
rel_edges <- relatedness_edges_from_kinship(kin, threshold = 0.1)

rel_meta <- data.frame(
  sample_id  = paste0("S", 1:4),
  subject_id = c("P1", "P2", "P3", "P4"),
  stringsAsFactors = FALSE
)
rel_graph <- build_dependency_graph(
  nodes = list(
    create_nodes(rel_meta, "Sample", "sample_id"),
    create_nodes(rel_meta, "Subject", "subject_id")
  ),
  edges = list(
    create_edges(rel_meta, "sample_id", "subject_id",
                 "Sample", "Subject", "sample_belongs_to_subject"),
    rel_edges
  )
)

grouping_vector(derive_split_constraints(rel_graph, mode = "relatedness"))

## ----modeling-pointer, eval = FALSE-------------------------------------------
# vignette("modeling-structure", package = "splitGraph")

## ----precedence-only----------------------------------------------------------
precedence_meta <- data.frame(
  sample_id = c("S1", "S2", "S3"),
  subject_id = c("P1", "P1", "P2"),
  study_id = c("ST1", "ST1", "ST2"),
  timepoint_id = c("T0", "T1", "T2"),
  stringsAsFactors = FALSE
)

precedence_graph <- build_dependency_graph(
  nodes = list(
    create_nodes(precedence_meta, type = "Sample", id_col = "sample_id"),
    create_nodes(precedence_meta, type = "Subject", id_col = "subject_id"),
    create_nodes(precedence_meta, type = "Study", id_col = "study_id"),
    create_nodes(
      data.frame(timepoint_id = c("T0", "T1", "T2"), stringsAsFactors = FALSE),
      type = "Timepoint",
      id_col = "timepoint_id"
    )
  ),
  edges = list(
    create_edges(
      precedence_meta, "sample_id", "subject_id",
      "Sample", "Subject", "sample_belongs_to_subject"
    ),
    create_edges(
      precedence_meta, "sample_id", "study_id",
      "Sample", "Study", "sample_from_study"
    ),
    create_edges(
      precedence_meta, "sample_id", "timepoint_id",
      "Sample", "Timepoint", "sample_collected_at_timepoint"
    ),
    create_edges(
      data.frame(
        from_timepoint = c("T0", "T1"),
        to_timepoint = c("T1", "T2"),
        stringsAsFactors = FALSE
      ),
      "from_timepoint", "to_timepoint",
      "Timepoint", "Timepoint", "timepoint_precedes"
    )
  ),
  graph_name = "precedence_only_graph"
)

precedence_time_constraint <- derive_split_constraints(precedence_graph, mode = "time")

precedence_time_constraint$metadata$time_order_source
as.data.frame(precedence_time_constraint)[, c("sample_id", "timepoint_id", "time_index", "order_rank")]

## ----split-spec---------------------------------------------------------------
split_spec <- as_split_spec(strict_constraint, graph = graph)
split_spec

as.data.frame(split_spec)[, c(
  "sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank"
)]

split_spec_validation <- validate_split_spec(split_spec)
split_spec_validation
as.data.frame(split_spec_validation)

## ----risk-summary-------------------------------------------------------------
risk_summary <- summarize_leakage_risks(
  graph,
  constraint = strict_constraint,
  split_spec = split_spec
)

risk_summary
as.data.frame(risk_summary)[, c("source", "severity", "category", "severed", "message")]

## ----serialize, eval = requireNamespace("jsonlite", quietly = TRUE)-----------
spec_path <- tempfile(fileext = ".json")
write_split_spec(split_spec, spec_path)

# Validate the file against the shipped JSON Schema.
validate_split_spec_json(spec_path)$valid

# Round-trip it back into R unchanged.
spec_round_trip <- read_split_spec(spec_path)
identical(split_spec$sample_data$group_id, spec_round_trip$sample_data$group_id)

unlink(spec_path)

## ----cookbook-pointer, eval = FALSE-------------------------------------------
# vignette("adapter-cookbook", package = "splitGraph")       # R adapters
# vignette("cross-language-handoff", package = "splitGraph") # Python / sklearn

## ----case-study-1-------------------------------------------------------------
subject_groups <- grouping_vector(subject_constraint)
time_groups <- time_constraint$sample_map[, c("sample_id", "group_id", "timepoint_id", "order_rank")]

subject_groups
time_groups

## ----case-study-2-------------------------------------------------------------
cross_study_issues <- as.data.frame(validation)[
  as.data.frame(validation)$code == "subject_cross_study_overlap",
  c("severity", "code", "message")
]

p2_shared <- detect_shared_dependencies(
  graph,
  via = "Subject",
  samples = c("S3", "S6")
)

study_only_map <- study_constraint$sample_map[, c("sample_id", "group_id", "group_label")]
strict_map <- strict_constraint$sample_map[, c("sample_id", "group_id", "constraint_type")]

cross_study_issues
as.data.frame(p2_shared)
study_only_map[study_only_map$sample_id %in% c("S3", "S6"), ]
strict_map[strict_map$sample_id %in% c("S3", "S6"), ]

## ----case-study-3-------------------------------------------------------------
batch_missing <- batch_constraint$sample_map[
  batch_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "group_label", "explanation")
]

rule_based_missing <- rule_based_constraint$sample_map[
  rule_based_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "constraint_type", "group_label", "explanation")
]

split_spec_missing <- as.data.frame(split_spec)[
  as.data.frame(split_spec)$sample_id == "S5",
  c("sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank")
]

batch_missing
rule_based_missing
split_spec_missing

## ----case-study-4-------------------------------------------------------------
strategy_summary <- data.frame(
  constraint = c("subject", "batch", "study", "time", "composite_strict", "composite_rule"),
  groups = c(
    length(unique(subject_constraint$sample_map$group_id)),
    length(unique(batch_constraint$sample_map$group_id)),
    length(unique(study_constraint$sample_map$group_id)),
    length(unique(time_constraint$sample_map$group_id)),
    length(unique(strict_constraint$sample_map$group_id)),
    length(unique(rule_based_constraint$sample_map$group_id))
  ),
  warnings = c(
    length(or_empty(subject_constraint$metadata$warnings)),
    length(or_empty(batch_constraint$metadata$warnings)),
    length(or_empty(study_constraint$metadata$warnings)),
    length(or_empty(time_constraint$metadata$warnings)),
    length(or_empty(strict_constraint$metadata$warnings)),
    length(or_empty(rule_based_constraint$metadata$warnings))
  ),
  recommended_resampling = c(
    as_split_spec(subject_constraint, graph = graph)$recommended_resampling,
    as_split_spec(batch_constraint, graph = graph)$recommended_resampling,
    as_split_spec(study_constraint, graph = graph)$recommended_resampling,
    as_split_spec(time_constraint, graph = graph)$recommended_resampling,
    as_split_spec(strict_constraint, graph = graph)$recommended_resampling,
    as_split_spec(rule_based_constraint, graph = graph)$recommended_resampling
  ),
  stringsAsFactors = FALSE
)

strategy_summary

# Graph assembly helpers.

# Build a helpful error message when edge endpoints don't match any node ID.
# If the missing references look like prefixed IDs (e.g. "sample:S1") while
# the node table uses unprefixed keys (or vice versa), point at that as a
# likely cause — this is the most common first-time-user mistake.
.depgraph_missing_reference_message <- function(missing, node_ids, side) {
  base_msg <- paste0(
    "Edge data contains `", side, "` references that are not present in the node table: ",
    paste(utils::head(missing, 5L), collapse = ", "),
    if (length(missing) > 5L) paste0(" ... (+", length(missing) - 5L, " more)") else ""
  )

  missing_has_prefix <- any(grepl(":", missing, fixed = TRUE))
  nodes_have_prefix <- any(grepl(":", node_ids, fixed = TRUE))

  hint <- ""
  if (missing_has_prefix && !nodes_have_prefix) {
    hint <- paste0(
      "\nHint: edge endpoints look prefixed (e.g. `sample:S1`) but node IDs are not. ",
      "Either rebuild nodes with `prefix = TRUE` (the default) or rebuild edges with ",
      "`from_prefix = FALSE` and `to_prefix = FALSE`."
    )
  } else if (!missing_has_prefix && nodes_have_prefix) {
    hint <- paste0(
      "\nHint: node IDs are prefixed (e.g. `sample:S1`) but edge endpoints are not. ",
      "Rebuild edges with `from_prefix = TRUE` and `to_prefix = TRUE`, or rebuild nodes ",
      "with `prefix = FALSE`."
    )
  }

  paste0(base_msg, hint)
}

#' Assemble and Validate Dependency Graphs
#'
#' Combine canonical node and edge tables into a typed dependency graph and
#' perform structural, semantic, and graph-local leakage-aware validation.
#'
#' @param nodes,edges Lists of \code{graph_node_set} and \code{graph_edge_set}
#'   objects.
#' @param graph_name,dataset_name Optional metadata labels.
#' @param validate If \code{TRUE}, run \code{validate_graph()} before returning.
#' @param validation_overrides Optional named list of explicit validation
#'   exceptions. Currently supported keys:
#'   \describe{
#'     \item{\code{allow_multi_subject_samples}}{If \code{TRUE}, the semantic
#'       validator does not flag samples linked to multiple subjects, and
#'       \code{derive_split_constraints(mode = "subject")} silently keeps the
#'       first listed subject assignment (recording the ambiguity in
#'       \code{metadata$warnings}). Defaults to \code{FALSE}.}
#'   }
#'   When passed to \code{validate_graph()} or \code{validate_depgraph()},
#'   the override is merged into the graph's existing
#'   \code{validation_overrides} for the duration of the call only.
#' @param graph A \code{dependency_graph}.
#' @param checks \strong{Deprecated.} Use \code{levels} and \code{severities}
#'   instead. Retained for backward compatibility with 0.1.0 callers.
#' @param error_on_fail If \code{TRUE}, stop when validation errors are found
#'   across all detected issues from the selected validation levels, even if
#'   those errors are hidden from \code{issues} by \code{severities}.
#' @param levels Optional validation layers to run.
#' @param severities Optional severities to retain in the returned
#'   \code{issues} table. This filter does not change whether the graph is
#'   considered valid.
#' @param x A \code{dependency_graph}.
#' @return For \code{build_dependency_graph()}, a \code{dependency_graph}. For
#'   \code{validate_graph()} and \code{validate_depgraph()}, a
#'   \code{depgraph_validation_report}. For \code{as_igraph()}, the underlying
#'   \code{igraph} object.
#' @examples
#' meta <- data.frame(
#'   sample_id = c("S1", "S2"),
#'   subject_id = c("P1", "P2")
#' )
#'
#' samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
#' subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
#' edges <- create_edges(
#'   meta,
#'   "sample_id",
#'   "subject_id",
#'   "Sample",
#'   "Subject",
#'   "sample_belongs_to_subject"
#' )
#'
#' g <- build_dependency_graph(list(samples, subjects), list(edges))
#' validate_graph(g)
#' @export
build_dependency_graph <- function(nodes, edges, graph_name = NULL, dataset_name = NULL, validate = TRUE, validation_overrides = list()) {
  node_data <- .depgraph_bind_data(nodes, "data")
  edge_data <- .depgraph_bind_data(edges, "data")

  .depgraph_assert(any(node_data$node_type == "Sample"), "A dependency graph must contain at least one `Sample` node.")
  .depgraph_assert(anyDuplicated(node_data$node_id) == 0L, "Duplicate `node_id` values found in node sets.")
  .depgraph_assert(anyDuplicated(edge_data$edge_id) == 0L, "Duplicate `edge_id` values found in edge sets.")
  .depgraph_assert(
    all(edge_data$from %in% node_data$node_id),
    .depgraph_missing_reference_message(
      missing = setdiff(unique(edge_data$from), node_data$node_id),
      node_ids = node_data$node_id,
      side = "from"
    )
  )
  .depgraph_assert(
    all(edge_data$to %in% node_data$node_id),
    .depgraph_missing_reference_message(
      missing = setdiff(unique(edge_data$to), node_data$node_id),
      node_ids = node_data$node_id,
      side = "to"
    )
  )

  graph_obj <- dependency_graph(
    nodes = graph_node_set(node_data),
    edges = graph_edge_set(edge_data),
    graph = NULL,
    metadata = list(
      graph_name = graph_name,
      dataset_name = dataset_name,
      created_at = Sys.time(),
      schema_version = .depgraph_schema_version,
      validation_overrides = validation_overrides
    )
  )

  if (isTRUE(validate)) {
    validation <- validate_graph(graph_obj)
    if (!isTRUE(validation$valid)) {
      stop(
        paste(
          c("Graph validation failed.", validation$errors, validation$warnings),
          collapse = "\n"
        ),
        call. = FALSE
      )
    }
  }

  graph_obj
}

#' @rdname build_dependency_graph
#' @export
build_depgraph <- function(nodes, edges, graph_name = NULL, dataset_name = NULL, validate = TRUE, validation_overrides = list()) {
  .Deprecated(
    new = "build_dependency_graph",
    package = "splitGraph",
    msg = "`build_depgraph()` is deprecated. Use `build_dependency_graph()` instead."
  )
  build_dependency_graph(
    nodes = nodes,
    edges = edges,
    graph_name = graph_name,
    dataset_name = dataset_name,
    validate = validate,
    validation_overrides = validation_overrides
  )
}

#' @rdname build_dependency_graph
#' @export
as_igraph <- function(x) {
  .depgraph_assert(inherits(x, "dependency_graph"), "`x` must be a `dependency_graph`.")
  x$graph
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)
library(splitGraph)

## ----cluster------------------------------------------------------------------
meta <- data.frame(
  sample_id   = paste0("S", 1:6),
  subject_id  = c("P1", "P1", "P2", "P2", "P3", "P3"),
  site_id     = c("NYC", "NYC", "BOS", "BOS", "NYC", "BOS"),
  platform_id = c("illumina", "illumina", "nanopore", "nanopore", "illumina", "nanopore"),
  assay_id    = c("rnaseq", "rnaseq", "rnaseq", "wgs", "wgs", "wgs"),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "structure-demo")

grouping_vector(derive_split_constraints(g, mode = "site"))
grouping_vector(derive_split_constraints(g, mode = "platform"))
grouping_vector(derive_split_constraints(g, mode = "assay"))

## ----block-annotations--------------------------------------------------------
spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
spec$block_vars
head(spec$sample_data[, c("sample_id", "group_id",
                          "site_group", "platform_group", "assay_group")])

## ----composite----------------------------------------------------------------
constraint <- derive_split_constraints(
  g, mode = "composite", strategy = "strict",
  via = c("Subject", "Site", "Platform")
)
grouping_vector(constraint)

## ----relatedness--------------------------------------------------------------
# A kinship table over subject pairs (one sample per subject here for clarity).
# P1-P2 and P2-P3 clear the threshold and chain together; P5-P6 form a second
# related pair; P1-P4 is too weak to count.
kin <- data.frame(
  id1     = c("P1", "P2", "P1", "P5"),
  id2     = c("P2", "P3", "P4", "P6"),
  kinship = c(0.25, 0.20, 0.02, 0.30),
  stringsAsFactors = FALSE
)
rel_edges <- relatedness_edges_from_kinship(kin, threshold = 0.1)

meta_r <- data.frame(
  sample_id  = paste0("S", 1:6),
  subject_id = paste0("P", 1:6),
  stringsAsFactors = FALSE
)
samples  <- create_nodes(meta_r, "Sample", "sample_id")
subjects <- create_nodes(meta_r, "Subject", "subject_id")
belongs  <- create_edges(meta_r, "sample_id", "subject_id",
                         "Sample", "Subject", "sample_belongs_to_subject")

g_rel <- build_dependency_graph(list(samples, subjects), list(belongs, rel_edges))

rel_groups <- grouping_vector(derive_split_constraints(g_rel, mode = "relatedness"))
rel_groups

## ----rel-plot, fig.width = 6.5, fig.height = 4.5------------------------------
subject_group <- setNames(rel_groups[meta_r$sample_id], meta_r$subject_id)
kept_pairs <- kin[kin$kinship >= 0.1, c("id1", "id2")]
rel_net <- igraph::graph_from_data_frame(
  kept_pairs, directed = FALSE,
  vertices = data.frame(name = meta_r$subject_id)
)

palette_rel <- c("#4C78A8", "#F58518", "#54A24B", "#B279A2")
set.seed(1)
plot(rel_net,
     vertex.color       = palette_rel[as.integer(factor(subject_group[igraph::V(rel_net)$name]))],
     vertex.size        = 34,
     vertex.label.color = "white",
     vertex.label.font  = 2,
     edge.color         = "grey60",
     edge.width         = 2,
     main               = "Relatedness clusters (kinship >= 0.1)")

## ----rel-threshold------------------------------------------------------------
rel_strict <- relatedness_edges_from_kinship(kin, threshold = 0.22)
g_rel_strict <- build_dependency_graph(list(samples, subjects), list(belongs, rel_strict))

grouping_vector(derive_split_constraints(g_rel_strict, mode = "relatedness"))

## ----spatial------------------------------------------------------------------
# Two spatial clusters. Cluster 1 (S1-S3) is a chain: neighbouring pairs are
# within the radius, but the endpoints are not.
coords <- data.frame(
  sample_id = paste0("S", 1:6),
  x = c(0, 1, 2,  6.0, 6.9, 6.2),
  y = c(0, 1, 0,  6.0, 6.6, 5.3),
  stringsAsFactors = FALSE
)
adj_edges <- spatial_edges_from_coords(coords, radius = 1.5)

meta_s <- data.frame(
  sample_id  = paste0("S", 1:6),
  subject_id = paste0("P", 1:6),
  stringsAsFactors = FALSE
)
samples_s  <- create_nodes(meta_s, "Sample", "sample_id")
subjects_s <- create_nodes(meta_s, "Subject", "subject_id")
belongs_s  <- create_edges(meta_s, "sample_id", "subject_id",
                           "Sample", "Subject", "sample_belongs_to_subject")

g_sp <- build_dependency_graph(list(samples_s, subjects_s), list(belongs_s, adj_edges))

sp_groups <- grouping_vector(derive_split_constraints(g_sp, mode = "spatial"))
sp_groups

## ----sp-plot, fig.width = 6.5, fig.height = 5---------------------------------
sp_grp <- factor(sp_groups[coords$sample_id])
row_of <- setNames(seq_len(nrow(coords)), coords$sample_id)
from_i <- row_of[sub("^sample:", "", adj_edges$data$from)]
to_i   <- row_of[sub("^sample:", "", adj_edges$data$to)]
palette_sp <- c("#4C78A8", "#F58518")

plot(coords$x, coords$y, type = "n", asp = 1, xlab = "x", ylab = "y",
     main = "Spatial groups (radius = 1.5)")
segments(coords$x[from_i], coords$y[from_i],
         coords$x[to_i],   coords$y[to_i], col = "grey60", lwd = 2)
points(coords$x, coords$y, pch = 19, cex = 3.5, col = palette_sp[as.integer(sp_grp)])
text(coords$x, coords$y, labels = coords$sample_id, col = "white", cex = 0.8, font = 2)
legend("topleft", legend = levels(sp_grp), pch = 19,
       col = palette_sp[seq_along(levels(sp_grp))], title = "Spatial group", bty = "n")

## ----subset-scoping-----------------------------------------------------------
grouping_vector(
  derive_split_constraints(g_sp, mode = "spatial", samples = c("S1", "S3"))
)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}

## ----build-spec---------------------------------------------------------------
meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
  batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
  timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
  time_index   = c(0, 1, 0, 1, 0, 1),
  outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl"),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "cookbook")
subject_constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(subject_constraint, graph = g)
spec

## ----show-sample-data---------------------------------------------------------
as.data.frame(spec)[, c("sample_id", "group_id", "batch_group", "order_rank")]

## ----adapter-base-r-----------------------------------------------------------
logo_folds <- function(spec, observation_data, sample_id_col = "sample_id") {
  stopifnot(inherits(spec, "split_spec"))
  if (!sample_id_col %in% names(observation_data)) {
    stop("`observation_data` must contain a `", sample_id_col, "` column.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$group_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )
  joined$.row <- seq_len(nrow(joined))
  groups <- split(joined$.row, joined[[spec$group_var]])

  lapply(names(groups), function(g) {
    list(
      group   = g,
      train   = unlist(groups[setdiff(names(groups), g)], use.names = FALSE),
      assess  = groups[[g]]
    )
  })
}

# Pretend we have an observation frame keyed by sample_id.
obs <- data.frame(
  sample_id = meta$sample_id,
  x = rnorm(nrow(meta)),
  y = rbinom(nrow(meta), 1, 0.5)
)

folds <- logo_folds(spec, obs)
length(folds)
folds[[1]]

## ----block-vars---------------------------------------------------------------
spec$block_vars
head(spec$sample_data[, c("sample_id", spec$group_var, spec$block_vars)])

## ----block-audit--------------------------------------------------------------
block <- spec$block_vars[[1]]
block_of <- setNames(spec$sample_data[[block]], spec$sample_data$sample_id)

do.call(rbind, lapply(folds, function(f) {
  data.frame(
    held_out_group    = f$group,
    straddling_batches = paste(
      intersect(block_of[obs$sample_id[f$train]],
                block_of[obs$sample_id[f$assess]]),
      collapse = ", "
    )
  )
}))

## ----adapter-rsample-group, eval=FALSE----------------------------------------
# spec_to_group_vfold <- function(spec, observation_data,
#                                 v = NULL,
#                                 sample_id_col = "sample_id") {
#   stopifnot(inherits(spec, "split_spec"))
#   if (!requireNamespace("rsample", quietly = TRUE)) {
#     stop("Install rsample to use this adapter.")
#   }
# 
#   joined <- merge(
#     observation_data,
#     spec$sample_data[, c("sample_id", spec$group_var)],
#     by.x = sample_id_col, by.y = "sample_id", sort = FALSE
#   )
# 
#   n_groups <- length(unique(joined[[spec$group_var]]))
#   if (is.null(v)) v <- n_groups
# 
#   rsample::group_vfold_cv(
#     data  = joined,
#     group = !!spec$group_var,
#     v     = v
#   )
# }

## ----adapter-rsample-rolling, eval=FALSE--------------------------------------
# spec_to_rolling_origin <- function(spec, observation_data,
#                                    sample_id_col = "sample_id",
#                                    initial = NULL,
#                                    assess = 1L) {
#   stopifnot(inherits(spec, "split_spec"))
#   if (is.null(spec$time_var)) {
#     stop("This split_spec has no `time_var`; ordered evaluation is not available.")
#   }
#   if (!requireNamespace("rsample", quietly = TRUE)) {
#     stop("Install rsample to use this adapter.")
#   }
# 
#   joined <- merge(
#     observation_data,
#     spec$sample_data[, c("sample_id", spec$time_var)],
#     by.x = sample_id_col, by.y = "sample_id", sort = FALSE
#   )
#   ordered <- joined[order(joined[[spec$time_var]]), , drop = FALSE]
# 
#   if (is.null(initial)) initial <- max(1L, floor(nrow(ordered) * 0.6))
#   rsample::rolling_origin(ordered, initial = initial, assess = assess)
# }

## ----serialize, eval = requireNamespace("jsonlite", quietly = TRUE)-----------
tmp <- tempfile(fileext = ".json")
write_split_spec(spec, tmp)

# The file opens with its $schema reference and schema_version.
cat(readLines(tmp, n = 5), sep = "\n")

# Validate the file against the shipped JSON Schema, then read it back exactly.
validate_split_spec_json(tmp)$valid
spec2 <- read_split_spec(tmp)
identical(spec$sample_data$group_id, spec2$sample_data$group_id)

unlink(tmp)

## ----xlang-pointer, eval = FALSE----------------------------------------------
# vignette("cross-language-handoff", package = "splitGraph")

## ----dispatch-----------------------------------------------------------------
recommend_adapter <- function(spec) {
  switch(
    spec$recommended_resampling,
    grouped_cv          = "group_vfold_cv (group = group_id)",
    blocked_cv          = "group_vfold_cv (group = group_id)",
    custom_grouped_cv   = "group_vfold_cv (group = group_id)",
    leave_one_group_out = "leave-one-group-out over group_id",
    ordered_split       = "rolling_origin (order by order_rank)",
    "group_vfold_cv (default)"
  )
}

# The subject spec recommends grouped CV; a time-mode spec recommends ordering.
recommend_adapter(spec)
time_spec <- as_split_spec(derive_split_constraints(g, mode = "time"), graph = g)
recommend_adapter(time_spec)

# JSON serialization for splitGraph's two handoff objects:
# `dependency_graph` and `split_spec`. The on-disk format is documented in
# the @section JSON format blocks below and is intended to be stable across
# patch releases. Schema bumps follow the rules described in
# `R/splitGraph-package.R` (`.depgraph_schema_version`).

.depgraph_require_jsonlite <- function() {
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop(
      "Package 'jsonlite' is required for splitGraph JSON serialization. ",
      "Install it with: install.packages(\"jsonlite\").",
      call. = FALSE
    )
  }
  invisible(TRUE)
}

.depgraph_check_writable_path <- function(path) {
  parent <- dirname(path)
  if (!nzchar(parent) || identical(parent, ".")) return(invisible())
  if (!dir.exists(parent)) {
    stop(
      "Parent directory does not exist: ", parent,
      ". Create it before writing.",
      call. = FALSE
    )
  }
  invisible()
}

# Convert a POSIXct (or NULL/NA) to an ISO 8601 string; round-trip via
# .depgraph_iso_to_posix.
.depgraph_posix_to_iso <- function(x) {
  if (is.null(x) || (length(x) == 1L && is.na(x))) {
    return(NA_character_)
  }
  format(as.POSIXct(x), "%Y-%m-%dT%H:%M:%OS6%z")
}

.depgraph_iso_to_posix <- function(x) {
  na_posix <- .POSIXct(NA_real_, tz = "UTC")
  if (is.null(x) || length(x) == 0L) {
    return(na_posix)
  }
  if (length(x) == 1L && is.na(x)) {
    return(na_posix)
  }
  try_parse <- function(call_expr) {
    tryCatch(
      suppressWarnings(call_expr),
      error = function(e) na_posix
    )
  }
  parsed <- try_parse(as.POSIXct(x, format = "%Y-%m-%dT%H:%M:%OS%z"))
  if (length(parsed) == 0L || is.na(parsed)) {
    parsed <- try_parse(as.POSIXct(x))
  }
  if (length(parsed) == 0L || is.na(parsed)) na_posix else parsed
}

# Public `$id` of the shipped JSON Schema for an object type, referenced from
# written JSON via the `$schema` key so external consumers can locate the
# formal contract. Mirrors the file names under `inst/schema/`.
.depgraph_schema_url <- function(object_type) {
  paste0(
    "https://raw.githubusercontent.com/selcukorkmaz/splitGraph/main/inst/schema/",
    object_type, ".schema.json"
  )
}

# Major-version component of a "X.Y.Z" schema string; NA if unparseable.
.depgraph_schema_major <- function(version) {
  major <- suppressWarnings(as.integer(sub("\\..*$", "", as.character(version)[1L])))
  major
}

.depgraph_check_schema_version <- function(observed, what) {
  if (is.null(observed) || !nzchar(observed)) {
    warning(
      "Reading ", what, ": no `schema_version` recorded in JSON. ",
      "Assuming current schema (", .depgraph_schema_version, ").",
      call. = FALSE
    )
    return(invisible())
  }

  observed <- as.character(observed)
  if (identical(observed, .depgraph_schema_version)) {
    return(invisible())
  }

  observed_major <- .depgraph_schema_major(observed)
  installed_major <- .depgraph_schema_major(.depgraph_schema_version)

  # Same MAJOR is read-compatible: additive-only differences, load silently.
  if (!is.na(observed_major) && identical(observed_major, installed_major)) {
    return(invisible())
  }

  migrator <- if (identical(what, "split_spec")) {
    "migrate_split_spec_json()"
  } else {
    "migrate_dependency_graph_json()"
  }
  warning(
    "Reading ", what, ": JSON schema_version `", observed,
    "` differs in major version from installed splitGraph schema_version `",
    .depgraph_schema_version, "`. Loading anyway; consider `", migrator,
    "` to upgrade the file.",
    call. = FALSE
  )
  invisible()
}

# Convert an attrs list (one row's named list) into a JSON-friendly list.
# Empty attrs become an empty named list (jsonlite emits `{}`); scalars stay
# as scalars.
.depgraph_attrs_to_json <- function(x) {
  if (is.null(x) || length(x) == 0L) {
    return(structure(list(), names = character()))
  }
  as.list(x)
}

# Reverse: a parsed JSON object becomes a named list (empty -> list()).
.depgraph_attrs_from_json <- function(x) {
  if (is.null(x)) return(list())
  if (!is.list(x)) return(list())
  if (length(x) == 0L) return(list())
  x
}

.depgraph_node_row_to_list <- function(row) {
  list(
    node_id   = row$node_id,
    node_type = row$node_type,
    node_key  = row$node_key,
    label     = row$label,
    attrs     = .depgraph_attrs_to_json(row$attrs[[1L]])
  )
}

.depgraph_edge_row_to_list <- function(row) {
  list(
    edge_id   = row$edge_id,
    from      = row$from,
    to        = row$to,
    edge_type = row$edge_type,
    attrs     = .depgraph_attrs_to_json(row$attrs[[1L]])
  )
}

# Serialize graph metadata, dropping fields that are not portable
# (validation_overrides IS preserved; igraph object is not stored).
.depgraph_metadata_to_json <- function(metadata) {
  out <- list(
    graph_name           = metadata$graph_name %||% NA_character_,
    dataset_name         = metadata$dataset_name %||% NA_character_,
    created_at           = .depgraph_posix_to_iso(metadata$created_at),
    schema_version       = metadata$schema_version %||% .depgraph_schema_version,
    validation_overrides = if (is.null(metadata$validation_overrides)) {
      structure(list(), names = character())
    } else {
      as.list(metadata$validation_overrides)
    }
  )
  out
}

.depgraph_metadata_from_json <- function(meta_list) {
  if (is.null(meta_list)) meta_list <- list()
  out <- list(
    graph_name     = if (is.null(meta_list$graph_name) || is.na(meta_list$graph_name)) NULL else as.character(meta_list$graph_name),
    dataset_name   = if (is.null(meta_list$dataset_name) || is.na(meta_list$dataset_name)) NULL else as.character(meta_list$dataset_name),
    created_at     = .depgraph_iso_to_posix(meta_list$created_at),
    schema_version = meta_list$schema_version %||% .depgraph_schema_version
  )
  if (!is.null(meta_list$validation_overrides) && length(meta_list$validation_overrides) > 0L) {
    out$validation_overrides <- as.list(meta_list$validation_overrides)
  }
  out
}

# ---- Public API: dependency_graph -------------------------------------------

#' Serialize a Dependency Graph to JSON
#'
#' Write a \code{dependency_graph} to a JSON file and read it back. The on-disk
#' format is intentionally simple and stable: it captures the canonical node
#' table, the canonical edge table (each with their list-column of
#' attributes), the graph metadata (including \code{validation_overrides}),
#' and the data-model \code{schema_version}. The internal \code{igraph}
#' representation is not stored; it is rebuilt on read via
#' \code{dependency_graph()}.
#'
#' This makes \code{split_spec}/\code{dependency_graph} objects portable
#' across R sessions, and across language boundaries (any consumer that can
#' read JSON can interpret the format).
#'
#' @section JSON format:
#' \preformatted{
#' {
#'   "$schema": "https://.../inst/schema/dependency_graph.schema.json",
#'   "splitGraph_object": "dependency_graph",
#'   "schema_version": "0.2.0",
#'   "metadata": {
#'     "graph_name": "...",
#'     "dataset_name": "...",
#'     "created_at": "2026-04-29T10:11:12.000000+0000",
#'     "schema_version": "0.2.0",
#'     "validation_overrides": { ... }
#'   },
#'   "nodes": [
#'     { "node_id": "sample:S1", "node_type": "Sample",
#'       "node_key": "S1", "label": "S1", "attrs": { ... } },
#'     ...
#'   ],
#'   "edges": [
#'     { "edge_id": "sample_belongs_to_subject:1",
#'       "from": "sample:S1", "to": "subject:P1",
#'       "edge_type": "sample_belongs_to_subject", "attrs": { ... } },
#'     ...
#'   ]
#' }
#' }
#' Reading a file whose \code{schema_version} shares the installed major
#' version loads silently (additive-only differences); a differing major
#' version loads with a warning suggesting \code{migrate_dependency_graph_json()}.
#' The written JSON also carries a \code{$schema} reference to the formal JSON
#' Schema shipped in \code{inst/schema/}; validate a file against it with
#' \code{validate_graph_json()}.
#'
#' @param graph A \code{dependency_graph} produced by
#'   \code{build_dependency_graph()} or \code{graph_from_metadata()}.
#' @param path Path to write to or read from.
#' @param pretty If \code{TRUE} (default), the JSON is indented for human
#'   inspection. Set \code{FALSE} for a compact representation.
#' @return \code{write_dependency_graph()} invisibly returns \code{path}.
#'   \code{read_dependency_graph()} returns a validated
#'   \code{dependency_graph}.
#' @examples
#' if (requireNamespace("jsonlite", quietly = TRUE)) {
#'   meta <- data.frame(
#'     sample_id  = c("S1", "S2"),
#'     subject_id = c("P1", "P2")
#'   )
#'   g <- graph_from_metadata(meta, graph_name = "demo")
#'
#'   tmp <- tempfile(fileext = ".json")
#'   write_dependency_graph(g, tmp)
#'   g2 <- read_dependency_graph(tmp)
#'   identical(g$nodes$data$node_id, g2$nodes$data$node_id)
#'   unlink(tmp)
#' }
#' @export
write_dependency_graph <- function(graph, path, pretty = TRUE) {
  .depgraph_require_jsonlite()
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  .depgraph_assert(is.character(path) && length(path) == 1L && nzchar(path), "`path` must be a single non-empty file path.")
  .depgraph_check_writable_path(path)

  node_rows <- if (nrow(graph$nodes$data) == 0L) list() else lapply(
    seq_len(nrow(graph$nodes$data)),
    function(i) .depgraph_node_row_to_list(graph$nodes$data[i, , drop = FALSE])
  )
  edge_rows <- if (nrow(graph$edges$data) == 0L) list() else lapply(
    seq_len(nrow(graph$edges$data)),
    function(i) .depgraph_edge_row_to_list(graph$edges$data[i, , drop = FALSE])
  )

  payload <- list(
    `$schema`         = .depgraph_schema_url("dependency_graph"),
    splitGraph_object = "dependency_graph",
    schema_version    = .depgraph_schema_version,
    metadata          = .depgraph_metadata_to_json(graph$metadata),
    nodes             = node_rows,
    edges             = edge_rows
  )

  json <- jsonlite::toJSON(
    payload,
    auto_unbox = TRUE,
    null = "null",
    na = "null",
    pretty = isTRUE(pretty)
  )
  writeLines(json, path)
  invisible(path)
}

#' @rdname write_dependency_graph
#' @export
read_dependency_graph <- function(path) {
  .depgraph_require_jsonlite()
  .depgraph_assert(is.character(path) && length(path) == 1L && nzchar(path), "`path` must be a single non-empty file path.")
  .depgraph_assert(file.exists(path), paste0("File not found: ", path))

  parsed <- tryCatch(
    jsonlite::fromJSON(path, simplifyVector = FALSE),
    error = function(e) {
      stop("Failed to parse JSON at `", path, "`: ", conditionMessage(e), call. = FALSE)
    }
  )

  obj_type <- parsed$splitGraph_object %||% NA_character_
  if (!identical(as.character(obj_type), "dependency_graph")) {
    stop(
      "JSON at `", path, "` is not a serialized dependency_graph (found `",
      obj_type, "`). Use `read_split_spec()` for split_spec files.",
      call. = FALSE
    )
  }

  .depgraph_check_schema_version(parsed$schema_version, "dependency_graph")

  node_rows <- parsed$nodes %||% list()
  edge_rows <- parsed$edges %||% list()

  node_data <- if (length(node_rows) == 0L) {
    .depgraph_empty_node_data()
  } else {
    data.frame(
      node_id   = vapply(node_rows, function(r) as.character(r$node_id), character(1)),
      node_type = vapply(node_rows, function(r) as.character(r$node_type), character(1)),
      node_key  = vapply(node_rows, function(r) as.character(r$node_key), character(1)),
      label     = vapply(node_rows, function(r) as.character(r$label), character(1)),
      attrs     = I(lapply(node_rows, function(r) .depgraph_attrs_from_json(r$attrs))),
      stringsAsFactors = FALSE
    )
  }

  edge_data <- if (length(edge_rows) == 0L) {
    .depgraph_empty_edge_data()
  } else {
    data.frame(
      edge_id   = vapply(edge_rows, function(r) as.character(r$edge_id), character(1)),
      from      = vapply(edge_rows, function(r) as.character(r$from), character(1)),
      to        = vapply(edge_rows, function(r) as.character(r$to), character(1)),
      edge_type = vapply(edge_rows, function(r) as.character(r$edge_type), character(1)),
      attrs     = I(lapply(edge_rows, function(r) .depgraph_attrs_from_json(r$attrs))),
      stringsAsFactors = FALSE
    )
  }

  metadata <- .depgraph_metadata_from_json(parsed$metadata)

  dependency_graph(
    nodes    = graph_node_set(node_data),
    edges    = graph_edge_set(edge_data),
    graph    = NULL,
    metadata = metadata
  )
}

# ---- Public API: split_spec -------------------------------------------------

# Convert a split_spec's sample_data row to a JSON-friendly list, preserving
# NA as null in the output stream (na = "null" in toJSON does this for us).
.depgraph_split_spec_row_to_list <- function(row) {
  list(
    sample_id      = row$sample_id,
    sample_node_id = row$sample_node_id,
    group_id       = row$group_id,
    primary_group  = row$primary_group,
    batch_group    = row$batch_group,
    study_group    = row$study_group,
    site_group     = row$site_group,
    region_group   = row$region_group,
    platform_group = row$platform_group,
    assay_group    = row$assay_group,
    timepoint_id   = row$timepoint_id,
    time_index     = row$time_index,
    order_rank     = row$order_rank
  )
}

.depgraph_metadata_split_spec_to_json <- function(metadata) {
  if (is.null(metadata)) return(structure(list(), names = character()))
  out <- as.list(metadata)
  out
}

.depgraph_metadata_split_spec_from_json <- function(meta_list) {
  if (is.null(meta_list) || length(meta_list) == 0L) return(list())
  out <- as.list(meta_list)
  # warnings/relations_used are character vectors; jsonlite keeps them as
  # lists when simplifyVector = FALSE — coerce back.
  for (k in c("warnings", "relations_used")) {
    if (!is.null(out[[k]])) {
      out[[k]] <- as.character(unlist(out[[k]]))
    }
  }
  out
}

#' Serialize a Split Specification to JSON
#'
#' Write a \code{split_spec} to a JSON file and read it back. The on-disk
#' format captures the canonical sample-level table (\code{sample_data}) plus
#' all spec-level fields needed by a downstream resampling adapter
#' (\code{group_var}, \code{block_vars}, \code{time_var},
#' \code{ordering_required}, \code{constraint_mode},
#' \code{constraint_strategy}, \code{recommended_resampling}) and the spec
#' metadata.
#'
#' \code{NA} values in \code{sample_data} are written as JSON \code{null} and
#' read back as \code{NA}.
#'
#' @section JSON format:
#' \preformatted{
#' {
#'   "$schema": "https://.../inst/schema/split_spec.schema.json",
#'   "splitGraph_object": "split_spec",
#'   "schema_version": "0.2.0",
#'   "group_var": "group_id",
#'   "block_vars": ["batch_group", "study_group"],
#'   "time_var": "order_rank",
#'   "ordering_required": false,
#'   "constraint_mode": "subject",
#'   "constraint_strategy": "subject",
#'   "recommended_resampling": "grouped_cv",
#'   "metadata": { ... },
#'   "sample_data": [
#'     { "sample_id": "S1", "group_id": "subject:P1", ... },
#'     ...
#'   ]
#' }
#' }
#'
#' @param spec A \code{split_spec} produced by \code{as_split_spec()}.
#' @param path Path to write to or read from.
#' @param pretty If \code{TRUE} (default), the JSON is indented.
#' @return \code{write_split_spec()} invisibly returns \code{path}.
#'   \code{read_split_spec()} returns a \code{split_spec}.
#' @examples
#' if (requireNamespace("jsonlite", quietly = TRUE)) {
#'   meta <- data.frame(
#'     sample_id  = c("S1", "S2"),
#'     subject_id = c("P1", "P2")
#'   )
#'   g <- graph_from_metadata(meta)
#'   constraint <- derive_split_constraints(g, mode = "subject")
#'   spec <- as_split_spec(constraint, graph = g)
#'
#'   tmp <- tempfile(fileext = ".json")
#'   write_split_spec(spec, tmp)
#'   spec2 <- read_split_spec(tmp)
#'   identical(spec$sample_data$group_id, spec2$sample_data$group_id)
#'   unlink(tmp)
#' }
#' @export
write_split_spec <- function(spec, path, pretty = TRUE) {
  .depgraph_require_jsonlite()
  .depgraph_assert(inherits(spec, "split_spec"), "`spec` must be a `split_spec`.")
  .depgraph_assert(is.character(path) && length(path) == 1L && nzchar(path), "`path` must be a single non-empty file path.")
  .depgraph_check_writable_path(path)

  sample_rows <- if (nrow(spec$sample_data) == 0L) list() else lapply(
    seq_len(nrow(spec$sample_data)),
    function(i) .depgraph_split_spec_row_to_list(spec$sample_data[i, , drop = FALSE])
  )

  payload <- list(
    `$schema`               = .depgraph_schema_url("split_spec"),
    splitGraph_object       = "split_spec",
    schema_version          = .depgraph_schema_version,
    group_var               = spec$group_var,
    block_vars              = if (length(spec$block_vars) == 0L) list() else as.list(spec$block_vars),
    time_var                = spec$time_var %||% NA_character_,
    ordering_required       = isTRUE(spec$ordering_required),
    constraint_mode         = spec$constraint_mode %||% NA_character_,
    constraint_strategy     = spec$constraint_strategy %||% NA_character_,
    recommended_resampling  = spec$recommended_resampling %||% NA_character_,
    metadata                = .depgraph_metadata_split_spec_to_json(spec$metadata),
    sample_data             = sample_rows
  )

  json <- jsonlite::toJSON(
    payload,
    auto_unbox = TRUE,
    null = "null",
    na = "null",
    pretty = isTRUE(pretty)
  )
  writeLines(json, path)
  invisible(path)
}

#' @rdname write_split_spec
#' @export
read_split_spec <- function(path) {
  .depgraph_require_jsonlite()
  .depgraph_assert(is.character(path) && length(path) == 1L && nzchar(path), "`path` must be a single non-empty file path.")
  .depgraph_assert(file.exists(path), paste0("File not found: ", path))

  parsed <- tryCatch(
    jsonlite::fromJSON(path, simplifyVector = FALSE),
    error = function(e) {
      stop("Failed to parse JSON at `", path, "`: ", conditionMessage(e), call. = FALSE)
    }
  )

  obj_type <- parsed$splitGraph_object %||% NA_character_
  if (!identical(as.character(obj_type), "split_spec")) {
    stop(
      "JSON at `", path, "` is not a serialized split_spec (found `",
      obj_type, "`). Use `read_dependency_graph()` for dependency_graph files.",
      call. = FALSE
    )
  }

  .depgraph_check_schema_version(parsed$schema_version, "split_spec")

  sample_rows <- parsed$sample_data %||% list()
  sample_data <- if (length(sample_rows) == 0L) {
    data.frame(
      sample_id      = character(),
      sample_node_id = character(),
      group_id       = character(),
      primary_group  = character(),
      batch_group    = character(),
      study_group    = character(),
      site_group     = character(),
      region_group   = character(),
      platform_group = character(),
      assay_group    = character(),
      timepoint_id   = character(),
      time_index     = numeric(),
      order_rank     = integer(),
      stringsAsFactors = FALSE
    )
  } else {
    .depgraph_chr <- function(rows, key) {
      vapply(rows, function(r) {
        v <- r[[key]]
        if (is.null(v) || (length(v) == 1L && is.na(v))) NA_character_ else as.character(v)
      }, character(1))
    }
    .depgraph_num <- function(rows, key) {
      vapply(rows, function(r) {
        v <- r[[key]]
        if (is.null(v) || (length(v) == 1L && is.na(v))) NA_real_ else as.numeric(v)
      }, numeric(1))
    }
    .depgraph_int <- function(rows, key) {
      vapply(rows, function(r) {
        v <- r[[key]]
        if (is.null(v) || (length(v) == 1L && is.na(v))) NA_integer_ else as.integer(v)
      }, integer(1))
    }

    data.frame(
      sample_id      = .depgraph_chr(sample_rows, "sample_id"),
      sample_node_id = .depgraph_chr(sample_rows, "sample_node_id"),
      group_id       = .depgraph_chr(sample_rows, "group_id"),
      primary_group  = .depgraph_chr(sample_rows, "primary_group"),
      batch_group    = .depgraph_chr(sample_rows, "batch_group"),
      study_group    = .depgraph_chr(sample_rows, "study_group"),
      site_group     = .depgraph_chr(sample_rows, "site_group"),
      region_group   = .depgraph_chr(sample_rows, "region_group"),
      platform_group = .depgraph_chr(sample_rows, "platform_group"),
      assay_group    = .depgraph_chr(sample_rows, "assay_group"),
      timepoint_id   = .depgraph_chr(sample_rows, "timepoint_id"),
      time_index     = .depgraph_num(sample_rows, "time_index"),
      order_rank     = .depgraph_int(sample_rows, "order_rank"),
      stringsAsFactors = FALSE
    )
  }

  block_vars_raw <- parsed$block_vars
  block_vars <- if (is.null(block_vars_raw)) {
    character()
  } else if (length(block_vars_raw) == 0L) {
    character()
  } else {
    as.character(unlist(block_vars_raw))
  }

  time_var_raw <- parsed$time_var
  time_var <- if (is.null(time_var_raw) || (length(time_var_raw) == 1L && is.na(time_var_raw))) NULL else as.character(time_var_raw)
  constraint_mode <- if (is.null(parsed$constraint_mode) || is.na(parsed$constraint_mode)) NULL else as.character(parsed$constraint_mode)
  constraint_strategy <- if (is.null(parsed$constraint_strategy) || is.na(parsed$constraint_strategy)) NULL else as.character(parsed$constraint_strategy)
  recommended_resampling <- if (is.null(parsed$recommended_resampling) || is.na(parsed$recommended_resampling)) NULL else as.character(parsed$recommended_resampling)

  split_spec(
    sample_data            = sample_data,
    group_var              = parsed$group_var %||% "group_id",
    block_vars             = block_vars,
    time_var               = time_var,
    ordering_required      = isTRUE(parsed$ordering_required),
    constraint_mode        = constraint_mode,
    constraint_strategy    = constraint_strategy,
    recommended_resampling = recommended_resampling,
    metadata               = .depgraph_metadata_split_spec_from_json(parsed$metadata)
  )
}

# Shared builders -------------------------------------------------------------

make_related_graph <- function(pairs, threshold = 0.1,
                               subject_ids = c("P1", "P2", "P3", "P4")) {
  meta <- data.frame(
    sample_id  = paste0("S", seq_along(subject_ids)),
    subject_id = subject_ids,
    stringsAsFactors = FALSE
  )
  samples  <- create_nodes(meta, "Sample", "sample_id")
  subjects <- create_nodes(meta, "Subject", "subject_id")
  belongs  <- create_edges(meta, "sample_id", "subject_id",
                           "Sample", "Subject", "sample_belongs_to_subject")
  rel_edges <- relatedness_edges_from_kinship(pairs, threshold = threshold)
  build_dependency_graph(list(samples, subjects), list(belongs, rel_edges))
}

make_spatial_graph <- function(coords, radius) {
  meta <- data.frame(
    sample_id  = coords$sample_id,
    subject_id = paste0("P", seq_len(nrow(coords))),
    stringsAsFactors = FALSE
  )
  samples  <- create_nodes(meta, "Sample", "sample_id")
  subjects <- create_nodes(meta, "Subject", "subject_id")
  belongs  <- create_edges(meta, "sample_id", "subject_id",
                           "Sample", "Subject", "sample_belongs_to_subject")
  adj <- spatial_edges_from_coords(coords, radius = radius)
  build_dependency_graph(list(samples, subjects), list(belongs, adj))
}

# Edge-building helpers -------------------------------------------------------

test_that("relatedness_edges_from_kinship keeps only pairs at or above threshold", {
  pairs <- data.frame(
    id1 = c("P1", "P1", "P2"),
    id2 = c("P2", "P3", "P3"),
    kinship = c(0.25, 0.02, 0.30),
    stringsAsFactors = FALSE
  )
  es <- relatedness_edges_from_kinship(pairs, threshold = 0.1)
  ed <- es$data
  expect_equal(nrow(ed), 2L)
  expect_true(all(ed$edge_type == "subject_related_to"))
  expect_setequal(ed$from, c("subject:P1", "subject:P2"))
  # kinship carried as an edge attribute
  expect_equal(sort(vapply(ed$attrs, function(a) a$kinship, numeric(1))), c(0.25, 0.30))
})

test_that("relatedness_edges_from_kinship drops self-pairs and NA metrics", {
  pairs <- data.frame(
    id1 = c("P1", "P2"),
    id2 = c("P1", "P3"),
    kinship = c(0.9, NA_real_),
    stringsAsFactors = FALSE
  )
  es <- relatedness_edges_from_kinship(pairs, threshold = 0.1)
  expect_equal(nrow(es$data), 0L)
})

test_that("spatial_edges_from_coords keeps only pairs within radius", {
  coords <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    x = c(0, 1, 9),
    y = c(0, 1, 9),
    stringsAsFactors = FALSE
  )
  es <- spatial_edges_from_coords(coords, radius = 2)
  ed <- es$data
  expect_equal(nrow(ed), 1L)
  expect_true(all(ed$edge_type == "sample_adjacent_to"))
  expect_setequal(c(ed$from, ed$to), c("sample:S1", "sample:S2"))
})

test_that("spatial_edges_from_coords errors when no coordinate columns are present", {
  coords <- data.frame(sample_id = c("S1", "S2"), stringsAsFactors = FALSE)
  expect_error(spatial_edges_from_coords(coords, radius = 1), "coordinate")
})

# Relatedness derivation ------------------------------------------------------

test_that("mode='relatedness' groups related subjects and separates the rest", {
  pairs <- data.frame(id1 = "P1", id2 = "P2", kinship = 0.25, stringsAsFactors = FALSE)
  g <- make_related_graph(pairs)
  constraint <- derive_split_constraints(g, mode = "relatedness")
  expect_s3_class(constraint, "split_constraint")

  gv <- grouping_vector(constraint)
  expect_identical(unname(gv[["S1"]]), unname(gv[["S2"]]))   # related
  expect_false(identical(gv[["S1"]], gv[["S3"]]))            # unrelated
  expect_false(identical(gv[["S3"]], gv[["S4"]]))
  expect_equal(length(unique(gv)), 3L)
  expect_identical(constraint$metadata$mode, "relatedness")
  expect_identical(constraint$metadata$relations_used, "subject_related_to")
})

test_that("relatedness grouping is transitive (P1-P2, P2-P3 => one group)", {
  pairs <- data.frame(
    id1 = c("P1", "P2"),
    id2 = c("P2", "P3"),
    kinship = c(0.25, 0.25),
    stringsAsFactors = FALSE
  )
  g <- make_related_graph(pairs)
  gv <- grouping_vector(derive_split_constraints(g, mode = "relatedness"))
  expect_identical(unname(gv[["S1"]]), unname(gv[["S2"]]))
  expect_identical(unname(gv[["S2"]]), unname(gv[["S3"]]))
  expect_false(identical(gv[["S1"]], gv[["S4"]]))
})

test_that("samples sharing a subject are always grouped together", {
  meta <- data.frame(
    sample_id  = c("S1", "S2", "S3"),
    subject_id = c("P1", "P1", "P2"),
    stringsAsFactors = FALSE
  )
  samples  <- create_nodes(meta, "Sample", "sample_id")
  subjects <- create_nodes(meta, "Subject", "subject_id")
  belongs  <- create_edges(meta, "sample_id", "subject_id",
                           "Sample", "Subject", "sample_belongs_to_subject")
  # No relatedness edges at all.
  pairs <- data.frame(id1 = character(), id2 = character(), kinship = numeric())
  rel_edges <- relatedness_edges_from_kinship(pairs, threshold = 0.1)
  g <- build_dependency_graph(list(samples, subjects), list(belongs, rel_edges))

  gv <- grouping_vector(derive_split_constraints(g, mode = "relatedness"))
  expect_identical(unname(gv[["S1"]]), unname(gv[["S2"]]))  # same subject
  expect_false(identical(gv[["S1"]], gv[["S3"]]))
})

# Spatial derivation ----------------------------------------------------------

test_that("mode='spatial' groups adjacent samples via connected components", {
  coords <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    x = c(0, 1, 20),
    y = c(0, 0, 0),
    stringsAsFactors = FALSE
  )
  g <- make_spatial_graph(coords, radius = 2)
  gv <- grouping_vector(derive_split_constraints(g, mode = "spatial"))
  expect_identical(unname(gv[["S1"]]), unname(gv[["S2"]]))   # 1 unit apart
  expect_false(identical(gv[["S1"]], gv[["S3"]]))            # far away
  expect_equal(length(unique(gv)), 2L)
})

test_that("spatial grouping respects the samples= subset (bridge excluded)", {
  # Chain S1 - S2 - S3 by adjacency; excluding the S2 bridge must split S1/S3.
  coords <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    x = c(0, 1, 2),
    y = c(0, 0, 0),
    stringsAsFactors = FALSE
  )
  g <- make_spatial_graph(coords, radius = 1.5)

  gv_all <- grouping_vector(derive_split_constraints(g, mode = "spatial"))
  expect_equal(length(unique(gv_all)), 1L)  # all chained together

  gv_sub <- grouping_vector(
    derive_split_constraints(g, mode = "spatial", samples = c("S1", "S3"))
  )
  expect_equal(length(gv_sub), 2L)
  expect_false(identical(gv_sub[["S1"]], gv_sub[["S3"]]))
})

# split_spec + serialization --------------------------------------------------

test_that("as_split_spec carries the component grouping for pairwise modes", {
  coords <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    x = c(0, 1, 20), y = c(0, 0, 0),
    stringsAsFactors = FALSE
  )
  g <- make_spatial_graph(coords, radius = 2)
  spec <- as_split_spec(derive_split_constraints(g, mode = "spatial"), graph = g)
  expect_s3_class(spec, "split_spec")
  expect_identical(spec$group_var, "group_id")
  expect_identical(spec$recommended_resampling, "grouped_cv")
  expect_true(validate_split_spec(spec)$valid)
})

test_that("a spatial split_spec survives a JSON round-trip", {
  testthat::skip_if_not_installed("jsonlite")
  coords <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    x = c(0, 1, 20), y = c(0, 0, 0),
    stringsAsFactors = FALSE
  )
  g <- make_spatial_graph(coords, radius = 2)
  spec <- as_split_spec(derive_split_constraints(g, mode = "spatial"), graph = g)

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_split_spec(spec, tmp)
  spec2 <- read_split_spec(tmp)
  expect_identical(spec2$sample_data$group_id, spec$sample_data$group_id)
  expect_identical(names(spec2$sample_data), names(spec$sample_data))
})

test_that("plot renders a graph carrying pairwise edges", {
  pairs <- data.frame(id1 = "P1", id2 = "P2", kinship = 0.25, stringsAsFactors = FALSE)
  g <- make_related_graph(pairs)
  expect_silent(plot(g))
})

make_region_graph <- function() {
  meta <- data.frame(
    sample_id  = c("S1", "S2", "S3", "S4"),
    subject_id = c("P1", "P1", "P2", "P2"),
    region_id  = c("cortex", "cortex", "hippocampus", "hippocampus"),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "region-demo")
}

test_that("graph_from_metadata auto-detects region_id and builds Region nodes and edges", {
  g <- make_region_graph()
  nd <- g$nodes$data
  expect_equal(sum(nd$node_type == "Region"), 2L)
  expect_setequal(nd$node_key[nd$node_type == "Region"], c("cortex", "hippocampus"))
  expect_equal(sum(g$edges$data$edge_type == "sample_located_in_region"), 4L)
})

test_that("query_node_type returns Region nodes", {
  g <- make_region_graph()
  res <- as.data.frame(query_node_type(g, "Region"))
  expect_equal(nrow(res), 2L)
})

test_that("graph_from_metadata builds no Region nodes when region_id is absent", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  expect_s3_class(g, "dependency_graph")
  expect_equal(sum(g$nodes$data$node_type == "Region"), 0L)
})

test_that("derive_split_constraints(mode='region') groups samples by region", {
  g <- make_region_graph()
  constraint <- derive_split_constraints(g, mode = "region")
  expect_s3_class(constraint, "split_constraint")
  gv <- grouping_vector(constraint)
  expect_identical(unname(gv[c("S1", "S2")]), c("region:cortex", "region:cortex"))
  expect_identical(unname(gv[c("S3", "S4")]), c("region:hippocampus", "region:hippocampus"))
  expect_equal(length(unique(gv)), 2L)
  expect_identical(constraint$metadata$mode, "region")
  expect_identical(constraint$metadata$relations_used, "sample_located_in_region")
})

test_that("as_split_spec carries region_group and registers it as a block variable", {
  g <- make_region_graph()
  constraint <- derive_split_constraints(g, mode = "region")
  spec <- as_split_spec(constraint, graph = g)
  expect_true("region_group" %in% names(spec$sample_data))
  expect_false(all(is.na(spec$sample_data$region_group)))
  expect_true("region_group" %in% spec$block_vars)
  expect_silent(validate_split_spec(spec))
})

test_that("region assignments enrich a subject-mode split_spec as a blocking annotation", {
  g <- make_region_graph()
  subject_constraint <- derive_split_constraints(g, mode = "subject")
  spec <- as_split_spec(subject_constraint, graph = g)
  expect_false(all(is.na(spec$sample_data$region_group)))
  expect_true("region_group" %in% spec$block_vars)
})

test_that("composite strict grouping can traverse Region dependencies", {
  g <- make_region_graph()
  constraint <- derive_split_constraints(
    g, mode = "composite", strategy = "strict", via = c("Subject", "Region")
  )
  expect_s3_class(constraint, "split_constraint")
  expect_equal(length(unique(grouping_vector(constraint))), 2L)
})

test_that("semantic validation flags multi-region samples as an error", {
  samples <- create_nodes(data.frame(sample_id = "S1"), "Sample", "sample_id")
  regions <- create_nodes(data.frame(region_id = c("cortex", "hippocampus")), "Region", "region_id")
  region_edges <- create_edges(
    data.frame(sample_id = "S1", region_id = c("cortex", "hippocampus")),
    "sample_id", "region_id", "Sample", "Region", "sample_located_in_region"
  )
  g_multi <- build_dependency_graph(list(samples, regions), list(region_edges), validate = FALSE)
  report <- validate_graph(g_multi)
  expect_false(report$valid)
  expect_true(any(grepl("sample_located_in_region", report$errors)))
})

test_that("derive_split_constraints(mode='region') rejects ambiguous region assignments", {
  samples <- create_nodes(data.frame(sample_id = "S1"), "Sample", "sample_id")
  regions <- create_nodes(data.frame(region_id = c("cortex", "hippocampus")), "Region", "region_id")
  region_edges <- create_edges(
    data.frame(sample_id = "S1", region_id = c("cortex", "hippocampus")),
    "sample_id", "region_id", "Sample", "Region", "sample_located_in_region"
  )
  g_multi <- build_dependency_graph(list(samples, regions), list(region_edges), validate = FALSE)
  expect_error(
    derive_split_constraints(g_multi, mode = "region"),
    "Multiple region assignments found",
    fixed = TRUE
  )
})

test_that("plot renders a graph containing Region nodes", {
  g <- make_region_graph()
  expect_silent(plot(g))
})

test_that("region_group survives a split_spec JSON round-trip", {
  testthat::skip_if_not_installed("jsonlite")
  g <- make_region_graph()
  constraint <- derive_split_constraints(g, mode = "region")
  spec <- as_split_spec(constraint, graph = g)

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_split_spec(spec, tmp)
  spec2 <- read_split_spec(tmp)

  expect_true("region_group" %in% names(spec2$sample_data))
  expect_identical(spec2$sample_data$region_group, spec$sample_data$region_group)
  expect_identical(names(spec2$sample_data), names(spec$sample_data))
})

skip_if_no_jsonlite <- function() {
  testthat::skip_if_not_installed("jsonlite")
}

make_round_trip_graph <- function() {
  meta <- data.frame(
    sample_id    = c("S1", "S2", "S3"),
    subject_id   = c("P1", "P1", "P2"),
    batch_id     = c("B1", "B2", "B1"),
    timepoint_id = c("T0", "T1", "T0"),
    time_index   = c(0, 1, 0),
    outcome_id   = c("ctrl", "case", "ctrl"),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "rt_demo", dataset_name = "rt_dataset")
}

# ---- dependency_graph round-trip --------------------------------------------

test_that("write_dependency_graph + read_dependency_graph round-trip preserves structure", {
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)

  written <- write_dependency_graph(g, tmp)
  expect_identical(written, tmp)
  expect_true(file.exists(tmp))

  g2 <- read_dependency_graph(tmp)
  expect_s3_class(g2, "dependency_graph")

  # Same nodes (sorted, comparing structural columns).
  n1 <- g$nodes$data[order(g$nodes$data$node_id), c("node_id", "node_type", "node_key", "label")]
  n2 <- g2$nodes$data[order(g2$nodes$data$node_id), c("node_id", "node_type", "node_key", "label")]
  row.names(n1) <- NULL; row.names(n2) <- NULL
  expect_identical(n1, n2)

  # Same edges (sorted).
  e1 <- g$edges$data[order(g$edges$data$edge_id), c("edge_id", "from", "to", "edge_type")]
  e2 <- g2$edges$data[order(g2$edges$data$edge_id), c("edge_id", "from", "to", "edge_type")]
  row.names(e1) <- NULL; row.names(e2) <- NULL
  expect_identical(e1, e2)

  # Validation status preserved (the rebuilt graph must validate).
  expect_true(validate_graph(g2)$valid)
})

test_that("round-trip preserves graph metadata labels", {
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)

  write_dependency_graph(g, tmp)
  g2 <- read_dependency_graph(tmp)

  expect_identical(g2$metadata$graph_name, "rt_demo")
  expect_identical(g2$metadata$dataset_name, "rt_dataset")
  expect_identical(g2$metadata$schema_version, splitGraph:::.depgraph_schema_version)
})

test_that("round-trip preserves node and edge attribute list-columns", {
  skip_if_no_jsonlite()
  # Build a graph with non-trivial node attributes via the explicit path.
  meta <- data.frame(sample_id = c("S1", "S2"),
                     subject_id = c("P1", "P2"),
                     stringsAsFactors = FALSE)
  samples <- create_nodes(meta, "Sample", "sample_id")
  subjects <- create_nodes(meta, "Subject", "subject_id")
  time_nodes <- create_nodes(
    data.frame(timepoint_id = c("T0", "T1"),
               time_index = c(0L, 1L),
               visit_label = c("baseline", "follow_up"),
               stringsAsFactors = FALSE),
    type = "Timepoint", id_col = "timepoint_id",
    attr_cols = c("time_index", "visit_label")
  )
  edges <- create_edges(meta, "sample_id", "subject_id",
                        "Sample", "Subject", "sample_belongs_to_subject")
  g <- build_dependency_graph(list(samples, subjects, time_nodes), list(edges))

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)
  g2 <- read_dependency_graph(tmp)

  t0_idx <- which(g2$nodes$data$node_id == "timepoint:T0")
  expect_length(t0_idx, 1L)
  attrs <- g2$nodes$data$attrs[[t0_idx]]
  expect_true(is.list(attrs))
  expect_equal(unname(attrs$time_index), 0)
  expect_equal(unname(attrs$visit_label), "baseline")

  # Empty-attrs nodes round-trip as list(). Subject nodes here have no
  # optional schema attributes populated from `meta`, so they're a clean test.
  p1_idx <- which(g2$nodes$data$node_id == "subject:P1")
  expect_length(p1_idx, 1L)
  expect_identical(g2$nodes$data$attrs[[p1_idx]], list())
})

test_that("round-trip preserves validation_overrides on the graph", {
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  g$metadata$validation_overrides <- list(allow_multi_subject_samples = TRUE)
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)
  g2 <- read_dependency_graph(tmp)
  expect_true(isTRUE(g2$metadata$validation_overrides$allow_multi_subject_samples))
})

test_that("read_dependency_graph returns NA for unparseable or missing created_at (no silent Sys.time fallback)", {
  # Regression test for C6: .depgraph_iso_to_posix used to fall back to
  # Sys.time() on parse failure and on NULL/NA input, silently turning a
  # corrupted or missing timestamp into "now". That destroys provenance.
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)

  payload <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)

  # Case 1: malformed timestamp must not become "now".
  bad <- payload
  bad$metadata$created_at <- "not-a-real-date"
  bad_path <- tempfile(fileext = ".json")
  on.exit(unlink(bad_path), add = TRUE)
  jsonlite::write_json(bad, bad_path, auto_unbox = TRUE, null = "null", na = "null")

  g_bad <- read_dependency_graph(bad_path)
  expect_true(is.na(g_bad$metadata$created_at),
              info = "Malformed created_at must read back as NA, not Sys.time().")
  expect_s3_class(g_bad$metadata$created_at, "POSIXct")

  # Case 2: missing/null timestamp must also be NA, not Sys.time().
  missing <- payload
  missing$metadata$created_at <- NULL
  missing_path <- tempfile(fileext = ".json")
  on.exit(unlink(missing_path), add = TRUE)
  jsonlite::write_json(missing, missing_path, auto_unbox = TRUE, null = "null", na = "null")

  g_missing <- read_dependency_graph(missing_path)
  expect_true(is.na(g_missing$metadata$created_at),
              info = "Missing created_at must read back as NA, not Sys.time().")
  expect_s3_class(g_missing$metadata$created_at, "POSIXct")
})

test_that("read_dependency_graph rejects malformed or wrong-type JSON", {
  skip_if_no_jsonlite()
  bad <- tempfile(fileext = ".json")
  on.exit(unlink(bad), add = TRUE)

  writeLines("not json at all {", bad)
  expect_error(read_dependency_graph(bad), "JSON|parse", ignore.case = TRUE)

  # Right shape but wrong object type.
  spec_like <- list(
    splitGraph_object = "split_spec",
    schema_version = "0.1.0",
    sample_data = list()
  )
  writeLines(jsonlite::toJSON(spec_like, auto_unbox = TRUE, null = "null"), bad)
  expect_error(read_dependency_graph(bad), "dependency_graph", ignore.case = TRUE)
})

test_that("read_dependency_graph warns on unknown future schema_version but still loads", {
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)

  raw <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)
  raw$schema_version <- "99.0.0"
  writeLines(jsonlite::toJSON(raw, auto_unbox = TRUE, null = "null"), tmp)

  expect_warning(g2 <- read_dependency_graph(tmp), "schema", ignore.case = TRUE)
  expect_s3_class(g2, "dependency_graph")
})

# ---- split_spec round-trip --------------------------------------------------

test_that("write_split_spec + read_split_spec round-trip preserves structure", {
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  constraint <- derive_split_constraints(g, mode = "subject")
  spec <- as_split_spec(constraint, graph = g)

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  written <- write_split_spec(spec, tmp)
  expect_identical(written, tmp)
  expect_true(file.exists(tmp))

  spec2 <- read_split_spec(tmp)
  expect_s3_class(spec2, "split_spec")

  # Sample-data table identical (column order / NA preservation).
  d1 <- spec$sample_data
  d2 <- spec2$sample_data
  expect_identical(names(d1), names(d2))
  expect_identical(nrow(d1), nrow(d2))
  expect_identical(d1$sample_id, d2$sample_id)
  expect_identical(d1$group_id, d2$group_id)
  expect_identical(d1$batch_group, d2$batch_group)
  expect_identical(d1$study_group, d2$study_group)
  expect_identical(d1$timepoint_id, d2$timepoint_id)
  expect_identical(d1$time_index, d2$time_index)
  expect_identical(d1$order_rank, d2$order_rank)

  # Top-level scalars.
  expect_identical(spec2$group_var, spec$group_var)
  expect_identical(spec2$block_vars, spec$block_vars)
  expect_identical(spec2$time_var, spec$time_var)
  expect_identical(spec2$ordering_required, spec$ordering_required)
  expect_identical(spec2$constraint_mode, spec$constraint_mode)
  expect_identical(spec2$recommended_resampling, spec$recommended_resampling)

  # Validation passes after round-trip.
  expect_true(validate_split_spec(spec2)$valid)
})

test_that("read_split_spec round-trip preserves NA in sample_data", {
  skip_if_no_jsonlite()
  # A graph where some samples have no batch / timepoint -> NA after as_split_spec.
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  constraint <- derive_split_constraints(g, mode = "subject")
  spec <- as_split_spec(constraint, graph = g)

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_split_spec(spec, tmp)
  spec2 <- read_split_spec(tmp)

  # batch_group / order_rank should still be all-NA.
  expect_true(all(is.na(spec2$sample_data$batch_group)))
  expect_true(all(is.na(spec2$sample_data$order_rank)))
})

test_that("read_split_spec rejects malformed or wrong-type JSON", {
  skip_if_no_jsonlite()
  bad <- tempfile(fileext = ".json")
  on.exit(unlink(bad), add = TRUE)

  writeLines("[1,2,3", bad)
  expect_error(read_split_spec(bad), "JSON|parse", ignore.case = TRUE)

  graph_like <- list(
    splitGraph_object = "dependency_graph",
    schema_version = "0.1.0",
    nodes = list(), edges = list()
  )
  writeLines(jsonlite::toJSON(graph_like, auto_unbox = TRUE, null = "null"), bad)
  expect_error(read_split_spec(bad), "split_spec", ignore.case = TRUE)
})

test_that("write_* errors when target directory does not exist", {
  skip_if_no_jsonlite()
  g <- make_round_trip_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
  bad_path <- file.path(tempdir(), "splitGraph_no_such_dir_xyz", "x.json")

  expect_error(write_dependency_graph(g, bad_path), "directory", ignore.case = TRUE)
  expect_error(write_split_spec(spec, bad_path), "directory", ignore.case = TRUE)
})

test_that("write_* / read_* error helpfully if jsonlite is missing", {
  # Simulate missing jsonlite without actually unloading it: we test the
  # internal guard directly.
  expect_silent(splitGraph:::.depgraph_require_jsonlite())
})

make_site_graph <- function() {
  meta <- data.frame(
    sample_id  = c("S1", "S2", "S3", "S4"),
    subject_id = c("P1", "P1", "P2", "P2"),
    site_id    = c("NYC", "NYC", "BOS", "BOS"),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "site-demo")
}

test_that("graph_from_metadata auto-detects site_id and builds Site nodes and edges", {
  g <- make_site_graph()
  nd <- g$nodes$data
  expect_equal(sum(nd$node_type == "Site"), 2L)
  expect_setequal(nd$node_key[nd$node_type == "Site"], c("NYC", "BOS"))
  expect_equal(sum(g$edges$data$edge_type == "sample_collected_at_site"), 4L)
})

test_that("query_node_type returns Site nodes", {
  g <- make_site_graph()
  res <- as.data.frame(query_node_type(g, "Site"))
  expect_equal(nrow(res), 2L)
})

test_that("graph_from_metadata builds no Site nodes when site_id is absent", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  expect_s3_class(g, "dependency_graph")
  expect_equal(sum(g$nodes$data$node_type == "Site"), 0L)
})

test_that("derive_split_constraints(mode='site') groups samples by site", {
  g <- make_site_graph()
  constraint <- derive_split_constraints(g, mode = "site")
  expect_s3_class(constraint, "split_constraint")
  gv <- grouping_vector(constraint)
  expect_identical(unname(gv[c("S1", "S2")]), c("site:NYC", "site:NYC"))
  expect_identical(unname(gv[c("S3", "S4")]), c("site:BOS", "site:BOS"))
  expect_equal(length(unique(gv)), 2L)
  expect_identical(constraint$metadata$mode, "site")
  expect_identical(constraint$metadata$relations_used, "sample_collected_at_site")
})

test_that("as_split_spec carries site_group and registers it as a block variable", {
  g <- make_site_graph()
  constraint <- derive_split_constraints(g, mode = "site")
  spec <- as_split_spec(constraint, graph = g)
  expect_true("site_group" %in% names(spec$sample_data))
  expect_false(all(is.na(spec$sample_data$site_group)))
  expect_true("site_group" %in% spec$block_vars)
  expect_silent(validate_split_spec(spec))
})

test_that("site assignments enrich a subject-mode split_spec as a blocking annotation", {
  g <- make_site_graph()
  subject_constraint <- derive_split_constraints(g, mode = "subject")
  spec <- as_split_spec(subject_constraint, graph = g)
  # site_group is carried even though the primary grouping is subject
  expect_false(all(is.na(spec$sample_data$site_group)))
  expect_true("site_group" %in% spec$block_vars)
})

test_that("composite strict grouping can traverse Site dependencies", {
  g <- make_site_graph()
  constraint <- derive_split_constraints(
    g, mode = "composite", strategy = "strict", via = c("Subject", "Site")
  )
  expect_s3_class(constraint, "split_constraint")
  expect_equal(length(unique(grouping_vector(constraint))), 2L)
})

test_that("semantic validation flags multi-site samples as an error", {
  samples <- create_nodes(data.frame(sample_id = "S1"), "Sample", "sample_id")
  sites   <- create_nodes(data.frame(site_id = c("NYC", "BOS")), "Site", "site_id")
  site_edges <- create_edges(
    data.frame(sample_id = "S1", site_id = c("NYC", "BOS")),
    "sample_id", "site_id", "Sample", "Site", "sample_collected_at_site"
  )
  g_multi <- build_dependency_graph(list(samples, sites), list(site_edges), validate = FALSE)
  report <- validate_graph(g_multi)
  expect_false(report$valid)
  expect_true(any(grepl("sample_collected_at_site", report$errors)))
})

test_that("derive_split_constraints(mode='site') rejects ambiguous site assignments", {
  samples <- create_nodes(data.frame(sample_id = "S1"), "Sample", "sample_id")
  sites   <- create_nodes(data.frame(site_id = c("NYC", "BOS")), "Site", "site_id")
  site_edges <- create_edges(
    data.frame(sample_id = "S1", site_id = c("NYC", "BOS")),
    "sample_id", "site_id", "Sample", "Site", "sample_collected_at_site"
  )
  g_multi <- build_dependency_graph(list(samples, sites), list(site_edges), validate = FALSE)
  expect_error(
    derive_split_constraints(g_multi, mode = "site"),
    "Multiple site assignments found",
    fixed = TRUE
  )
})

test_that("plot renders a graph containing Site nodes", {
  g <- make_site_graph()
  expect_silent(plot(g))
})

test_that("site_group survives a split_spec JSON round-trip", {
  testthat::skip_if_not_installed("jsonlite")
  g <- make_site_graph()
  constraint <- derive_split_constraints(g, mode = "site")
  spec <- as_split_spec(constraint, graph = g)

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_split_spec(spec, tmp)
  spec2 <- read_split_spec(tmp)

  expect_true("site_group" %in% names(spec2$sample_data))
  expect_identical(spec2$sample_data$site_group, spec$sample_data$site_group)
  expect_identical(names(spec2$sample_data), names(spec$sample_data))
})

make_platform_graph <- function() {
  meta <- data.frame(
    sample_id   = c("S1", "S2", "S3", "S4"),
    subject_id  = c("P1", "P1", "P2", "P2"),
    platform_id = c("illumina", "illumina", "nanopore", "nanopore"),
    assay_id    = c("rnaseq", "rnaseq", "wgs", "wgs"),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "platform-demo")
}

test_that("graph_from_metadata auto-detects platform_id and builds Platform nodes and edges", {
  g <- make_platform_graph()
  nd <- g$nodes$data
  expect_equal(sum(nd$node_type == "Platform"), 2L)
  expect_setequal(nd$node_key[nd$node_type == "Platform"], c("illumina", "nanopore"))
  expect_equal(sum(g$edges$data$edge_type == "sample_run_on_platform"), 4L)
})

test_that("query_node_type returns Platform nodes", {
  g <- make_platform_graph()
  res <- as.data.frame(query_node_type(g, "Platform"))
  expect_equal(nrow(res), 2L)
})

test_that("graph_from_metadata builds no Platform nodes when platform_id is absent", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  expect_s3_class(g, "dependency_graph")
  expect_equal(sum(g$nodes$data$node_type == "Platform"), 0L)
})

test_that("derive_split_constraints(mode='platform') groups samples by platform", {
  g <- make_platform_graph()
  constraint <- derive_split_constraints(g, mode = "platform")
  expect_s3_class(constraint, "split_constraint")
  gv <- grouping_vector(constraint)
  expect_identical(unname(gv[c("S1", "S2")]), c("platform:illumina", "platform:illumina"))
  expect_identical(unname(gv[c("S3", "S4")]), c("platform:nanopore", "platform:nanopore"))
  expect_equal(length(unique(gv)), 2L)
  expect_identical(constraint$metadata$mode, "platform")
  expect_identical(constraint$metadata$relations_used, "sample_run_on_platform")
})

test_that("derive_split_constraints(mode='assay') groups samples by assay", {
  g <- make_platform_graph()
  constraint <- derive_split_constraints(g, mode = "assay")
  expect_s3_class(constraint, "split_constraint")
  gv <- grouping_vector(constraint)
  expect_identical(unname(gv[c("S1", "S2")]), c("assay:rnaseq", "assay:rnaseq"))
  expect_identical(unname(gv[c("S3", "S4")]), c("assay:wgs", "assay:wgs"))
  expect_equal(length(unique(gv)), 2L)
  expect_identical(constraint$metadata$mode, "assay")
  expect_identical(constraint$metadata$relations_used, "sample_measured_by_assay")
})

test_that("as_split_spec carries platform_group / assay_group as block variables", {
  g <- make_platform_graph()

  spec_p <- as_split_spec(derive_split_constraints(g, mode = "platform"), graph = g)
  expect_true("platform_group" %in% names(spec_p$sample_data))
  expect_false(all(is.na(spec_p$sample_data$platform_group)))
  expect_true("platform_group" %in% spec_p$block_vars)

  spec_a <- as_split_spec(derive_split_constraints(g, mode = "assay"), graph = g)
  expect_true("assay_group" %in% names(spec_a$sample_data))
  expect_false(all(is.na(spec_a$sample_data$assay_group)))
  expect_true("assay_group" %in% spec_a$block_vars)
})

test_that("platform / assay assignments enrich a subject-mode split_spec", {
  g <- make_platform_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
  expect_false(all(is.na(spec$sample_data$platform_group)))
  expect_false(all(is.na(spec$sample_data$assay_group)))
  expect_true(all(c("platform_group", "assay_group") %in% spec$block_vars))
})

test_that("composite strict grouping can traverse Platform dependencies", {
  g <- make_platform_graph()
  constraint <- derive_split_constraints(
    g, mode = "composite", strategy = "strict", via = c("Subject", "Platform")
  )
  expect_s3_class(constraint, "split_constraint")
  expect_equal(length(unique(grouping_vector(constraint))), 2L)
})

test_that("semantic validation flags multi-platform samples as an error", {
  samples   <- create_nodes(data.frame(sample_id = "S1"), "Sample", "sample_id")
  platforms <- create_nodes(data.frame(platform_id = c("illumina", "nanopore")), "Platform", "platform_id")
  edges <- create_edges(
    data.frame(sample_id = "S1", platform_id = c("illumina", "nanopore")),
    "sample_id", "platform_id", "Sample", "Platform", "sample_run_on_platform"
  )
  g_multi <- build_dependency_graph(list(samples, platforms), list(edges), validate = FALSE)
  report <- validate_graph(g_multi)
  expect_false(report$valid)
  expect_true(any(grepl("sample_run_on_platform", report$errors)))
})

test_that("assay_uses_platform is a valid manually-constructed edge", {
  samples   <- create_nodes(data.frame(sample_id = "S1"), "Sample", "sample_id")
  assays    <- create_nodes(data.frame(assay_id = "rnaseq"), "Assay", "assay_id")
  platforms <- create_nodes(data.frame(platform_id = "illumina"), "Platform", "platform_id")
  assay_edges <- create_edges(
    data.frame(sample_id = "S1", assay_id = "rnaseq"),
    "sample_id", "assay_id", "Sample", "Assay", "sample_measured_by_assay"
  )
  platform_edges <- create_edges(
    data.frame(assay_id = "rnaseq", platform_id = "illumina"),
    "assay_id", "platform_id", "Assay", "Platform", "assay_uses_platform"
  )
  g <- build_dependency_graph(
    list(samples, assays, platforms), list(assay_edges, platform_edges), validate = TRUE
  )
  expect_s3_class(g, "dependency_graph")
  expect_equal(sum(g$edges$data$edge_type == "assay_uses_platform"), 1L)
})

test_that("platform_group and assay_group survive a split_spec JSON round-trip", {
  testthat::skip_if_not_installed("jsonlite")
  g <- make_platform_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "platform"), graph = g)

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_split_spec(spec, tmp)
  spec2 <- read_split_spec(tmp)

  expect_true(all(c("platform_group", "assay_group") %in% names(spec2$sample_data)))
  expect_identical(spec2$sample_data$platform_group, spec$sample_data$platform_group)
  expect_identical(spec2$sample_data$assay_group, spec$sample_data$assay_group)
  expect_identical(names(spec2$sample_data), names(spec$sample_data))
})

test_that("plot renders a graph containing Platform nodes", {
  g <- make_platform_graph()
  expect_silent(plot(g))
})

# Constructors for core splitGraph S3 classes.

#' Construct Core splitGraph S3 Objects
#'
#' Low-level constructors for the core S3 classes used throughout
#' \pkg{splitGraph}.
#'
#' @param data A data frame matching the canonical schema for nodes or edges.
#' @param schema_version Schema version string stored on the object.
#' @param source Optional source metadata.
#' @param nodes,edges A \code{graph_node_set} and \code{graph_edge_set}.
#' @param graph An internal \code{igraph} object.
#' @param metadata,caches,params,recommended_downstream_args Named lists with
#'   auxiliary metadata.
#' @param query Query label stored on a \code{graph_query_result}.
#' @param table Tabular query result payload.
#' @param constraint_id,relation_types,transitive Fields describing a
#'   dependency constraint.
#' @param sample_map Sample-level mapping table for constraints.
#' @param strategy Split strategy identifier.
#' @return An S3 object corresponding to the constructor that was called.
#' @examples
#' meta <- data.frame(
#'   sample_id = c("S1", "S2"),
#'   subject_id = c("P1", "P2")
#' )
#'
#' samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
#' subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
#' edges <- create_edges(
#'   meta,
#'   from_col = "sample_id",
#'   to_col = "subject_id",
#'   from_type = "Sample",
#'   to_type = "Subject",
#'   relation = "sample_belongs_to_subject"
#' )
#'
#' nodes_set <- graph_node_set(rbind(samples$data, subjects$data))
#' edges_set <- graph_edge_set(edges$data)
#' nodes_set
#' edges_set
#' @export
graph_node_set <- function(data = NULL, schema_version = .depgraph_schema_version, source = list()) {
  data <- .depgraph_normalize_node_data(data)
  structure(
    list(
      data = data,
      schema_version = as.character(schema_version),
      source = source
    ),
    class = "graph_node_set"
  )
}

#' @rdname graph_node_set
#' @export
graph_edge_set <- function(data = NULL, schema_version = .depgraph_schema_version, source = list()) {
  data <- .depgraph_normalize_edge_data(data)
  structure(
    list(
      data = data,
      schema_version = as.character(schema_version),
      source = source
    ),
    class = "graph_edge_set"
  )
}

#' @rdname graph_node_set
#' @export
dependency_graph <- function(nodes, edges, graph, metadata = list(), caches = list()) {
  .depgraph_assert(inherits(nodes, "graph_node_set"), "`nodes` must be a `graph_node_set`.")
  .depgraph_assert(inherits(edges, "graph_edge_set"), "`edges` must be a `graph_edge_set`.")
  if (missing(graph) || is.null(graph)) {
    graph <- .depgraph_build_igraph(nodes$data, edges$data)
  }
  .depgraph_assert(inherits(graph, "igraph"), "`graph` must be an `igraph` object.")
  alignment <- .depgraph_graph_table_alignment(graph, nodes$data, edges$data)
  .depgraph_assert(
    isTRUE(alignment$valid),
    "The embedded `igraph` does not match the supplied node and edge tables."
  )

  if (length(caches) == 0L) {
    caches <- .depgraph_build_caches(nodes$data, edges$data)
  }

  structure(
    list(
      nodes = nodes,
      edges = edges,
      graph = graph,
      metadata = utils::modifyList(
        list(
          graph_name = NULL,
          dataset_name = NULL,
          created_at = Sys.time(),
          schema_version = .depgraph_schema_version
        ),
        metadata
      ),
      caches = caches
    ),
    class = "dependency_graph"
  )
}

#' @rdname graph_node_set
#' @export
new_depgraph_nodes <- function(data = NULL, schema_version = .depgraph_schema_version, source = list()) {
  .Deprecated(
    new = "graph_node_set",
    package = "splitGraph",
    msg = "`new_depgraph_nodes()` is deprecated. Use `graph_node_set()` instead."
  )
  graph_node_set(data = data, schema_version = schema_version, source = source)
}

#' @rdname graph_node_set
#' @export
new_depgraph_edges <- function(data = NULL, schema_version = .depgraph_schema_version, source = list()) {
  .Deprecated(
    new = "graph_edge_set",
    package = "splitGraph",
    msg = "`new_depgraph_edges()` is deprecated. Use `graph_edge_set()` instead."
  )
  graph_edge_set(data = data, schema_version = schema_version, source = source)
}

#' @rdname graph_node_set
#' @export
new_depgraph <- function(nodes, edges, graph = NULL, metadata = list(), caches = list()) {
  .Deprecated(
    new = "dependency_graph",
    package = "splitGraph",
    msg = "`new_depgraph()` is deprecated. Use `dependency_graph()` instead."
  )
  dependency_graph(nodes = nodes, edges = edges, graph = graph, metadata = metadata, caches = caches)
}

#' @rdname graph_node_set
#' @export
graph_query_result <- function(query = "", params = list(), nodes = NULL, edges = NULL, table = NULL, metadata = list()) {
  structure(
    list(
      query = as.character(query)[1L],
      params = params,
      nodes = if (is.null(nodes)) data.frame(stringsAsFactors = FALSE) else nodes,
      edges = if (is.null(edges)) data.frame(stringsAsFactors = FALSE) else edges,
      table = if (is.null(table)) data.frame(stringsAsFactors = FALSE) else table,
      metadata = metadata
    ),
    class = "graph_query_result"
  )
}

#' @rdname graph_node_set
#' @export
dependency_constraint <- function(constraint_id, relation_types, sample_map, transitive = TRUE, metadata = list()) {
  .depgraph_assert(is.data.frame(sample_map), "`sample_map` must be a data.frame.")
  structure(
    list(
      constraint_id = as.character(constraint_id)[1L],
      relation_types = as.character(relation_types),
      sample_map = sample_map,
      transitive = isTRUE(transitive),
      metadata = metadata
    ),
    class = "dependency_constraint"
  )
}

#' @rdname graph_node_set
#' @export
split_constraint <- function(strategy, sample_map, recommended_downstream_args = list(), metadata = list()) {
  .depgraph_assert(is.data.frame(sample_map), "`sample_map` must be a data.frame.")
  structure(
    list(
      strategy = as.character(strategy)[1L],
      sample_map = sample_map,
      recommended_downstream_args = recommended_downstream_args,
      metadata = metadata
    ),
    class = "split_constraint"
  )
}

#' Validation Report Object for splitGraph Graphs
#'
#' \code{depgraph_validation_report} is the structured return type produced by
#' \code{validate_graph()} and \code{validate_depgraph()}.
#'
#' The report contains:
#'
#' \itemize{
#'   \item \code{graph_name}: graph label when available
#'   \item \code{valid}: whether any \code{error}-severity issues were found
#'   \item \code{issues}: canonical issue table
#'   \item \code{summary}: counts by level, severity, and code
#'   \item \code{metadata}: report metadata
#'   \item \code{errors}, \code{warnings}, \code{advisories}: backward-compatible
#'     message vectors
#'   \item \code{metrics}: graph and issue counts
#' }
#'
#' The canonical issue table includes the columns:
#' \code{issue_id}, \code{level}, \code{severity}, \code{code}, \code{message},
#' \code{node_ids}, \code{edge_ids}, and \code{details}.
#'
#' @param graph_name Graph label stored on the report.
#' @param issues Canonical issue table. When \code{NULL}, an empty skeleton is
#'   constructed.
#' @param metrics Named list of graph- and issue-level counts.
#' @param metadata Named list of report metadata.
#' @param valid Optional logical override for the overall validity flag.
#' @param errors,warnings,advisories Optional character vectors of
#'   severity-specific messages.
#' @param sample_data Sample-level mapping table carried by a
#'   \code{split_spec}.
#' @param group_var Name of the grouping column.
#' @param block_vars Optional blocking variable names.
#' @param time_var Optional ordering column name.
#' @param ordering_required Whether ordering is required for downstream
#'   evaluation.
#' @param constraint_mode,constraint_strategy Constraint-derivation metadata.
#' @param recommended_resampling Optional recommended resampling routine.
#' @param overview Character vector of human-readable overview lines.
#' @param diagnostics Diagnostics data frame for leakage risks.
#' @param validation_summary,constraint_summary,split_spec_summary Named lists
#'   carrying pre-computed summaries.
#' @return An S3 object corresponding to the constructor that was called.
#' @seealso \code{\link{validate_graph}}
#' @examples
#' meta <- data.frame(
#'   sample_id  = c("S1", "S2"),
#'   subject_id = c("P1", "P2")
#' )
#' g <- graph_from_metadata(meta)
#'
#' report <- validate_graph(g)
#' report$valid
#' summary(report)
#' @export
depgraph_validation_report <- function(graph_name = NULL, issues = NULL, metrics = list(), metadata = list(), valid = NULL, errors = NULL, warnings = NULL, advisories = NULL) {
  if (is.null(issues)) {
    issues <- data.frame(
      issue_id = character(),
      level = character(),
      severity = character(),
      code = character(),
      message = character(),
      node_ids = I(list()),
      edge_ids = I(list()),
      details = I(list()),
      stringsAsFactors = FALSE
    )
  }

  summary <- .summarize_validation_issues(issues)
  if (is.null(errors)) {
    errors <- unique(issues$message[issues$severity == "error"])
  }
  if (is.null(warnings)) {
    warnings <- unique(issues$message[issues$severity == "warning"])
  }
  if (is.null(advisories)) {
    advisories <- unique(issues$message[issues$severity == "advisory"])
  }
  if (is.null(valid)) {
    valid <- length(errors) == 0L
  }

  structure(
    list(
      graph_name = graph_name,
      valid = isTRUE(valid),
      issues = issues,
      summary = summary,
      metadata = utils::modifyList(
        list(
          created_at = Sys.time(),
          schema_version = .depgraph_schema_version
        ),
        metadata
      ),
      errors = errors,
      warnings = warnings,
      advisories = advisories,
      metrics = utils::modifyList(
        list(
          n_issues = nrow(issues),
          n_errors = length(errors),
          n_warnings = length(warnings),
          n_advisories = length(advisories)
        ),
        metrics
      )
    ),
    class = "depgraph_validation_report"
  )
}

#' @rdname depgraph_validation_report
#' @export
split_spec <- function(sample_data = NULL, group_var = "group_id", block_vars = character(), time_var = NULL, ordering_required = FALSE, constraint_mode = NULL, constraint_strategy = NULL, recommended_resampling = NULL, metadata = list()) {
  if (is.null(sample_data)) {
    sample_data <- data.frame(
      sample_id = character(),
      sample_node_id = character(),
      group_id = character(),
      primary_group = character(),
      batch_group = character(),
      study_group = character(),
      timepoint_id = character(),
      time_index = numeric(),
      order_rank = integer(),
      stringsAsFactors = FALSE
    )
  }

  .depgraph_assert(is.data.frame(sample_data), "`sample_data` must be a data.frame.")

  structure(
    list(
      sample_data = sample_data,
      group_var = as.character(group_var)[1L],
      block_vars = as.character(block_vars),
      time_var = if (is.null(time_var)) NULL else as.character(time_var)[1L],
      ordering_required = isTRUE(ordering_required),
      constraint_mode = if (is.null(constraint_mode)) NULL else as.character(constraint_mode)[1L],
      constraint_strategy = if (is.null(constraint_strategy)) NULL else as.character(constraint_strategy)[1L],
      recommended_resampling = if (is.null(recommended_resampling)) NULL else as.character(recommended_resampling)[1L],
      metadata = metadata
    ),
    class = "split_spec"
  )
}

#' @rdname depgraph_validation_report
#' @export
split_spec_validation <- function(issues = NULL, metadata = list()) {
  if (is.null(issues)) {
    issues <- data.frame(
      issue_id = character(),
      severity = character(),
      code = character(),
      message = character(),
      n_affected = integer(),
      details = I(list()),
      stringsAsFactors = FALSE
    )
  }

  summary <- list(
    by_severity = .depgraph_count_table(if (nrow(issues) == 0L) character() else issues$severity),
    by_code = .depgraph_count_table(if (nrow(issues) == 0L) character() else issues$code)
  )
  valid <- !any(issues$severity %in% "error")

  structure(
    list(
      valid = valid,
      issues = issues,
      summary = summary,
      metadata = metadata
    ),
    class = "split_spec_validation"
  )
}

#' @rdname depgraph_validation_report
#' @export
leakage_risk_summary <- function(overview = character(), diagnostics = NULL, validation_summary = list(), constraint_summary = list(), split_spec_summary = list(), metadata = list()) {
  if (is.null(diagnostics)) {
    diagnostics <- data.frame(
      severity = character(),
      category = character(),
      message = character(),
      source = character(),
      n_affected = integer(),
      stringsAsFactors = FALSE
    )
  }

  structure(
    list(
      overview = as.character(overview),
      diagnostics = diagnostics,
      validation_summary = validation_summary,
      constraint_summary = constraint_summary,
      split_spec_summary = split_spec_summary,
      metadata = metadata
    ),
    class = "leakage_risk_summary"
  )
}

make_constraint_graph <- function(time_source = c("index", "precedence", "missing")) {
  time_source <- match.arg(time_source)

  meta <- data.frame(
    sample_id = c("S1", "S2", "S3", "S4", "S5"),
    subject_id = c("P1", "P1", "P2", "P3", "P4"),
    batch_id = c("B1", "B2", "B1", "B3", NA),
    study_id = c("ST1", "ST1", "ST2", "ST2", "ST3"),
    timepoint_id = c("T0", "T1", "T0", "T2", NA),
    stringsAsFactors = FALSE
  )

  sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
  study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")

  time_data <- unique(meta[!is.na(meta$timepoint_id), "timepoint_id", drop = FALSE])
  time_nodes <- if (identical(time_source, "index")) {
    create_nodes(
      transform(time_data, time_index = c(0L, 1L, 2L)),
      type = "Timepoint",
      id_col = "timepoint_id",
      attr_cols = c("time_index")
    )
  } else {
    create_nodes(time_data, type = "Timepoint", id_col = "timepoint_id")
  }

  subject_edges <- create_edges(
    meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject"
  )
  batch_edges <- create_edges(
    meta, "sample_id", "batch_id", "Sample", "Batch", "sample_processed_in_batch", allow_missing = TRUE
  )
  study_edges <- create_edges(
    meta, "sample_id", "study_id", "Sample", "Study", "sample_from_study"
  )
  time_edges <- create_edges(
    meta, "sample_id", "timepoint_id", "Sample", "Timepoint", "sample_collected_at_timepoint", allow_missing = TRUE
  )

  edge_sets <- list(subject_edges, batch_edges, study_edges, time_edges)
  if (identical(time_source, "precedence")) {
    precedence_edges <- create_edges(
      data.frame(
        from_timepoint = c("T0", "T1"),
        to_timepoint = c("T1", "T2"),
        stringsAsFactors = FALSE
      ),
      "from_timepoint", "to_timepoint", "Timepoint", "Timepoint", "timepoint_precedes"
    )
    edge_sets <- c(edge_sets, list(precedence_edges))
  }

  build_dependency_graph(
    nodes = list(sample_nodes, subject_nodes, batch_nodes, study_nodes, time_nodes),
    edges = edge_sets,
    graph_name = paste0("constraint_graph_", time_source),
    dataset_name = "ConstraintDemo"
  )
}

make_ambiguous_assignment_graph <- function(mode = c("batch", "study", "time")) {
  mode <- match.arg(mode)

  sample_nodes <- graph_node_set(
    data.frame(
      node_id = "sample:S1",
      node_type = "Sample",
      node_key = "S1",
      label = "S1",
      attrs = I(list(list())),
      stringsAsFactors = FALSE
    )
  )

  if (identical(mode, "batch")) {
    target_nodes <- graph_node_set(
      data.frame(
        node_id = c("batch:B1", "batch:B2"),
        node_type = c("Batch", "Batch"),
        node_key = c("B1", "B2"),
        label = c("B1", "B2"),
        attrs = I(list(list(), list())),
        stringsAsFactors = FALSE
      )
    )
    edge_type <- "sample_processed_in_batch"
    edge_targets <- c("batch:B1", "batch:B2")
  } else if (identical(mode, "study")) {
    target_nodes <- graph_node_set(
      data.frame(
        node_id = c("study:ST1", "study:ST2"),
        node_type = c("Study", "Study"),
        node_key = c("ST1", "ST2"),
        label = c("ST1", "ST2"),
        attrs = I(list(list(), list())),
        stringsAsFactors = FALSE
      )
    )
    edge_type <- "sample_from_study"
    edge_targets <- c("study:ST1", "study:ST2")
  } else {
    target_nodes <- graph_node_set(
      data.frame(
        node_id = c("timepoint:T1", "timepoint:T2"),
        node_type = c("Timepoint", "Timepoint"),
        node_key = c("T1", "T2"),
        label = c("T1", "T2"),
        attrs = I(list(list(time_index = 1), list(time_index = 2))),
        stringsAsFactors = FALSE
      )
    )
    edge_type <- "sample_collected_at_timepoint"
    edge_targets <- c("timepoint:T1", "timepoint:T2")
  }

  edges <- graph_edge_set(
    data.frame(
      edge_id = paste0(edge_type, ":", 1:2),
      from = c("sample:S1", "sample:S1"),
      to = edge_targets,
      edge_type = c(edge_type, edge_type),
      attrs = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )

  dependency_graph(
    nodes = graph_node_set(rbind(sample_nodes$data, target_nodes$data)),
    edges = edges,
    graph = NULL
  )
}

test_that("subject constraints group samples by subject and expose grouping vectors", {
  graph <- make_constraint_graph("index")
  result <- derive_split_constraints(graph, mode = "subject")

  expect_s3_class(result, "split_constraint")
  expect_true(all(c(
    "sample_id", "sample_node_id", "group_id", "constraint_type", "group_label", "explanation"
  ) %in% names(result$sample_map)))
  expect_true(all(result$sample_map$constraint_type == "subject"))
  expect_equal(result$sample_map$group_id[result$sample_map$sample_id == "S1"], result$sample_map$group_id[result$sample_map$sample_id == "S2"])
  expect_equal(result$recommended_downstream_args$group_var, "group_id")

  groups <- grouping_vector(result)
  expect_true(is.character(groups))
  expect_identical(unname(groups[["S1"]]), result$sample_map$group_id[result$sample_map$sample_id == "S1"])
})

test_that("batch constraints warn on missing batches and keep assignments deterministic", {
  graph <- make_constraint_graph("index")
  result <- derive_split_constraints(graph, mode = "batch")

  expect_s3_class(result, "split_constraint")
  expect_equal(result$sample_map$group_id[result$sample_map$sample_id == "S1"], result$sample_map$group_id[result$sample_map$sample_id == "S3"])
  expect_false(result$sample_map$group_id[result$sample_map$sample_id == "S2"] == result$sample_map$group_id[result$sample_map$sample_id == "S1"])
  expect_true(any(grepl("missing batch", result$metadata$warnings, ignore.case = TRUE)))
  expect_true(grepl("unlinked", result$sample_map$explanation[result$sample_map$sample_id == "S5"], ignore.case = TRUE))
})

test_that("build_dependency_graph rejects ambiguous batch assignments", {
  edge_meta <- data.frame(
    sample_id = "S1",
    batch_id = c("B1", "B2"),
    study_id = c("ST1", "ST1"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(data.frame(sample_id = "S1", stringsAsFactors = FALSE), type = "Sample", id_col = "sample_id")
  batches <- create_nodes(data.frame(batch_id = c("B1", "B2"), stringsAsFactors = FALSE), type = "Batch", id_col = "batch_id")
  studies <- create_nodes(data.frame(study_id = "ST1", stringsAsFactors = FALSE), type = "Study", id_col = "study_id")
  batch_edges <- create_edges(edge_meta, "sample_id", "batch_id", "Sample", "Batch", "sample_processed_in_batch")
  study_edges <- create_edges(edge_meta, "sample_id", "study_id", "Sample", "Study", "sample_from_study")

  expect_error(
    build_dependency_graph(list(samples, batches, studies), list(batch_edges, study_edges)),
    "sample_processed_in_batch",
    fixed = TRUE
  )
})

test_that("study constraints respect sample subsets", {
  graph <- make_constraint_graph("index")
  result <- derive_split_constraints(graph, mode = "study", samples = c("S1", "sample:S4"))

  expect_s3_class(result, "split_constraint")
  expect_setequal(result$sample_map$sample_id, c("S1", "S4"))
  expect_equal(nrow(result$sample_map), 2L)
  expect_true(all(result$sample_map$constraint_type == "study"))
})

test_that("constraint derivation rejects partially unresolved sample subsets", {
  graph <- make_constraint_graph("index")

  expect_error(
    derive_split_constraints(graph, mode = "subject", samples = c("S1", "BAD")),
    "Unknown sample IDs",
    fixed = TRUE
  )
})

test_that("time constraints use time_index ordering when available", {
  graph <- make_constraint_graph("index")
  result <- derive_split_constraints(graph, mode = "time")

  expect_s3_class(result, "split_constraint")
  expect_true(all(c("time_index", "timepoint_id", "order_rank") %in% names(result$sample_map)))
  expect_equal(result$sample_map$time_index[result$sample_map$sample_id == "S1"], 0)
  expect_equal(result$sample_map$order_rank[result$sample_map$sample_id == "S1"], 1)
  expect_equal(result$sample_map$order_rank[result$sample_map$sample_id == "S4"], 3)
  expect_true(any(grepl("missing timepoint", result$metadata$warnings, ignore.case = TRUE)))
})

test_that("time constraints can derive ordering from timepoint precedence edges", {
  graph <- make_constraint_graph("precedence")
  result <- derive_split_constraints(graph, mode = "time")

  expect_s3_class(result, "split_constraint")
  expect_equal(result$sample_map$order_rank[result$sample_map$sample_id == "S1"], 1)
  expect_equal(result$sample_map$order_rank[result$sample_map$sample_id == "S2"], 2)
  expect_equal(result$sample_map$order_rank[result$sample_map$sample_id == "S4"], 3)
  expect_true(all(is.na(result$sample_map$time_index[result$sample_map$sample_id %in% c("S1", "S2", "S4")])))
})

test_that("time constraint derivation rejects contradictory ordering metadata", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "timepoint:T1", "timepoint:T2"),
      node_type = c("Sample", "Sample", "Timepoint", "Timepoint"),
      node_key = c("S1", "S2", "T1", "T2"),
      label = c("S1", "S2", "T1", "T2"),
      attrs = I(list(
        list(),
        list(),
        list(time_index = 2),
        list(time_index = 1)
      )),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c(
        "sample_collected_at_timepoint:1",
        "sample_collected_at_timepoint:2",
        "timepoint_precedes:1"
      ),
      from = c("sample:S1", "sample:S2", "timepoint:T1"),
      to = c("timepoint:T1", "timepoint:T2", "timepoint:T2"),
      edge_type = c(
        "sample_collected_at_timepoint",
        "sample_collected_at_timepoint",
        "timepoint_precedes"
      ),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)

  expect_error(
    derive_split_constraints(graph, mode = "time"),
    "Time ordering metadata conflict",
    fixed = TRUE
  )
})

test_that("constraint derivation rejects ambiguous direct assignments", {
  expect_error(
    derive_split_constraints(make_ambiguous_assignment_graph("batch"), mode = "batch"),
    "Multiple batch assignments found",
    fixed = TRUE
  )

  expect_error(
    derive_split_constraints(make_ambiguous_assignment_graph("study"), mode = "study"),
    "Multiple study assignments found",
    fixed = TRUE
  )

  expect_error(
    derive_split_constraints(make_ambiguous_assignment_graph("time"), mode = "time"),
    "Multiple time assignments found",
    fixed = TRUE
  )
})

test_that("composite strict constraints use transitive dependency closure", {
  graph <- make_constraint_graph("index")
  result <- derive_split_constraints(graph, mode = "composite", strategy = "strict", via = c("Subject", "Batch"))

  expect_s3_class(result, "split_constraint")
  expect_equal(result$sample_map$group_id[result$sample_map$sample_id == "S1"], result$sample_map$group_id[result$sample_map$sample_id == "S2"])
  expect_equal(result$sample_map$group_id[result$sample_map$sample_id == "S1"], result$sample_map$group_id[result$sample_map$sample_id == "S3"])
  expect_false(result$sample_map$group_id[result$sample_map$sample_id == "S4"] == result$sample_map$group_id[result$sample_map$sample_id == "S1"])
  expect_equal(result$metadata$strategy, "strict")
})

test_that("composite strict respects sample subset (no bridging through excluded samples)", {
  # Regression test for C1: .derive_composite_strict_constraints used to
  # compute components over the WHOLE graph, then filter to `samples` — so two
  # in-subset samples connected only through an out-of-subset sample silently
  # landed in the same group, leaking out-of-subset structure into the split.
  graph <- make_constraint_graph("index")
  # Full-graph projection via Subject + Batch links S1, S2, S3 (S1-P1-S2;
  # S1-B1-S3). Within the subset c("S2","S3") alone there is no shared Subject
  # and no shared Batch, so the two samples must NOT share a group_id.
  result <- derive_split_constraints(
    graph,
    mode = "composite",
    strategy = "strict",
    via = c("Subject", "Batch"),
    samples = c("S2", "S3")
  )

  expect_equal(nrow(result$sample_map), 2L)
  g_s2 <- result$sample_map$group_id[result$sample_map$sample_id == "S2"]
  g_s3 <- result$sample_map$group_id[result$sample_map$sample_id == "S3"]
  expect_false(
    identical(g_s2, g_s3),
    info = "S2 and S3 share no Subject or Batch within the subset; they must not be bridged through excluded S1."
  )
  expect_equal(result$metadata$n_groups, 2L)
})

test_that("composite rule-based constraints follow priority fallback rules", {
  graph <- make_constraint_graph("index")
  result <- derive_split_constraints(
    graph,
    mode = "composite",
    strategy = "rule_based",
    priority = c("batch", "study", "subject", "time")
  )

  expect_s3_class(result, "split_constraint")
  expect_equal(result$sample_map$group_id[result$sample_map$sample_id == "S1"], result$sample_map$group_id[result$sample_map$sample_id == "S3"])
  expect_false(result$sample_map$group_id[result$sample_map$sample_id == "S2"] == result$sample_map$group_id[result$sample_map$sample_id == "S1"])
  expect_equal(result$sample_map$constraint_type[result$sample_map$sample_id == "S5"], "study")
  expect_equal(result$sample_map$group_label[result$sample_map$sample_id == "S5"], "ST3")
  expect_equal(result$metadata$strategy, "rule_based")
  expect_false(isTRUE(result$recommended_downstream_args$ordering_required))
  expect_equal(result$recommended_downstream_args$time_var, "order_rank")
})

test_that("create_nodes and create_edges build valid core objects", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    batch_id = c("B1", "B1"),
    study_id = c("ST1", "ST1"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")

  graph <- build_dependency_graph(list(samples, subjects), list(edges))
  validation <- validate_graph(graph)

  expect_s3_class(samples, "graph_node_set")
  expect_s3_class(edges, "graph_edge_set")
  expect_s3_class(graph, "dependency_graph")
  expect_true(validation$valid)
  expect_named(samples$data$attrs[[1]], c("subject_id", "batch_id", "study_id"))
  expect_equal(igraph::vcount(as_igraph(graph)), nrow(graph$nodes$data))
  expect_equal(igraph::ecount(as_igraph(graph)), nrow(graph$edges$data))
})

test_that("compatibility aliases delegate to the primary constructors", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  # The aliases are deprecated as of 0.2.0; deprecation warnings are tested
  # in test-deprecations.R, so silence them here while we verify delegation.
  suppressWarnings({
    subjects <- new_depgraph_nodes(create_nodes(meta, type = "Subject", id_col = "subject_id")$data)
    edges <- new_depgraph_edges(
      create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")$data
    )
    full_nodes <- new_depgraph_nodes(rbind(samples$data, subjects$data))

    graph <- new_depgraph(nodes = full_nodes, edges = edges)
    graph2 <- build_depgraph(nodes = list(samples, subjects), edges = list(edges))
  })

  expect_s3_class(subjects, "graph_node_set")
  expect_s3_class(edges, "graph_edge_set")
  expect_s3_class(graph, "dependency_graph")
  expect_s3_class(graph2, "dependency_graph")
  expect_equal(igraph::vcount(as_igraph(graph2)), 4)
})

test_that("build_dependency_graph stores validation overrides in metadata", {
  meta <- data.frame(
    sample_id = c("S1"),
    subject_id = c("P1"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")

  graph <- build_dependency_graph(
    list(samples, subjects),
    list(edges),
    validation_overrides = list(allow_multi_subject_samples = TRUE)
  )

  expect_true(isTRUE(graph$metadata$validation_overrides$allow_multi_subject_samples))
})

test_that("create_edges can target unprefixed node identifiers when requested", {
  meta <- data.frame(
    sample_id = "S1",
    subject_id = "P1",
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id", prefix = FALSE)
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(
    meta,
    "sample_id",
    "subject_id",
    "Sample",
    "Subject",
    "sample_belongs_to_subject",
    from_prefix = FALSE
  )

  graph <- build_dependency_graph(list(samples, subjects), list(edges))

  expect_s3_class(graph, "dependency_graph")
  expect_true(validate_graph(graph)$valid)
  expect_equal(graph$edges$data$from, "S1")
  expect_equal(graph$edges$data$to, "subject:P1")
})

test_that("create_nodes collapses exact duplicate definitions deterministically", {
  nodes <- create_nodes(
    data.frame(
      subject_id = c("P1", "P1"),
      subject_type = c("case", "case"),
      stringsAsFactors = FALSE
    ),
    type = "Subject",
    id_col = "subject_id",
    attr_cols = "subject_type"
  )

  expect_s3_class(nodes, "graph_node_set")
  expect_equal(nrow(nodes$data), 1L)
  expect_equal(nodes$data$node_id, "subject:P1")
  expect_identical(nodes$data$attrs[[1]]$subject_type, "case")
})

test_that("create_nodes rejects conflicting duplicate definitions", {
  expect_error(
    create_nodes(
      data.frame(
        subject_id = c("P1", "P1"),
        subject_type = c("case", "control"),
        stringsAsFactors = FALSE
      ),
      type = "Subject",
      id_col = "subject_id",
      attr_cols = "subject_type"
    ),
    "Conflicting node definitions found for IDs",
    fixed = TRUE
  )
})

test_that("create_edges collapses exact duplicate definitions deterministically", {
  edges <- create_edges(
    data.frame(
      sample_id = c("S1", "S1"),
      subject_id = c("P1", "P1"),
      note = c("same", "same"),
      stringsAsFactors = FALSE
    ),
    from_col = "sample_id",
    to_col = "subject_id",
    from_type = "Sample",
    to_type = "Subject",
    relation = "sample_belongs_to_subject",
    attr_cols = "note"
  )

  expect_s3_class(edges, "graph_edge_set")
  expect_equal(nrow(edges$data), 1L)
  expect_equal(edges$data$from, "sample:S1")
  expect_equal(edges$data$to, "subject:P1")
  expect_identical(edges$data$attrs[[1]]$note, "same")
})

test_that("create_edges rejects conflicting duplicate definitions", {
  expect_error(
    create_edges(
      data.frame(
        sample_id = c("S1", "S1"),
        subject_id = c("P1", "P1"),
        note = c("first", "second"),
        stringsAsFactors = FALSE
      ),
      from_col = "sample_id",
      to_col = "subject_id",
      from_type = "Sample",
      to_type = "Subject",
      relation = "sample_belongs_to_subject",
      attr_cols = "note"
    ),
    "Conflicting edge definitions found for relations",
    fixed = TRUE
  )
})

test_that("conflicting duplicate timepoint definitions cannot change derived ordering", {
  make_graph <- function(time_index_values) {
    meta <- data.frame(
      sample_id = c("S1", "S2"),
      timepoint_id = c("T1", "T2"),
      stringsAsFactors = FALSE
    )

    sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
    time_nodes <- create_nodes(
      data.frame(
        timepoint_id = c("T1", "T1", "T2"),
        time_index = time_index_values,
        stringsAsFactors = FALSE
      ),
      type = "Timepoint",
      id_col = "timepoint_id",
      attr_cols = "time_index"
    )
    time_edges <- create_edges(
      meta,
      "sample_id",
      "timepoint_id",
      "Sample",
      "Timepoint",
      "sample_collected_at_timepoint"
    )

    build_dependency_graph(list(sample_nodes, time_nodes), list(time_edges))
  }

  expect_error(
    make_graph(c(1, 9, 2)),
    "Conflicting node definitions found for IDs",
    fixed = TRUE
  )
  expect_error(
    make_graph(c(9, 1, 2)),
    "Conflicting node definitions found for IDs",
    fixed = TRUE
  )
})

test_that("low-level constructors reject malformed attrs", {
  expect_error(
    graph_node_set(
      data.frame(
        node_id = "sample:S1",
        node_type = "Sample",
        node_key = "S1",
        label = "S1",
        attrs = I(list(list("bad"))),
        stringsAsFactors = FALSE
      )
    ),
    "named lists"
  )

  expect_error(
    graph_edge_set(
      data.frame(
        edge_id = "sample_belongs_to_subject:1",
        from = "sample:S1",
        to = "subject:P1",
        edge_type = "sample_belongs_to_subject",
        attrs = I(list(list("bad"))),
        stringsAsFactors = FALSE
      )
    ),
    "named lists"
  )
})

test_that("dependency_graph rejects igraph objects that do not match node and edge tables", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "subject:P1"),
      node_type = c("Sample", "Subject"),
      node_key = c("S1", "P1"),
      label = c("S1", "P1"),
      attrs = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = "sample_belongs_to_subject:1",
      from = "sample:S1",
      to = "subject:P1",
      edge_type = "sample_belongs_to_subject",
      attrs = I(list(list())),
      stringsAsFactors = FALSE
    )
  )

  bad_graph <- igraph::make_empty_graph(n = 1, directed = TRUE)
  igraph::V(bad_graph)$name <- "sample:S1"

  expect_error(
    dependency_graph(nodes = nodes, edges = edges, graph = bad_graph),
    "does not match the supplied node and edge tables",
    fixed = TRUE
  )
})

test_that("validate_graph catches unsupported edge types", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "subject:P1"),
      node_type = c("Sample", "Subject"),
      node_key = c("S1", "P1"),
      label = c("S1", "P1"),
      attrs = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = "mystery:1",
      from = "sample:S1",
      to = "subject:P1",
      edge_type = "mystery_relation",
      attrs = I(list(list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_false(validation$valid)
  expect_match(validation$errors[[1]], "Unknown edge type", fixed = TRUE)
})

test_that("validate_graph returns a structured validation report", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")
  graph <- build_dependency_graph(list(samples, subjects), list(edges))

  validation <- validate_graph(graph)

  expect_s3_class(validation, "depgraph_validation_report")
  expect_true(is.data.frame(validation$issues))
  expect_true(all(c("issue_id", "level", "severity", "code", "message", "node_ids", "edge_ids", "details") %in% names(validation$issues)))
  expect_true(is.logical(validation$valid))
  expect_true(is.list(validation$summary))
  expect_true(is.list(validation$metrics))
  expect_equal(as.data.frame(validation), validation$issues)
})

test_that("validate_graph detects igraph objects that drift out of sync with graph tables", {
  meta <- data.frame(
    sample_id = c("S1"),
    subject_id = c("P1"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")
  graph <- build_dependency_graph(list(samples, subjects), list(edges))

  graph$graph <- igraph::make_empty_graph(n = 1, directed = TRUE)
  igraph::V(graph$graph)$name <- "sample:S1"

  validation <- validate_graph(graph)

  expect_false(validation$valid)
  expect_true(any(validation$issues$code == "igraph_table_mismatch"))
})

test_that("validate_graph enforces structural single-target cardinality", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "assay:A1", "assay:A2"),
      node_type = c("Sample", "Assay", "Assay"),
      node_key = c("S1", "A1", "A2"),
      label = c("S1", "A1", "A2"),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c("sample_measured_by_assay:1", "sample_measured_by_assay:2"),
      from = c("sample:S1", "sample:S1"),
      to = c("assay:A1", "assay:A2"),
      edge_type = c("sample_measured_by_assay", "sample_measured_by_assay"),
      attrs = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_false(validation$valid)
  expect_true(any(validation$issues$code == "single_target_violation"))
})

test_that("severity filtering does not change validation truth", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "assay:A1", "assay:A2", "study:ST1"),
      node_type = c("Sample", "Sample", "Assay", "Assay", "Study"),
      node_key = c("S1", "S2", "A1", "A2", "ST1"),
      label = c("S1", "S2", "A1", "A2", "ST1"),
      attrs = I(list(list(), list(), list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c(
        "sample_measured_by_assay:1",
        "sample_measured_by_assay:2",
        "sample_from_study:1"
      ),
      from = c("sample:S1", "sample:S1", "sample:S1"),
      to = c("assay:A1", "assay:A2", "study:ST1"),
      edge_type = c(
        "sample_measured_by_assay",
        "sample_measured_by_assay",
        "sample_from_study"
      ),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)

  validation <- validate_graph(graph, severities = "warning")

  expect_false(validation$valid)
  expect_equal(nrow(validation$issues), 1L)
  expect_true(all(validation$issues$severity == "warning"))
  expect_equal(validation$metrics$n_errors, 1L)
  expect_true(any(validation$issues$code == "sample_missing_study_assignment"))
  expect_true(any(grepl("sample_measured_by_assay", validation$errors, fixed = TRUE)))
})

test_that("error_on_fail still errors when severity filtering hides visible errors", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "assay:A1", "assay:A2", "study:ST1"),
      node_type = c("Sample", "Sample", "Assay", "Assay", "Study"),
      node_key = c("S1", "S2", "A1", "A2", "ST1"),
      label = c("S1", "S2", "A1", "A2", "ST1"),
      attrs = I(list(list(), list(), list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c(
        "sample_measured_by_assay:1",
        "sample_measured_by_assay:2",
        "sample_from_study:1"
      ),
      from = c("sample:S1", "sample:S1", "sample:S1"),
      to = c("assay:A1", "assay:A2", "study:ST1"),
      edge_type = c(
        "sample_measured_by_assay",
        "sample_measured_by_assay",
        "sample_from_study"
      ),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)

  expect_error(
    validate_graph(graph, severities = "warning", error_on_fail = TRUE),
    "Graph validation failed.",
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(validate_depgraph(graph, severities = "warning", error_on_fail = TRUE)),
    "Graph validation failed.",
    fixed = TRUE
  )
})

test_that("validate_graph warns on non-canonical node attributes", {
  nodes <- graph_node_set(
    data.frame(
      node_id = "sample:S1",
      node_type = "Sample",
      node_key = "S1",
      label = "S1",
      attrs = I(list(list(custom_field = "value"))),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = character(),
      from = character(),
      to = character(),
      edge_type = character(),
      attrs = I(list()),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_true(validation$valid)
  expect_true(any(grepl("non-canonical attributes", validation$warnings, fixed = TRUE)))
})

test_that("semantic validation does not require study assignments when no study structure exists", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")
  graph <- build_dependency_graph(list(samples, subjects), list(edges))

  validation <- validate_graph(graph)

  expect_false(any(validation$issues$code == "sample_missing_study_assignment"))
})

test_that("validate_depgraph is a compatibility alias", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "subject:P1", "featureset:FS1"),
      node_type = c("Sample", "Sample", "Subject", "FeatureSet"),
      node_key = c("S1", "S2", "P1", "FS1"),
      label = c("S1", "S2", "P1", "FS1"),
      attrs = I(list(
        list(),
        list(),
        list(),
        list(derivation_scope = "per_dataset")
      )),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c(
        "sample_belongs_to_subject:1",
        "sample_belongs_to_subject:2",
        "sample_uses_featureset:1",
        "sample_uses_featureset:2"
      ),
      from = c("sample:S1", "sample:S2", "sample:S1", "sample:S2"),
      to = c("subject:P1", "subject:P1", "featureset:FS1", "featureset:FS1"),
      edge_type = c(
        "sample_belongs_to_subject",
        "sample_belongs_to_subject",
        "sample_uses_featureset",
        "sample_uses_featureset"
      ),
      attrs = I(list(list(), list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  # validate_depgraph is deprecated as of 0.2.0; deprecation warnings are
  # tested in test-deprecations.R, so silence them here while we verify
  # delegation parity with validate_graph().
  validation <- suppressWarnings(validate_depgraph(graph))
  graph_validation <- validate_graph(graph)

  expect_s3_class(validation, "depgraph_validation_report")
  expect_identical(validation$valid, graph_validation$valid)
  expect_identical(validation$issues$level, graph_validation$issues$level)
  expect_identical(validation$issues$severity, graph_validation$issues$severity)
  expect_identical(validation$issues$code, graph_validation$issues$code)
  expect_identical(validation$issues$message, graph_validation$issues$message)
})

test_that("semantic validation flags invalid FeatureSet derivation scope", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "featureset:FS1"),
      node_type = c("Sample", "FeatureSet"),
      node_key = c("S1", "FS1"),
      label = c("S1", "FS1"),
      attrs = I(list(list(), list(derivation_scope = "invalid_scope"))),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = "sample_uses_featureset:1",
      from = "sample:S1",
      to = "featureset:FS1",
      edge_type = "sample_uses_featureset",
      attrs = I(list(list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_true(any(validation$issues$code == "invalid_featureset_derivation_scope"))
  expect_true(any(validation$issues$severity == "warning"))
})

test_that("semantic validation flags missing study assignment when study structure exists", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "study:ST1"),
      node_type = c("Sample", "Sample", "Study"),
      node_key = c("S1", "S2", "ST1"),
      label = c("S1", "S2", "ST1"),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = "sample_from_study:1",
      from = "sample:S1",
      to = "study:ST1",
      edge_type = "sample_from_study",
      attrs = I(list(list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_true(any(validation$issues$code == "sample_missing_study_assignment"))
  expect_true(any(validation$issues$severity == "warning"))
})

test_that("semantic validation rejects contradictory time ordering metadata", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "timepoint:T1", "timepoint:T2"),
      node_type = c("Sample", "Sample", "Timepoint", "Timepoint"),
      node_key = c("S1", "S2", "T1", "T2"),
      label = c("S1", "S2", "T1", "T2"),
      attrs = I(list(
        list(),
        list(),
        list(time_index = 2),
        list(time_index = 1)
      )),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c(
        "sample_collected_at_timepoint:1",
        "sample_collected_at_timepoint:2",
        "timepoint_precedes:1"
      ),
      from = c("sample:S1", "sample:S2", "timepoint:T1"),
      to = c("timepoint:T1", "timepoint:T2", "timepoint:T2"),
      edge_type = c(
        "sample_collected_at_timepoint",
        "sample_collected_at_timepoint",
        "timepoint_precedes"
      ),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_false(validation$valid)
  expect_true(any(validation$issues$code == "time_order_conflict"))
})

test_that("semantic validation rejects cyclic timepoint precedence graphs", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("timepoint:T1", "timepoint:T2", "timepoint:T3"),
      node_type = c("Timepoint", "Timepoint", "Timepoint"),
      node_key = c("T1", "T2", "T3"),
      label = c("T1", "T2", "T3"),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c("timepoint_precedes:1", "timepoint_precedes:2", "timepoint_precedes:3"),
      from = c("timepoint:T1", "timepoint:T2", "timepoint:T3"),
      to = c("timepoint:T2", "timepoint:T3", "timepoint:T1"),
      edge_type = c("timepoint_precedes", "timepoint_precedes", "timepoint_precedes"),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_false(validation$valid)
  expect_true(any(validation$issues$code == "timepoint_precedence_cycle"))
})

test_that("semantic validation flags multi-subject samples unless explicitly allowed", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "subject:P1", "subject:P2"),
      node_type = c("Sample", "Subject", "Subject"),
      node_key = c("S1", "P1", "P2"),
      label = c("S1", "P1", "P2"),
      attrs = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c("sample_belongs_to_subject:1", "sample_belongs_to_subject:2"),
      from = c("sample:S1", "sample:S1"),
      to = c("subject:P1", "subject:P2"),
      edge_type = c("sample_belongs_to_subject", "sample_belongs_to_subject"),
      attrs = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_false(validation$valid)
  expect_true(any(validation$issues$code == "sample_multiple_subject_assignments"))

  allowed_graph <- dependency_graph(
    nodes = nodes,
    edges = edges,
    graph = NULL,
    metadata = list(validation_overrides = list(allow_multi_subject_samples = TRUE))
  )
  allowed_validation <- validate_graph(allowed_graph)

  expect_false(any(allowed_validation$issues$code == "sample_multiple_subject_assignments"))
})

test_that("leakage validation reports repeated-subject and per-dataset feature risks", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "sample:S2", "subject:P1", "featureset:FS1"),
      node_type = c("Sample", "Sample", "Subject", "FeatureSet"),
      node_key = c("S1", "S2", "P1", "FS1"),
      label = c("S1", "S2", "P1", "FS1"),
      attrs = I(list(
        list(),
        list(),
        list(),
        list(derivation_scope = "per_dataset")
      )),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c(
        "sample_belongs_to_subject:1",
        "sample_belongs_to_subject:2",
        "sample_uses_featureset:1",
        "sample_uses_featureset:2"
      ),
      from = c("sample:S1", "sample:S2", "sample:S1", "sample:S2"),
      to = c("subject:P1", "subject:P1", "featureset:FS1", "featureset:FS1"),
      edge_type = c(
        "sample_belongs_to_subject",
        "sample_belongs_to_subject",
        "sample_uses_featureset",
        "sample_uses_featureset"
      ),
      attrs = I(list(list(), list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  validation <- validate_graph(graph)

  expect_true(any(validation$issues$code == "repeated_subject_samples"))
  expect_true(any(validation$issues$code == "per_dataset_featureset"))
  expect_true(any(validation$issues$severity == "advisory"))
})

make_multi_subject_graph <- function() {
  nodes <- splitGraph::graph_node_set(
    data.frame(
      node_id   = c("sample:S1", "subject:P1", "subject:P2"),
      node_type = c("Sample", "Subject", "Subject"),
      node_key  = c("S1", "P1", "P2"),
      label     = c("S1", "P1", "P2"),
      attrs     = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- splitGraph::graph_edge_set(
    data.frame(
      edge_id   = c("sample_belongs_to_subject:1", "sample_belongs_to_subject:2"),
      from      = c("sample:S1", "sample:S1"),
      to        = c("subject:P1", "subject:P2"),
      edge_type = c("sample_belongs_to_subject", "sample_belongs_to_subject"),
      attrs     = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )

  splitGraph::dependency_graph(nodes = nodes, edges = edges, graph = NULL)
}

test_that("validate_graph() accepts validation_overrides as a first-class argument", {
  g <- make_multi_subject_graph()

  # Default: multi-subject samples are an error.
  default_report <- validate_graph(g)
  expect_false(default_report$valid)
  expect_true(any(default_report$issues$code == "sample_multiple_subject_assignments"))

  # With the override, the same situation is no longer flagged.
  permissive_report <- validate_graph(
    g,
    validation_overrides = list(allow_multi_subject_samples = TRUE)
  )
  expect_true(permissive_report$valid)
  expect_false(any(permissive_report$issues$code == "sample_multiple_subject_assignments"))
})

test_that("validate_graph() validates the type of validation_overrides", {
  g <- make_multi_subject_graph()

  expect_error(
    validate_graph(g, validation_overrides = "not a list"),
    "validation_overrides"
  )
  expect_error(
    validate_graph(g, validation_overrides = list(TRUE)),
    "validation_overrides"
  )
})

test_that("validation_overrides on validate_graph() temporarily overrides graph metadata", {
  g <- make_multi_subject_graph()
  # Embed a permissive override on the graph itself.
  g$metadata$validation_overrides <- list(allow_multi_subject_samples = TRUE)

  # The graph-level override applies when no explicit override is passed.
  expect_true(validate_graph(g)$valid)

  # An explicit FALSE in the call beats the graph metadata.
  strict_report <- validate_graph(
    g,
    validation_overrides = list(allow_multi_subject_samples = FALSE)
  )
  expect_false(strict_report$valid)
})

test_that("derive_split_constraints(mode='subject') respects allow_multi_subject_samples", {
  g <- make_multi_subject_graph()

  # Without the override, derivation cannot pick a single subject -> error.
  expect_error(
    derive_split_constraints(g, mode = "subject"),
    "Multiple subject assignments|allow_multi_subject_samples",
    perl = TRUE
  )

  # With the override, derivation succeeds and records that S1 was assigned
  # to one of its subjects deterministically (the first one in edge order).
  g$metadata$validation_overrides <- list(allow_multi_subject_samples = TRUE)
  constraint <- derive_split_constraints(g, mode = "subject")
  expect_s3_class(constraint, "split_constraint")
  expect_equal(nrow(constraint$sample_map), 1L)
  expect_true(constraint$sample_map$sample_id == "S1")
  expect_true(constraint$sample_map$group_id %in% c("subject:P1", "subject:P2"))
  # And we record a warning so the ambiguity is not invisible.
  warnings <- constraint$metadata$warnings %||% character()
  expect_true(any(grepl("multiple subject", warnings, ignore.case = TRUE)))
})

test_that("override does not leak across modes (batch/study/time still error on multi-assignment)", {
  nodes <- splitGraph::graph_node_set(
    data.frame(
      node_id   = c("sample:S1", "batch:B1", "batch:B2"),
      node_type = c("Sample", "Batch", "Batch"),
      node_key  = c("S1", "B1", "B2"),
      label     = c("S1", "B1", "B2"),
      attrs     = I(list(list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )
  edges <- splitGraph::graph_edge_set(
    data.frame(
      edge_id   = c("sample_processed_in_batch:1", "sample_processed_in_batch:2"),
      from      = c("sample:S1", "sample:S1"),
      to        = c("batch:B1", "batch:B2"),
      edge_type = c("sample_processed_in_batch", "sample_processed_in_batch"),
      attrs     = I(list(list(), list())),
      stringsAsFactors = FALSE
    )
  )
  g <- splitGraph::dependency_graph(nodes = nodes, edges = edges, graph = NULL)
  g$metadata$validation_overrides <- list(allow_multi_subject_samples = TRUE)

  expect_error(
    derive_split_constraints(g, mode = "batch"),
    "Multiple batch assignments"
  )
})

make_simple_graph <- function() {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject"
  )
  build_dependency_graph(list(samples, subjects), list(edges))
}

test_that("validate_graph(checks = ...) is deprecated but still works", {
  g <- make_simple_graph()

  # Default call (no `checks` supplied) must NOT warn.
  expect_silent(validate_graph(g))

  # Explicit `checks` triggers a deprecation warning.
  expect_warning(
    report <- validate_graph(g, checks = c("ids", "references")),
    "deprecated",
    ignore.case = TRUE
  )
  expect_s3_class(report, "depgraph_validation_report")
  expect_true(report$valid)
})

test_that("validate_graph() recommended path (levels=) is silent and equivalent", {
  g <- make_simple_graph()

  expect_silent(report <- validate_graph(g, levels = c("structural", "semantic")))
  expect_s3_class(report, "depgraph_validation_report")
  expect_true(report$valid)
})

test_that("alias functions are deprecated but still functional", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject"
  )

  expect_warning(
    g <- build_depgraph(list(samples, subjects), list(edges)),
    "deprecated", ignore.case = TRUE
  )
  expect_s3_class(g, "dependency_graph")

  expect_warning(
    report <- validate_depgraph(g),
    "deprecated", ignore.case = TRUE
  )
  expect_s3_class(report, "depgraph_validation_report")

  # Constructor aliases.
  expect_warning(
    n <- new_depgraph_nodes(),
    "deprecated", ignore.case = TRUE
  )
  expect_s3_class(n, "graph_node_set")

  expect_warning(
    e <- new_depgraph_edges(),
    "deprecated", ignore.case = TRUE
  )
  expect_s3_class(e, "graph_edge_set")

  # new_depgraph wraps dependency_graph(), which expects already-bound
  # node/edge sets — we mimic build_dependency_graph()'s assembly:
  bound_nodes <- graph_node_set(rbind(
    as.data.frame(create_nodes(meta, "Sample", "sample_id")),
    as.data.frame(create_nodes(meta, "Subject", "subject_id"))
  ))
  bound_edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject"
  )
  expect_warning(
    nd <- new_depgraph(nodes = bound_nodes, edges = bound_edges, graph = NULL),
    "deprecated", ignore.case = TRUE
  )
  expect_s3_class(nd, "dependency_graph")
})

test_that("the canonical (non-deprecated) constructors produce no warnings", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  expect_silent(graph_node_set(data.frame(
    node_id = c("sample:S1"), node_type = "Sample",
    node_key = "S1", label = "S1",
    attrs = I(list(list())), stringsAsFactors = FALSE
  )))
  expect_silent(graph_edge_set())

  samples <- create_nodes(meta, "Sample", "sample_id")
  subjects <- create_nodes(meta, "Subject", "subject_id")
  edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject"
  )
  expect_silent(build_dependency_graph(list(samples, subjects), list(edges)))
})

# S3 methods for splitGraph core classes.

#' @export
print.graph_node_set <- function(x, ...) {
  cat(
    "<graph_node_set>",
    nrow(x$data), "nodes across",
    length(unique(x$data$node_type)), "types",
    "(schema", x$schema_version, ")\n"
  )
  invisible(x)
}

#' @export
summary.graph_node_set <- function(object, ...) {
  list(
    schema_version = object$schema_version,
    n_nodes = nrow(object$data),
    node_types = .depgraph_count_table(object$data$node_type)
  )
}

#' @export
as.data.frame.graph_node_set <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$data
}

#' @export
print.graph_edge_set <- function(x, ...) {
  cat(
    "<graph_edge_set>",
    nrow(x$data), "edges across",
    length(unique(x$data$edge_type)), "relations",
    "(schema", x$schema_version, ")\n"
  )
  invisible(x)
}

#' @export
summary.graph_edge_set <- function(object, ...) {
  list(
    schema_version = object$schema_version,
    n_edges = nrow(object$data),
    edge_types = .depgraph_count_table(object$data$edge_type)
  )
}

#' @export
as.data.frame.graph_edge_set <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$data
}

#' @export
print.dependency_graph <- function(x, ...) {
  graph_name <- x$metadata$graph_name %||% "<unnamed>"
  cat(
    "<dependency_graph>",
    graph_name,
    "\n  Nodes:", igraph::vcount(x$graph),
    "\n  Edges:", igraph::ecount(x$graph),
    "\n"
  )
  invisible(x)
}

#' @export
summary.dependency_graph <- function(object, ...) {
  list(
    graph_name = object$metadata$graph_name,
    dataset_name = object$metadata$dataset_name,
    schema_version = object$metadata$schema_version,
    n_nodes = nrow(object$nodes$data),
    n_edges = nrow(object$edges$data),
    node_types = .depgraph_count_table(object$nodes$data$node_type),
    edge_types = .depgraph_count_table(object$edges$data$edge_type)
  )
}

.depgraph_type_rows <- c(
  Sample     = 0L,
  Subject    = 1L,
  Batch      = 1L,
  Study      = 1L,
  Timepoint  = 1L,
  Site       = 1L,
  Region     = 1L,
  Assay      = 2L,
  FeatureSet = 2L,
  Platform   = 2L,
  Outcome    = 3L
)

.depgraph_type_palette <- function() {
  c(
    Sample     = "#4C78A8",
    Subject    = "#F58518",
    Batch      = "#54A24B",
    Study      = "#B279A2",
    Timepoint  = "#E45756",
    Assay      = "#72B7B2",
    FeatureSet = "#EECA3B",
    Outcome    = "#9D755D",
    Site       = "#FF9DA6",
    Region     = "#8CD17D",
    Platform   = "#D37295",
    `_other_`  = "#B0B0B0"
  )
}

.depgraph_typed_layout <- function(graph) {
  node_types <- igraph::vertex_attr(graph, "node_type")
  if (is.null(node_types)) {
    node_types <- rep("Sample", igraph::vcount(graph))
  }

  rows <- .depgraph_type_rows[node_types]
  rows[is.na(rows)] <- max(.depgraph_type_rows) + 1L

  order_in_row <- stats::ave(seq_along(rows), rows, FUN = seq_along)
  row_widths <- stats::ave(seq_along(rows), rows, FUN = length)

  x <- ifelse(row_widths > 1L,
              (order_in_row - 1L) / (row_widths - 1L) - 0.5,
              0)
  y <- -as.numeric(rows)

  cbind(x, y)
}

.depgraph_node_colors <- function(graph, node_colors = NULL) {
  palette <- .depgraph_type_palette()
  if (!is.null(node_colors)) {
    for (nm in names(node_colors)) palette[[nm]] <- node_colors[[nm]]
  }
  node_types <- igraph::vertex_attr(graph, "node_type")
  if (is.null(node_types)) {
    node_types <- rep("Sample", igraph::vcount(graph))
  }
  colors <- palette[node_types]
  colors[is.na(colors)] <- palette[["_other_"]]
  unname(colors)
}

#' @export
plot.dependency_graph <- function(x,
                                  layout = c("typed", "sugiyama", "auto"),
                                  node_colors = NULL,
                                  show_labels = TRUE,
                                  legend = TRUE,
                                  legend_position = "topleft",
                                  ...) {
  layout_choice <- if (is.character(layout)) match.arg(layout) else layout
  g <- x$graph

  user_args <- list(...)
  plot_args <- list(x = g)

  if (is.character(layout_choice)) {
    if (identical(layout_choice, "typed")) {
      plot_args$layout <- .depgraph_typed_layout(g)
    } else if (identical(layout_choice, "sugiyama")) {
      plot_args$layout <- igraph::layout_with_sugiyama(g)$layout
    }
  } else if (is.matrix(layout_choice)) {
    plot_args$layout <- layout_choice
  } else if (is.function(layout_choice)) {
    plot_args$layout <- layout_choice(g)
  }

  if (is.null(user_args[["vertex.color"]])) {
    plot_args$vertex.color <- .depgraph_node_colors(g, node_colors = node_colors)
  }
  if (is.null(user_args[["vertex.label"]])) {
    plot_args$vertex.label <- if (isTRUE(show_labels)) {
      lbl <- igraph::vertex_attr(g, "label")
      if (is.null(lbl) || all(is.na(lbl) | !nzchar(lbl))) {
        igraph::vertex_attr(g, "name")
      } else {
        ifelse(is.na(lbl) | !nzchar(lbl), igraph::vertex_attr(g, "name"), lbl)
      }
    } else {
      NA
    }
  }
  if (is.null(user_args[["vertex.label.cex"]])) plot_args$vertex.label.cex <- 0.7
  if (is.null(user_args[["vertex.size"]]))       plot_args$vertex.size <- 14
  if (is.null(user_args[["edge.arrow.size"]]))   plot_args$edge.arrow.size <- 0.4
  if (is.null(user_args[["edge.color"]]))        plot_args$edge.color <- "#888888"

  plot_args[names(user_args)] <- user_args

  do.call(graphics::plot, plot_args)

  if (isTRUE(legend)) {
    palette <- .depgraph_type_palette()
    if (!is.null(node_colors)) {
      for (nm in names(node_colors)) palette[[nm]] <- node_colors[[nm]]
    }
    node_types <- igraph::vertex_attr(g, "node_type")
    present <- intersect(names(palette), unique(node_types))
    if (length(present) > 0L) {
      graphics::legend(
        legend_position,
        legend = present,
        pt.bg = unname(palette[present]),
        pch = 21,
        pt.cex = 1.4,
        col = "#333333",
        bty = "n",
        cex = 0.75,
        title = "Node type"
      )
    }
  }

  invisible(x)
}

#' @export
print.graph_query_result <- function(x, ...) {
  cat("<graph_query_result>", x$query, "\n")
  cat("  Rows:", nrow(x$table), "\n")
  invisible(x)
}

#' @export
summary.graph_query_result <- function(object, ...) {
  list(
    query = object$query,
    n_nodes = nrow(object$nodes),
    n_edges = nrow(object$edges),
    n_rows = nrow(object$table)
  )
}

#' @export
as.data.frame.graph_query_result <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$table
}

#' @export
print.dependency_constraint <- function(x, ...) {
  cat("<dependency_constraint>", x$constraint_id, "\n")
  cat("  Samples:", nrow(x$sample_map), "\n")
  invisible(x)
}

#' @export
summary.dependency_constraint <- function(object, ...) {
  group_col <- intersect(c("group_id", "split_unit", "block_id"), names(object$sample_map))
  n_groups <- if (length(group_col) == 0L) NA_integer_ else length(unique(object$sample_map[[group_col[[1L]]]]))
  list(
    constraint_id = object$constraint_id,
    relation_types = object$relation_types,
    n_samples = nrow(object$sample_map),
    n_groups = n_groups,
    transitive = object$transitive
  )
}

#' @export
as.data.frame.dependency_constraint <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$sample_map
}

#' @export
print.split_constraint <- function(x, ...) {
  cat("<split_constraint>", x$strategy, "\n")
  cat("  Samples:", nrow(x$sample_map), "\n")
  if ("group_id" %in% names(x$sample_map)) {
    cat("  Groups:", length(unique(x$sample_map$group_id)), "\n")
  }
  warnings <- x$metadata$warnings %||% character()
  if (length(warnings) > 0L) {
    cat("  Warnings:", length(warnings), "\n")
  }
  invisible(x)
}

#' @export
summary.split_constraint <- function(object, ...) {
  warnings <- object$metadata$warnings %||% character()
  list(
    strategy = object$strategy,
    mode = object$metadata$mode %||% NA_character_,
    n_samples = nrow(object$sample_map),
    n_groups = if ("group_id" %in% names(object$sample_map)) length(unique(object$sample_map$group_id)) else NA_integer_,
    n_warnings = length(warnings),
    columns = names(object$sample_map)
  )
}

#' @export
as.data.frame.split_constraint <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$sample_map
}

#' @export
print.depgraph_validation_report <- function(x, ...) {
  cat("<depgraph_validation_report>")
  if (!is.null(x$graph_name) && nzchar(x$graph_name)) {
    cat(" ", x$graph_name)
  }
  cat(
    "\n  Valid:", x$valid,
    "\n  Issues:", nrow(x$issues),
    "\n"
  )

  severity_counts <- x$summary$by_severity
  if (nrow(severity_counts) > 0L) {
    cat("  By severity:\n")
    for (i in seq_len(nrow(severity_counts))) {
      cat("   -", severity_counts$value[[i]], ":", severity_counts$n[[i]], "\n")
    }
  }
  invisible(x)
}

#' @export
summary.depgraph_validation_report <- function(object, ...) {
  object$summary
}

#' @export
as.data.frame.depgraph_validation_report <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$issues
}

#' @export
print.split_spec <- function(x, ...) {
  cat("<split_spec>", x$constraint_mode %||% "<unknown>", "\n")
  cat("  Samples:", nrow(x$sample_data), "\n")
  cat("  Groups:", length(unique(x$sample_data[[x$group_var]])), "\n")
  if (!is.null(x$recommended_resampling)) {
    cat("  Recommended resampling:", x$recommended_resampling, "\n")
  }
  invisible(x)
}

#' @export
summary.split_spec <- function(object, ...) {
  list(
    constraint_mode = object$constraint_mode,
    constraint_strategy = object$constraint_strategy,
    n_samples = nrow(object$sample_data),
    n_groups = length(unique(object$sample_data[[object$group_var]])),
    block_vars = object$block_vars,
    time_var = object$time_var,
    ordering_required = object$ordering_required,
    recommended_resampling = object$recommended_resampling
  )
}

#' @export
as.data.frame.split_spec <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$sample_data
}

#' @export
print.split_spec_validation <- function(x, ...) {
  cat("<split_spec_validation>\n")
  cat("  Valid:", x$valid, "\n")
  cat("  Issues:", nrow(x$issues), "\n")
  invisible(x)
}

#' @export
summary.split_spec_validation <- function(object, ...) {
  object$summary
}

#' @export
as.data.frame.split_spec_validation <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$issues
}

#' @export
print.leakage_risk_summary <- function(x, ...) {
  cat("<leakage_risk_summary>\n")
  cat("  Overview:", paste(x$overview, collapse = " "), "\n")
  cat("  Diagnostics:", nrow(x$diagnostics), "\n")
  invisible(x)
}

#' @export
summary.leakage_risk_summary <- function(object, ...) {
  list(
    overview = object$overview,
    n_diagnostics = nrow(object$diagnostics),
    by_source = .depgraph_count_table(if (nrow(object$diagnostics) == 0L) character() else object$diagnostics$source),
    by_severity = .depgraph_count_table(if (nrow(object$diagnostics) == 0L) character() else object$diagnostics$severity)
  )
}

#' @export
as.data.frame.leakage_risk_summary <- function(x, row.names = NULL, optional = FALSE, ...) {
  x$diagnostics
}

`%||%` <- function(x, y) {
  if (is.null(x) || length(x) == 0L || (length(x) == 1L && is.na(x))) {
    return(y)
  }
  x
}

# Split-constraint derivation for dependency_graph objects.

.depgraph_constraint_mode_map <- c(
  subject = "Subject",
  batch = "Batch",
  study = "Study",
  time = "Timepoint",
  site = "Site",
  region = "Region",
  platform = "Platform",
  assay = "Assay"
)

.depgraph_constraint_edge_map <- c(
  subject = "sample_belongs_to_subject",
  batch = "sample_processed_in_batch",
  study = "sample_from_study",
  time = "sample_collected_at_timepoint",
  site = "sample_collected_at_site",
  region = "sample_located_in_region",
  platform = "sample_run_on_platform",
  assay = "sample_measured_by_assay"
)

.depgraph_normalize_constraint_mode <- function(mode) {
  mode <- tolower(as.character(mode)[1L])
  .depgraph_assert(mode %in% c("subject", "batch", "study", "time", "site", "region", "platform", "assay", "relatedness", "spatial", "composite"), paste0("Unsupported constraint mode: ", mode))
  mode
}

.depgraph_normalize_constraint_modes <- function(modes) {
  modes <- unique(tolower(as.character(modes)))
  .depgraph_assert(length(modes) > 0L, "At least one constraint mode is required.")
  .depgraph_assert(all(modes %in% names(.depgraph_constraint_mode_map)), paste0(
    "Unsupported constraint mode(s): ",
    paste(setdiff(modes, names(.depgraph_constraint_mode_map)), collapse = ", ")
  ))
  modes
}

.depgraph_constraint_samples <- function(graph, samples = NULL) {
  sample_ids <- .depgraph_resolve_sample_node_ids(graph, samples)
  sample_nodes <- graph$nodes$data[graph$nodes$data$node_type == "Sample", , drop = FALSE]
  sample_nodes[match(sample_ids, sample_nodes$node_id), , drop = FALSE]
}

.depgraph_direct_assignment <- function(graph, mode, samples = NULL) {
  mode <- .depgraph_normalize_constraint_mode(mode)
  .depgraph_assert(mode != "composite", "Composite assignments must be derived separately.")

  sample_nodes <- .depgraph_constraint_samples(graph, samples)
  node_data <- graph$nodes$data
  edge_data <- graph$edges$data
  edge_type <- .depgraph_constraint_edge_map[[mode]]
  target_type <- .depgraph_constraint_mode_map[[mode]]

  edges <- edge_data[
    edge_data$edge_type == edge_type &
      edge_data$from %in% sample_nodes$node_id,
    c("from", "to", "edge_type"),
    drop = FALSE
  ]

  duplicate_sources <- unique(edges$from[duplicated(edges$from)])
  if (length(duplicate_sources) > 0L) {
    duplicate_samples <- sample_nodes$node_key[match(duplicate_sources, sample_nodes$node_id)]
    allow_multi <- identical(mode, "subject") &&
      .depgraph_validation_override(graph, "allow_multi_subject_samples", default = FALSE)
    if (!allow_multi) {
      hint <- if (identical(mode, "subject")) {
        " (set `validation_overrides = list(allow_multi_subject_samples = TRUE)` to allow this and pick the first listed assignment.)"
      } else {
        ""
      }
      stop(
        paste0(
          "Multiple ", mode, " assignments found for sample(s): ",
          paste(duplicate_samples, collapse = ", "),
          hint
        ),
        call. = FALSE
      )
    }
    multi_assignment_samples <- duplicate_samples
  } else {
    multi_assignment_samples <- character()
  }

  edges <- edges[!duplicated(edges$from), , drop = FALSE]

  target_rows <- node_data[node_data$node_id %in% edges$to, c("node_id", "node_key", "label", "attrs"), drop = FALSE]
  target_map <- stats::setNames(split(target_rows, target_rows$node_id), names(split(target_rows, target_rows$node_id)))

  rows <- lapply(seq_len(nrow(sample_nodes)), function(i) {
    sample_row <- sample_nodes[i, , drop = FALSE]
    edge_row <- edges[edges$from == sample_row$node_id, , drop = FALSE]

    if (nrow(edge_row) == 0L) {
      return(data.frame(
        sample_id = sample_row$node_key,
        sample_node_id = sample_row$node_id,
        linked_node_id = NA_character_,
        linked_node_type = target_type,
        linked_key = NA_character_,
        linked_label = NA_character_,
        edge_type = edge_type,
        stringsAsFactors = FALSE
      ))
    }

    target_row <- target_map[[edge_row$to[[1L]]]]
    data.frame(
      sample_id = sample_row$node_key,
      sample_node_id = sample_row$node_id,
      linked_node_id = target_row$node_id[[1L]],
      linked_node_type = target_type,
      linked_key = target_row$node_key[[1L]],
      linked_label = target_row$label[[1L]],
      edge_type = edge_row$edge_type[[1L]],
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, rows)
  row.names(out) <- NULL
  attr(out, "multi_assignment_samples") <- multi_assignment_samples
  out
}

.depgraph_warning_if_missing <- function(assignments, mode) {
  missing_samples <- assignments$sample_id[is.na(assignments$linked_node_id)]
  if (length(missing_samples) == 0L) {
    return(character())
  }

  paste0(
    "Samples missing ", mode, " assignments were retained as singleton groups: ",
    paste(missing_samples, collapse = ", ")
  )
}

.depgraph_build_sample_map <- function(assignments, mode, explanation_prefix = NULL) {
  mode <- .depgraph_normalize_constraint_mode(mode)

  rows <- lapply(seq_len(nrow(assignments)), function(i) {
    row <- assignments[i, , drop = FALSE]
    linked <- !is.na(row$linked_node_id[[1L]]) && nzchar(row$linked_node_id[[1L]])
    group_id <- if (linked) {
      paste0(mode, ":", row$linked_key[[1L]])
    } else {
      paste0(mode, ":unlinked:", row$sample_id[[1L]])
    }
    group_label <- if (linked) row$linked_key[[1L]] else paste0("unlinked_", row$sample_id[[1L]])
    explanation <- if (linked) {
      paste0(
        explanation_prefix %||% paste0("Grouped by ", mode),
        " through ", row$edge_type[[1L]],
        " -> ", row$linked_key[[1L]], "."
      )
    } else {
      paste0(
        "No ", mode, " assignment was available; sample retained as an unlinked singleton group."
      )
    }

    data.frame(
      sample_id = row$sample_id,
      sample_node_id = row$sample_node_id,
      group_id = group_id,
      constraint_type = mode,
      group_label = group_label,
      explanation = explanation,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, rows)
  row.names(out) <- NULL
  out
}

.depgraph_compute_time_order <- function(graph, time_node_ids) {
  time_nodes <- graph$nodes$data[match(time_node_ids, graph$nodes$data$node_id), , drop = FALSE]
  raw_time_index <- vapply(
    time_nodes$attrs,
    function(x) {
      value <- .depgraph_extract_attr(x, "time_index")
      if (length(value) == 0L || is.na(value)) {
        return(NA_real_)
      }
      suppressWarnings(as.numeric(as.character(value)))
    },
    FUN.VALUE = numeric(1)
  )

  warnings <- character()
  order_rank <- rep(NA_integer_, length(time_node_ids))
  names(order_rank) <- time_node_ids
  names(raw_time_index) <- time_node_ids

  if (length(time_node_ids) == 0L) {
    return(list(
      source = "none",
      time_index = raw_time_index,
      order_rank = order_rank,
      warnings = warnings
    ))
  }

  time_consistency <- .depgraph_time_order_consistency(graph$nodes$data, graph$edges$data)
  if (length(time_consistency$self_loop_edge_ids) > 0L || length(time_consistency$cycle_edge_ids) > 0L) {
    stop(
      "Time ordering metadata are inconsistent: `timepoint_precedes` must define an acyclic ordering.",
      call. = FALSE
    )
  }
  if (nrow(time_consistency$conflicting_pairs) > 0L) {
    pair_labels <- paste0(
      time_consistency$conflicting_pairs$from,
      " -> ",
      time_consistency$conflicting_pairs$to
    )
    stop(
      paste0(
        "Time ordering metadata conflict between `time_index` and `timepoint_precedes`: ",
        paste(pair_labels, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  if (all(!is.na(raw_time_index))) {
    unique_order <- sort(unique(raw_time_index))
    rank_map <- stats::setNames(seq_along(unique_order), as.character(unique_order))
    order_rank <- as.integer(rank_map[as.character(raw_time_index)])
    names(order_rank) <- time_node_ids
    return(list(
      source = "time_index",
      time_index = raw_time_index,
      order_rank = order_rank,
      warnings = warnings
    ))
  }

  precedence_edges <- graph$edges$data[
    graph$edges$data$edge_type == "timepoint_precedes" &
      graph$edges$data$from %in% time_node_ids &
      graph$edges$data$to %in% time_node_ids,
    c("from", "to"),
    drop = FALSE
  ]

  if (nrow(precedence_edges) > 0L) {
    precedence_graph <- igraph::graph_from_data_frame(
      d = precedence_edges,
      vertices = data.frame(name = time_node_ids, stringsAsFactors = FALSE),
      directed = TRUE
    )
    if (igraph::is_dag(precedence_graph)) {
      topo <- igraph::as_ids(igraph::topo_sort(precedence_graph, mode = "out"))
      order_rank[topo] <- seq_along(topo)
      return(list(
        source = "timepoint_precedes",
        time_index = raw_time_index,
        order_rank = order_rank,
        warnings = warnings
      ))
    }

    warnings <- c(warnings, "Timepoint precedence edges are cyclic; ordering metadata could not be derived from graph structure.")
  }

  if (any(!is.na(raw_time_index))) {
    available <- sort(unique(raw_time_index[!is.na(raw_time_index)]))
    rank_map <- stats::setNames(seq_along(available), as.character(available))
    order_rank[!is.na(raw_time_index)] <- as.integer(rank_map[as.character(raw_time_index[!is.na(raw_time_index)])])
    warnings <- c(warnings, "Time ordering is only partially available because some timepoints are missing `time_index` values.")
  } else {
    warnings <- c(warnings, "Time ordering is unavailable because neither `time_index` metadata nor usable `timepoint_precedes` edges were found.")
  }

  list(
    source = "partial",
    time_index = raw_time_index,
    order_rank = order_rank,
    warnings = warnings
  )
}

.derive_subject_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "subject", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "subject")
  multi_subject_samples <- attr(assignments, "multi_assignment_samples", exact = TRUE)
  if (!is.null(multi_subject_samples) && length(multi_subject_samples) > 0L) {
    warnings <- c(
      warnings,
      paste0(
        "Samples linked to multiple subjects were assigned to their first listed subject ",
        "(allow_multi_subject_samples override): ",
        paste(multi_subject_samples, collapse = ", ")
      )
    )
  }
  sample_map <- .depgraph_build_sample_map(assignments, "subject")

  split_constraint(
    strategy = "subject",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = NULL,
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "subject",
      strategy = "subject",
      relations_used = "sample_belongs_to_subject",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_batch_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "batch", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "batch")
  sample_map <- .depgraph_build_sample_map(assignments, "batch")

  split_constraint(
    strategy = "batch",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "batch",
      strategy = "batch",
      relations_used = "sample_processed_in_batch",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_study_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "study", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "study")
  sample_map <- .depgraph_build_sample_map(assignments, "study")

  split_constraint(
    strategy = "study",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "study",
      strategy = "study",
      relations_used = "sample_from_study",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_site_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "site", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "site")
  sample_map <- .depgraph_build_sample_map(assignments, "site")

  split_constraint(
    strategy = "site",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "site",
      strategy = "site",
      relations_used = "sample_collected_at_site",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_region_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "region", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "region")
  sample_map <- .depgraph_build_sample_map(assignments, "region")

  split_constraint(
    strategy = "region",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "region",
      strategy = "region",
      relations_used = "sample_located_in_region",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_platform_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "platform", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "platform")
  sample_map <- .depgraph_build_sample_map(assignments, "platform")

  split_constraint(
    strategy = "platform",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "platform",
      strategy = "platform",
      relations_used = "sample_run_on_platform",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_assay_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "assay", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "assay")
  sample_map <- .depgraph_build_sample_map(assignments, "assay")

  split_constraint(
    strategy = "assay",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "assay",
      strategy = "assay",
      relations_used = "sample_measured_by_assay",
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

.derive_time_constraints <- function(graph, samples = NULL) {
  assignments <- .depgraph_direct_assignment(graph, "time", samples = samples)
  warnings <- .depgraph_warning_if_missing(assignments, "timepoint")
  valid_time_nodes <- stats::na.omit(assignments$linked_node_id)
  order_info <- .depgraph_compute_time_order(graph, unique(valid_time_nodes))
  warnings <- unique(c(warnings, order_info$warnings))

  order_rank <- order_info$order_rank[assignments$linked_node_id]
  order_rank[is.na(assignments$linked_node_id)] <- NA_integer_
  time_index <- order_info$time_index[assignments$linked_node_id]
  time_index[is.na(assignments$linked_node_id)] <- NA_real_
  timepoint_id <- assignments$linked_key

  sample_map <- .depgraph_build_sample_map(assignments, "time")
  sample_map$time_index <- unname(time_index)
  sample_map$timepoint_id <- timepoint_id
  sample_map$order_rank <- unname(as.integer(order_rank))
  has_complete_ordering <- nrow(sample_map) > 0L && all(!is.na(sample_map$order_rank))

  split_constraint(
    strategy = "time",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = NULL,
      time_var = if (any(!is.na(sample_map$order_rank))) "order_rank" else NULL,
      ordering_required = has_complete_ordering
    ),
    metadata = list(
      mode = "time",
      strategy = "time",
      relations_used = c("sample_collected_at_timepoint", if (identical(order_info$source, "timepoint_precedes")) "timepoint_precedes" else NULL),
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      time_order_source = order_info$source,
      warnings = warnings
    )
  )
}

.depgraph_normalize_via_modes <- function(via) {
  if (is.null(via)) {
    return(c("subject", "batch", "study", "time"))
  }

  via <- as.character(via)
  via_modes <- vapply(via, function(x) {
    x_lower <- tolower(x)
    if (x_lower %in% names(.depgraph_constraint_mode_map)) {
      return(x_lower)
    }

    match_idx <- match(x_lower, tolower(.depgraph_constraint_mode_map))
    .depgraph_assert(!is.na(match_idx), paste0("Unsupported composite dependency source: ", x))
    names(.depgraph_constraint_mode_map)[[match_idx]]
  }, character(1), USE.NAMES = FALSE)

  unique(via_modes)
}

.derive_composite_strict_constraints <- function(graph, samples = NULL, via = NULL) {
  via_modes <- .depgraph_normalize_via_modes(via)
  via_types <- unname(.depgraph_constraint_mode_map[via_modes])
  components <- detect_dependency_components(graph, via = via_types, min_size = 1)
  table <- components$table

  if (!is.null(samples)) {
    sample_nodes <- .depgraph_constraint_samples(graph, samples)
    keep_ids <- sample_nodes$node_id

    # Recompute components within the requested subset only. Without this,
    # two in-subset samples connected only through an out-of-subset sample
    # would inherit a shared component_id from the full-graph projection,
    # silently leaking out-of-subset structure into the produced split.
    projection <- components$metadata$projection_edges
    if (is.null(projection) || nrow(projection) == 0L) {
      projection <- data.frame(
        sample_node_id_1 = character(0),
        sample_node_id_2 = character(0),
        stringsAsFactors = FALSE
      )
    } else {
      projection <- projection[
        projection$sample_node_id_1 %in% keep_ids &
          projection$sample_node_id_2 %in% keep_ids,
        , drop = FALSE
      ]
    }

    subset_graph <- if (nrow(projection) == 0L) {
      igraph::make_empty_graph(n = length(keep_ids), directed = FALSE)
    } else {
      igraph::graph_from_data_frame(
        d = data.frame(
          from = projection$sample_node_id_1,
          to = projection$sample_node_id_2,
          stringsAsFactors = FALSE
        ),
        vertices = data.frame(name = keep_ids, stringsAsFactors = FALSE),
        directed = FALSE
      )
    }
    igraph::V(subset_graph)$name <- keep_ids

    sub_components <- igraph::components(subset_graph)
    membership_idx <- as.integer(sub_components$membership[keep_ids])
    table <- data.frame(
      sample_id = sample_nodes$node_key,
      sample_node_id = keep_ids,
      component_id = paste0("component_", membership_idx),
      component_size = as.integer(sub_components$csize[membership_idx]),
      stringsAsFactors = FALSE
    )
    components$metadata$projection_edges <- projection
  }

  sample_map <- data.frame(
    sample_id = table$sample_id,
    sample_node_id = table$sample_node_id,
    group_id = table$component_id,
    constraint_type = "composite_strict",
    group_label = table$component_id,
    explanation = paste0(
      "Strict composite grouping via transitive closure over: ",
      paste(via_types, collapse = ", "),
      "."
    ),
    stringsAsFactors = FALSE
  )

  warnings <- character()
  if (nrow(sample_map) > 0L) {
    singleton_fraction <- mean(table$component_size == 1L)
    if (singleton_fraction > 0.5) {
      warnings <- c(warnings, "Most strict composite groups are singletons; dependency coverage may be sparse.")
    }
  }

  split_constraint(
    strategy = "strict",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = NULL,
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = "composite",
      strategy = "strict",
      relations_used = unname(.depgraph_constraint_edge_map[via_modes]),
      via = via_types,
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings,
      projection_edges = components$metadata$projection_edges
    )
  )
}

.derive_rule_based_constraints <- function(graph, samples = NULL, priority = NULL, via = NULL) {
  via_modes <- .depgraph_normalize_via_modes(via)
  priority <- if (is.null(priority)) via_modes else .depgraph_normalize_constraint_modes(priority)
  .depgraph_assert(all(priority %in% via_modes), "`priority` must be a subset of the selected composite dependency sources.")

  sample_nodes <- .depgraph_constraint_samples(graph, samples)
  assignments <- lapply(priority, function(mode) .depgraph_direct_assignment(graph, mode, samples = sample_nodes$node_id))
  names(assignments) <- priority
  time_info <- if ("time" %in% priority) .derive_time_constraints(graph, samples = sample_nodes$node_id)$sample_map else NULL

  rows <- lapply(seq_len(nrow(sample_nodes)), function(i) {
    sample_row <- sample_nodes[i, , drop = FALSE]
    sample_id <- sample_row$node_key[[1L]]
    sample_node_id <- sample_row$node_id[[1L]]

    available <- lapply(assignments, function(x) x[x$sample_node_id == sample_node_id, , drop = FALSE])
    available_modes <- names(available)[vapply(available, function(x) nrow(x) == 1L && !is.na(x$linked_node_id[[1L]]), logical(1))]

    if (length(available_modes) == 0L) {
      return(data.frame(
        sample_id = sample_id,
        sample_node_id = sample_node_id,
        group_id = paste0("composite:unlinked:", sample_id),
        constraint_type = "unlinked",
        group_label = paste0("unlinked_", sample_id),
        explanation = "No prioritized dependency assignment was available; sample retained as a singleton group.",
        time_index = NA_real_,
        timepoint_id = NA_character_,
        order_rank = NA_integer_,
        stringsAsFactors = FALSE
      ))
    }

    chosen_mode <- available_modes[[1L]]
    chosen_row <- available[[chosen_mode]]
    additional <- setdiff(available_modes, chosen_mode)
    explanation <- paste0(
      "Composite rule-based grouping selected ", chosen_mode,
      " based on priority order ", paste(priority, collapse = " > "),
      " -> ", chosen_row$linked_key[[1L]], "."
    )
    if (length(additional) > 0L) {
      detail <- vapply(additional, function(mode) {
        paste0(mode, "=", available[[mode]]$linked_key[[1L]])
      }, character(1), USE.NAMES = FALSE)
      explanation <- paste0(explanation, " Additional available dependencies: ", paste(detail, collapse = ", "), ".")
    }

    time_row <- if (!is.null(time_info)) time_info[time_info$sample_node_id == sample_node_id, , drop = FALSE] else NULL

    data.frame(
      sample_id = sample_id,
      sample_node_id = sample_node_id,
      group_id = paste0("composite_", chosen_mode, ":", chosen_row$linked_key[[1L]]),
      constraint_type = chosen_mode,
      group_label = chosen_row$linked_key[[1L]],
      explanation = explanation,
      time_index = if (!is.null(time_row) && nrow(time_row) == 1L) time_row$time_index[[1L]] else NA_real_,
      timepoint_id = if (!is.null(time_row) && nrow(time_row) == 1L) time_row$timepoint_id[[1L]] else NA_character_,
      order_rank = if (!is.null(time_row) && nrow(time_row) == 1L) time_row$order_rank[[1L]] else NA_integer_,
      stringsAsFactors = FALSE
    )
  })

  sample_map <- do.call(rbind, rows)
  row.names(sample_map) <- NULL
  warnings <- character()
  if (any(sample_map$constraint_type == "unlinked")) {
    warnings <- c(warnings, paste0(
      "Some samples did not match any prioritized dependency and were retained as singleton groups: ",
      paste(sample_map$sample_id[sample_map$constraint_type == "unlinked"], collapse = ", ")
    ))
  }

  split_constraint(
    strategy = "rule_based",
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = NULL,
      time_var = if (any(!is.na(sample_map$order_rank))) "order_rank" else NULL,
      ordering_required = nrow(sample_map) > 0L && all(!is.na(sample_map$order_rank))
    ),
    metadata = list(
      mode = "composite",
      strategy = "rule_based",
      relations_used = unname(.depgraph_constraint_edge_map[priority]),
      priority = priority,
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings
    )
  )
}

#' Derive Split Constraints from Dependency Graphs
#'
#' Convert dataset dependency structure into deterministic sample-level
#' grouping constraints suitable for leakage-aware evaluation design.
#'
#' Constraint derivation rules:
#'
#' \describe{
#'   \item{\code{mode = "subject"}}{Groups samples by the target of
#'   \code{sample_belongs_to_subject}. All samples linked to the same
#'   \code{Subject} receive the same \code{group_id}.}
#'
#'   \item{\code{mode = "batch"}}{Groups samples by the target of
#'   \code{sample_processed_in_batch}. Samples with no batch assignment are
#'   retained as singleton unlinked groups and recorded in metadata warnings.}
#'
#'   \item{\code{mode = "study"}}{Groups samples by the target of
#'   \code{sample_from_study}.}
#'
#'   \item{\code{mode = "site"}}{Groups samples by the target of
#'   \code{sample_collected_at_site}. Samples with no site assignment are
#'   retained as singleton unlinked groups and recorded in metadata warnings.}
#'
#'   \item{\code{mode = "region"}}{Groups samples by the target of
#'   \code{sample_located_in_region} (e.g. a categorical tissue or anatomical
#'   region). Samples with no region assignment are retained as singleton
#'   unlinked groups and recorded in metadata warnings.}
#'
#'   \item{\code{mode = "platform"}}{Groups samples by the target of
#'   \code{sample_run_on_platform} (the sequencing / measurement platform or
#'   instrument). Samples with no platform assignment are retained as singleton
#'   unlinked groups and recorded in metadata warnings.}
#'
#'   \item{\code{mode = "assay"}}{Groups samples by the target of
#'   \code{sample_measured_by_assay} (the assay / modality). Samples with no
#'   assay assignment are retained as singleton unlinked groups and recorded in
#'   metadata warnings.}
#'
#'   \item{\code{mode = "relatedness"}}{Groups samples by transitive closure
#'   over thresholded \code{subject_related_to} edges (genetic relatedness).
#'   Samples that share a subject, or whose subjects are directly or indirectly
#'   related above threshold, land in the same connected-component group. Build
#'   the edges with \code{\link{relatedness_edges_from_kinship}}. Samples with
#'   no subject are retained as singleton groups (recorded in metadata
#'   warnings).}
#'
#'   \item{\code{mode = "spatial"}}{Groups samples by transitive closure over
#'   thresholded \code{sample_adjacent_to} edges (spatial proximity). Build the
#'   edges with \code{\link{spatial_edges_from_coords}}. Isolated samples form
#'   singleton groups.}
#'
#'   \item{\code{mode = "time"}}{Groups samples by the target of
#'   \code{sample_collected_at_timepoint}. When \code{Timepoint} nodes have
#'   \code{time_index} metadata, that value is used to derive
#'   \code{order_rank}. If \code{time_index} is unavailable, the function
#'   attempts to derive ordering from \code{timepoint_precedes} edges over the
#'   timepoint subgraph.}
#'
#'   \item{\code{mode = "composite"}, \code{strategy = "strict"}}{Projects the
#'   selected dependency relations onto a sample graph and assigns one
#'   \code{group_id} per connected component. This is the transitive-closure
#'   interpretation of composite dependency grouping.}
#'
#'   \item{\code{mode = "composite"}, \code{strategy = "rule_based"}}{Evaluates
#'   dependency assignments in deterministic priority order and groups each
#'   sample by the highest-priority available dependency source.
#'   Lower-priority available dependencies are retained in the explanation
#'   field.}
#' }
#'
#' The returned \code{split_constraint$sample_map} always contains
#' \code{sample_id}, \code{sample_node_id}, \code{group_id},
#' \code{constraint_type}, \code{group_label}, and \code{explanation}.
#' Time-aware constraints also include \code{time_index}, \code{timepoint_id},
#' and \code{order_rank} when available.
#'
#' Ambiguous direct assignments are rejected. A sample cannot be assigned to
#' multiple batches, studies, or timepoints when deriving direct split
#' constraints.
#'
#' @param graph A \code{dependency_graph}.
#' @param mode Constraint derivation mode.
#' @param samples Optional sample identifiers or sample node IDs used to
#'   restrict the returned \code{sample_map}. All requested samples must
#'   resolve successfully.
#' @param strategy Composite grouping strategy. Ignored for non-composite
#'   modes.
#' @param via Optional dependency sources used for composite grouping. May be
#'   given as lower-case modes such as \code{"subject"} or node types such as
#'   \code{"Subject"}.
#' @param priority Optional priority order used for
#'   \code{strategy = "rule_based"}.
#' @param include_warnings Whether to retain human-readable warnings in the
#'   returned metadata.
#' @param x A \code{split_constraint}.
#' @return \code{derive_split_constraints()} returns a \code{split_constraint}
#'   whose \code{sample_map} contains grouping assignments and, for time-aware
#'   constraints, ordering metadata. \code{grouping_vector()} returns a named
#'   character vector of \code{group_id} values keyed by \code{sample_id}.
#' @examples
#' meta <- data.frame(
#'   sample_id  = c("S1", "S2", "S3", "S4"),
#'   subject_id = c("P1", "P1", "P2", "P2"),
#'   batch_id   = c("B1", "B2", "B1", "B2")
#' )
#' g <- graph_from_metadata(meta)
#'
#' constraint <- derive_split_constraints(g, mode = "subject")
#' grouping_vector(constraint)
#' @export
derive_split_constraints <- function(graph, mode = c("subject", "batch", "study", "time", "site", "region", "platform", "assay", "relatedness", "spatial", "composite"), samples = NULL, strategy = c("strict", "rule_based"), via = NULL, priority = NULL, include_warnings = TRUE) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  mode <- .depgraph_normalize_constraint_mode(match.arg(mode))
  strategy <- match.arg(strategy)

  result <- switch(
    mode,
    subject = .derive_subject_constraints(graph, samples = samples),
    batch = .derive_batch_constraints(graph, samples = samples),
    study = .derive_study_constraints(graph, samples = samples),
    time = .derive_time_constraints(graph, samples = samples),
    site = .derive_site_constraints(graph, samples = samples),
    region = .derive_region_constraints(graph, samples = samples),
    platform = .derive_platform_constraints(graph, samples = samples),
    assay = .derive_assay_constraints(graph, samples = samples),
    relatedness = .derive_pairwise_constraints(graph, "relatedness", samples = samples),
    spatial = .derive_pairwise_constraints(graph, "spatial", samples = samples),
    composite = if (identical(strategy, "strict")) {
      .derive_composite_strict_constraints(graph, samples = samples, via = via)
    } else {
      .derive_rule_based_constraints(graph, samples = samples, priority = priority, via = via)
    }
  )

  if (!isTRUE(include_warnings)) {
    result$metadata$warnings <- character()
  }

  result
}

#' @rdname derive_split_constraints
#' @export
grouping_vector <- function(x) {
  .depgraph_assert(inherits(x, "split_constraint"), "`x` must be a `split_constraint`.")
  groups <- as.character(x$sample_map$group_id)
  names(groups) <- x$sample_map$sample_id
  groups
}

# Boundary contract with bioLeak, the primary downstream consumer. splitGraph
# emits a `split_spec`; bioLeak::as_leaksplits() consumes it. This test pins the
# seam so neither side can silently break it: it asserts that a splitGraph
# split_spec satisfies the columns and fields bioLeak reads
# (`sample_id`, `group_var`, and the optional `batch_group` / `study_group` /
# `timepoint_id` / `order_rank` block columns). Skipped when bioLeak is absent.

skip_if_no_bioleak <- function() {
  testthat::skip_if_not_installed("bioLeak")
}

make_contract_graph <- function() {
  meta <- data.frame(
    sample_id  = paste0("S", 1:12),
    subject_id = rep(paste0("P", 1:6), each = 2),
    batch_id   = rep(c("B1", "B2", "B3"), times = 4),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "contract-demo")
}

contract_data <- function(n = 12) {
  set.seed(1)
  data.frame(
    sample_id = paste0("S", seq_len(n)),
    y = rbinom(n, 1, 0.5),
    stringsAsFactors = FALSE
  )
}

test_that("a subject-mode split_spec is accepted by bioLeak::as_leaksplits()", {
  skip_if_no_bioleak()
  g <- make_contract_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)

  ls <- bioLeak::as_leaksplits(spec, data = contract_data(), outcome = "y", v = 3)
  expect_s4_class(ls, "LeakSplits")
})

test_that("a batch-mode split_spec forwards its block column to bioLeak", {
  skip_if_no_bioleak()
  g <- make_contract_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "batch"), graph = g)

  # batch_group must be present and populated for bioLeak to pick it up.
  expect_true("batch_group" %in% names(spec$sample_data))
  expect_false(all(is.na(spec$sample_data$batch_group)))

  ls <- bioLeak::as_leaksplits(spec, data = contract_data(), outcome = "y", v = 3)
  expect_s4_class(ls, "LeakSplits")
})

test_that("newer additive columns do not break the bioLeak seam", {
  skip_if_no_bioleak()
  # A subject-mode spec on a graph that ALSO carries site structure: the
  # constraint_mode stays in bioLeak's supported set, while sample_data gains a
  # site_group annotation (a column bioLeak does not read). The seam must hold:
  # bioLeak keys on sample_id + group_var and ignores the extra annotation.
  meta <- data.frame(
    sample_id  = paste0("S", 1:12),
    subject_id = rep(paste0("P", 1:6), each = 2),
    site_id    = rep(c("NYC", "BOS", "SFO"), times = 4),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
  expect_true("site_group" %in% names(spec$sample_data))
  expect_false(all(is.na(spec$sample_data$site_group)))

  ls <- bioLeak::as_leaksplits(spec, data = contract_data(), outcome = "y", v = 3)
  expect_s4_class(ls, "LeakSplits")
})

# Cross-language conformance: the shipped Python reference consumer must
# recover exactly the grouping and ordering that R produced. Skipped unless a
# python3 interpreter is available (and never run on CRAN), so the R package
# check stays self-contained while CI / local runs still exercise the seam.

skip_if_no_python_conformance <- function() {
  testthat::skip_if_not_installed("jsonlite")
  testthat::skip_on_cran()
  py <- Sys.which("python3")
  if (!nzchar(py)) testthat::skip("python3 not available")
  script <- system.file("python", "conformance.py", package = "splitGraph")
  if (!nzchar(script) || !file.exists(script)) testthat::skip("conformance.py not found")
  list(py = py, script = script)
}

test_that("Python reference consumer reproduces R grouping and order_rank", {
  cfg <- skip_if_no_python_conformance()

  meta <- data.frame(
    sample_id    = c("S1", "S2", "S3", "S4"),
    subject_id   = c("P1", "P1", "P2", "P3"),
    timepoint_id = c("T0", "T1", "T0", "T2"),
    time_index   = c(0, 1, 0, 2),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  constraint <- derive_split_constraints(g, mode = "time")
  spec <- as_split_spec(constraint, graph = g)

  in_path  <- tempfile(fileext = ".json")
  out_path <- tempfile(fileext = ".json")
  on.exit(unlink(c(in_path, out_path)), add = TRUE)
  write_split_spec(spec, in_path)

  status <- suppressWarnings(system2(
    cfg$py, c("-B", shQuote(cfg$script), shQuote(in_path), shQuote(out_path)),
    stdout = TRUE, stderr = TRUE
  ))
  exit_code <- attr(status, "status")
  if (!is.null(exit_code) && exit_code != 0L) {
    testthat::skip(paste("python conformance run failed:", paste(status, collapse = " ")))
  }
  expect_true(file.exists(out_path))

  py <- jsonlite::fromJSON(out_path, simplifyVector = TRUE)

  # Schema version agrees.
  expect_identical(py$schema_version, splitGraph:::.depgraph_schema_version)

  # Grouping matches R's grouping_vector() exactly.
  r_grouping <- grouping_vector(constraint)
  py_grouping <- unlist(py$grouping)
  expect_identical(sort(names(py_grouping)), sort(names(r_grouping)))
  expect_identical(py_grouping[names(r_grouping)], r_grouping[names(r_grouping)])

  # order_rank matches spec sample_data exactly.
  r_order <- stats::setNames(spec$sample_data$order_rank, spec$sample_data$sample_id)
  py_order <- unlist(py$order_ranks)
  expect_equal(py_order[names(r_order)], r_order[names(r_order)],
               ignore_attr = TRUE)
})

# Validation of serialized splitGraph JSON against the shipped contract, plus a
# one-shot upgrader for files written under an older schema_version. The formal
# JSON Schemas live in `inst/schema/` (Draft 2020-12) and are the portable
# contract for non-R consumers; the checks below are a dependency-free R
# encoding of the same structural invariants so `read_*()` callers can validate
# a handoff file without pulling in a JSON Schema engine.

# Path to a shipped schema file within the installed package (or source tree
# under pkgload). Returns "" if it cannot be located.
.depgraph_schema_path <- function(object_type) {
  file <- paste0(object_type, ".schema.json")
  p <- system.file("schema", file, package = "splitGraph")
  if (nzchar(p)) p else ""
}

.depgraph_json_report <- function(object_type, path, issues) {
  structure(
    list(
      object_type = object_type,
      path = path,
      valid = length(issues) == 0L,
      issues = issues,
      schema = .depgraph_schema_url(object_type)
    ),
    class = "splitgraph_json_report"
  )
}

#' @export
print.splitgraph_json_report <- function(x, ...) {
  cat("<splitGraph JSON validation>\n")
  cat("  object:  ", x$object_type, "\n", sep = "")
  cat("  path:    ", x$path, "\n", sep = "")
  cat("  valid:   ", x$valid, "\n", sep = "")
  if (!x$valid) {
    cat("  issues:\n")
    for (msg in x$issues) cat("    - ", msg, "\n", sep = "")
  }
  invisible(x)
}

.depgraph_parse_json_file <- function(path) {
  .depgraph_require_jsonlite()
  .depgraph_assert(is.character(path) && length(path) == 1L && nzchar(path),
                   "`path` must be a single non-empty file path.")
  .depgraph_assert(file.exists(path), paste0("File not found: ", path))
  tryCatch(
    jsonlite::fromJSON(path, simplifyVector = FALSE),
    error = function(e) {
      stop("Failed to parse JSON at `", path, "`: ", conditionMessage(e), call. = FALSE)
    }
  )
}

.depgraph_is_string <- function(x) is.character(x) && length(x) == 1L && !is.na(x)

.depgraph_valid_version <- function(x) {
  .depgraph_is_string(x) && grepl("^[0-9]+\\.[0-9]+\\.[0-9]+$", x)
}

#' Validate Serialized splitGraph JSON Against the Shipped Schema
#'
#' Check that a JSON file written by \code{write_dependency_graph()} or
#' \code{write_split_spec()} conforms to the splitGraph on-disk contract. The
#' formal JSON Schemas (Draft 2020-12) ship in \code{inst/schema/} and are
#' referenced from the written JSON via the \code{$schema} key; these functions
#' apply a dependency-free structural check of the same invariants (required
#' fields, value types, node/edge-type enumerations, and referential integrity
#' of edge endpoints) so a handoff file can be validated without a JSON Schema
#' engine.
#'
#' @param path Path to a serialized \code{dependency_graph} or
#'   \code{split_spec} JSON file.
#' @return A \code{splitgraph_json_report}: a list with \code{valid} (logical),
#'   \code{issues} (character vector of failures), the detected
#'   \code{object_type}, and the schema \code{$id}.
#' @examples
#' if (requireNamespace("jsonlite", quietly = TRUE)) {
#'   meta <- data.frame(sample_id = c("S1", "S2"), subject_id = c("P1", "P2"))
#'   g <- graph_from_metadata(meta)
#'   tmp <- tempfile(fileext = ".json")
#'   write_dependency_graph(g, tmp)
#'   validate_graph_json(tmp)
#'   unlink(tmp)
#' }
#' @name validate_json
#' @export
validate_graph_json <- function(path) {
  parsed <- .depgraph_parse_json_file(path)
  issues <- character()

  if (!identical(as.character(parsed$splitGraph_object %||% NA_character_), "dependency_graph")) {
    issues <- c(issues, paste0(
      "`splitGraph_object` must be \"dependency_graph\" (found \"",
      parsed$splitGraph_object %||% "<missing>", "\")."
    ))
    return(.depgraph_json_report("dependency_graph", path, issues))
  }

  if (!.depgraph_valid_version(parsed$schema_version)) {
    issues <- c(issues, "`schema_version` must be a \"X.Y.Z\" string.")
  }

  node_rows <- parsed$nodes
  edge_rows <- parsed$edges
  if (!is.null(node_rows) && !is.list(node_rows)) {
    issues <- c(issues, "`nodes` must be an array.")
    node_rows <- list()
  }
  if (!is.null(edge_rows) && !is.list(edge_rows)) {
    issues <- c(issues, "`edges` must be an array.")
    edge_rows <- list()
  }

  node_ids <- character()
  for (i in seq_along(node_rows)) {
    r <- node_rows[[i]]
    if (!.depgraph_is_string(r$node_id) || !.depgraph_is_string(r$node_type) ||
        !.depgraph_is_string(r$node_key)) {
      issues <- c(issues, paste0("node[", i, "]: `node_id`, `node_type`, `node_key` are required strings."))
      next
    }
    node_ids <- c(node_ids, r$node_id)
    if (!r$node_type %in% .depgraph_node_types) {
      issues <- c(issues, paste0("node[", i, "]: unknown `node_type` \"", r$node_type, "\"."))
    }
  }

  valid_edge_types <- .depgraph_edge_schema$edge_type
  for (i in seq_along(edge_rows)) {
    r <- edge_rows[[i]]
    if (!.depgraph_is_string(r$edge_id) || !.depgraph_is_string(r$from) ||
        !.depgraph_is_string(r$to) || !.depgraph_is_string(r$edge_type)) {
      issues <- c(issues, paste0("edge[", i, "]: `edge_id`, `from`, `to`, `edge_type` are required strings."))
      next
    }
    if (!r$edge_type %in% valid_edge_types) {
      issues <- c(issues, paste0("edge[", i, "]: unknown `edge_type` \"", r$edge_type, "\"."))
    }
    if (length(node_ids) > 0L) {
      if (!r$from %in% node_ids) {
        issues <- c(issues, paste0("edge[", i, "]: `from` \"", r$from, "\" does not reference a declared node."))
      }
      if (!r$to %in% node_ids) {
        issues <- c(issues, paste0("edge[", i, "]: `to` \"", r$to, "\" does not reference a declared node."))
      }
    }
  }

  .depgraph_json_report("dependency_graph", path, issues)
}

#' @rdname validate_json
#' @export
validate_split_spec_json <- function(path) {
  parsed <- .depgraph_parse_json_file(path)
  issues <- character()

  if (!identical(as.character(parsed$splitGraph_object %||% NA_character_), "split_spec")) {
    issues <- c(issues, paste0(
      "`splitGraph_object` must be \"split_spec\" (found \"",
      parsed$splitGraph_object %||% "<missing>", "\")."
    ))
    return(.depgraph_json_report("split_spec", path, issues))
  }

  if (!.depgraph_valid_version(parsed$schema_version)) {
    issues <- c(issues, "`schema_version` must be a \"X.Y.Z\" string.")
  }
  if (!.depgraph_is_string(parsed$group_var)) {
    issues <- c(issues, "`group_var` must be a string.")
  }
  if (!is.null(parsed$block_vars) && !is.list(parsed$block_vars)) {
    issues <- c(issues, "`block_vars` must be an array of strings.")
  }

  sample_rows <- parsed$sample_data
  if (is.null(sample_rows) || !is.list(sample_rows)) {
    issues <- c(issues, "`sample_data` must be an array.")
    sample_rows <- list()
  }
  for (i in seq_along(sample_rows)) {
    r <- sample_rows[[i]]
    if (!.depgraph_is_string(r$sample_id)) {
      issues <- c(issues, paste0("sample_data[", i, "]: `sample_id` is a required string."))
    }
    if (!.depgraph_is_string(r$group_id)) {
      issues <- c(issues, paste0("sample_data[", i, "]: `group_id` is a required string."))
    }
  }

  .depgraph_json_report("split_spec", path, issues)
}

#' Upgrade Serialized splitGraph JSON to the Current Schema Version
#'
#' Read a \code{dependency_graph} or \code{split_spec} JSON file written under
#' an older \code{schema_version} and rewrite it at the installed version. The
#' round-trip fills any field introduced since the file was written with its
#' default (\code{NA} for missing \code{sample_data} columns), stamps the
#' current \code{schema_version}, and adds the \code{$schema} reference. Files
#' already at the current version are rewritten unchanged.
#'
#' @param path Path to the JSON file to upgrade.
#' @param out Path to write the upgraded file to. Defaults to \code{path}
#'   (in-place upgrade).
#' @return The output path, invisibly.
#' @examples
#' if (requireNamespace("jsonlite", quietly = TRUE)) {
#'   meta <- data.frame(sample_id = c("S1", "S2"), subject_id = c("P1", "P2"))
#'   g <- graph_from_metadata(meta)
#'   tmp <- tempfile(fileext = ".json")
#'   write_dependency_graph(g, tmp)
#'   migrate_dependency_graph_json(tmp)
#'   unlink(tmp)
#' }
#' @name migrate_json
#' @export
migrate_dependency_graph_json <- function(path, out = path) {
  g <- read_dependency_graph(path)
  write_dependency_graph(g, out)
  invisible(out)
}

#' @rdname migrate_json
#' @export
migrate_split_spec_json <- function(path, out = path) {
  spec <- read_split_spec(path)
  write_split_spec(spec, out)
  invisible(out)
}

skip_if_no_jsonlite <- function() testthat::skip_if_not_installed("jsonlite")

make_spec_graph <- function() {
  meta <- data.frame(
    sample_id  = c("S1", "S2", "S3"),
    subject_id = c("P1", "P1", "P2"),
    batch_id   = c("B1", "B2", "B1"),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "schema-demo")
}

# ---- shipped schema files ---------------------------------------------------

test_that("shipped JSON Schema files exist and are valid JSON", {
  skip_if_no_jsonlite()
  for (obj in c("dependency_graph", "split_spec")) {
    p <- splitGraph:::.depgraph_schema_path(obj)
    expect_true(nzchar(p), info = obj)
    schema <- jsonlite::fromJSON(p, simplifyVector = FALSE)
    expect_identical(schema$`$schema`, "https://json-schema.org/draft/2020-12/schema")
    expect_true(nzchar(schema$`$id`))
  }
})

# ---- $schema reference + version stamping -----------------------------------

test_that("written JSON carries a $schema reference and the current version", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)

  raw <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)
  expect_match(raw$`$schema`, "dependency_graph\\.schema\\.json$")
  expect_identical(raw$schema_version, "0.2.0")
  expect_identical(raw$schema_version, splitGraph:::.depgraph_schema_version)
})

# ---- validate_graph_json ----------------------------------------------------

test_that("validate_graph_json passes a well-formed graph", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)

  report <- validate_graph_json(tmp)
  expect_s3_class(report, "splitgraph_json_report")
  expect_true(report$valid)
  expect_length(report$issues, 0L)
})

test_that("validate_graph_json flags dangling edge endpoints and unknown types", {
  skip_if_no_jsonlite()
  bad <- list(
    splitGraph_object = "dependency_graph",
    schema_version = "0.2.0",
    nodes = list(list(node_id = "sample:S1", node_type = "Sample", node_key = "S1")),
    edges = list(list(
      edge_id = "e1", from = "sample:S1", to = "subject:P9",
      edge_type = "sample_belongs_to_subject"
    ))
  )
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(jsonlite::toJSON(bad, auto_unbox = TRUE, null = "null"), tmp)

  report <- validate_graph_json(tmp)
  expect_false(report$valid)
  expect_true(any(grepl("does not reference a declared node", report$issues)))
})

test_that("validate_graph_json rejects a split_spec file", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_split_spec(spec, tmp)

  report <- validate_graph_json(tmp)
  expect_false(report$valid)
  expect_true(any(grepl("dependency_graph", report$issues)))
})

# ---- validate_split_spec_json -----------------------------------------------

test_that("validate_split_spec_json passes a well-formed spec and flags missing group_id", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
  good <- tempfile(fileext = ".json")
  on.exit(unlink(good), add = TRUE)
  write_split_spec(spec, good)
  expect_true(validate_split_spec_json(good)$valid)

  bad <- list(
    splitGraph_object = "split_spec",
    schema_version = "0.2.0",
    group_var = "group_id",
    sample_data = list(list(sample_id = "S1"))  # missing group_id
  )
  bad_path <- tempfile(fileext = ".json")
  on.exit(unlink(bad_path), add = TRUE)
  writeLines(jsonlite::toJSON(bad, auto_unbox = TRUE, null = "null"), bad_path)

  report <- validate_split_spec_json(bad_path)
  expect_false(report$valid)
  expect_true(any(grepl("group_id", report$issues)))
})

# ---- schema-version compatibility -------------------------------------------

test_that("reading a same-major (0.1.0) file loads silently", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)

  raw <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)
  raw$schema_version <- "0.1.0"
  writeLines(jsonlite::toJSON(raw, auto_unbox = TRUE, null = "null"), tmp)

  expect_silent(g2 <- read_dependency_graph(tmp))
  expect_s3_class(g2, "dependency_graph")
})

# ---- migration --------------------------------------------------------------

test_that("migrate_split_spec_json upgrades an old-version file to current", {
  skip_if_no_jsonlite()
  # A hand-built 0.1.0 split_spec: old version, no $schema, and none of the
  # site/region/platform/assay columns that were added later.
  old <- list(
    splitGraph_object = "split_spec",
    schema_version = "0.1.0",
    group_var = "group_id",
    block_vars = list("batch_group"),
    time_var = NA,
    ordering_required = FALSE,
    constraint_mode = "subject",
    constraint_strategy = "subject",
    recommended_resampling = "grouped_cv",
    metadata = list(source_mode = "subject"),
    sample_data = list(
      list(sample_id = "S1", sample_node_id = "sample:S1",
           group_id = "subject:P1", primary_group = "subject:P1",
           batch_group = "B1", study_group = NA,
           timepoint_id = NA, time_index = NA, order_rank = NA)
    )
  )
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(jsonlite::toJSON(old, auto_unbox = TRUE, null = "null"), tmp)

  migrate_split_spec_json(tmp)

  raw <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)
  expect_identical(raw$schema_version, "0.2.0")
  expect_match(raw$`$schema`, "split_spec\\.schema\\.json$")
  # New columns are now present in every row.
  row1 <- raw$sample_data[[1L]]
  expect_true(all(c("site_group", "region_group", "platform_group", "assay_group") %in% names(row1)))
  # And the migrated file validates.
  expect_true(validate_split_spec_json(tmp)$valid)
})

# ---- provenance -------------------------------------------------------------

test_that("as_split_spec records derivation provenance in metadata", {
  g <- make_spec_graph()
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
  expect_identical(spec$metadata$source_mode, "subject")
  expect_false(is.null(spec$metadata$splitgraph_version))
  expect_false(is.null(spec$metadata$derived_at))
})

# ---- validator error branches -----------------------------------------------

test_that("validate_graph_json flags unknown node/edge types and a bad version", {
  skip_if_no_jsonlite()
  bad <- list(
    splitGraph_object = "dependency_graph",
    schema_version = "not-a-version",
    nodes = list(list(node_id = "x:1", node_type = "Bogus", node_key = "1")),
    edges = list(list(edge_id = "e1", from = "x:1", to = "x:1",
                      edge_type = "not_a_relation"))
  )
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(jsonlite::toJSON(bad, auto_unbox = TRUE, null = "null"), tmp)

  report <- validate_graph_json(tmp)
  expect_false(report$valid)
  expect_true(any(grepl("schema_version", report$issues)))
  expect_true(any(grepl("node_type", report$issues)))
  expect_true(any(grepl("edge_type", report$issues)))
})

test_that("validate_graph_json flags malformed and non-array node collections", {
  skip_if_no_jsonlite()
  # Node missing required fields.
  missing_fields <- list(
    splitGraph_object = "dependency_graph", schema_version = "0.2.0",
    nodes = list(list(node_id = "sample:S1")), edges = list()
  )
  tmp1 <- tempfile(fileext = ".json"); on.exit(unlink(tmp1), add = TRUE)
  writeLines(jsonlite::toJSON(missing_fields, auto_unbox = TRUE, null = "null"), tmp1)
  expect_true(any(grepl("required strings", validate_graph_json(tmp1)$issues)))

  # `nodes` is a scalar, not an array.
  not_array <- list(
    splitGraph_object = "dependency_graph", schema_version = "0.2.0",
    nodes = "oops", edges = list()
  )
  tmp2 <- tempfile(fileext = ".json"); on.exit(unlink(tmp2), add = TRUE)
  writeLines(jsonlite::toJSON(not_array, auto_unbox = TRUE, null = "null"), tmp2)
  expect_true(any(grepl("`nodes` must be an array", validate_graph_json(tmp2)$issues)))
})

test_that("validate_split_spec_json flags version, group_var, and block_vars problems", {
  skip_if_no_jsonlite()
  bad <- list(
    splitGraph_object = "split_spec",
    schema_version = "x",            # invalid version
    block_vars = "not-an-array",     # must be an array
    sample_data = list(list(sample_id = "S1", group_id = "g1"))
    # group_var missing entirely
  )
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  writeLines(jsonlite::toJSON(bad, auto_unbox = TRUE, null = "null"), tmp)

  report <- validate_split_spec_json(tmp)
  expect_false(report$valid)
  expect_true(any(grepl("schema_version", report$issues)))
  expect_true(any(grepl("group_var", report$issues)))
  expect_true(any(grepl("block_vars", report$issues)))
})

# ---- migrate_dependency_graph_json ------------------------------------------

test_that("migrate_dependency_graph_json upgrades an old-version graph file", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  write_dependency_graph(g, tmp)

  # Downgrade the recorded version and strip the $schema reference.
  raw <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)
  raw$schema_version <- "0.1.0"
  raw[["$schema"]] <- NULL
  writeLines(jsonlite::toJSON(raw, auto_unbox = TRUE, null = "null"), tmp)

  migrate_dependency_graph_json(tmp)

  out <- jsonlite::fromJSON(tmp, simplifyVector = FALSE)
  expect_identical(out$schema_version, "0.2.0")
  expect_match(out[["$schema"]], "dependency_graph\\.schema\\.json$")
  expect_true(validate_graph_json(tmp)$valid)
})

# ---- print method -----------------------------------------------------------

test_that("print.splitgraph_json_report shows issues only when invalid", {
  skip_if_no_jsonlite()
  g <- make_spec_graph()
  good_path <- tempfile(fileext = ".json"); on.exit(unlink(good_path), add = TRUE)
  write_dependency_graph(g, good_path)
  valid_out <- capture.output(print(validate_graph_json(good_path)))
  expect_true(any(grepl("valid:.*TRUE", valid_out)))
  expect_false(any(grepl("issues:", valid_out)))

  bad <- list(splitGraph_object = "split_spec", schema_version = "0.2.0",
              group_var = "group_id",
              sample_data = list(list(sample_id = "S1")))  # missing group_id
  bad_path <- tempfile(fileext = ".json"); on.exit(unlink(bad_path), add = TRUE)
  writeLines(jsonlite::toJSON(bad, auto_unbox = TRUE, null = "null"), bad_path)
  invalid_out <- capture.output(print(validate_split_spec_json(bad_path)))
  expect_true(any(grepl("valid:.*FALSE", invalid_out)))
  expect_true(any(grepl("issues:", invalid_out)))
})

# Pairwise (thresholded) leakage relations. Unlike the direct-assignment modes
# (subject / batch / site / ...), these relations are *pairwise and continuous*:
# leakage risk is a graded property of a pair of subjects (genetic relatedness)
# or samples (spatial proximity). They are modelled as undirected, thresholded
# edges, then collapsed into groups via transitive closure (connected
# components) over the thresholded edge set. The thresholding happens up front
# in the edge-building helpers below; the derivation modes only run the
# component search over whatever edges the graph already carries.

.depgraph_pairwise_relation <- c(
  relatedness = "subject_related_to",
  spatial     = "sample_adjacent_to"
)

# Build sample-sample projection edges for a pairwise mode, restricted to the
# in-scope sample node ids. For "spatial" the graph already carries
# sample-sample edges. For "relatedness" the graph carries subject-subject
# edges, so we expand them onto samples: two samples are linked when they share
# a subject (same individual) or when their subjects are directly related.
.depgraph_pairwise_projection_edges <- function(graph, mode, sample_node_ids) {
  relation <- .depgraph_pairwise_relation[[mode]]
  edge_data <- graph$edges$data

  empty <- data.frame(
    sample_node_id_1 = character(),
    sample_node_id_2 = character(),
    stringsAsFactors = FALSE
  )

  if (identical(mode, "spatial")) {
    e <- edge_data[
      edge_data$edge_type == relation &
        edge_data$from %in% sample_node_ids &
        edge_data$to %in% sample_node_ids,
      c("from", "to"),
      drop = FALSE
    ]
    if (nrow(e) == 0L) return(empty)
    return(unique(data.frame(
      sample_node_id_1 = e$from,
      sample_node_id_2 = e$to,
      stringsAsFactors = FALSE
    )))
  }

  # relatedness: sample -> subject for in-scope samples.
  belongs <- edge_data[
    edge_data$edge_type == "sample_belongs_to_subject" &
      edge_data$from %in% sample_node_ids,
    c("from", "to"),
    drop = FALSE
  ]
  if (nrow(belongs) == 0L) return(empty)

  samples_of_subject <- split(belongs$from, belongs$to)
  parts <- list()

  # within-subject: samples from the same individual are always grouped.
  for (subject in names(samples_of_subject)) {
    s <- unique(samples_of_subject[[subject]])
    if (length(s) >= 2L) {
      cmb <- utils::combn(s, 2L)
      parts[[length(parts) + 1L]] <- data.frame(
        sample_node_id_1 = cmb[1L, ],
        sample_node_id_2 = cmb[2L, ],
        stringsAsFactors = FALSE
      )
    }
  }

  # across related subjects (undirected subject_related_to edges).
  rel_edges <- edge_data[edge_data$edge_type == relation, c("from", "to"), drop = FALSE]
  for (i in seq_len(nrow(rel_edges))) {
    sa <- samples_of_subject[[rel_edges$from[[i]]]]
    sb <- samples_of_subject[[rel_edges$to[[i]]]]
    if (length(sa) > 0L && length(sb) > 0L) {
      grid <- expand.grid(a = sa, b = sb, KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)
      parts[[length(parts) + 1L]] <- data.frame(
        sample_node_id_1 = grid$a,
        sample_node_id_2 = grid$b,
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(parts) == 0L) return(empty)
  unique(do.call(rbind, parts))
}

.derive_pairwise_constraints <- function(graph, mode, samples = NULL) {
  relation <- .depgraph_pairwise_relation[[mode]]
  sample_nodes <- .depgraph_constraint_samples(graph, samples)
  keep_ids <- sample_nodes$node_id

  projection <- .depgraph_pairwise_projection_edges(graph, mode, keep_ids)

  subset_graph <- if (nrow(projection) == 0L) {
    igraph::make_empty_graph(n = length(keep_ids), directed = FALSE)
  } else {
    igraph::graph_from_data_frame(
      d = data.frame(
        from = projection$sample_node_id_1,
        to = projection$sample_node_id_2,
        stringsAsFactors = FALSE
      ),
      vertices = data.frame(name = keep_ids, stringsAsFactors = FALSE),
      directed = FALSE
    )
  }
  igraph::V(subset_graph)$name <- keep_ids

  comps <- igraph::components(subset_graph)
  membership_idx <- as.integer(comps$membership[keep_ids])
  component_size <- as.integer(comps$csize[membership_idx])

  sample_map <- data.frame(
    sample_id = sample_nodes$node_key,
    sample_node_id = keep_ids,
    group_id = paste0(mode, ":component_", membership_idx),
    constraint_type = mode,
    group_label = paste0("component_", membership_idx),
    explanation = paste0(
      "Grouped by transitive closure over thresholded ", relation,
      " edges (connected component ", membership_idx, ")."
    ),
    stringsAsFactors = FALSE
  )
  row.names(sample_map) <- NULL

  warnings <- character()
  if (identical(mode, "relatedness")) {
    linked_samples <- unique(c(projection$sample_node_id_1, projection$sample_node_id_2))
    without_subject <- setdiff(
      keep_ids,
      graph$edges$data$from[graph$edges$data$edge_type == "sample_belongs_to_subject"]
    )
    missing_subject <- intersect(keep_ids, without_subject)
    if (length(missing_subject) > 0L) {
      warnings <- c(warnings, paste0(
        "Samples without a subject assignment were retained as singleton groups ",
        "(relatedness cannot be assessed): ",
        paste(sample_nodes$node_key[match(missing_subject, keep_ids)], collapse = ", ")
      ))
    }
  }
  if (nrow(sample_map) > 0L && mean(component_size == 1L) > 0.5) {
    warnings <- c(warnings, paste0(
      "Most ", mode, " groups are singletons; pairwise coverage may be sparse ",
      "or the threshold may be too strict."
    ))
  }

  split_constraint(
    strategy = mode,
    sample_map = sample_map,
    recommended_downstream_args = list(
      group_var = "group_id",
      block_var = "group_id",
      time_var = NULL,
      ordering_required = FALSE
    ),
    metadata = list(
      mode = mode,
      strategy = mode,
      relations_used = relation,
      n_groups = length(unique(sample_map$group_id)),
      n_samples = nrow(sample_map),
      warnings = warnings,
      projection_edges = projection
    )
  )
}

#' Build Pairwise Leakage Edges from Continuous Similarity
#'
#' Helpers that turn a continuous, pairwise similarity signal into the
#' thresholded, undirected edges consumed by
#' \code{derive_split_constraints(mode = "relatedness")} and
#' \code{derive_split_constraints(mode = "spatial")}. Only pairs that pass the
#' threshold become edges; the derivation modes then form groups as connected
#' components over those edges (transitive closure), so a chain of individually
#' below-radius neighbours can still land in one group.
#'
#' \code{relatedness_edges_from_kinship()} keeps subject pairs whose kinship (or
#' relatedness) coefficient is \emph{at least} \code{threshold} and emits
#' \code{subject_related_to} edges (\code{Subject} -> \code{Subject}).
#'
#' \code{spatial_edges_from_coords()} keeps sample pairs whose Euclidean
#' distance over the coordinate columns is \emph{at most} \code{radius} and
#' emits \code{sample_adjacent_to} edges (\code{Sample} -> \code{Sample}).
#'
#' Both return a \code{graph_edge_set} that can be combined with the other node
#' and edge sets in \code{build_dependency_graph()}. The passing metric value is
#' carried on each edge as an attribute (\code{kinship} / \code{distance}).
#'
#' @param pairs A data.frame of subject pairs with two id columns and a metric
#'   column.
#' @param threshold Minimum kinship value (inclusive) for a pair to be kept.
#' @param id1,id2 Column names in \code{pairs} holding the two subject ids.
#' @param kinship Column name in \code{pairs} holding the kinship / relatedness
#'   value.
#' @param coords A data.frame with one row per sample: a sample id column plus
#'   the numeric coordinate columns.
#' @param radius Maximum distance (inclusive) for two samples to be adjacent.
#' @param id Column name in \code{coords} holding the sample id.
#' @param coord_cols Character vector of coordinate columns in \code{coords}.
#'   Defaults to every numeric column other than \code{id}.
#' @return A \code{graph_edge_set}.
#' @examples
#' pairs <- data.frame(
#'   id1 = c("P1", "P1", "P2"),
#'   id2 = c("P2", "P3", "P3"),
#'   kinship = c(0.25, 0.02, 0.30)
#' )
#' relatedness_edges_from_kinship(pairs, threshold = 0.1)
#'
#' coords <- data.frame(
#'   sample_id = c("S1", "S2", "S3"),
#'   x = c(0, 1, 9),
#'   y = c(0, 1, 9)
#' )
#' spatial_edges_from_coords(coords, radius = 2)
#' @name pairwise_edges
#' @export
relatedness_edges_from_kinship <- function(pairs, threshold, id1 = "id1", id2 = "id2", kinship = "kinship") {
  .depgraph_assert(is.data.frame(pairs), "`pairs` must be a data.frame.")
  .depgraph_assert(length(threshold) == 1L && is.numeric(threshold) && !is.na(threshold),
                   "`threshold` must be a single numeric value.")
  for (col in c(id1, id2, kinship)) {
    .depgraph_assert(col %in% names(pairs), paste0("Missing column in `pairs`: ", col))
  }

  value <- suppressWarnings(as.numeric(pairs[[kinship]]))
  keep <- !is.na(value) & value >= threshold &
    !is.na(pairs[[id1]]) & !is.na(pairs[[id2]]) &
    as.character(pairs[[id1]]) != as.character(pairs[[id2]])

  kept <- data.frame(
    from_id = as.character(pairs[[id1]])[keep],
    to_id = as.character(pairs[[id2]])[keep],
    kinship = value[keep],
    stringsAsFactors = FALSE
  )
  if (nrow(kept) == 0L) return(graph_edge_set())

  create_edges(
    kept,
    from_col = "from_id", to_col = "to_id",
    from_type = "Subject", to_type = "Subject",
    relation = "subject_related_to",
    attr_cols = "kinship"
  )
}

#' @rdname pairwise_edges
#' @export
spatial_edges_from_coords <- function(coords, radius, id = "sample_id", coord_cols = NULL) {
  .depgraph_assert(is.data.frame(coords), "`coords` must be a data.frame.")
  .depgraph_assert(length(radius) == 1L && is.numeric(radius) && !is.na(radius),
                   "`radius` must be a single numeric value.")
  .depgraph_assert(id %in% names(coords), paste0("Missing id column in `coords`: ", id))

  if (is.null(coord_cols)) {
    numeric_cols <- names(coords)[vapply(coords, is.numeric, logical(1))]
    coord_cols <- setdiff(numeric_cols, id)
  }
  .depgraph_assert(length(coord_cols) >= 1L,
                   "No coordinate columns found; supply `coord_cols`.")
  for (col in coord_cols) {
    .depgraph_assert(col %in% names(coords), paste0("Missing coordinate column in `coords`: ", col))
  }

  ids <- as.character(coords[[id]])
  mat <- as.matrix(coords[, coord_cols, drop = FALSE])
  storage.mode(mat) <- "double"

  empty <- data.frame(
    from_id = character(), to_id = character(), distance = numeric(),
    stringsAsFactors = FALSE
  )

  n <- nrow(mat)
  kept <- if (n < 2L) {
    empty
  } else {
    dmat <- as.matrix(stats::dist(mat))
    rows <- list()
    for (i in seq_len(n - 1L)) {
      for (j in seq(i + 1L, n)) {
        d <- dmat[i, j]
        if (!is.na(d) && d <= radius && ids[[i]] != ids[[j]]) {
          rows[[length(rows) + 1L]] <- data.frame(
            from_id = ids[[i]], to_id = ids[[j]], distance = d,
            stringsAsFactors = FALSE
          )
        }
      }
    }
    if (length(rows) == 0L) empty else do.call(rbind, rows)
  }
  if (nrow(kept) == 0L) return(graph_edge_set())

  create_edges(
    kept,
    from_col = "from_id", to_col = "to_id",
    from_type = "Sample", to_type = "Sample",
    relation = "sample_adjacent_to",
    attr_cols = "distance"
  )
}

# Query helpers for dependency_graph objects.

# Resolve the user-supplied `max_length` to an `igraph::all_simple_paths`
# `cutoff` value. NULL -> documented default cap; Inf -> no cap (cutoff=-1);
# any non-negative integer is passed through. Any other input errors.
.depgraph_resolve_path_cutoff <- function(max_length) {
  if (is.null(max_length)) {
    return(.depgraph_default_path_cap)
  }
  if (length(max_length) != 1L || !is.numeric(max_length)) {
    stop("`max_length` must be a single numeric value, Inf, or NULL.", call. = FALSE)
  }
  if (is.infinite(max_length)) {
    return(-1)
  }
  if (is.na(max_length) || max_length < 0) {
    stop("`max_length` must be non-negative.", call. = FALSE)
  }
  as.integer(max_length)
}

.depgraph_resolve_node_ids <- function(graph, node_ids) {
  node_data <- graph$nodes$data
  node_ids <- unique(as.character(node_ids))
  .depgraph_assert(length(node_ids) > 0L, "`node_ids` must contain at least one value.")
  .depgraph_assert(
    all(node_ids %in% node_data$node_id),
    paste0("Unknown node IDs: ", paste(setdiff(node_ids, node_data$node_id), collapse = ", "))
  )
  node_ids
}

.depgraph_resolve_sample_node_ids <- function(graph, samples = NULL) {
  sample_nodes <- graph$nodes$data[graph$nodes$data$node_type == "Sample", , drop = FALSE]
  if (is.null(samples)) {
    return(sample_nodes$node_id)
  }

  samples <- unique(as.character(samples))
  valid_inputs <- unique(c(sample_nodes$node_id, sample_nodes$node_key))
  unknown <- setdiff(samples, valid_inputs)
  .depgraph_assert(
    length(unknown) == 0L,
    paste0("Unknown sample IDs: ", paste(unknown, collapse = ", "))
  )

  matched <- sample_nodes$node_id[sample_nodes$node_id %in% samples]
  matched <- unique(c(
    matched,
    sample_nodes$node_id[match(samples, sample_nodes$node_key, nomatch = 0L)]
  ))
  matched <- matched[!is.na(matched) & nzchar(matched)]

  .depgraph_assert(length(matched) > 0L, "No matching sample nodes found.")
  matched
}

.depgraph_subset_edges <- function(graph, edge_types = NULL, node_ids = NULL) {
  edge_data <- graph$edges$data

  if (!is.null(edge_types)) {
    edge_types <- unique(as.character(edge_types))
    edge_data <- edge_data[edge_data$edge_type %in% edge_types, , drop = FALSE]
  }

  if (!is.null(node_ids)) {
    node_ids <- unique(as.character(node_ids))
    edge_data <- edge_data[edge_data$from %in% node_ids | edge_data$to %in% node_ids, , drop = FALSE]
  }

  edge_data
}

.depgraph_query_result <- function(graph, query, params, node_ids = NULL, edge_ids = NULL, table = NULL, metadata = list()) {
  node_data <- graph$nodes$data
  edge_data <- graph$edges$data

  nodes <- if (is.null(node_ids)) {
    node_data[0, , drop = FALSE]
  } else {
    node_data[node_data$node_id %in% unique(node_ids), , drop = FALSE]
  }

  edges <- if (is.null(edge_ids)) {
    edge_data[0, , drop = FALSE]
  } else {
    edge_data[edge_data$edge_id %in% unique(edge_ids), , drop = FALSE]
  }

  graph_query_result(
    query = query,
    params = params,
    nodes = nodes,
    edges = edges,
    table = if (is.null(table)) data.frame(stringsAsFactors = FALSE) else table,
    metadata = metadata
  )
}

.depgraph_filter_graph_by_edge_type <- function(graph, edge_types = NULL) {
  if (is.null(edge_types)) {
    return(graph$graph)
  }

  edge_types <- unique(as.character(edge_types))
  keep_ids <- graph$edges$data$edge_id[graph$edges$data$edge_type %in% edge_types]
  keep_edges <- igraph::E(graph$graph)[igraph::E(graph$graph)$edge_id %in% keep_ids]
  igraph::subgraph_from_edges(graph$graph, eids = keep_edges, delete.vertices = FALSE)
}

.depgraph_edge_between <- function(edge_data, from, to, undirected = FALSE) {
  matches <- edge_data$from == from & edge_data$to == to
  if (isTRUE(undirected)) {
    matches <- matches | (edge_data$from == to & edge_data$to == from)
  }
  edge_data[matches, , drop = FALSE]
}

.depgraph_format_paths <- function(graph, path_node_ids, edge_data, query_name, params, metadata = list()) {
  node_data <- graph$nodes$data
  rows <- list()
  used_node_ids <- character()
  used_edge_ids <- character()

  if (length(path_node_ids) == 0L) {
    return(.depgraph_query_result(graph, query_name, params, table = data.frame(
      path_id = character(),
      step = integer(),
      node_id = character(),
      node_type = character(),
      edge_id = character(),
      edge_type = character(),
      stringsAsFactors = FALSE
    ), metadata = metadata))
  }

  row_idx <- 1L
  for (i in seq_along(path_node_ids)) {
    nodes_in_path <- path_node_ids[[i]]
    if (length(nodes_in_path) == 0L) {
      next
    }

    for (step_idx in seq_along(nodes_in_path)) {
      node_id <- nodes_in_path[[step_idx]]
      node_row <- node_data[node_data$node_id == node_id, , drop = FALSE]
      edge_id <- NA_character_
      edge_type <- NA_character_

      if (step_idx > 1L) {
        prev_node <- nodes_in_path[[step_idx - 1L]]
        edge_row <- .depgraph_edge_between(edge_data, prev_node, node_id, undirected = identical(params$mode, "all"))
        if (nrow(edge_row) > 0L) {
          edge_id <- edge_row$edge_id[[1L]]
          edge_type <- edge_row$edge_type[[1L]]
          used_edge_ids <- c(used_edge_ids, edge_id)
        }
      }

      rows[[row_idx]] <- data.frame(
        path_id = paste0("path_", i),
        step = step_idx,
        node_id = node_id,
        node_type = node_row$node_type[[1L]],
        edge_id = edge_id,
        edge_type = edge_type,
        stringsAsFactors = FALSE
      )
      used_node_ids <- c(used_node_ids, node_id)
      row_idx <- row_idx + 1L
    }
  }

  table <- do.call(rbind, rows)
  row.names(table) <- NULL

  .depgraph_query_result(
    graph = graph,
    query = query_name,
    params = params,
    node_ids = unique(used_node_ids),
    edge_ids = unique(used_edge_ids),
    table = table,
    metadata = metadata
  )
}

.depgraph_edge_types_for_via <- function(via, edge_types = NULL) {
  if (!is.null(edge_types)) {
    return(unique(as.character(edge_types)))
  }

  via <- unique(vapply(via, .depgraph_match_node_type, character(1), USE.NAMES = FALSE))
  .depgraph_edge_schema$edge_type[
    .depgraph_edge_schema$from_type == "Sample" &
      .depgraph_edge_schema$to_type %in% via
  ]
}

.depgraph_shared_dependency_table <- function(graph, via, samples = NULL, edge_types = NULL) {
  node_data <- graph$nodes$data
  edge_data <- graph$edges$data
  sample_nodes <- .depgraph_resolve_sample_node_ids(graph, samples)
  selected_edge_types <- .depgraph_edge_types_for_via(via, edge_types)

  edges <- edge_data[
    edge_data$edge_type %in% selected_edge_types &
      edge_data$from %in% sample_nodes,
    ,
    drop = FALSE
  ]

  rows <- list()
  row_idx <- 1L
  sample_key_map <- stats::setNames(node_data$node_key, node_data$node_id)
  node_type_map <- stats::setNames(node_data$node_type, node_data$node_id)

  if (nrow(edges) == 0L) {
    return(data.frame(
      sample_id_1 = character(),
      sample_id_2 = character(),
      sample_node_id_1 = character(),
      sample_node_id_2 = character(),
      shared_node_id = character(),
      shared_node_type = character(),
      edge_type = character(),
      stringsAsFactors = FALSE
    ))
  }

  split_targets <- split(edges, list(edges$to, edges$edge_type), drop = TRUE)
  for (group in split_targets) {
    sample_group <- sort(unique(group$from))
    if (length(sample_group) < 2L) {
      next
    }

    pairs <- utils::combn(sample_group, 2L, simplify = FALSE)
    for (pair in pairs) {
      rows[[row_idx]] <- data.frame(
        sample_id_1 = sample_key_map[[pair[[1L]]]],
        sample_id_2 = sample_key_map[[pair[[2L]]]],
        sample_node_id_1 = pair[[1L]],
        sample_node_id_2 = pair[[2L]],
        shared_node_id = group$to[[1L]],
        shared_node_type = node_type_map[[group$to[[1L]]]],
        edge_type = group$edge_type[[1L]],
        stringsAsFactors = FALSE
      )
      row_idx <- row_idx + 1L
    }
  }

  if (length(rows) == 0L) {
    return(data.frame(
      sample_id_1 = character(),
      sample_id_2 = character(),
      sample_node_id_1 = character(),
      sample_node_id_2 = character(),
      shared_node_id = character(),
      shared_node_type = character(),
      edge_type = character(),
      stringsAsFactors = FALSE
    ))
  }

  out <- do.call(rbind, rows)
  row.names(out) <- NULL
  out
}

.depgraph_project_sample_dependencies <- function(graph, via, edge_types = NULL) {
  shared <- .depgraph_shared_dependency_table(graph, via = via, edge_types = edge_types)
  if (nrow(shared) == 0L) {
    return(shared)
  }

  projection <- unique(shared[, c("sample_node_id_1", "sample_node_id_2"), drop = FALSE])
  projection$projection_edge_id <- paste0("projection_", seq_len(nrow(projection)))
  projection
}

.depgraph_shared_dependency_edge_ids <- function(graph, shared_table) {
  if (nrow(shared_table) == 0L) {
    return(character())
  }

  edge_data <- graph$edges$data
  edge_ids <- character()
  for (i in seq_len(nrow(shared_table))) {
    relevant_edges <- edge_data[
      edge_data$edge_type == shared_table$edge_type[[i]] &
        edge_data$to == shared_table$shared_node_id[[i]] &
        edge_data$from %in% c(shared_table$sample_node_id_1[[i]], shared_table$sample_node_id_2[[i]]),
      "edge_id",
      drop = TRUE
    ]
    edge_ids <- c(edge_ids, relevant_edges)
  }

  unique(edge_ids)
}

#' Query Dependency Graph Structure
#'
#' Query graph neighborhoods, typed nodes and edges, path structure, projected
#' sample dependency components, and direct shared dependencies within a
#' \code{dependency_graph}.
#'
#' When a \code{samples} subset is supplied, partial matching is not allowed:
#' unknown sample identifiers raise an error rather than being silently
#' dropped.
#'
#' @param graph A \code{dependency_graph}.
#' @param node_ids,from,to Node identifiers to use as query seeds or endpoints.
#' @param edge_types Optional edge types used to filter the traversal graph or
#'   edge table.
#' @param node_types Optional node types used to filter node results or allowed
#'   path members.
#' @param direction,mode Traversal direction.
#' @param ids Optional node identifiers used to further restrict
#'   \code{query_node_type()}.
#' @param max_length Maximum path length (number of edges) for
#'   \code{query_paths()}. Defaults to a documented finite cap
#'   (\code{8}) so that \code{igraph::all_simple_paths()} cannot blow up
#'   on dense graphs. Pass \code{Inf} to opt out and search exhaustively;
#'   pass any non-negative integer for an explicit cap. Negative values and
#'   non-numeric inputs are rejected.
#' @param via Dependency node types used for sample-level dependency detection.
#' @param min_size Minimum component size retained by
#'   \code{detect_dependency_components()}.
#' @param samples Optional sample identifiers or sample node IDs used to
#'   restrict direct shared-dependency detection. All requested samples must
#'   resolve successfully.
#' @return Each function returns a \code{graph_query_result}. Use
#'   \code{as.data.frame()} to obtain the tidy result table.
#' @examples
#' meta <- data.frame(
#'   sample_id  = c("S1", "S2", "S3"),
#'   subject_id = c("P1", "P1", "P2"),
#'   batch_id   = c("B1", "B2", "B1")
#' )
#' g <- graph_from_metadata(meta)
#'
#' query_node_type(g, "Sample")
#' query_neighbors(g, "sample:S1", direction = "out")
#' detect_shared_dependencies(g, via = "Subject")
#' @export
query_node_type <- function(graph, node_types, ids = NULL) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  node_types <- unique(vapply(node_types, .depgraph_match_node_type, character(1), USE.NAMES = FALSE))

  node_data <- graph$nodes$data
  table <- node_data[node_data$node_type %in% node_types, , drop = FALSE]
  if (!is.null(ids)) {
    ids <- unique(as.character(ids))
    table <- table[table$node_id %in% ids, , drop = FALSE]
  }

  .depgraph_query_result(
    graph = graph,
    query = "query_node_type",
    params = list(node_types = node_types, ids = ids),
    node_ids = table$node_id,
    table = table
  )
}

#' @rdname query_node_type
#' @export
query_edge_type <- function(graph, edge_types, node_ids = NULL) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  edge_types <- unique(as.character(edge_types))
  if (!is.null(node_ids)) {
    node_ids <- .depgraph_resolve_node_ids(graph, node_ids)
  }

  table <- .depgraph_subset_edges(graph, edge_types = edge_types, node_ids = node_ids)
  touched_nodes <- unique(c(table$from, table$to))

  .depgraph_query_result(
    graph = graph,
    query = "query_edge_type",
    params = list(edge_types = edge_types, node_ids = node_ids),
    node_ids = touched_nodes,
    edge_ids = table$edge_id,
    table = table
  )
}

#' @rdname query_node_type
#' @export
query_neighbors <- function(graph, node_ids, edge_types = NULL, node_types = NULL, direction = c("out", "in", "all")) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  direction <- match.arg(direction)
  seed_ids <- .depgraph_resolve_node_ids(graph, node_ids)
  g <- .depgraph_filter_graph_by_edge_type(graph, edge_types)

  rows <- list()
  row_idx <- 1L
  used_nodes <- character()
  used_edges <- character()
  node_data <- graph$nodes$data
  edge_data <- .depgraph_subset_edges(graph, edge_types = edge_types)

  for (seed in seed_ids) {
    neigh <- igraph::neighbors(g, v = seed, mode = direction)
    neigh_ids <- igraph::as_ids(neigh)
    if (!is.null(node_types) && length(neigh_ids) > 0L) {
      node_types <- unique(vapply(node_types, .depgraph_match_node_type, character(1), USE.NAMES = FALSE))
      keep <- node_data$node_type[match(neigh_ids, node_data$node_id)] %in% node_types
      neigh_ids <- neigh_ids[keep]
    }

    for (neighbor in neigh_ids) {
      edge_rows <- if (identical(direction, "out")) {
        edge_data[edge_data$from == seed & edge_data$to == neighbor, , drop = FALSE]
      } else if (identical(direction, "in")) {
        edge_data[edge_data$from == neighbor & edge_data$to == seed, , drop = FALSE]
      } else {
        .depgraph_edge_between(edge_data, seed, neighbor, undirected = TRUE)
      }

      if (nrow(edge_rows) == 0L) {
        next
      }

      neighbor_row <- node_data[node_data$node_id == neighbor, , drop = FALSE]
      for (j in seq_len(nrow(edge_rows))) {
        rows[[row_idx]] <- data.frame(
          seed_node_id = seed,
          node_id = neighbor,
          node_type = neighbor_row$node_type[[1L]],
          edge_id = edge_rows$edge_id[[j]],
          edge_type = edge_rows$edge_type[[j]],
          direction = direction,
          stringsAsFactors = FALSE
        )
        used_nodes <- c(used_nodes, seed, neighbor)
        used_edges <- c(used_edges, edge_rows$edge_id[[j]])
        row_idx <- row_idx + 1L
      }
    }
  }

  table <- if (length(rows) == 0L) {
    data.frame(
      seed_node_id = character(),
      node_id = character(),
      node_type = character(),
      edge_id = character(),
      edge_type = character(),
      direction = character(),
      stringsAsFactors = FALSE
    )
  } else {
    out <- do.call(rbind, rows)
    row.names(out) <- NULL
    out
  }

  .depgraph_query_result(
    graph = graph,
    query = "query_neighbors",
    params = list(node_ids = seed_ids, edge_types = edge_types, node_types = node_types, direction = direction),
    node_ids = unique(used_nodes),
    edge_ids = unique(used_edges),
    table = table
  )
}

#' @rdname query_node_type
#' @export
query_paths <- function(graph, from, to, edge_types = NULL, node_types = NULL, mode = c("out", "in", "all"), max_length = NULL) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  mode <- match.arg(mode)
  from_ids <- .depgraph_resolve_node_ids(graph, from)
  to_ids <- .depgraph_resolve_node_ids(graph, to)
  cutoff_value <- .depgraph_resolve_path_cutoff(max_length)
  filtered_graph <- .depgraph_filter_graph_by_edge_type(graph, edge_types)
  edge_data <- .depgraph_subset_edges(graph, edge_types = edge_types)
  path_nodes <- list()
  path_idx <- 1L
  # Truncation tracking: TRUE whenever a returned path reaches the cap in
  # edges, which means longer simple paths between the same endpoints may
  # have been suppressed. Conservative — never a false negative when a
  # finite cap was binding, may flag false positives (returned paths at the
  # boundary that happen to have no longer siblings).
  cap_binding <- (cutoff_value != -1L)
  hit_cap <- FALSE

  for (from_id in from_ids) {
    for (to_id in to_ids) {
      paths <- igraph::all_simple_paths(
        filtered_graph,
        from = from_id,
        to = to_id,
        mode = mode,
        cutoff = cutoff_value
      )
      if (length(paths) == 0L) {
        next
      }

      for (path in paths) {
        node_ids_in_path <- igraph::as_ids(path)
        if (!is.null(node_types)) {
          allowed_types <- unique(vapply(node_types, .depgraph_match_node_type, character(1), USE.NAMES = FALSE))
          path_types <- graph$nodes$data$node_type[match(node_ids_in_path, graph$nodes$data$node_id)]
          if (!all(path_types %in% allowed_types)) {
            next
          }
        }
        if (cap_binding && (length(node_ids_in_path) - 1L) >= cutoff_value) {
          hit_cap <- TRUE
        }
        path_nodes[[path_idx]] <- node_ids_in_path
        path_idx <- path_idx + 1L
      }
    }
  }

  .depgraph_format_paths(
    graph = graph,
    path_node_ids = path_nodes,
    edge_data = edge_data,
    query_name = "query_paths",
    params = list(from = from_ids, to = to_ids, edge_types = edge_types, node_types = node_types, mode = mode, max_length = max_length),
    metadata = list(
      truncated = isTRUE(cap_binding && hit_cap),
      max_length = if (cap_binding) as.integer(cutoff_value) else Inf
    )
  )
}

#' @rdname query_node_type
#' @export
query_shortest_paths <- function(graph, from, to, edge_types = NULL, node_types = NULL, mode = c("out", "in", "all")) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  mode <- match.arg(mode)
  from_ids <- .depgraph_resolve_node_ids(graph, from)
  to_ids <- .depgraph_resolve_node_ids(graph, to)
  filtered_graph <- .depgraph_filter_graph_by_edge_type(graph, edge_types)
  edge_data <- .depgraph_subset_edges(graph, edge_types = edge_types)
  allowed_types <- NULL
  if (!is.null(node_types)) {
    allowed_types <- unique(vapply(node_types, .depgraph_match_node_type, character(1), USE.NAMES = FALSE))
    allowed_nodes <- graph$nodes$data$node_id[graph$nodes$data$node_type %in% allowed_types]
    filtered_graph <- igraph::induced_subgraph(filtered_graph, vids = allowed_nodes)
  }
  path_nodes <- list()
  path_idx <- 1L

  for (from_id in from_ids) {
    for (to_id in to_ids) {
      graph_nodes <- igraph::V(filtered_graph)$name
      if (!from_id %in% graph_nodes || !to_id %in% graph_nodes) {
        next
      }
      shortest <- igraph::shortest_paths(filtered_graph, from = from_id, to = to_id, mode = mode)$vpath
      if (length(shortest) == 0L || length(shortest[[1L]]) == 0L) {
        next
      }
      node_ids_in_path <- igraph::as_ids(shortest[[1L]])
      if (!is.null(allowed_types)) {
        path_types <- graph$nodes$data$node_type[match(node_ids_in_path, graph$nodes$data$node_id)]
        if (!all(path_types %in% allowed_types)) {
          next
        }
      }
      path_nodes[[path_idx]] <- node_ids_in_path
      path_idx <- path_idx + 1L
    }
  }

  .depgraph_format_paths(
    graph = graph,
    path_node_ids = path_nodes,
    edge_data = edge_data,
    query_name = "query_shortest_paths",
    params = list(from = from_ids, to = to_ids, edge_types = edge_types, node_types = node_types, mode = mode)
  )
}

#' @rdname query_node_type
#' @export
detect_dependency_components <- function(graph, via = c("Subject", "Batch", "Study", "Timepoint", "Assay", "FeatureSet", "Outcome"), edge_types = NULL, min_size = 1) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  shared <- .depgraph_shared_dependency_table(graph, via = via, edge_types = edge_types)
  projection <- .depgraph_project_sample_dependencies(graph, via = via, edge_types = edge_types)
  sample_nodes <- graph$nodes$data[graph$nodes$data$node_type == "Sample", , drop = FALSE]

  sample_graph <- if (nrow(projection) == 0L) {
    igraph::make_empty_graph(n = nrow(sample_nodes), directed = FALSE)
  } else {
    g <- igraph::graph_from_data_frame(
      d = data.frame(
        from = projection$sample_node_id_1,
        to = projection$sample_node_id_2,
        stringsAsFactors = FALSE
      ),
      vertices = data.frame(name = sample_nodes$node_id, stringsAsFactors = FALSE),
      directed = FALSE
    )
    g
  }
  igraph::V(sample_graph)$name <- sample_nodes$node_id

  components <- igraph::components(sample_graph)
  table <- data.frame(
    sample_id = sample_nodes$node_key,
    sample_node_id = sample_nodes$node_id,
    component_id = paste0("component_", components$membership[match(sample_nodes$node_id, names(components$membership))]),
    component_size = components$csize[components$membership[match(sample_nodes$node_id, names(components$membership))]],
    stringsAsFactors = FALSE
  )

  table <- table[table$component_size >= min_size, , drop = FALSE]
  keep_nodes <- table$sample_node_id
  projection <- projection[
    projection$sample_node_id_1 %in% keep_nodes &
      projection$sample_node_id_2 %in% keep_nodes,
    ,
    drop = FALSE
  ]
  shared <- shared[
    shared$sample_node_id_1 %in% keep_nodes &
      shared$sample_node_id_2 %in% keep_nodes,
    ,
    drop = FALSE
  ]
  edge_ids <- .depgraph_shared_dependency_edge_ids(graph, shared)

  .depgraph_query_result(
    graph = graph,
    query = "detect_dependency_components",
    params = list(via = via, edge_types = edge_types, min_size = min_size),
    node_ids = keep_nodes,
    table = table,
    metadata = list(
      n_components = length(unique(table$component_id)),
      projection_edges = projection
    ),
    edge_ids = edge_ids
  )
}

#' @rdname query_node_type
#' @export
detect_shared_dependencies <- function(graph, via = c("Subject", "Batch", "Study", "Timepoint"), samples = NULL) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  table <- .depgraph_shared_dependency_table(graph, via = via, samples = samples)

  node_ids <- unique(c(
    table$sample_node_id_1,
    table$sample_node_id_2,
    table$shared_node_id
  ))

  edge_ids <- .depgraph_shared_dependency_edge_ids(graph, table)

  .depgraph_query_result(
    graph = graph,
    query = "detect_shared_dependencies",
    params = list(via = via, samples = samples),
    node_ids = node_ids,
    edge_ids = unique(edge_ids),
    table = table
  )
}

make_query_graph <- function() {
  meta <- data.frame(
    sample_id = c("S1", "S2", "S3", "S4"),
    subject_id = c("P1", "P1", "P2", "P3"),
    batch_id = c("B1", "B2", "B1", "B3"),
    study_id = c("ST1", "ST1", "ST2", "ST2"),
    timepoint_id = c("T0", "T1", "T0", "T0"),
    outcome_id = c("O1", "O1", "O2", "O3"),
    stringsAsFactors = FALSE
  )

  sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
  study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")
  time_nodes <- create_nodes(
    transform(meta, time_index = c(0L, 1L, 0L, 0L)),
    type = "Timepoint",
    id_col = "timepoint_id",
    attr_cols = c("time_index")
  )
  outcome_nodes <- create_nodes(
    data.frame(
      outcome_id = c("O1", "O2", "O3"),
      observation_level = c("subject", "subject", "subject"),
      stringsAsFactors = FALSE
    ),
    type = "Outcome",
    id_col = "outcome_id",
    attr_cols = c("observation_level")
  )

  subject_edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")
  batch_edges <- create_edges(meta, "sample_id", "batch_id", "Sample", "Batch", "sample_processed_in_batch")
  study_edges <- create_edges(meta, "sample_id", "study_id", "Sample", "Study", "sample_from_study")
  time_edges <- create_edges(meta, "sample_id", "timepoint_id", "Sample", "Timepoint", "sample_collected_at_timepoint")
  subject_outcome_edges <- create_edges(
    data.frame(
      subject_id = c("P1", "P2", "P3"),
      outcome_id = c("O1", "O2", "O3"),
      stringsAsFactors = FALSE
    ),
    "subject_id", "outcome_id", "Subject", "Outcome", "subject_has_outcome"
  )

  build_dependency_graph(
    nodes = list(sample_nodes, subject_nodes, batch_nodes, study_nodes, time_nodes, outcome_nodes),
    edges = list(subject_edges, batch_edges, study_edges, time_edges, subject_outcome_edges),
    graph_name = "query_graph",
    dataset_name = "QueryDemo"
  )
}

test_that("query_node_type returns graph_query_result with tidy node rows", {
  graph <- make_query_graph()
  result <- query_node_type(graph, node_types = "Subject")

  expect_s3_class(result, "graph_query_result")
  expect_true(all(result$table$node_type == "Subject"))
  expect_equal(nrow(result$table), 3L)
  expect_identical(as.data.frame(result), result$table)
})

test_that("query_edge_type returns graph_query_result with tidy edge rows", {
  graph <- make_query_graph()
  result <- query_edge_type(graph, edge_types = "sample_from_study")

  expect_s3_class(result, "graph_query_result")
  expect_true(all(result$table$edge_type == "sample_from_study"))
  expect_equal(nrow(result$table), 4L)
})

test_that("query_neighbors respects direction and edge type filtering", {
  graph <- make_query_graph()
  out_result <- query_neighbors(
    graph,
    node_ids = "sample:S1",
    edge_types = "sample_belongs_to_subject",
    direction = "out"
  )
  in_result <- query_neighbors(
    graph,
    node_ids = "subject:P1",
    edge_types = "sample_belongs_to_subject",
    direction = "in"
  )

  expect_s3_class(out_result, "graph_query_result")
  expect_equal(out_result$table$node_id, "subject:P1")
  expect_true(all(in_result$table$seed_node_id == "subject:P1"))
  expect_setequal(in_result$table$node_id, c("sample:S1", "sample:S2"))
})

test_that("query_paths returns stepwise path rows", {
  graph <- make_query_graph()
  result <- query_paths(
    graph,
    from = "sample:S1",
    to = "outcome:O1",
    edge_types = c("sample_belongs_to_subject", "subject_has_outcome"),
    max_length = 2
  )

  expect_s3_class(result, "graph_query_result")
  expect_true(all(c("path_id", "step", "node_id", "node_type", "edge_id", "edge_type") %in% names(result$table)))
  expect_true(any(result$table$node_id == "subject:P1"))
  expect_true(any(result$table$node_id == "outcome:O1"))
})

test_that("query_shortest_paths returns shortest path rows", {
  graph <- make_query_graph()
  result <- query_shortest_paths(
    graph,
    from = "sample:S1",
    to = "outcome:O1",
    edge_types = c("sample_belongs_to_subject", "subject_has_outcome")
  )

  expect_s3_class(result, "graph_query_result")
  expect_equal(length(unique(result$table$path_id)), 1L)
  expect_true(any(result$table$node_id == "subject:P1"))
})

test_that("query_shortest_paths respects node type filters across alternative shortest paths", {
  nodes <- graph_node_set(
    data.frame(
      node_id = c("sample:S1", "subject:P1", "study:ST1", "outcome:O1"),
      node_type = c("Sample", "Subject", "Study", "Outcome"),
      node_key = c("S1", "P1", "ST1", "O1"),
      label = c("S1", "P1", "ST1", "O1"),
      attrs = I(list(list(), list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  edges <- graph_edge_set(
    data.frame(
      edge_id = c("r:1", "r:2", "r:3", "r:4"),
      from = c("sample:S1", "subject:P1", "sample:S1", "study:ST1"),
      to = c("subject:P1", "outcome:O1", "study:ST1", "outcome:O1"),
      edge_type = c("r1", "r2", "r3", "r4"),
      attrs = I(list(list(), list(), list(), list())),
      stringsAsFactors = FALSE
    )
  )

  graph <- dependency_graph(nodes = nodes, edges = edges, graph = NULL)

  result <- query_shortest_paths(
    graph,
    from = "sample:S1",
    to = "outcome:O1",
    edge_types = c("r1", "r2", "r3", "r4"),
    node_types = c("Sample", "Study", "Outcome")
  )

  expect_s3_class(result, "graph_query_result")
  expect_equal(length(unique(result$table$path_id)), 1L)
  expect_true(any(result$table$node_id == "study:ST1"))
  expect_false(any(result$table$node_id == "subject:P1"))
})

test_that("detect_dependency_components uses projected sample dependencies", {
  graph <- make_query_graph()
  result <- detect_dependency_components(graph, via = c("Subject", "Batch"))

  expect_s3_class(result, "graph_query_result")
  expect_true(all(c("sample_id", "sample_node_id", "component_id", "component_size") %in% names(result$table)))

  comp_sizes <- stats::setNames(result$table$component_size, result$table$sample_id)
  expect_equal(comp_sizes[["S1"]], 3L)
  expect_equal(comp_sizes[["S2"]], 3L)
  expect_equal(comp_sizes[["S3"]], 3L)
  expect_equal(comp_sizes[["S4"]], 1L)
  expect_true(all(result$edges$edge_id %in% graph$edges$data$edge_id))
  expect_true(is.data.frame(result$metadata$projection_edges))
  expect_true(all(c("sample_node_id_1", "sample_node_id_2", "projection_edge_id") %in% names(result$metadata$projection_edges)))
})

test_that("detect_shared_dependencies finds direct shared subject and batch links", {
  graph <- make_query_graph()
  subject_result <- detect_shared_dependencies(graph, via = "Subject")
  batch_result <- detect_shared_dependencies(graph, via = "Batch")

  expect_s3_class(subject_result, "graph_query_result")
  expect_true(all(c(
    "sample_id_1", "sample_id_2", "sample_node_id_1", "sample_node_id_2",
    "shared_node_id", "shared_node_type", "edge_type"
  ) %in% names(subject_result$table)))

  expect_true(any(
    subject_result$table$sample_id_1 == "S1" &
      subject_result$table$sample_id_2 == "S2" &
      subject_result$table$shared_node_id == "subject:P1"
  ))

  expect_true(any(
    batch_result$table$sample_id_1 == "S1" &
      batch_result$table$sample_id_2 == "S3" &
      batch_result$table$shared_node_id == "batch:B1"
  ))
})

test_that("sample-scoped query APIs reject partially unresolved sample subsets", {
  graph <- make_query_graph()

  expect_error(
    detect_shared_dependencies(graph, via = "Subject", samples = c("S1", "BAD")),
    "Unknown sample IDs",
    fixed = TRUE
  )
})

make_path_graph <- function() {
  meta <- data.frame(
    sample_id    = c("S1", "S2", "S3"),
    subject_id   = c("P1", "P2", "P3"),
    timepoint_id = c("T0", "T1", "T2"),
    time_index   = c(0, 1, 2),
    stringsAsFactors = FALSE
  )
  graph_from_metadata(meta, graph_name = "path_demo")
}

test_that("query_paths uses a finite default safety cap and returns finite results", {
  g <- make_path_graph()
  # The default max_length should be finite — no Inf, no -1.
  res <- query_paths(g, from = "sample:S1", to = "subject:P1")
  expect_s3_class(res, "graph_query_result")
  # We should reach P1 from S1 within the cap.
  expect_true(nrow(res$table) >= 1L)
})

test_that("query_paths(max_length = NULL) is equivalent to the documented default cap", {
  g <- make_path_graph()
  default_res <- query_paths(g, from = "sample:S1", to = "subject:P1")
  explicit_res <- query_paths(
    g,
    from = "sample:S1",
    to = "subject:P1",
    max_length = splitGraph:::.depgraph_default_path_cap
  )
  expect_identical(default_res$table, explicit_res$table)
})

test_that("query_paths(max_length = Inf) opts out of the cap", {
  g <- make_path_graph()
  uncapped <- query_paths(
    g,
    from = "sample:S1",
    to = "subject:P1",
    max_length = Inf
  )
  expect_s3_class(uncapped, "graph_query_result")
  expect_true(nrow(uncapped$table) >= 1L)
})

test_that("query_paths rejects negative or non-numeric max_length", {
  g <- make_path_graph()
  expect_error(
    query_paths(g, from = "sample:S1", to = "subject:P1", max_length = -2),
    "max_length"
  )
  expect_error(
    query_paths(g, from = "sample:S1", to = "subject:P1", max_length = "five"),
    "max_length"
  )
})

test_that("query_paths flags truncated = TRUE when the path-length cap was binding", {
  # Regression test for C3: query_paths silently capped traversal at
  # max_length = 8 (default) without telling the caller. A user who got an
  # empty/partial result could not distinguish "no such path" from
  # "no such path within 8 hops." The result must carry a `truncated` field
  # on its metadata that is TRUE whenever the cap was binding.

  # Chain T0 -> T1 -> ... -> T8 via timepoint_precedes (auto-built from
  # time_index). Shortest path T0 -> T8 traverses exactly 8 edges = cap.
  meta <- data.frame(
    sample_id    = paste0("S", 0:8),
    subject_id   = rep("P1", 9),
    timepoint_id = paste0("T", 0:8),
    time_index   = 0:8,
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta, graph_name = "long_chain")

  # Default cap (8) is binding for T0 -> T8.
  res_capped <- query_paths(
    g, from = "timepoint:T0", to = "timepoint:T8",
    edge_types = "timepoint_precedes", mode = "out"
  )
  expect_true("truncated" %in% names(res_capped$metadata),
              info = "query_paths result must carry a `truncated` flag.")
  expect_true(isTRUE(res_capped$metadata$truncated),
              info = "When a returned path's edge-length equals the cap, truncated must be TRUE.")

  # Same query with max_length = Inf opts out of the cap -> not truncated.
  res_uncapped <- query_paths(
    g, from = "timepoint:T0", to = "timepoint:T8",
    edge_types = "timepoint_precedes", mode = "out",
    max_length = Inf
  )
  expect_false(isTRUE(res_uncapped$metadata$truncated),
               info = "max_length = Inf disables the cap; truncated must be FALSE.")

  # Short query well within the cap -> not truncated.
  res_short <- query_paths(
    g, from = "timepoint:T0", to = "timepoint:T2",
    edge_types = "timepoint_precedes", mode = "out"
  )
  expect_false(isTRUE(res_short$metadata$truncated),
               info = "Returned paths all shorter than the cap -> truncated must be FALSE.")
})

test_that("query_paths(max_length = 0) returns no paths (boundary)", {
  g <- make_path_graph()
  zero <- query_paths(g, from = "sample:S1", to = "subject:P1", max_length = 0)
  # cutoff = 0 means no edges -> no paths between distinct nodes.
  expect_s3_class(zero, "graph_query_result")
  expect_equal(nrow(zero$table), 0L)
})

# Helpers for translating splitGraph constraints into stable, tool-agnostic
# sample-level split specifications. Downstream packages (bioLeak, fastml,
# rsample, ...) consume `split_spec` objects through their own adapters.

.split_spec_sample_data_template <- function(n = 0L) {
  data.frame(
    sample_id = rep(NA_character_, n),
    sample_node_id = rep(NA_character_, n),
    group_id = rep(NA_character_, n),
    primary_group = rep(NA_character_, n),
    batch_group = rep(NA_character_, n),
    study_group = rep(NA_character_, n),
    site_group = rep(NA_character_, n),
    region_group = rep(NA_character_, n),
    platform_group = rep(NA_character_, n),
    assay_group = rep(NA_character_, n),
    timepoint_id = rep(NA_character_, n),
    time_index = rep(NA_real_, n),
    order_rank = rep(NA_integer_, n),
    stringsAsFactors = FALSE
  )
}

.split_spec_new_issue <- function(severity, code, message, n_affected = 0L, details = list()) {
  data.frame(
    issue_id = NA_character_,
    severity = as.character(severity)[1L],
    code = as.character(code)[1L],
    message = as.character(message)[1L],
    n_affected = as.integer(n_affected)[1L],
    details = I(list(details)),
    stringsAsFactors = FALSE
  )
}

.split_spec_bind_issues <- function(issues) {
  issues <- Filter(Negate(is.null), issues)
  if (length(issues) == 0L) {
    return(data.frame(
      issue_id = character(),
      severity = character(),
      code = character(),
      message = character(),
      n_affected = integer(),
      details = I(list()),
      stringsAsFactors = FALSE
    ))
  }

  out <- do.call(rbind, issues)
  row.names(out) <- NULL
  out$issue_id <- paste0("split_spec_issue_", seq_len(nrow(out)))
  out
}

.split_spec_resampling_hint <- function(mode, strategy = NULL) {
  mode <- tolower(as.character(mode)[1L])
  strategy <- if (is.null(strategy)) NULL else tolower(as.character(strategy)[1L])

  if (identical(mode, "subject")) return("grouped_cv")
  if (identical(mode, "batch")) return("blocked_cv")
  if (identical(mode, "study")) return("leave_one_group_out")
  if (identical(mode, "site")) return("leave_one_group_out")
  if (identical(mode, "region")) return("grouped_cv")
  if (identical(mode, "platform")) return("blocked_cv")
  if (identical(mode, "assay")) return("grouped_cv")
  if (identical(mode, "relatedness")) return("grouped_cv")
  if (identical(mode, "spatial")) return("grouped_cv")
  if (identical(mode, "time")) return("ordered_split")
  if (identical(mode, "composite") && identical(strategy, "strict")) return("custom_grouped_cv")
  if (identical(mode, "composite") && identical(strategy, "rule_based")) return("grouped_cv")
  "grouped_cv"
}

.split_spec_match_assignment_key <- function(assignments, sample_node_ids) {
  matched <- assignments[match(sample_node_ids, assignments$sample_node_id), , drop = FALSE]
  matched$linked_key
}

.split_spec_enrich_from_graph <- function(sample_data, graph) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  sample_ids <- sample_data$sample_node_id

  batch_assignments <- .depgraph_direct_assignment(graph, "batch", samples = sample_ids)
  study_assignments <- .depgraph_direct_assignment(graph, "study", samples = sample_ids)
  site_assignments <- .depgraph_direct_assignment(graph, "site", samples = sample_ids)
  time_constraint <- .derive_time_constraints(graph, samples = sample_ids)$sample_map
  time_constraint <- time_constraint[match(sample_ids, time_constraint$sample_node_id), , drop = FALSE]

  missing_batch <- is.na(sample_data$batch_group) | !nzchar(sample_data$batch_group)
  sample_data$batch_group[missing_batch] <- .split_spec_match_assignment_key(batch_assignments, sample_ids)[missing_batch]

  missing_study <- is.na(sample_data$study_group) | !nzchar(sample_data$study_group)
  sample_data$study_group[missing_study] <- .split_spec_match_assignment_key(study_assignments, sample_ids)[missing_study]

  missing_site <- is.na(sample_data$site_group) | !nzchar(sample_data$site_group)
  sample_data$site_group[missing_site] <- .split_spec_match_assignment_key(site_assignments, sample_ids)[missing_site]

  region_assignments <- .depgraph_direct_assignment(graph, "region", samples = sample_ids)
  missing_region <- is.na(sample_data$region_group) | !nzchar(sample_data$region_group)
  sample_data$region_group[missing_region] <- .split_spec_match_assignment_key(region_assignments, sample_ids)[missing_region]

  platform_assignments <- .depgraph_direct_assignment(graph, "platform", samples = sample_ids)
  missing_platform <- is.na(sample_data$platform_group) | !nzchar(sample_data$platform_group)
  sample_data$platform_group[missing_platform] <- .split_spec_match_assignment_key(platform_assignments, sample_ids)[missing_platform]

  assay_assignments <- .depgraph_direct_assignment(graph, "assay", samples = sample_ids)
  missing_assay <- is.na(sample_data$assay_group) | !nzchar(sample_data$assay_group)
  sample_data$assay_group[missing_assay] <- .split_spec_match_assignment_key(assay_assignments, sample_ids)[missing_assay]

  missing_timepoint <- is.na(sample_data$timepoint_id) | !nzchar(sample_data$timepoint_id)
  sample_data$timepoint_id[missing_timepoint] <- time_constraint$timepoint_id[missing_timepoint]

  missing_time_index <- is.na(sample_data$time_index)
  sample_data$time_index[missing_time_index] <- time_constraint$time_index[missing_time_index]

  missing_order <- is.na(sample_data$order_rank)
  sample_data$order_rank[missing_order] <- time_constraint$order_rank[missing_order]

  sample_data
}

.split_spec_constraint_summary <- function(constraint) {
  if (is.null(constraint)) {
    return(list())
  }

  sample_map <- constraint$sample_map
  group_counts <- table(sample_map$group_id)
  list(
    mode = constraint$metadata$mode %||% NA_character_,
    strategy = constraint$metadata$strategy %||% constraint$strategy,
    n_samples = nrow(sample_map),
    n_groups = length(group_counts),
    n_singleton_groups = sum(group_counts == 1L),
    warnings = constraint$metadata$warnings %||% character()
  )
}

#' Translate splitGraph Constraints into Stable Split Specifications
#'
#' Translate graph-derived split constraints into a stable, inspectable
#' structure for sample-level grouping, blocking, and ordering, perform
#' preflight structural checks on that translation, and summarize structural
#' leakage risks.
#'
#' The translation layer always produces canonical sample-level columns
#' including \code{sample_id}, \code{sample_node_id}, \code{group_id}, and
#' \code{primary_group}. When available, it also carries \code{batch_group},
#' \code{study_group}, \code{timepoint_id}, \code{time_index}, and
#' \code{order_rank}. Missing but relevant fields are retained as \code{NA}
#' columns rather than omitted.
#'
#' When only a subset of samples has ordering metadata, the translated split
#' spec still exposes that partial ordering through \code{time_var}, but
#' \code{ordering_required} remains \code{FALSE}. Ordering is only marked as
#' required when the constraint implies complete ordering coverage.
#'
#' The split-spec validator checks:
#' \itemize{
#'   \item missing required columns
#'   \item missing or duplicated sample identifiers
#'   \item missing grouping assignments
#'   \item singleton-only grouping structures
#'   \item missing ordering when ordering is required
#'   \item invalid or empty block variables
#' }
#'
#' Repeated validation of the same split spec yields deterministic issue IDs
#' and diagnostics, which makes the returned validation object stable across
#' runs.
#'
#' The produced \code{split_spec} is tool-agnostic. Downstream consumers are
#' expected to provide their own adapters to convert a \code{split_spec} into
#' their native split representation, so \pkg{splitGraph} has no runtime
#' dependency on any of them.
#'
#' \code{summarize_leakage_risks()} reuses \code{validate_graph()} and
#' \code{split_constraint} metadata rather than duplicating downstream
#' evaluation logic.
#'
#' @param constraint A \code{split_constraint}.
#' @param graph A \code{dependency_graph}.
#' @param split_spec An optional \code{split_spec}.
#' @param validation An optional \code{depgraph_validation_report}.
#' @param x A \code{split_spec}.
#' @return \code{as_split_spec()} returns a \code{split_spec}.
#'   \code{validate_split_spec()} returns a \code{split_spec_validation}.
#'   \code{summarize_leakage_risks()} returns a \code{leakage_risk_summary}.
#' @examples
#' meta <- data.frame(
#'   sample_id  = c("S1", "S2", "S3", "S4"),
#'   subject_id = c("P1", "P1", "P2", "P2")
#' )
#' g <- graph_from_metadata(meta)
#'
#' constraint <- derive_split_constraints(g, mode = "subject")
#' spec <- as_split_spec(constraint, graph = g)
#' validate_split_spec(spec)
#' summarize_leakage_risks(g, constraint = constraint, split_spec = spec)
#' @export
as_split_spec <- function(constraint, graph = NULL) {
  .depgraph_assert(inherits(constraint, "split_constraint"), "`constraint` must be a `split_constraint`.")

  sample_map <- constraint$sample_map
  sample_data <- .split_spec_sample_data_template(nrow(sample_map))
  sample_data$sample_id <- as.character(sample_map$sample_id)
  sample_data$sample_node_id <- as.character(sample_map$sample_node_id)
  sample_data$group_id <- as.character(sample_map$group_id)
  sample_data$primary_group <- sample_data$group_id

  mode <- constraint$metadata$mode %||% NA_character_
  strategy <- constraint$metadata$strategy %||% constraint$strategy

  if (identical(mode, "batch")) {
    sample_data$batch_group <- as.character(sample_map$group_label)
  }
  if (identical(mode, "study")) {
    sample_data$study_group <- as.character(sample_map$group_label)
  }
  if (identical(mode, "site")) {
    sample_data$site_group <- as.character(sample_map$group_label)
  }
  if (identical(mode, "region")) {
    sample_data$region_group <- as.character(sample_map$group_label)
  }
  if (identical(mode, "platform")) {
    sample_data$platform_group <- as.character(sample_map$group_label)
  }
  if (identical(mode, "assay")) {
    sample_data$assay_group <- as.character(sample_map$group_label)
  }

  if ("timepoint_id" %in% names(sample_map)) {
    sample_data$timepoint_id <- as.character(sample_map$timepoint_id)
  }
  if ("time_index" %in% names(sample_map)) {
    sample_data$time_index <- suppressWarnings(as.numeric(sample_map$time_index))
  }
  if ("order_rank" %in% names(sample_map)) {
    sample_data$order_rank <- suppressWarnings(as.integer(sample_map$order_rank))
  }

  enrichment_used <- FALSE
  if (!is.null(graph)) {
    sample_data <- .split_spec_enrich_from_graph(sample_data, graph)
    enrichment_used <- TRUE
  }

  block_vars <- c()
  if (!all(is.na(sample_data$batch_group))) {
    block_vars <- c(block_vars, "batch_group")
  }
  if (!all(is.na(sample_data$study_group))) {
    block_vars <- c(block_vars, "study_group")
  }
  if (!all(is.na(sample_data$site_group))) {
    block_vars <- c(block_vars, "site_group")
  }
  if (!all(is.na(sample_data$region_group))) {
    block_vars <- c(block_vars, "region_group")
  }
  if (!all(is.na(sample_data$platform_group))) {
    block_vars <- c(block_vars, "platform_group")
  }
  if (!all(is.na(sample_data$assay_group))) {
    block_vars <- c(block_vars, "assay_group")
  }

  time_var <- if (!all(is.na(sample_data$order_rank))) "order_rank" else NULL
  ordering_required <- isTRUE(constraint$recommended_downstream_args$ordering_required)

  split_spec(
    sample_data = sample_data,
    group_var = "group_id",
    block_vars = block_vars,
    time_var = time_var,
    ordering_required = ordering_required,
    constraint_mode = mode,
    constraint_strategy = strategy,
    recommended_resampling = .split_spec_resampling_hint(mode, strategy),
    metadata = list(
      graph_name = if (is.null(graph)) NULL else graph$metadata$graph_name,
      dataset_name = if (is.null(graph)) NULL else graph$metadata$dataset_name,
      source_mode = mode,
      source_strategy = strategy,
      relations_used = constraint$metadata$relations_used %||% character(),
      splitgraph_version = .depgraph_package_version(),
      derived_at = format(Sys.time(), "%Y-%m-%dT%H:%M:%OS6%z"),
      n_samples = nrow(sample_data),
      n_groups = length(unique(sample_data$group_id)),
      warnings = constraint$metadata$warnings %||% character(),
      enriched_from_graph = enrichment_used
    )
  )
}

#' @rdname as_split_spec
#' @export
validate_split_spec <- function(x) {
  .depgraph_assert(inherits(x, "split_spec"), "`x` must be a `split_spec`.")
  issues <- list()
  data <- x$sample_data

  required_cols <- c("sample_id", "sample_node_id", x$group_var)
  missing_cols <- setdiff(required_cols, names(data))
  if (length(missing_cols) > 0L) {
    issues[[length(issues) + 1L]] <- .split_spec_new_issue(
      severity = "error",
      code = "missing_required_columns",
      message = paste0("Split spec is missing required columns: ", paste(missing_cols, collapse = ", ")),
      n_affected = length(missing_cols),
      details = list(columns = missing_cols)
    )
  }

  if ("sample_id" %in% names(data)) {
    if (any(is.na(data$sample_id) | !nzchar(as.character(data$sample_id)))) {
      issues[[length(issues) + 1L]] <- .split_spec_new_issue(
        severity = "error",
        code = "missing_sample_id",
        message = "Split spec contains missing `sample_id` values.",
        n_affected = sum(is.na(data$sample_id) | !nzchar(as.character(data$sample_id)))
      )
    }

    dup_sample_ids <- unique(data$sample_id[duplicated(data$sample_id)])
    if (length(dup_sample_ids) > 0L) {
      issues[[length(issues) + 1L]] <- .split_spec_new_issue(
        severity = "error",
        code = "duplicate_sample_id",
        message = paste0("Split spec contains duplicated sample IDs: ", paste(dup_sample_ids, collapse = ", ")),
        n_affected = length(dup_sample_ids),
        details = list(sample_ids = dup_sample_ids)
      )
    }
  }

  if (x$group_var %in% names(data)) {
    group_values <- data[[x$group_var]]
    if (any(is.na(group_values) | !nzchar(as.character(group_values)))) {
      issues[[length(issues) + 1L]] <- .split_spec_new_issue(
        severity = "error",
        code = "missing_group_id",
        message = paste0("Split spec contains missing values in `", x$group_var, "`."),
        n_affected = sum(is.na(group_values) | !nzchar(as.character(group_values)))
      )
    } else {
      group_sizes <- table(group_values)
      if (length(group_sizes) > 0L && all(group_sizes == 1L)) {
        issues[[length(issues) + 1L]] <- .split_spec_new_issue(
          severity = "advisory",
          code = "singleton_groups_only",
          message = "All split groups are singletons; grouping may provide little protection against structural leakage.",
          n_affected = length(group_sizes)
        )
      }
    }
  } else {
    issues[[length(issues) + 1L]] <- .split_spec_new_issue(
      severity = "error",
      code = "invalid_group_var",
      message = paste0("Declared `group_var` is not present in `sample_data`: ", x$group_var)
    )
  }

  if (isTRUE(x$ordering_required)) {
    if (is.null(x$time_var) || !x$time_var %in% names(data)) {
      issues[[length(issues) + 1L]] <- .split_spec_new_issue(
        severity = "error",
        code = "invalid_time_var",
        message = "Ordering is required but the declared `time_var` is missing from `sample_data`."
      )
    } else if (any(is.na(data[[x$time_var]]))) {
      issues[[length(issues) + 1L]] <- .split_spec_new_issue(
        severity = "error",
        code = "missing_required_ordering",
        message = "Ordering is required but some samples are missing ordering values.",
        n_affected = sum(is.na(data[[x$time_var]]))
      )
    }
  }

  if (length(x$block_vars) > 0L) {
    for (block_var in x$block_vars) {
      if (!block_var %in% names(data)) {
        issues[[length(issues) + 1L]] <- .split_spec_new_issue(
          severity = "error",
          code = "invalid_block_var",
          message = paste0("Declared block variable is not present in `sample_data`: ", block_var)
        )
      } else if (all(is.na(data[[block_var]]) | !nzchar(as.character(data[[block_var]])))) {
        issues[[length(issues) + 1L]] <- .split_spec_new_issue(
          severity = "warning",
          code = "empty_block_var",
          message = paste0("Block variable `", block_var, "` is present but empty for all samples.")
        )
      }
    }
  }

  split_spec_validation(
    issues = .split_spec_bind_issues(issues),
    metadata = list(
      n_samples = nrow(data),
      n_groups = if (x$group_var %in% names(data)) length(unique(data[[x$group_var]])) else NA_integer_
    )
  )
}

.depgraph_constraint_via_modes <- function(constraint) {
  # Constraints store composite `via` as capitalized node types
  # (e.g., "Subject"); convert to lower-case mode names for severance lookup.
  via <- constraint$metadata$via %||% character()
  if (length(via) == 0L) return(character())
  reverse <- stats::setNames(
    names(.depgraph_constraint_mode_map),
    unname(.depgraph_constraint_mode_map)
  )
  matched <- reverse[as.character(via)]
  unname(matched[!is.na(matched)])
}

# Map a validation issue code to whether the chosen constraint mode (+ optional
# composite `via`) severs that leakage path. Returns TRUE/FALSE for codes the
# package can reason about, NA for codes whose severance is not a function of
# split-mode (structural errors, etc.) or when no constraint is supplied.
.leakage_severed_by_constraint <- function(code, mode = NULL, via_modes = character()) {
  if (is.null(mode) || !nzchar(mode)) return(NA)
  via_modes <- as.character(via_modes %||% character())
  composite_covers <- function(needed) {
    identical(mode, "composite") && needed %in% via_modes
  }
  switch(
    as.character(code),
    repeated_subject_samples     = identical(mode, "subject") || composite_covers("subject"),
    subject_cross_study_overlap  = mode %in% c("subject", "study") ||
                                     composite_covers("subject") ||
                                     composite_covers("study"),
    heavy_batch_reuse            = identical(mode, "batch") || composite_covers("batch"),
    missing_time_ordering        = identical(mode, "time") || composite_covers("time"),
    per_dataset_featureset       = FALSE,
    shared_featureset_provenance = FALSE,
    NA
  )
}

.leakage_summary_from_validation <- function(validation, constraint = NULL) {
  if (nrow(validation$issues) == 0L) {
    return(data.frame(
      severity = character(),
      category = character(),
      message = character(),
      source = character(),
      n_affected = integer(),
      severed = logical(),
      stringsAsFactors = FALSE
    ))
  }

  mode <- if (!is.null(constraint)) as.character(constraint$metadata$mode %||% NA) else NA_character_
  via_modes <- if (!is.null(constraint)) .depgraph_constraint_via_modes(constraint) else character()

  severed <- vapply(
    validation$issues$code,
    function(code) {
      .leakage_severed_by_constraint(code, mode = if (is.na(mode)) NULL else mode, via_modes = via_modes)
    },
    logical(1)
  )

  data.frame(
    severity = validation$issues$severity,
    category = validation$issues$code,
    message = validation$issues$message,
    source = "validation",
    n_affected = vapply(validation$issues$node_ids, length, integer(1)),
    severed = severed,
    stringsAsFactors = FALSE
  )
}

.leakage_summary_from_constraint <- function(constraint) {
  empty <- data.frame(
    severity = character(),
    category = character(),
    message = character(),
    source = character(),
    n_affected = integer(),
    severed = logical(),
    stringsAsFactors = FALSE
  )
  if (is.null(constraint)) {
    return(empty)
  }

  diagnostics <- list()
  warnings <- constraint$metadata$warnings %||% character()
  if (length(warnings) > 0L) {
    for (warning_msg in warnings) {
      diagnostics[[length(diagnostics) + 1L]] <- data.frame(
        severity = "warning",
        category = "constraint_warning",
        message = warning_msg,
        source = "constraint",
        n_affected = nrow(constraint$sample_map),
        severed = NA,
        stringsAsFactors = FALSE
      )
    }
  }

  if ("group_id" %in% names(constraint$sample_map)) {
    group_sizes <- table(constraint$sample_map$group_id)
    if (length(group_sizes) > 0L && mean(group_sizes == 1L) > 0.5) {
      diagnostics[[length(diagnostics) + 1L]] <- data.frame(
        severity = "advisory",
        category = "singleton_heavy_constraint",
        message = "The derived split constraint is dominated by singleton groups.",
        source = "constraint",
        n_affected = sum(group_sizes == 1L),
        severed = NA,
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(diagnostics) == 0L) {
    return(empty)
  }

  out <- do.call(rbind, diagnostics)
  row.names(out) <- NULL
  out
}

.leakage_summary_from_split_spec <- function(split_spec) {
  empty <- data.frame(
    severity = character(),
    category = character(),
    message = character(),
    source = character(),
    n_affected = integer(),
    severed = logical(),
    stringsAsFactors = FALSE
  )
  if (is.null(split_spec)) {
    return(list(diagnostics = empty, summary = list()))
  }

  validation <- validate_split_spec(split_spec)
  diagnostics <- list()
  if (nrow(validation$issues) == 0L) {
    diagnostics[[length(diagnostics) + 1L]] <- data.frame(
      severity = "advisory",
      category = "split_spec_ready",
      message = "Split spec passed preflight validation.",
      source = "split_spec",
      n_affected = nrow(split_spec$sample_data),
      severed = NA,
      stringsAsFactors = FALSE
    )
  } else {
    diagnostics[[length(diagnostics) + 1L]] <- data.frame(
      severity = validation$issues$severity,
      category = validation$issues$code,
      message = validation$issues$message,
      source = "split_spec",
      n_affected = validation$issues$n_affected,
      severed = NA,
      stringsAsFactors = FALSE
    )
  }

  if (!is.null(split_spec$time_var) && split_spec$time_var %in% names(split_spec$sample_data)) {
    complete_ordering <- sum(!is.na(split_spec$sample_data[[split_spec$time_var]]))
    diagnostics[[length(diagnostics) + 1L]] <- data.frame(
      severity = "advisory",
      category = "ordering_available",
      message = paste0(
        "Split spec provides ordering through `", split_spec$time_var,
        "` for ", complete_ordering, " of ", nrow(split_spec$sample_data), " samples."
      ),
      source = "split_spec",
      n_affected = complete_ordering,
      severed = NA,
      stringsAsFactors = FALSE
    )
  }

  if (length(split_spec$block_vars) > 0L) {
    for (block_var in split_spec$block_vars) {
      available <- if (block_var %in% names(split_spec$sample_data)) {
        sum(!is.na(split_spec$sample_data[[block_var]]) & nzchar(as.character(split_spec$sample_data[[block_var]])))
      } else {
        0L
      }
      diagnostics[[length(diagnostics) + 1L]] <- data.frame(
        severity = "advisory",
        category = "blocking_available",
        message = paste0(
          "Split spec provides blocking variable `", block_var,
          "` for ", available, " of ", nrow(split_spec$sample_data), " samples."
        ),
        source = "split_spec",
        n_affected = available,
        severed = NA,
        stringsAsFactors = FALSE
      )
    }
  }

  group_sizes <- table(split_spec$sample_data[[split_spec$group_var]])
  if (length(group_sizes) > 0L && mean(group_sizes == 1L) > 0.5) {
    diagnostics[[length(diagnostics) + 1L]] <- data.frame(
      severity = "advisory",
      category = "split_spec_singleton_heavy",
      message = "Split spec grouping is dominated by singleton groups.",
      source = "split_spec",
      n_affected = sum(group_sizes == 1L),
      severed = NA,
      stringsAsFactors = FALSE
    )
  }

  diagnostics <- do.call(rbind, diagnostics)
  row.names(diagnostics) <- NULL

  list(
    diagnostics = diagnostics,
    summary = list(
      valid = validation$valid,
      n_samples = nrow(split_spec$sample_data),
      n_groups = length(unique(split_spec$sample_data[[split_spec$group_var]])),
      recommended_resampling = split_spec$recommended_resampling,
      time_var = split_spec$time_var,
      ordering_required = split_spec$ordering_required,
      block_vars = split_spec$block_vars,
      singleton_groups = if (length(group_sizes) == 0L) 0L else sum(group_sizes == 1L)
    )
  )
}

#' @rdname as_split_spec
#' @export
summarize_leakage_risks <- function(graph, constraint = NULL, split_spec = NULL, validation = NULL) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")
  if (!is.null(constraint)) {
    .depgraph_assert(inherits(constraint, "split_constraint"), "`constraint` must be a `split_constraint`.")
  }
  if (!is.null(split_spec)) {
    .depgraph_assert(inherits(split_spec, "split_spec"), "`split_spec` must be a `split_spec`.")
  }

  validation <- validation %||% validate_graph(graph)
  validation_diag <- .leakage_summary_from_validation(validation, constraint = constraint)
  constraint_diag <- .leakage_summary_from_constraint(constraint)
  split_spec_info <- .leakage_summary_from_split_spec(split_spec)

  diagnostics <- do.call(rbind, Filter(function(x) is.data.frame(x) && nrow(x) > 0L, list(
    validation_diag,
    constraint_diag,
    split_spec_info$diagnostics
  )))
  if (is.null(diagnostics)) {
    diagnostics <- data.frame(
      severity = character(),
      category = character(),
      message = character(),
      source = character(),
      n_affected = integer(),
      stringsAsFactors = FALSE
    )
  } else {
    row.names(diagnostics) <- NULL
  }

  overview <- if (nrow(diagnostics) == 0L) {
    "No structural leakage risks were detected."
  } else {
    paste0(
      "Detected ", nrow(diagnostics), " structural leakage diagnostics across validation",
      if (!is.null(constraint)) ", constraint" else "",
      if (!is.null(split_spec)) ", and split-spec readiness" else "",
      "."
    )
  }

  leakage_risk_summary(
    overview = overview,
    diagnostics = diagnostics,
    validation_summary = validation$summary,
    constraint_summary = .split_spec_constraint_summary(constraint),
    split_spec_summary = split_spec_info$summary,
    metadata = list(
      graph_name = graph$metadata$graph_name,
      dataset_name = graph$metadata$dataset_name
    )
  )
}

make_split_spec_graph <- function(time_source = c("index", "precedence")) {
  time_source <- match.arg(time_source)

  meta <- data.frame(
    sample_id = c("S1", "S2", "S3", "S4", "S5"),
    subject_id = c("P1", "P1", "P2", "P3", "P4"),
    batch_id = c("B1", "B2", "B1", "B3", NA),
    study_id = c("ST1", "ST1", "ST2", "ST2", "ST3"),
    timepoint_id = c("T0", "T1", "T0", "T2", NA),
    stringsAsFactors = FALSE
  )

  sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
  study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")

  time_data <- unique(meta[!is.na(meta$timepoint_id), "timepoint_id", drop = FALSE])
  time_nodes <- if (identical(time_source, "index")) {
    create_nodes(
      transform(time_data, time_index = c(0L, 1L, 2L)),
      type = "Timepoint",
      id_col = "timepoint_id",
      attr_cols = c("time_index")
    )
  } else {
    create_nodes(time_data, type = "Timepoint", id_col = "timepoint_id")
  }

  edge_sets <- list(
    create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject"),
    create_edges(meta, "sample_id", "batch_id", "Sample", "Batch", "sample_processed_in_batch", allow_missing = TRUE),
    create_edges(meta, "sample_id", "study_id", "Sample", "Study", "sample_from_study"),
    create_edges(meta, "sample_id", "timepoint_id", "Sample", "Timepoint", "sample_collected_at_timepoint", allow_missing = TRUE)
  )

  if (identical(time_source, "precedence")) {
    edge_sets <- c(edge_sets, list(
      create_edges(
        data.frame(
          from_timepoint = c("T0", "T1"),
          to_timepoint = c("T1", "T2"),
          stringsAsFactors = FALSE
        ),
        "from_timepoint", "to_timepoint", "Timepoint", "Timepoint", "timepoint_precedes"
      )
    ))
  }

  build_dependency_graph(
    nodes = list(sample_nodes, subject_nodes, batch_nodes, study_nodes, time_nodes),
    edges = edge_sets,
    graph_name = paste0("split_spec_graph_", time_source),
    dataset_name = "SplitSpecDemo"
  )
}

test_that("as_split_spec translates constraints into canonical sample data", {
  graph <- make_split_spec_graph("index")
  constraint <- derive_split_constraints(graph, mode = "batch")
  spec <- as_split_spec(constraint, graph = graph)
  validation <- validate_split_spec(spec)

  expect_s3_class(spec, "split_spec")
  expect_true(all(c(
    "sample_id", "sample_node_id", "group_id", "primary_group", "batch_group",
    "study_group", "timepoint_id", "time_index", "order_rank"
  ) %in% names(spec$sample_data)))
  expect_equal(spec$group_var, "group_id")
  expect_true("batch_group" %in% spec$block_vars)
  expect_equal(spec$recommended_resampling, "blocked_cv")
  expect_equal(spec$sample_data$batch_group[spec$sample_data$sample_id == "S1"], "B1")
  expect_equal(spec$sample_data$study_group[spec$sample_data$sample_id == "S1"], "ST1")
  expect_equal(spec$sample_data$order_rank[spec$sample_data$sample_id == "S4"], 3L)
  expect_equal(spec$metadata$source_mode, "batch")
  expect_equal(spec$metadata$source_strategy, "batch")
  expect_true(validation$valid)
})

test_that("as_split_spec preserves time-aware ordering semantics", {
  graph <- make_split_spec_graph("precedence")
  constraint <- derive_split_constraints(graph, mode = "time")
  spec <- as_split_spec(constraint, graph = graph)
  validation <- validate_split_spec(spec)

  expect_s3_class(spec, "split_spec")
  expect_identical(spec$time_var, "order_rank")
  expect_false(isTRUE(spec$ordering_required))
  expect_equal(spec$sample_data$order_rank[spec$sample_data$sample_id == "S2"], 2L)
  expect_equal(spec$recommended_resampling, "ordered_split")
  expect_true(validation$valid)
})

test_that("partial time coverage does not force invalid ordering requirements in generated split specs", {
  graph <- make_split_spec_graph("index")
  constraint <- derive_split_constraints(graph, mode = "time")
  spec <- as_split_spec(constraint, graph = graph)
  validation <- validate_split_spec(spec)

  expect_false(isTRUE(spec$ordering_required))
  expect_identical(spec$time_var, "order_rank")
  expect_true(any(is.na(spec$sample_data$order_rank)))
  expect_true(validation$valid)
  expect_false(any(validation$issues$code == "missing_required_ordering"))
})

test_that("validate_split_spec detects missing ordering and duplicate samples", {
  graph <- make_split_spec_graph("index")
  constraint <- derive_split_constraints(graph, mode = "time")
  spec <- as_split_spec(constraint, graph = graph)
  broken <- spec
  broken$ordering_required <- TRUE
  broken$sample_data$order_rank[] <- NA_integer_
  broken$sample_data <- rbind(broken$sample_data, broken$sample_data[1, , drop = FALSE])

  validation <- validate_split_spec(broken)

  expect_s3_class(validation, "split_spec_validation")
  expect_false(validation$valid)
  expect_true(any(validation$issues$code == "duplicate_sample_id"))
  expect_true(any(validation$issues$code == "missing_required_ordering"))
  expect_true(length(capture.output(print(validation))) > 0L)
})

test_that("validate_split_spec is deterministic across repeated runs", {
  graph <- make_split_spec_graph("index")
  constraint <- derive_split_constraints(graph, mode = "time")
  spec <- as_split_spec(constraint, graph = graph)
  broken <- spec
  broken$sample_data$order_rank[] <- NA_integer_
  broken$sample_data <- rbind(broken$sample_data, broken$sample_data[1, , drop = FALSE])

  validation_a <- validate_split_spec(broken)
  validation_b <- validate_split_spec(broken)

  expect_identical(validation_a$issues$code, validation_b$issues$code)
  expect_identical(validation_a$issues$message, validation_b$issues$message)
  expect_identical(validation_a$issues$n_affected, validation_b$issues$n_affected)
  expect_identical(validation_a$issues$issue_id, validation_b$issues$issue_id)
})

test_that("summarize_leakage_risks marks validation risks not severed by the chosen constraint", {
  # Regression test for C2: summarize_leakage_risks reused validate_graph()
  # but never asked whether the CHOSEN constraint mode actually severs each
  # reported leakage path. A user picking `mode = "subject"` on a graph with
  # heavy batch reuse could ship a spec whose splits leak batch effects, and
  # the summary would not flag the mismatch.
  meta <- data.frame(
    sample_id  = paste0("S", 1:6),
    subject_id = paste0("P", 1:6),                          # each subject distinct
    batch_id   = c("B1", "B1", "B1", "B2", "B2", "B3"),     # B1 holds 3 of 6
    stringsAsFactors = FALSE
  )
  graph <- graph_from_metadata(meta, graph_name = "heavy_batch_demo")

  subject_c <- derive_split_constraints(graph, mode = "subject")
  summary_subject <- summarize_leakage_risks(graph, constraint = subject_c)

  expect_true("severed" %in% names(summary_subject$diagnostics),
              info = "Diagnostics must carry a `severed` column when a constraint is provided.")
  heavy_row_subject <- summary_subject$diagnostics[
    summary_subject$diagnostics$category == "heavy_batch_reuse", , drop = FALSE
  ]
  expect_equal(nrow(heavy_row_subject), 1L,
               info = "heavy_batch_reuse must appear when one batch holds >= ceiling(n*0.5) samples.")
  expect_false(isTRUE(heavy_row_subject$severed[[1L]]),
               info = "mode = 'subject' does not block batch leakage; row must be marked not severed.")

  # Same graph, but mode = 'batch' should sever the same risk.
  batch_c <- derive_split_constraints(graph, mode = "batch")
  summary_batch <- summarize_leakage_risks(graph, constraint = batch_c)
  heavy_row_batch <- summary_batch$diagnostics[
    summary_batch$diagnostics$category == "heavy_batch_reuse", , drop = FALSE
  ]
  expect_true(isTRUE(heavy_row_batch$severed[[1L]]),
              info = "mode = 'batch' groups samples by batch; heavy_batch_reuse must be marked severed.")

  # Without a constraint, `severed` should be NA (no claim about severance).
  summary_no_c <- summarize_leakage_risks(graph)
  expect_true("severed" %in% names(summary_no_c$diagnostics))
  heavy_row_nc <- summary_no_c$diagnostics[
    summary_no_c$diagnostics$category == "heavy_batch_reuse", , drop = FALSE
  ]
  expect_true(is.na(heavy_row_nc$severed[[1L]]),
              info = "Without a constraint, severance is not knowable; column must be NA.")
})

test_that("summarize_leakage_risks combines graph, constraint, and split-spec diagnostics", {
  graph <- make_split_spec_graph("index")
  constraint <- derive_split_constraints(graph, mode = "composite", strategy = "strict", via = c("Subject", "Batch"))
  spec <- as_split_spec(constraint, graph = graph)
  summary_obj <- summarize_leakage_risks(graph, constraint = constraint, split_spec = spec)

  expect_s3_class(summary_obj, "leakage_risk_summary")
  expect_true(is.character(summary_obj$overview))
  expect_true(all(c("severity", "category", "message", "source", "n_affected") %in% names(summary_obj$diagnostics)))
  expect_true(any(summary_obj$diagnostics$source == "validation"))
  expect_true(any(summary_obj$diagnostics$source == "constraint"))
  expect_true(any(summary_obj$diagnostics$source == "split_spec"))
  expect_true(any(summary_obj$diagnostics$category == "blocking_available"))
  expect_true(any(summary_obj$diagnostics$category == "ordering_available"))
  expect_true("block_vars" %in% names(summary_obj$split_spec_summary))
  expect_true("ordering_required" %in% names(summary_obj$split_spec_summary))
  expect_true(length(capture.output(print(summary_obj))) > 0L)
})

# ---- leakage-summary edge branches ------------------------------------------

test_that("summarize_leakage_risks reports no risks for a structurally clean graph", {
  clean <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    study_id   = c("ST1", "ST2"),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(clean)
  expect_true(validate_graph(g)$valid)

  # No constraint and no split_spec: every summary source is empty.
  rs <- summarize_leakage_risks(g)
  expect_s3_class(rs, "leakage_risk_summary")
  expect_equal(nrow(as.data.frame(rs)), 0L)
  expect_match(rs$overview, "No structural leakage risks")
})

test_that("summarize_leakage_risks surfaces constraint warnings as diagnostics", {
  meta <- data.frame(
    sample_id  = c("S1", "S2", "S3"),
    subject_id = c("P1", "P2", "P3"),
    site_id    = c("NYC", "NYC", NA),   # S3 has no site -> constraint warning
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  constraint <- derive_split_constraints(g, mode = "site")
  expect_true(length(constraint$metadata$warnings) > 0L)

  df <- as.data.frame(summarize_leakage_risks(g, constraint = constraint))
  warn_rows <- df[df$category == "constraint_warning", , drop = FALSE]
  expect_true(nrow(warn_rows) >= 1L)
  expect_true(all(warn_rows$source == "constraint"))
  expect_true(all(is.na(warn_rows$severed)))
})

test_that("summarize_leakage_risks reports a failing preflight and phantom block vars", {
  meta <- data.frame(
    sample_id  = c("S1", "S2", "S3", "S4"),
    subject_id = c("P1", "P1", "P2", "P2"),
    batch_id   = c("B1", "B2", "B1", "B2"),
    stringsAsFactors = FALSE
  )
  g <- graph_from_metadata(meta)
  spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)

  # Blank a declared block variable so preflight validation flags it, and add a
  # phantom block variable that is absent from sample_data.
  spec$sample_data$batch_group <- NA_character_
  spec$block_vars <- c(spec$block_vars, "phantom_group")

  df <- as.data.frame(summarize_leakage_risks(g, split_spec = spec))
  # The preflight issue is carried into the split_spec diagnostics.
  expect_true(any(df$source == "split_spec" & df$severity != "advisory"))
  # The phantom block variable is reported as available for 0 samples.
  expect_true(any(grepl("phantom_group.*for 0 of", df$message)))
})

# One-shot builder that derives nodes, edges, and timepoint ordering from
# canonical metadata columns.

.depgraph_auto_specs <- list(
  list(type = "Subject",    id_col = "subject_id",    relation = "sample_belongs_to_subject"),
  list(type = "Batch",      id_col = "batch_id",      relation = "sample_processed_in_batch"),
  list(type = "Study",      id_col = "study_id",      relation = "sample_from_study"),
  list(type = "Timepoint",  id_col = "timepoint_id",  relation = "sample_collected_at_timepoint"),
  list(type = "Assay",      id_col = "assay_id",      relation = "sample_measured_by_assay"),
  list(type = "FeatureSet", id_col = "featureset_id", relation = "sample_uses_featureset"),
  list(type = "Site",       id_col = "site_id",       relation = "sample_collected_at_site"),
  list(type = "Region",     id_col = "region_id",     relation = "sample_located_in_region"),
  list(type = "Platform",   id_col = "platform_id",   relation = "sample_run_on_platform")
)

#' Build a Dependency Graph Directly from a Metadata Table
#'
#' One-shot convenience builder that auto-detects canonical columns in a
#' metadata table, creates the corresponding node and edge sets, optionally
#' derives timepoint ordering from \code{time_index}, and assembles a
#' \code{dependency_graph}. Columns that are absent or entirely missing are
#' silently skipped.
#'
#' @param meta A \code{data.frame} containing one row per sample and optional
#'   canonical columns: \code{sample_id} (required), \code{subject_id},
#'   \code{batch_id}, \code{study_id}, \code{timepoint_id}, \code{time_index},
#'   \code{assay_id}, \code{featureset_id}, \code{outcome_id}, or
#'   \code{outcome_value}.
#' @param columns Optional named character vector passed to
#'   \code{ingest_metadata()} to rename user columns to canonical names.
#' @param dataset_name,graph_name Optional metadata labels.
#' @param outcome_scope Either \code{"sample"} (default) or \code{"subject"}.
#'   Controls whether outcome edges attach to samples or subjects.
#' @param time_precedence If \code{TRUE} and \code{time_index} is present,
#'   derive \code{timepoint_precedes} edges from the ordering of
#'   \code{time_index}.
#' @param validate Forwarded to \code{build_dependency_graph()}.
#' @param validation_overrides Forwarded to \code{build_dependency_graph()}.
#' @return A validated \code{dependency_graph}.
#' @examples
#' meta <- data.frame(
#'   sample_id  = c("S1", "S2", "S3", "S4"),
#'   subject_id = c("P1", "P1", "P2", "P2"),
#'   batch_id   = c("B1", "B2", "B1", "B2"),
#'   timepoint_id = c("T1", "T2", "T1", "T2"),
#'   time_index = c(1, 2, 1, 2),
#'   outcome_id = c("ctrl", "case", "ctrl", "case")
#' )
#'
#' g <- graph_from_metadata(meta, graph_name = "demo")
#' g
#' @export
graph_from_metadata <- function(meta,
                                columns = NULL,
                                dataset_name = NULL,
                                graph_name = NULL,
                                outcome_scope = c("sample", "subject"),
                                time_precedence = TRUE,
                                validate = TRUE,
                                validation_overrides = list()) {
  .depgraph_assert(is.data.frame(meta), "`meta` must be a data.frame.")
  outcome_scope <- match.arg(outcome_scope)

  meta <- ingest_metadata(meta, col_map = columns, dataset_name = dataset_name, strict = TRUE)

  sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  node_sets <- list(sample_nodes)
  edge_sets <- list()

  for (spec in .depgraph_auto_specs) {
    if (!(spec$id_col %in% names(meta))) next
    present <- !is.na(meta[[spec$id_col]]) & nzchar(as.character(meta[[spec$id_col]]))
    if (!any(present)) next
    sub <- meta[present, , drop = FALSE]

    node_sets[[length(node_sets) + 1L]] <- create_nodes(
      sub, type = spec$type, id_col = spec$id_col
    )
    edge_sets[[length(edge_sets) + 1L]] <- create_edges(
      sub,
      from_col = "sample_id", to_col = spec$id_col,
      from_type = "Sample", to_type = spec$type,
      relation = spec$relation
    )
  }

  has_timepoint <- "timepoint_id" %in% names(meta) &&
    any(!is.na(meta$timepoint_id) & nzchar(as.character(meta$timepoint_id)))
  if (isTRUE(time_precedence) && has_timepoint && "time_index" %in% names(meta)) {
    tp_rows <- !is.na(meta$timepoint_id) & nzchar(as.character(meta$timepoint_id)) &
      !is.na(meta$time_index)
    if (any(tp_rows)) {
      tp_df <- unique(data.frame(
        timepoint_id = as.character(meta$timepoint_id[tp_rows]),
        time_index = meta$time_index[tp_rows],
        stringsAsFactors = FALSE
      ))
      tp_df <- tp_df[order(tp_df$time_index, tp_df$timepoint_id), , drop = FALSE]
      tp_df <- tp_df[!duplicated(tp_df$timepoint_id), , drop = FALSE]
      if (nrow(tp_df) >= 2L) {
        precedes_df <- data.frame(
          from_id = tp_df$timepoint_id[-nrow(tp_df)],
          to_id = tp_df$timepoint_id[-1L],
          stringsAsFactors = FALSE
        )
        edge_sets[[length(edge_sets) + 1L]] <- create_edges(
          precedes_df,
          from_col = "from_id", to_col = "to_id",
          from_type = "Timepoint", to_type = "Timepoint",
          relation = "timepoint_precedes"
        )
      }
    }
  }

  outcome_col <- if ("outcome_id" %in% names(meta)) "outcome_id" else if ("outcome_value" %in% names(meta)) "outcome_value" else NULL
  if (!is.null(outcome_col)) {
    present <- !is.na(meta[[outcome_col]]) & nzchar(as.character(meta[[outcome_col]]))
    if (any(present)) {
      sub <- meta[present, , drop = FALSE]
      if (identical(outcome_col, "outcome_value") && is.numeric(meta[[outcome_col]])) {
        warning(
          "`outcome_value` is numeric: graph_from_metadata() will create one ",
          "Outcome node per distinct numeric value (e.g. `outcome:0`, `outcome:1`). ",
          "If you intended a class label, pass `outcome_id` (character) instead, ",
          "or coerce `outcome_value` to a character class label first.",
          call. = FALSE
        )
      }
      sub$outcome_id <- as.character(sub[[outcome_col]])
      node_sets[[length(node_sets) + 1L]] <- create_nodes(
        sub, type = "Outcome", id_col = "outcome_id"
      )
      if (identical(outcome_scope, "subject") && "subject_id" %in% names(sub)) {
        edge_sets[[length(edge_sets) + 1L]] <- create_edges(
          sub,
          from_col = "subject_id", to_col = "outcome_id",
          from_type = "Subject", to_type = "Outcome",
          relation = "subject_has_outcome"
        )
      } else {
        edge_sets[[length(edge_sets) + 1L]] <- create_edges(
          sub,
          from_col = "sample_id", to_col = "outcome_id",
          from_type = "Sample", to_type = "Outcome",
          relation = "sample_has_outcome"
        )
      }
    }
  }

  build_dependency_graph(
    nodes = node_sets,
    edges = edge_sets,
    graph_name = graph_name,
    dataset_name = dataset_name,
    validate = validate,
    validation_overrides = validation_overrides
  )
}

test_that("graph_from_metadata warns when outcome_value is numeric", {
  meta <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    subject_id = c("P1", "P2", "P3"),
    outcome_value = c(0, 1, 0),
    stringsAsFactors = FALSE
  )

  expect_warning(
    g <- graph_from_metadata(meta),
    "outcome_value",
    ignore.case = TRUE
  )
  expect_s3_class(g, "dependency_graph")
  # The Outcome nodes are still constructed (compatibility) but with numeric-derived keys.
  outcome_keys <- g$nodes$data$node_key[g$nodes$data$node_type == "Outcome"]
  expect_setequal(outcome_keys, c("0", "1"))
})

test_that("graph_from_metadata is silent when outcome_id is character", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    outcome_id = c("case", "ctrl"),
    stringsAsFactors = FALSE
  )

  expect_silent(g <- graph_from_metadata(meta))
  outcome_keys <- g$nodes$data$node_key[g$nodes$data$node_type == "Outcome"]
  expect_setequal(outcome_keys, c("case", "ctrl"))
})

test_that("graph_from_metadata is silent when outcome_value is character (case/ctrl)", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    outcome_value = c("case", "ctrl"),
    stringsAsFactors = FALSE
  )
  expect_silent(g <- graph_from_metadata(meta))
})

test_that("graph_from_metadata is silent when no outcome columns are present", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )
  expect_silent(graph_from_metadata(meta))
})

test_that("schema_version is stable and independent of package version", {
  # The data-model schema version must NOT change automatically when the
  # package version is bumped. Only an explicit, documented schema change
  # should bump it. This test locks the contract.
  expect_identical(splitGraph:::.depgraph_schema_version, "0.2.0")

  pkg_version <- as.character(utils::packageVersion("splitGraph"))
  expect_type(pkg_version, "character")
})

test_that("schema_version is exposed via the metadata of constructed objects", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  g <- graph_from_metadata(meta)
  expect_identical(g$metadata$schema_version, splitGraph:::.depgraph_schema_version)
  expect_identical(g$nodes$schema_version, splitGraph:::.depgraph_schema_version)
  expect_identical(g$edges$schema_version, splitGraph:::.depgraph_schema_version)
})

test_that("graph_from_metadata auto-detects canonical columns", {
  meta <- data.frame(
    sample_id    = c("S1", "S2", "S3", "S4"),
    subject_id   = c("P1", "P1", "P2", "P2"),
    batch_id     = c("B1", "B2", "B1", "B2"),
    timepoint_id = c("T1", "T2", "T1", "T2"),
    time_index   = c(1, 2, 1, 2),
    outcome_id   = c("ctrl", "case", "ctrl", "case"),
    stringsAsFactors = FALSE
  )

  g <- graph_from_metadata(meta, graph_name = "demo")

  expect_s3_class(g, "dependency_graph")
  node_types <- sort(unique(g$nodes$data$node_type))
  expect_true(all(c("Sample", "Subject", "Batch", "Timepoint", "Outcome") %in% node_types))

  edge_types <- sort(unique(g$edges$data$edge_type))
  expect_true("sample_belongs_to_subject" %in% edge_types)
  expect_true("sample_processed_in_batch" %in% edge_types)
  expect_true("sample_collected_at_timepoint" %in% edge_types)
  expect_true("sample_has_outcome" %in% edge_types)
  expect_true("timepoint_precedes" %in% edge_types)
})

test_that("graph_from_metadata skips absent columns silently", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  g <- graph_from_metadata(meta)
  expect_s3_class(g, "dependency_graph")
  expect_false("Batch" %in% g$nodes$data$node_type)
  expect_false("Timepoint" %in% g$nodes$data$node_type)
})

test_that("graph_from_metadata supports subject-scope outcome", {
  meta <- data.frame(
    sample_id     = c("S1", "S2", "S3"),
    subject_id    = c("P1", "P2", "P3"),
    outcome_id    = c("ctrl", "case", "ctrl"),
    stringsAsFactors = FALSE
  )

  g <- graph_from_metadata(meta, outcome_scope = "subject")
  edge_types <- unique(g$edges$data$edge_type)
  expect_true("subject_has_outcome" %in% edge_types)
  expect_false("sample_has_outcome" %in% edge_types)
})

# Metadata ingestion and canonical node/edge builders.

#' Standardize Sample Metadata
#'
#' Normalize user-provided metadata into the canonical column contract used by
#' \pkg{splitGraph}.
#'
#' @param data A sample-level \code{data.frame}.
#' @param col_map Optional named character vector mapping canonical names to
#'   user-provided columns.
#' @param dataset_name Optional dataset label stored as an attribute on the
#'   returned table.
#' @param strict If \code{TRUE}, error when required columns are missing.
#' @return A standardized \code{data.frame} with canonical identifier columns
#'   coerced to character.
#' @examples
#' meta <- ingest_metadata(
#'   data.frame(sample_id = c("S1", "S2"), subject_id = c("P1", "P2"))
#' )
#' @export
ingest_metadata <- function(data, col_map = NULL, dataset_name = NULL, strict = TRUE) {
  .depgraph_assert(is.data.frame(data), "`data` must be a data.frame.")

  out <- data
  if (!is.null(col_map)) {
    .depgraph_assert(is.character(col_map) && !is.null(names(col_map)), "`col_map` must be a named character vector.")
    missing_cols <- unname(col_map)[!unname(col_map) %in% names(out)]
    .depgraph_assert(
      length(missing_cols) == 0L,
      paste0("Mapped columns not found in `data`: ", paste(missing_cols, collapse = ", "))
    )
    rename_idx <- match(unname(col_map), names(out))
    names(out)[rename_idx] <- names(col_map)
  }

  required_cols <- "sample_id"
  if (isTRUE(strict)) {
    missing_required <- setdiff(required_cols, names(out))
    .depgraph_assert(
      length(missing_required) == 0L,
      paste0("Missing required metadata columns: ", paste(missing_required, collapse = ", "))
    )
  }

  id_like <- intersect(
    names(out),
    c(
      "sample_id", "subject_id", "batch_id", "study_id",
      "timepoint_id", "assay_id", "featureset_id", "outcome_id"
    )
  )

  for (col in id_like) {
    out[[col]] <- as.character(out[[col]])
  }

  attr(out, "dataset_name") <- dataset_name
  out
}

#' Create Canonical Node and Edge Tables
#'
#' Build canonical node and edge tables from ordinary metadata frames.
#'
#' The package uses typed node identifiers such as \code{sample:S1} as the
#' canonical graph representation. If you create node sets with
#' \code{prefix = FALSE}, the corresponding edge endpoints must use matching
#' prefix settings via \code{from_prefix} and \code{to_prefix}.
#'
#' When \code{dedupe = TRUE}, exact duplicate node or edge definitions are
#' collapsed, but conflicting definitions for the same canonical node
#' identifier or edge relation are rejected with an error.
#'
#' @param data A \code{data.frame} containing entity or relationship columns.
#' @param type,from_type,to_type Supported node types such as \code{"Sample"}
#'   or \code{"Subject"}.
#' @param id_col Column containing the source identifier for the node type.
#' @param label_col Optional column used for node labels.
#' @param attr_cols Optional columns stored in the \code{attrs} list-column.
#' @param prefix If \code{TRUE}, prepend typed prefixes such as \code{sample:}
#'   to node identifiers.
#' @param dedupe If \code{TRUE}, collapse duplicate identifiers or duplicate
#'   edges only when the retained definition is identical.
#' @param from_col,to_col Source and target identifier columns for edge
#'   creation.
#' @param relation Canonical edge type.
#' @param allow_missing If \code{TRUE}, drop rows with missing edge endpoints
#'   instead of erroring.
#' @param from_prefix,to_prefix Whether to prepend typed prefixes when
#'   constructing the edge endpoint identifiers. Defaults preserve the
#'   canonical prefixed-ID format.
#' @return For \code{create_nodes()}, a \code{graph_node_set}. For
#'   \code{create_edges()}, a \code{graph_edge_set}.
#' @examples
#' meta <- data.frame(
#'   sample_id = c("S1", "S2"),
#'   subject_id = c("P1", "P2")
#' )
#'
#' samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
#' edges <- create_edges(
#'   meta,
#'   from_col = "sample_id",
#'   to_col = "subject_id",
#'   from_type = "Sample",
#'   to_type = "Subject",
#'   relation = "sample_belongs_to_subject"
#' )
#' @export
create_nodes <- function(data, type, id_col, label_col = NULL, attr_cols = NULL, prefix = TRUE, dedupe = TRUE) {
  .depgraph_assert(is.data.frame(data), "`data` must be a data.frame.")
  .depgraph_assert(id_col %in% names(data), paste0("Missing `id_col`: ", id_col))

  type <- .depgraph_match_node_type(type)
  schema <- .depgraph_node_type_schema(type)

  if (identical(id_col, schema$key_field) && is.null(attr_cols)) {
    attr_cols <- intersect(schema$optional_attrs, names(data))
  }

  keep_cols <- unique(c(id_col, label_col, attr_cols))
  keep_cols <- keep_cols[!is.na(keep_cols) & nzchar(keep_cols)]
  selected <- data[, keep_cols, drop = FALSE]
  selected <- selected[!is.na(selected[[id_col]]) & nzchar(selected[[id_col]]), , drop = FALSE]

  if (isTRUE(dedupe) && anyDuplicated(selected[[id_col]]) > 0L) {
    dup_ids <- unique(selected[[id_col]][duplicated(selected[[id_col]])])
    dup_rows <- selected[selected[[id_col]] %in% dup_ids, , drop = FALSE]
    conflicting_ids <- dup_ids[vapply(dup_ids, function(dup_id) {
      id_rows <- selected[selected[[id_col]] == dup_id, , drop = FALSE]
      nrow(unique(id_rows)) > 1L
    }, logical(1))]

    if (length(conflicting_ids) > 0L) {
      stop(
        paste0(
          "Conflicting node definitions found for IDs: ",
          paste(conflicting_ids, collapse = ", ")
        ),
        call. = FALSE
      )
    }
    selected <- selected[!duplicated(selected[[id_col]]), , drop = FALSE]
  }

  labels <- if (is.null(label_col)) selected[[id_col]] else selected[[label_col]]
  label_missing <- is.na(labels) | !nzchar(as.character(labels))
  if (any(label_missing)) {
    labels[label_missing] <- selected[[id_col]][label_missing]
  }

  node_data <- data.frame(
    node_id = .depgraph_make_node_id(type, selected[[id_col]], prefix = prefix),
    node_type = type,
    node_key = as.character(selected[[id_col]]),
    label = as.character(labels),
    attrs = .depgraph_make_attr_list(selected, attr_cols),
    stringsAsFactors = FALSE
  )

  graph_node_set(node_data, source = list(type = type, id_col = id_col))
}

#' @rdname create_nodes
#' @export
create_edges <- function(data, from_col, to_col, from_type, to_type, relation, attr_cols = NULL, allow_missing = FALSE, dedupe = TRUE, from_prefix = TRUE, to_prefix = TRUE) {
  .depgraph_assert(is.data.frame(data), "`data` must be a data.frame.")
  .depgraph_assert(from_col %in% names(data), paste0("Missing `from_col`: ", from_col))
  .depgraph_assert(to_col %in% names(data), paste0("Missing `to_col`: ", to_col))

  from_type <- .depgraph_match_node_type(from_type)
  to_type <- .depgraph_match_node_type(to_type)
  relation <- as.character(relation)[1L]

  schema_row <- .depgraph_edge_type_schema(relation)
  if (nrow(schema_row) == 1L) {
    .depgraph_assert(
      identical(schema_row$from_type[[1L]], from_type) && identical(schema_row$to_type[[1L]], to_type),
      paste0("Relation `", relation, "` expects ", schema_row$from_type[[1L]], " -> ", schema_row$to_type[[1L]], ".")
    )
  }

  keep_cols <- unique(c(from_col, to_col, attr_cols))
  keep_cols <- keep_cols[!is.na(keep_cols) & nzchar(keep_cols)]
  selected <- data[, keep_cols, drop = FALSE]

  missing_endpoint <- is.na(selected[[from_col]]) | !nzchar(as.character(selected[[from_col]])) |
    is.na(selected[[to_col]]) | !nzchar(as.character(selected[[to_col]]))

  if (any(missing_endpoint)) {
    if (isTRUE(allow_missing)) {
      selected <- selected[!missing_endpoint, , drop = FALSE]
    } else {
      stop("Missing endpoint values found while creating edges.", call. = FALSE)
    }
  }

  edge_data <- data.frame(
    from = .depgraph_make_node_id(from_type, selected[[from_col]], prefix = from_prefix),
    to = .depgraph_make_node_id(to_type, selected[[to_col]], prefix = to_prefix),
    edge_type = relation,
    attrs = .depgraph_make_attr_list(selected, attr_cols),
    stringsAsFactors = FALSE
  )

  if (isTRUE(dedupe) && nrow(edge_data) > 0L) {
    edge_keys <- paste(edge_data$from, edge_data$to, edge_data$edge_type, sep = "\r")
    dup_keys <- unique(edge_keys[duplicated(edge_keys)])
    conflicting_keys <- dup_keys[vapply(dup_keys, function(edge_key) {
      key_rows <- which(edge_keys == edge_key)
      reference_attrs <- edge_data$attrs[[key_rows[[1L]]]]
      any(vapply(
        key_rows[-1L],
        function(idx) !identical(edge_data$attrs[[idx]], reference_attrs),
        logical(1)
      ))
    }, logical(1))]

    if (length(conflicting_keys) > 0L) {
      conflicting_labels <- vapply(conflicting_keys, function(edge_key) {
        parts <- strsplit(edge_key, "\r", fixed = TRUE)[[1L]]
        paste0(parts[[1L]], " -> ", parts[[2L]], " [", parts[[3L]], "]")
      }, character(1))
      stop(
        paste0(
          "Conflicting edge definitions found for relations: ",
          paste(conflicting_labels, collapse = ", ")
        ),
        call. = FALSE
      )
    }

    keep <- !duplicated(edge_data[, c("from", "to", "edge_type"), drop = FALSE])
    edge_data <- edge_data[keep, , drop = FALSE]
  }

  edge_data$edge_id <- paste0(edge_data$edge_type, ":", seq_len(nrow(edge_data)))
  edge_data <- edge_data[, c("edge_id", "from", "to", "edge_type", "attrs"), drop = FALSE]

  graph_edge_set(edge_data, source = list(relation = relation, from_col = from_col, to_col = to_col))
}

test_that("summary and print methods expose stable metadata", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")
  graph <- build_dependency_graph(list(samples, subjects), list(edges), graph_name = "demo")

  sample_summary <- summary(samples)
  edge_summary <- summary(edges)
  graph_summary <- summary(graph)

  expect_equal(sample_summary$schema_version, splitGraph:::.depgraph_schema_version)
  expect_equal(edge_summary$schema_version, splitGraph:::.depgraph_schema_version)
  expect_equal(graph_summary$schema_version, splitGraph:::.depgraph_schema_version)
  expect_equal(graph_summary$graph_name, "demo")
  expect_true(length(capture.output(print(graph))) > 0L)
  expect_true(length(capture.output(print(samples))) > 0L)
  expect_true(length(capture.output(print(edges))) > 0L)
  expect_s3_class(as.data.frame(samples), "data.frame")
  expect_s3_class(as.data.frame(edges), "data.frame")
})

test_that("plot.dependency_graph renders with the typed layout", {
  meta <- data.frame(
    sample_id = c("S1", "S2", "S3"),
    subject_id = c("P1", "P1", "P2"),
    batch_id = c("B1", "B2", "B1"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  batches <- create_nodes(meta, type = "Batch", id_col = "batch_id")
  e_sub <- create_edges(meta, "sample_id", "subject_id",
                        "Sample", "Subject", "sample_belongs_to_subject")
  e_bat <- create_edges(meta, "sample_id", "batch_id",
                        "Sample", "Batch", "sample_processed_in_batch")

  graph <- build_dependency_graph(
    nodes = list(samples, subjects, batches),
    edges = list(e_sub, e_bat),
    graph_name = "plot_demo"
  )

  pdf_file <- tempfile(fileext = ".pdf")
  grDevices::pdf(pdf_file)
  on.exit({
    grDevices::dev.off()
    unlink(pdf_file)
  }, add = TRUE)

  expect_silent(plot(graph))
  expect_silent(plot(graph, layout = "sugiyama"))
  expect_silent(plot(graph, show_labels = FALSE))
})

test_that("build_dependency_graph error mentions prefix mismatch when likely", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  # Build nodes WITHOUT prefixes ...
  samples <- create_nodes(meta, "Sample", "sample_id", prefix = FALSE)
  subjects <- create_nodes(meta, "Subject", "subject_id", prefix = FALSE)

  # ... but use the default (prefixed) edges, so endpoints will reference
  # `sample:S1` while the node table only has `S1`. This is the most common
  # first-time mistake.
  edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject"
  )

  expect_error(
    build_dependency_graph(list(samples, subjects), list(edges)),
    "prefix",
    ignore.case = TRUE
  )
})

test_that("build_dependency_graph error does NOT mention prefix when no mismatch pattern", {
  # Endpoints simply absent from the node table for a non-prefix reason.
  meta <- data.frame(
    sample_id  = "S1",
    subject_id = "P1",
    stringsAsFactors = FALSE
  )
  samples <- create_nodes(meta, "Sample", "sample_id")
  edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject"
  )
  # No subject node provided at all -> still errors, but the message should
  # be the generic missing-references error, not the prefix hint.
  expect_error(
    build_dependency_graph(list(samples), list(edges)),
    "not present",
    ignore.case = TRUE
  )
})

test_that("matching prefix conventions on both sides build successfully", {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, "Sample", "sample_id", prefix = FALSE)
  subjects <- create_nodes(meta, "Subject", "subject_id", prefix = FALSE)
  edges <- create_edges(
    meta, "sample_id", "subject_id",
    "Sample", "Subject", "sample_belongs_to_subject",
    from_prefix = FALSE, to_prefix = FALSE
  )

  expect_silent(g <- build_dependency_graph(list(samples, subjects), list(edges)))
  expect_s3_class(g, "dependency_graph")
})

test_that("validation report methods summarize issues cleanly", {
  meta <- data.frame(
    sample_id = c("S1", "S2"),
    subject_id = c("P1", "P1"),
    stringsAsFactors = FALSE
  )

  samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
  subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
  edges <- create_edges(meta, "sample_id", "subject_id", "Sample", "Subject", "sample_belongs_to_subject")
  graph <- build_dependency_graph(list(samples, subjects), list(edges))
  validation <- validate_graph(graph)

  validation_summary <- summary(validation)

  expect_true(length(capture.output(print(validation))) > 0L)
  expect_true(is.list(validation_summary))
  expect_true(all(c("by_level", "by_severity", "by_code") %in% names(validation_summary)))
  expect_s3_class(as.data.frame(validation), "data.frame")
})

# Core package constants and internal helpers for splitGraph.

#' splitGraph: Dataset Dependency Graphs for Leakage-Aware Evaluation
#'
#' The \pkg{splitGraph} package provides typed graph objects for representing
#' dataset structure, sample provenance, and leakage-relevant dependencies in
#' biomedical evaluation workflows. It makes dataset dependency structure
#' explicit enough to validate, query, and convert into a stable, tool-agnostic
#' split specification (\code{split_spec}) for leakage-aware evaluation.
#'
#' @section Scope (what splitGraph does):
#' \itemize{
#'   \item Model dataset dependency structure as a typed graph.
#'   \item Validate that structure (structural, semantic, leakage-relevant).
#'   \item Derive deterministic split \emph{constraints} from the structure.
#'   \item Emit and validate the tool-agnostic \code{split_spec} interchange
#'     format (with a formal JSON Schema and a Python reference consumer).
#' }
#'
#' @section Non-goals (what downstream consumers own):
#' splitGraph deliberately stops at the constraint / \code{split_spec} boundary.
#' It does \emph{not} generate resamples or folds, perform stratified splitting,
#' apply purge/embargo, fit or tune models, or produce statistical leakage
#' evidence. Those belong to downstream consumers. The reference consumer is
#' \pkg{bioLeak}, whose \code{as_leaksplits()} turns a \code{split_spec} into an
#' executable split plan; \code{split_spec} is neutral, so other tools (an
#' \pkg{rsample} adapter, the shipped Python reader, etc.) can consume it
#' equally. See the \code{split_spec} contract in \code{?as_split_spec} and the
#' "Scope & relationship to bioLeak" section of the README.
#'
#' @keywords internal
#' @importFrom graphics plot
#' @importFrom igraph ecount graph_from_data_frame vcount
"_PACKAGE"

# Schema version for the dataset-dependency-graph data model. This is
# intentionally decoupled from the package version: bump it only when the
# canonical node, edge, constraint, or split_spec table contracts change in a
# way that breaks round-trip compatibility with previously constructed
# objects. A package release that only adds functions or fixes bugs must NOT
# bump this string.
#
# Compatibility policy (enforced by `.depgraph_check_schema_version()`): the
# MAJOR component is the compatibility boundary. Files whose recorded version
# shares the installed MAJOR are read-compatible and load silently, because
# readers tolerate additive change -- `read_dependency_graph()` /
# `read_split_spec()` fill any field missing from an older file with `NA` and
# ignore fields they do not recognise. A differing MAJOR (older or newer) loads
# with a warning suggesting `migrate_dependency_graph_json()` /
# `migrate_split_spec_json()`. Only a change that would make an older file
# genuinely unreadable (renaming/removing a field, changing a field's type or
# meaning) warrants a MAJOR bump.
#
# History: "0.1.0" was the initial format. "0.2.0" formalises the on-disk
# contract with shipped JSON Schemas (`inst/schema/`), a `$schema` reference in
# output, split_spec provenance (`splitgraph_version`, `derived_at`), and the
# node/edge/column additions from the 0.3.0 development cycle (Site, Region,
# Platform, the pairwise relations, and their split_spec annotations). All of
# those are additive within MAJOR 0, so "0.1.0" files still load silently.
.depgraph_schema_version <- "0.2.0"

.depgraph_node_types <- c(
  "Sample", "Subject", "Batch", "Study",
  "Timepoint", "Assay", "FeatureSet", "Outcome", "Site", "Region", "Platform"
)

.depgraph_prefix_map <- c(
  Sample = "sample",
  Subject = "subject",
  Batch = "batch",
  Study = "study",
  Timepoint = "timepoint",
  Assay = "assay",
  FeatureSet = "featureset",
  Outcome = "outcome",
  Site = "site",
  Region = "region",
  Platform = "platform"
)

.depgraph_node_schema <- list(
  Sample = list(
    key_field = "sample_id",
    required_attrs = character(),
    optional_attrs = c(
      "subject_id", "batch_id", "study_id", "timepoint_id",
      "assay_id", "featureset_id", "outcome_id", "site_id", "region_id",
      "platform_id", "sample_role", "collection_date", "source_file"
    )
  ),
  Subject = list(
    key_field = "subject_id",
    required_attrs = character(),
    optional_attrs = c(
      "subject_type", "species", "sex", "site", "enrollment_study"
    )
  ),
  Batch = list(
    key_field = "batch_id",
    required_attrs = character(),
    optional_attrs = c(
      "batch_type", "platform", "processing_center", "run_date",
      "operator"
    )
  ),
  Study = list(
    key_field = "study_id",
    required_attrs = character(),
    optional_attrs = c(
      "study_name", "study_type", "institution", "country",
      "study_phase"
    )
  ),
  Timepoint = list(
    key_field = "timepoint_id",
    required_attrs = character(),
    optional_attrs = c(
      "time_index", "time_value", "time_unit", "collection_datetime",
      "visit_label"
    )
  ),
  Assay = list(
    key_field = "assay_id",
    required_attrs = character(),
    optional_attrs = c(
      "modality", "platform", "vendor", "protocol_version"
    )
  ),
  FeatureSet = list(
    key_field = "featureset_id",
    required_attrs = character(),
    optional_attrs = c(
      "featureset_name", "modality", "feature_count",
      "derivation_scope", "derivation_version", "generated_from"
    )
  ),
  Outcome = list(
    key_field = "outcome_id",
    required_attrs = character(),
    optional_attrs = c(
      "outcome_name", "outcome_type", "outcome_value",
      "outcome_scale", "observation_level"
    )
  ),
  Site = list(
    key_field = "site_id",
    required_attrs = character(),
    optional_attrs = c(
      "site_name", "institution", "city", "country", "site_type"
    )
  ),
  Region = list(
    key_field = "region_id",
    required_attrs = character(),
    optional_attrs = c(
      "region_name", "region_type", "tissue", "organ", "coordinate_system"
    )
  ),
  Platform = list(
    key_field = "platform_id",
    required_attrs = character(),
    optional_attrs = c(
      "platform_name", "vendor", "model", "technology", "chemistry"
    )
  )
)

.depgraph_allowed_derivation_scopes <- c(
  "external",
  "per_training_split",
  "per_dataset"
)

.depgraph_allowed_observation_levels <- c(
  "sample",
  "subject"
)

.depgraph_validation_levels <- c(
  "structural",
  "semantic",
  "leakage"
)

.depgraph_validation_severities <- c(
  "error",
  "warning",
  "advisory"
)

# Default upper bound on path length used by `query_paths()` when the caller
# does not pass `max_length`. Picked to keep `igraph::all_simple_paths()`
# tractable on typed dataset graphs (most useful chains span Sample ->
# Subject/Batch/... -> Outcome, i.e. <= 3 hops). Pass `max_length = Inf` to
# opt out of the cap.
.depgraph_default_path_cap <- 8L

.depgraph_edge_schema <- data.frame(
  edge_type = c(
    "sample_belongs_to_subject",
    "sample_processed_in_batch",
    "sample_from_study",
    "sample_collected_at_timepoint",
    "sample_measured_by_assay",
    "sample_uses_featureset",
    "sample_has_outcome",
    "subject_has_outcome",
    "timepoint_precedes",
    "featureset_generated_from_study",
    "featureset_generated_from_batch",
    "sample_collected_at_site",
    "sample_located_in_region",
    "sample_run_on_platform",
    "assay_uses_platform",
    "subject_related_to",
    "sample_adjacent_to"
  ),
  from_type = c(
    "Sample", "Sample", "Sample", "Sample", "Sample",
    "Sample", "Sample", "Subject", "Timepoint", "FeatureSet",
    "FeatureSet", "Sample", "Sample", "Sample", "Assay",
    "Subject", "Sample"
  ),
  to_type = c(
    "Subject", "Batch", "Study", "Timepoint", "Assay",
    "FeatureSet", "Outcome", "Outcome", "Timepoint", "Study",
    "Batch", "Site", "Region", "Platform", "Platform",
    "Subject", "Sample"
  ),
  stringsAsFactors = FALSE
)

.depgraph_single_target_edges <- c(
  "sample_collected_at_timepoint",
  "sample_measured_by_assay"
)

.depgraph_semantic_relation_rules <- list(
  sample_belongs_to_subject = list(
    min_targets = 0L,
    max_targets = 1L,
    missing_code = NULL,
    multiple_code = "sample_multiple_subject_assignments",
    missing_severity = NULL,
    multiple_severity = "error"
  ),
  sample_from_study = list(
    min_targets = 1L,
    max_targets = 1L,
    missing_code = "sample_missing_study_assignment",
    multiple_code = "sample_multiple_study_assignments",
    missing_severity = "warning",
    multiple_severity = "error"
  ),
  sample_processed_in_batch = list(
    min_targets = 0L,
    max_targets = 1L,
    missing_code = NULL,
    multiple_code = "sample_multiple_batch_assignments",
    missing_severity = NULL,
    multiple_severity = "error"
  ),
  sample_collected_at_site = list(
    min_targets = 0L,
    max_targets = 1L,
    missing_code = NULL,
    multiple_code = "sample_multiple_site_assignments",
    missing_severity = NULL,
    multiple_severity = "error"
  ),
  sample_located_in_region = list(
    min_targets = 0L,
    max_targets = 1L,
    missing_code = NULL,
    multiple_code = "sample_multiple_region_assignments",
    missing_severity = NULL,
    multiple_severity = "error"
  ),
  sample_run_on_platform = list(
    min_targets = 0L,
    max_targets = 1L,
    missing_code = NULL,
    multiple_code = "sample_multiple_platform_assignments",
    missing_severity = NULL,
    multiple_severity = "error"
  )
)

.depgraph_empty_node_data <- function() {
  data.frame(
    node_id = character(),
    node_type = character(),
    node_key = character(),
    label = character(),
    attrs = I(list()),
    stringsAsFactors = FALSE
  )
}

.depgraph_empty_edge_data <- function() {
  data.frame(
    edge_id = character(),
    from = character(),
    to = character(),
    edge_type = character(),
    attrs = I(list()),
    stringsAsFactors = FALSE
  )
}

.depgraph_empty_attrs <- function(n) {
  rep(list(list()), n)
}

.depgraph_assert <- function(condition, message) {
  if (!isTRUE(condition)) {
    stop(message, call. = FALSE)
  }
}

# Installed splitGraph package version as a string, for stamping provenance.
# Returns NA_character_ if the version cannot be resolved (e.g. exotic loads).
.depgraph_package_version <- function() {
  tryCatch(as.character(utils::packageVersion("splitGraph")), error = function(e) NA_character_)
}

.depgraph_match_node_type <- function(type) {
  .depgraph_assert(length(type) == 1L && !is.na(type), "`type` must be length 1.")
  match_idx <- match(tolower(type), tolower(.depgraph_node_types))
  .depgraph_assert(!is.na(match_idx), paste0("Unsupported node type: ", type))
  .depgraph_node_types[[match_idx]]
}

.depgraph_prefix_for_type <- function(type) {
  type <- .depgraph_match_node_type(type)
  .depgraph_prefix_map[[type]]
}

.depgraph_node_type_schema <- function(type) {
  type <- .depgraph_match_node_type(type)
  .depgraph_node_schema[[type]]
}

.depgraph_edge_type_schema <- function(edge_type) {
  edge_type <- as.character(edge_type)[1L]
  .depgraph_edge_schema[.depgraph_edge_schema$edge_type == edge_type, , drop = FALSE]
}

.depgraph_make_node_id <- function(type, key, prefix = TRUE) {
  key <- as.character(key)
  .depgraph_assert(all(!is.na(key) & nzchar(key)), "Node keys must be non-missing strings.")
  if (!isTRUE(prefix)) {
    return(key)
  }
  paste0(.depgraph_prefix_for_type(type), ":", key)
}

.depgraph_normalize_attr_entry <- function(x, context = "`attrs`") {
  if (is.null(x)) {
    return(list())
  }

  .depgraph_assert(is.list(x), paste0(context, " entries must be lists."))
  if (length(x) == 0L) {
    return(list())
  }

  attr_names <- names(x)
  .depgraph_assert(
    !is.null(attr_names) && all(!is.na(attr_names)) && all(nzchar(attr_names)),
    paste0(context, " entries must be named lists.")
  )
  .depgraph_assert(
    anyDuplicated(attr_names) == 0L,
    paste0(context, " entries cannot contain duplicate names.")
  )

  as.list(x)
}

.depgraph_make_attr_list <- function(data, attr_cols = NULL) {
  if (is.null(attr_cols) || length(attr_cols) == 0L) {
    return(I(.depgraph_empty_attrs(nrow(data))))
  }

  missing_cols <- setdiff(attr_cols, names(data))
  .depgraph_assert(
    length(missing_cols) == 0L,
    paste0("Missing attribute columns: ", paste(missing_cols, collapse = ", "))
  )

  attrs <- lapply(seq_len(nrow(data)), function(i) {
    as.list(data[i, attr_cols, drop = FALSE])
  })
  I(attrs)
}

.depgraph_validate_node_attrs <- function(node_type, attrs) {
  schema <- .depgraph_node_type_schema(node_type)
  attr_names <- names(attrs)

  missing_required <- setdiff(schema$required_attrs, attr_names)
  unknown_attrs <- setdiff(attr_names, c(schema$required_attrs, schema$optional_attrs))

  list(
    missing_required = missing_required,
    unknown_attrs = unknown_attrs
  )
}

.depgraph_build_igraph <- function(node_data, edge_data) {
  vertices <- node_data[, c("node_id", "node_type", "node_key", "label"), drop = FALSE]
  edges_for_graph <- edge_data[, c("from", "to", "edge_id", "edge_type"), drop = FALSE]

  igraph::graph_from_data_frame(
    d = edges_for_graph,
    vertices = vertices,
    directed = TRUE
  )
}

.depgraph_build_caches <- function(node_data, edge_data) {
  list(
    sample_ids = node_data$node_key[node_data$node_type == "Sample"],
    node_index = stats::setNames(seq_len(nrow(node_data)), node_data$node_id),
    edge_index = stats::setNames(seq_len(nrow(edge_data)), edge_data$edge_id)
  )
}

.depgraph_graph_table_alignment <- function(graph, node_data, edge_data) {
  if (!igraph::is_directed(graph)) {
    return(list(
      valid = FALSE,
      message = "The embedded `igraph` must be directed and match the supplied node and edge tables."
    ))
  }

  graph_nodes <- sort(igraph::V(graph)$name)
  table_nodes <- sort(node_data$node_id)
  if (!identical(graph_nodes, table_nodes)) {
    return(list(
      valid = FALSE,
      message = "The embedded `igraph` does not match the supplied node and edge tables."
    ))
  }

  graph_edges <- igraph::as_data_frame(graph, what = "edges")
  required_edge_cols <- c("from", "to", "edge_id", "edge_type")
  if (!all(required_edge_cols %in% names(graph_edges))) {
    return(list(
      valid = FALSE,
      message = "The embedded `igraph` does not match the supplied node and edge tables."
    ))
  }

  graph_edges <- graph_edges[, required_edge_cols, drop = FALSE]
  table_edges <- edge_data[, required_edge_cols, drop = FALSE]

  graph_edges <- graph_edges[do.call(order, unname(graph_edges)), , drop = FALSE]
  table_edges <- table_edges[do.call(order, unname(table_edges)), , drop = FALSE]
  row.names(graph_edges) <- NULL
  row.names(table_edges) <- NULL

  if (!identical(graph_edges, table_edges)) {
    return(list(
      valid = FALSE,
      message = "The embedded `igraph` does not match the supplied node and edge tables."
    ))
  }

  list(valid = TRUE, message = NULL)
}

.depgraph_normalize_node_data <- function(data) {
  if (is.null(data)) {
    return(.depgraph_empty_node_data())
  }

  .depgraph_assert(is.data.frame(data), "`data` must be a data.frame.")

  required_cols <- c("node_id", "node_type", "node_key", "label")
  missing_cols <- setdiff(required_cols, names(data))
  .depgraph_assert(
    length(missing_cols) == 0L,
    paste0("Node data is missing columns: ", paste(missing_cols, collapse = ", "))
  )

  if (!"attrs" %in% names(data)) {
    data$attrs <- I(.depgraph_empty_attrs(nrow(data)))
  }

  .depgraph_assert(is.list(data$attrs), "`attrs` must be a list column.")

  data$node_id <- as.character(data$node_id)
  data$node_type <- as.character(data$node_type)
  data$node_key <- as.character(data$node_key)
  data$label <- as.character(data$label)
  data$attrs <- I(lapply(data$attrs, .depgraph_normalize_attr_entry, context = "`attrs`"))

  .depgraph_assert(
    all(!is.na(data$node_id) & nzchar(data$node_id)),
    "Node data contains missing or empty `node_id` values."
  )
  .depgraph_assert(
    all(!is.na(data$node_key) & nzchar(data$node_key)),
    "Node data contains missing or empty `node_key` values."
  )
  .depgraph_assert(
    all(!is.na(data$label) & nzchar(data$label)),
    "Node data contains missing or empty `label` values."
  )

  node_types <- vapply(data$node_type, .depgraph_match_node_type, character(1), USE.NAMES = FALSE)
  data$node_type <- node_types

  data[, c("node_id", "node_type", "node_key", "label", "attrs"), drop = FALSE]
}

.depgraph_normalize_edge_data <- function(data) {
  if (is.null(data)) {
    return(.depgraph_empty_edge_data())
  }

  .depgraph_assert(is.data.frame(data), "`data` must be a data.frame.")

  required_cols <- c("edge_id", "from", "to", "edge_type")
  missing_cols <- setdiff(required_cols, names(data))
  .depgraph_assert(
    length(missing_cols) == 0L,
    paste0("Edge data is missing columns: ", paste(missing_cols, collapse = ", "))
  )

  if (!"attrs" %in% names(data)) {
    data$attrs <- I(.depgraph_empty_attrs(nrow(data)))
  }

  .depgraph_assert(is.list(data$attrs), "`attrs` must be a list column.")

  data$edge_id <- as.character(data$edge_id)
  data$from <- as.character(data$from)
  data$to <- as.character(data$to)
  data$edge_type <- as.character(data$edge_type)
  data$attrs <- I(lapply(data$attrs, .depgraph_normalize_attr_entry, context = "`attrs`"))

  .depgraph_assert(
    all(!is.na(data$edge_id) & nzchar(data$edge_id)),
    "Edge data contains missing or empty `edge_id` values."
  )
  .depgraph_assert(
    all(!is.na(data$from) & nzchar(data$from)),
    "Edge data contains missing or empty `from` values."
  )
  .depgraph_assert(
    all(!is.na(data$to) & nzchar(data$to)),
    "Edge data contains missing or empty `to` values."
  )
  .depgraph_assert(
    all(!is.na(data$edge_type) & nzchar(data$edge_type)),
    "Edge data contains missing or empty `edge_type` values."
  )

  data[, c("edge_id", "from", "to", "edge_type", "attrs"), drop = FALSE]
}

.depgraph_count_table <- function(x) {
  counts <- sort(table(x), decreasing = TRUE)
  data.frame(
    value = names(counts),
    n = as.integer(counts),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

.depgraph_bind_data <- function(objects, field) {
  if (inherits(objects, c("graph_node_set", "graph_edge_set"))) {
    objects <- list(objects)
  }
  .depgraph_assert(is.list(objects) && length(objects) > 0L, "`objects` must be a non-empty list.")
  pieces <- lapply(objects, function(x) x[[field]])
  out <- do.call(rbind, pieces)
  row.names(out) <- NULL
  out
}

.depgraph_extract_attr <- function(attrs, key) {
  value <- attrs[[key]]
  if (is.null(value) || length(value) == 0L) {
    return(NA)
  }
  value[[1L]]
}

.depgraph_time_order_consistency <- function(node_data, edge_data) {
  time_nodes <- node_data[node_data$node_type == "Timepoint", , drop = FALSE]
  precedes <- edge_data[
    edge_data$edge_type == "timepoint_precedes",
    c("edge_id", "from", "to"),
    drop = FALSE
  ]

  empty_pairs <- data.frame(
    edge_id = character(),
    from = character(),
    to = character(),
    stringsAsFactors = FALSE
  )

  if (nrow(time_nodes) == 0L || nrow(precedes) == 0L) {
    return(list(
      self_loop_edge_ids = character(),
      cycle_edge_ids = character(),
      conflicting_pairs = empty_pairs
    ))
  }

  self_loop_edge_ids <- precedes$edge_id[precedes$from == precedes$to]

  nonself_precedes <- precedes[precedes$from != precedes$to, , drop = FALSE]
  cycle_edge_ids <- character()
  if (nrow(nonself_precedes) > 0L) {
    precedence_graph <- igraph::graph_from_data_frame(
      d = nonself_precedes[, c("from", "to"), drop = FALSE],
      vertices = data.frame(name = time_nodes$node_id, stringsAsFactors = FALSE),
      directed = TRUE
    )
    if (!igraph::is_dag(precedence_graph)) {
      cycle_edge_ids <- nonself_precedes$edge_id
    }
  }

  raw_time_index <- vapply(
    time_nodes$attrs,
    function(x) {
      value <- .depgraph_extract_attr(x, "time_index")
      if (length(value) == 0L || is.na(value)) {
        return(NA_real_)
      }
      suppressWarnings(as.numeric(as.character(value)))
    },
    FUN.VALUE = numeric(1)
  )
  names(raw_time_index) <- time_nodes$node_id

  from_index <- raw_time_index[precedes$from]
  to_index <- raw_time_index[precedes$to]
  conflict_idx <- !is.na(from_index) & !is.na(to_index) & from_index >= to_index
  conflicting_pairs <- precedes[conflict_idx, , drop = FALSE]
  row.names(conflicting_pairs) <- NULL

  list(
    self_loop_edge_ids = self_loop_edge_ids,
    cycle_edge_ids = cycle_edge_ids,
    conflicting_pairs = conflicting_pairs
  )
}

.depgraph_validation_override <- function(graph, key, default = FALSE) {
  overrides <- graph$metadata$validation_overrides
  if (is.null(overrides) || is.null(overrides[[key]])) {
    return(default)
  }
  isTRUE(overrides[[key]])
}

# Returns a shallow copy of the graph with `validation_overrides` merged into
# its metadata. Explicit values in `overrides` win over any pre-existing
# graph-level overrides; unspecified keys are preserved.
.depgraph_with_overrides <- function(graph, overrides) {
  if (is.null(overrides) || length(overrides) == 0L) {
    return(graph)
  }
  existing <- graph$metadata$validation_overrides
  if (is.null(existing)) existing <- list()
  graph$metadata$validation_overrides <- utils::modifyList(existing, overrides)
  graph
}

# Validation helpers for dependency_graph objects.

.depgraph_empty_issue_data <- function() {
  data.frame(
    issue_id = character(),
    level = character(),
    severity = character(),
    code = character(),
    message = character(),
    node_ids = I(list()),
    edge_ids = I(list()),
    details = I(list()),
    stringsAsFactors = FALSE
  )
}

.new_validation_issue <- function(level, severity, code, message, node_ids = character(), edge_ids = character(), details = list()) {
  .depgraph_assert(level %in% .depgraph_validation_levels, paste0("Unsupported validation level: ", level))
  .depgraph_assert(severity %in% .depgraph_validation_severities, paste0("Unsupported validation severity: ", severity))

  data.frame(
    issue_id = NA_character_,
    level = level,
    severity = severity,
    code = code,
    message = as.character(message)[1L],
    node_ids = I(list(as.character(node_ids))),
    edge_ids = I(list(as.character(edge_ids))),
    details = I(list(.depgraph_normalize_attr_entry(details, context = "`details`"))),
    stringsAsFactors = FALSE
  )
}

.depgraph_bind_issues <- function(issues) {
  if (length(issues) == 0L) {
    return(.depgraph_empty_issue_data())
  }

  out <- do.call(rbind, issues)
  row.names(out) <- NULL
  if (nrow(out) == 0L) {
    return(.depgraph_empty_issue_data())
  }
  out$issue_id <- paste0("issue_", seq_len(nrow(out)))
  out
}

.depgraph_issue_count_table <- function(x, column) {
  if (nrow(x) == 0L) {
    return(data.frame(value = character(), n = integer(), stringsAsFactors = FALSE))
  }
  .depgraph_count_table(x[[column]])
}

.summarize_validation_issues <- function(issues) {
  list(
    n_issues = nrow(issues),
    by_level = .depgraph_issue_count_table(issues, "level"),
    by_severity = .depgraph_issue_count_table(issues, "severity"),
    by_code = .depgraph_issue_count_table(issues, "code")
  )
}

.depgraph_old_checks_to_levels <- function(checks) {
  checks <- unique(as.character(checks))
  levels <- character()

  if (any(checks %in% c("ids", "references", "cardinality"))) {
    levels <- c(levels, "structural")
  }
  if (any(checks %in% c("schema", "time", "cardinality"))) {
    levels <- c(levels, "semantic")
  }

  unique(levels)
}

.validate_structural <- function(graph) {
  node_data <- graph$nodes$data
  edge_data <- graph$edges$data
  issues <- list()

  alignment <- .depgraph_graph_table_alignment(graph$graph, node_data, edge_data)
  if (!isTRUE(alignment$valid)) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "igraph_table_mismatch",
      message = alignment$message
    )
  }

  if (anyNA(node_data$node_id) || any(!nzchar(node_data$node_id))) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "missing_node_id",
      message = "Node table contains missing or empty `node_id` values."
    )
  }

  if (anyDuplicated(node_data$node_id) > 0L) {
    dup_ids <- unique(node_data$node_id[duplicated(node_data$node_id)])
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "duplicate_node_id",
      message = paste0("Node table contains duplicated `node_id` values: ", paste(dup_ids, collapse = ", ")),
      node_ids = dup_ids
    )
  }

  if (anyNA(edge_data$edge_id) || any(!nzchar(edge_data$edge_id))) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "missing_edge_id",
      message = "Edge table contains missing or empty `edge_id` values."
    )
  }

  if (anyDuplicated(edge_data$edge_id) > 0L) {
    dup_ids <- unique(edge_data$edge_id[duplicated(edge_data$edge_id)])
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "duplicate_edge_id",
      message = paste0("Edge table contains duplicated `edge_id` values: ", paste(dup_ids, collapse = ", ")),
      edge_ids = dup_ids
    )
  }

  if (nrow(edge_data) > 0L) {
    triplets <- paste(edge_data$from, edge_data$to, edge_data$edge_type, sep = "||")
    if (anyDuplicated(triplets) > 0L) {
      dup_rows <- duplicated(triplets) | duplicated(triplets, fromLast = TRUE)
      dup_edges <- edge_data$edge_id[dup_rows]
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "structural",
        severity = "error",
        code = "duplicate_edge_triplet",
        message = "Edge table contains duplicate `(from, to, edge_type)` relations.",
        edge_ids = dup_edges
      )
    }
  }

  valid_nodes <- node_data$node_id
  bad_from <- setdiff(unique(edge_data$from), valid_nodes)
  bad_to <- setdiff(unique(edge_data$to), valid_nodes)
  if (length(bad_from) > 0L) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "missing_source_node",
      message = paste0("Edges reference missing source nodes: ", paste(bad_from, collapse = ", ")),
      node_ids = bad_from
    )
  }
  if (length(bad_to) > 0L) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "missing_target_node",
      message = paste0("Edges reference missing target nodes: ", paste(bad_to, collapse = ", ")),
      node_ids = bad_to
    )
  }

  bad_node_types <- setdiff(unique(node_data$node_type), .depgraph_node_types)
  if (length(bad_node_types) > 0L) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "structural",
      severity = "error",
      code = "unsupported_node_type",
      message = paste0("Unsupported node types found: ", paste(bad_node_types, collapse = ", "))
    )
  }

  type_map <- stats::setNames(node_data$node_type, node_data$node_id)
  if (nrow(edge_data) > 0L) {
    for (i in seq_len(nrow(edge_data))) {
      edge_type <- edge_data$edge_type[[i]]
      schema_row <- .depgraph_edge_type_schema(edge_type)
      if (nrow(schema_row) == 0L) {
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "structural",
          severity = "error",
          code = "unsupported_edge_type",
          message = paste0("Unknown edge type encountered: ", edge_type),
          edge_ids = edge_data$edge_id[[i]]
        )
        next
      }

      observed_from <- unname(type_map[[edge_data$from[[i]]]])
      observed_to <- unname(type_map[[edge_data$to[[i]]]])
      if (!identical(observed_from, schema_row$from_type[[1L]]) || !identical(observed_to, schema_row$to_type[[1L]])) {
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "structural",
          severity = "error",
          code = "invalid_edge_signature",
          message = paste0(
            "Edge `", edge_data$edge_id[[i]], "` violates schema for `", edge_type,
            "` (expected ", schema_row$from_type[[1L]], " -> ", schema_row$to_type[[1L]],
            ", found ", observed_from, " -> ", observed_to, ")."
          ),
          node_ids = c(edge_data$from[[i]], edge_data$to[[i]]),
          edge_ids = edge_data$edge_id[[i]]
        )
      }
    }
  }

  for (edge_type in .depgraph_single_target_edges) {
    subset_idx <- edge_data$edge_type == edge_type
    if (!any(subset_idx)) {
      next
    }
    edge_subset <- edge_data[subset_idx, c("edge_id", "from", "to"), drop = FALSE]
    counts <- stats::aggregate(
      edge_subset$to,
      by = list(from = edge_subset$from),
      FUN = function(x) length(unique(x))
    )
    bad <- counts[counts$x > 1L, , drop = FALSE]
    if (nrow(bad) > 0L) {
      offending_edges <- edge_subset$edge_id[edge_subset$from %in% bad$from]
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "structural",
        severity = "error",
        code = "single_target_violation",
        message = paste0(
          "Edge type `", edge_type, "` links a source node to multiple targets: ",
          paste(bad$from, collapse = ", ")
        ),
        node_ids = bad$from,
        edge_ids = offending_edges,
        details = list(edge_type = edge_type)
      )
    }
  }

  .depgraph_bind_issues(issues)
}

.validate_semantic <- function(graph) {
  node_data <- graph$nodes$data
  edge_data <- graph$edges$data
  issues <- list()

  for (i in seq_len(nrow(node_data))) {
    attr_check <- .depgraph_validate_node_attrs(
      node_type = node_data$node_type[[i]],
      attrs = node_data$attrs[[i]]
    )

    if (length(attr_check$missing_required) > 0L) {
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "semantic",
        severity = "error",
        code = "missing_required_node_attr",
        message = paste0(
          "Node `", node_data$node_id[[i]], "` is missing required attributes: ",
          paste(attr_check$missing_required, collapse = ", ")
        ),
        node_ids = node_data$node_id[[i]]
      )
    }

    if (length(attr_check$unknown_attrs) > 0L) {
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "semantic",
        severity = "warning",
        code = "non_canonical_node_attr",
        message = paste0(
          "Node `", node_data$node_id[[i]], "` contains non-canonical attributes: ",
          paste(attr_check$unknown_attrs, collapse = ", ")
        ),
        node_ids = node_data$node_id[[i]]
      )
    }
  }

  sample_ids <- node_data$node_id[node_data$node_type == "Sample"]
  for (relation in names(.depgraph_semantic_relation_rules)) {
    rule <- .depgraph_semantic_relation_rules[[relation]]
    schema_row <- .depgraph_edge_type_schema(relation)
    relation_edges <- edge_data[edge_data$edge_type == relation, c("edge_id", "from", "to"), drop = FALSE]

    counts <- stats::setNames(integer(length(sample_ids)), sample_ids)
    targets <- vector("list", length(sample_ids))
    names(targets) <- sample_ids

    if (nrow(relation_edges) > 0L) {
      observed_counts <- stats::aggregate(
        relation_edges$to,
        by = list(from = relation_edges$from),
        FUN = function(x) length(unique(x))
      )
      counts[observed_counts$from] <- observed_counts$x
      for (sample_id in unique(relation_edges$from)) {
        targets[[sample_id]] <- unique(relation_edges$to[relation_edges$from == sample_id])
      }
    }

    if (!is.null(rule$missing_code) && !is.null(rule$missing_severity) && rule$min_targets > 0L) {
      target_type_present <- nrow(schema_row) == 1L &&
        any(node_data$node_type == schema_row$to_type[[1L]])
      if (nrow(relation_edges) > 0L || target_type_present) {
        missing_samples <- names(counts)[counts < rule$min_targets]
        if (length(missing_samples) > 0L) {
          issues[[length(issues) + 1L]] <- .new_validation_issue(
            level = "semantic",
            severity = rule$missing_severity,
            code = rule$missing_code,
            message = paste0(
              "Samples missing required relation `", relation, "`: ",
              paste(missing_samples, collapse = ", ")
            ),
            node_ids = missing_samples,
            details = list(edge_type = relation)
          )
        }
      }
    }

    allow_multiple <- identical(relation, "sample_belongs_to_subject") &&
      .depgraph_validation_override(graph, "allow_multi_subject_samples", default = FALSE)

    if (!allow_multiple && !is.null(rule$multiple_code) && !is.null(rule$multiple_severity) && is.finite(rule$max_targets)) {
      multi_samples <- names(counts)[counts > rule$max_targets]
      if (length(multi_samples) > 0L) {
        edge_ids <- relation_edges$edge_id[relation_edges$from %in% multi_samples]
        related_nodes <- unique(c(
          multi_samples,
          unlist(targets[multi_samples], use.names = FALSE)
        ))
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "semantic",
          severity = rule$multiple_severity,
          code = rule$multiple_code,
          message = paste0(
            "Samples violate relation cardinality for `", relation, "`: ",
            paste(multi_samples, collapse = ", ")
          ),
          node_ids = related_nodes,
          edge_ids = edge_ids,
          details = list(edge_type = relation)
        )
      }
    }
  }

  feature_nodes <- node_data[node_data$node_type == "FeatureSet", , drop = FALSE]
  if (nrow(feature_nodes) > 0L) {
    for (i in seq_len(nrow(feature_nodes))) {
      derivation_scope <- .depgraph_extract_attr(feature_nodes$attrs[[i]], "derivation_scope")
      if (!is.na(derivation_scope) && !derivation_scope %in% .depgraph_allowed_derivation_scopes) {
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "semantic",
          severity = "warning",
          code = "invalid_featureset_derivation_scope",
          message = paste0(
            "FeatureSet node `", feature_nodes$node_id[[i]], "` has invalid `derivation_scope`: ",
            derivation_scope
          ),
          node_ids = feature_nodes$node_id[[i]]
        )
      }
    }
  }

  outcome_nodes <- node_data[node_data$node_type == "Outcome", , drop = FALSE]
  if (nrow(outcome_nodes) > 0L) {
    for (i in seq_len(nrow(outcome_nodes))) {
      observation_level <- .depgraph_extract_attr(outcome_nodes$attrs[[i]], "observation_level")
      if (!is.na(observation_level) && !observation_level %in% .depgraph_allowed_observation_levels) {
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "semantic",
          severity = "warning",
          code = "invalid_outcome_observation_level",
          message = paste0(
            "Outcome node `", outcome_nodes$node_id[[i]], "` has invalid `observation_level`: ",
            observation_level
          ),
          node_ids = outcome_nodes$node_id[[i]]
        )
      }
    }
  }

  time_nodes <- node_data[node_data$node_type == "Timepoint", , drop = FALSE]
  if (nrow(time_nodes) > 0L) {
    for (i in seq_len(nrow(time_nodes))) {
      value <- .depgraph_extract_attr(time_nodes$attrs[[i]], "time_index")
      converted <- suppressWarnings(as.numeric(value))
      if (!is.na(value) && is.na(converted)) {
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "semantic",
          severity = "warning",
          code = "invalid_time_index",
          message = paste0("Timepoint node `", time_nodes$node_id[[i]], "` has a non-numeric `time_index`."),
          node_ids = time_nodes$node_id[[i]]
        )
      }
    }
  }

  precedes <- edge_data[edge_data$edge_type == "timepoint_precedes", , drop = FALSE]
  if (nrow(precedes) > 0L && any(precedes$from == precedes$to)) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "semantic",
      severity = "error",
      code = "timepoint_self_loop",
      message = "`timepoint_precedes` contains self-loops.",
      edge_ids = precedes$edge_id[precedes$from == precedes$to]
    )
  }

  time_consistency <- .depgraph_time_order_consistency(node_data, edge_data)
  if (length(time_consistency$cycle_edge_ids) > 0L) {
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "semantic",
      severity = "error",
      code = "timepoint_precedence_cycle",
      message = "`timepoint_precedes` must define an acyclic ordering.",
      edge_ids = time_consistency$cycle_edge_ids
    )
  }
  if (nrow(time_consistency$conflicting_pairs) > 0L) {
    conflicting_nodes <- unique(c(
      time_consistency$conflicting_pairs$from,
      time_consistency$conflicting_pairs$to
    ))
    pair_labels <- paste0(
      time_consistency$conflicting_pairs$from,
      " -> ",
      time_consistency$conflicting_pairs$to
    )
    issues[[length(issues) + 1L]] <- .new_validation_issue(
      level = "semantic",
      severity = "error",
      code = "time_order_conflict",
      message = paste0(
        "Timepoint ordering metadata conflict between `time_index` and `timepoint_precedes`: ",
        paste(pair_labels, collapse = ", ")
      ),
      node_ids = conflicting_nodes,
      edge_ids = time_consistency$conflicting_pairs$edge_id
    )
  }

  .depgraph_bind_issues(issues)
}

.validate_leakage <- function(graph) {
  node_data <- graph$nodes$data
  edge_data <- graph$edges$data
  issues <- list()

  subject_edges <- edge_data[edge_data$edge_type == "sample_belongs_to_subject", c("edge_id", "from", "to"), drop = FALSE]
  if (nrow(subject_edges) > 0L) {
    sample_counts <- stats::aggregate(
      subject_edges$from,
      by = list(subject = subject_edges$to),
      FUN = function(x) length(unique(x))
    )
    repeated <- sample_counts[sample_counts$x > 1L, , drop = FALSE]
    for (i in seq_len(nrow(repeated))) {
      subject_id <- repeated$subject[[i]]
      sample_ids <- unique(subject_edges$from[subject_edges$to == subject_id])
      edge_ids <- subject_edges$edge_id[subject_edges$to == subject_id]
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "leakage",
        severity = "advisory",
        code = "repeated_subject_samples",
        message = paste0("Subject `", subject_id, "` is linked to multiple samples."),
        node_ids = c(subject_id, sample_ids),
        edge_ids = edge_ids
      )
    }
  }

  study_edges <- edge_data[edge_data$edge_type == "sample_from_study", c("edge_id", "from", "to"), drop = FALSE]
  if (nrow(subject_edges) > 0L && nrow(study_edges) > 0L) {
    subject_study <- merge(
      subject_edges[, c("from", "to")],
      study_edges[, c("from", "to")],
      by = "from",
      suffixes = c("_subject", "_study")
    )
    if (nrow(subject_study) > 0L) {
      study_counts <- stats::aggregate(
        subject_study$to_study,
        by = list(subject = subject_study$to_subject),
        FUN = function(x) length(unique(x))
      )
      cross_study <- study_counts[study_counts$x > 1L, , drop = FALSE]
      for (i in seq_len(nrow(cross_study))) {
        subject_id <- cross_study$subject[[i]]
        linked_samples <- unique(subject_study$from[subject_study$to_subject == subject_id])
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "leakage",
          severity = "warning",
          code = "subject_cross_study_overlap",
          message = paste0("Subject `", subject_id, "` appears across multiple studies."),
          node_ids = c(subject_id, linked_samples)
        )
      }
    }
  }

  feature_nodes <- node_data[node_data$node_type == "FeatureSet", , drop = FALSE]
  if (nrow(feature_nodes) > 0L) {
    for (i in seq_len(nrow(feature_nodes))) {
      derivation_scope <- .depgraph_extract_attr(feature_nodes$attrs[[i]], "derivation_scope")
      if (!is.na(derivation_scope) && identical(derivation_scope, "per_dataset")) {
        issues[[length(issues) + 1L]] <- .new_validation_issue(
          level = "leakage",
          severity = "advisory",
          code = "per_dataset_featureset",
          message = paste0("FeatureSet `", feature_nodes$node_id[[i]], "` was derived at the full-dataset scope."),
          node_ids = feature_nodes$node_id[[i]]
        )
      }
    }
  }

  feature_edges <- edge_data[edge_data$edge_type == "sample_uses_featureset", c("edge_id", "from", "to"), drop = FALSE]
  if (nrow(feature_edges) > 0L) {
    feature_counts <- stats::aggregate(
      feature_edges$from,
      by = list(featureset = feature_edges$to),
      FUN = function(x) length(unique(x))
    )
    shared <- feature_counts[feature_counts$x > 1L, , drop = FALSE]
    for (i in seq_len(nrow(shared))) {
      featureset_id <- shared$featureset[[i]]
      linked_samples <- unique(feature_edges$from[feature_edges$to == featureset_id])
      edge_ids <- feature_edges$edge_id[feature_edges$to == featureset_id]
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "leakage",
        severity = "advisory",
        code = "shared_featureset_provenance",
        message = paste0("FeatureSet `", featureset_id, "` is shared across multiple samples."),
        node_ids = c(featureset_id, linked_samples),
        edge_ids = edge_ids
      )
    }
  }

  has_time_edges <- any(edge_data$edge_type == "sample_collected_at_timepoint")
  if (has_time_edges) {
    has_precedes <- any(edge_data$edge_type == "timepoint_precedes")
    time_nodes <- node_data[node_data$node_type == "Timepoint", , drop = FALSE]
    time_indices <- vapply(
      time_nodes$attrs,
      function(x) {
        value <- .depgraph_extract_attr(x, "time_index")
        if (length(value) == 0L || is.na(value)) {
          return(NA_character_)
        }
        as.character(value)
      },
      FUN.VALUE = character(1)
    )
    if (!has_precedes && (length(time_indices) == 0L || all(is.na(time_indices)))) {
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "leakage",
        severity = "warning",
        code = "missing_time_ordering",
        message = "Timepoint-linked samples exist but no explicit time ordering metadata were found."
      )
    }
  }

  batch_edges <- edge_data[edge_data$edge_type == "sample_processed_in_batch", c("edge_id", "from", "to"), drop = FALSE]
  n_samples <- sum(node_data$node_type == "Sample")
  if (nrow(batch_edges) > 0L && n_samples > 0L) {
    batch_counts <- stats::aggregate(
      batch_edges$from,
      by = list(batch = batch_edges$to),
      FUN = function(x) length(unique(x))
    )
    threshold <- max(3L, ceiling(n_samples * 0.5))
    heavy <- batch_counts[batch_counts$x >= threshold, , drop = FALSE]
    for (i in seq_len(nrow(heavy))) {
      batch_id <- heavy$batch[[i]]
      issues[[length(issues) + 1L]] <- .new_validation_issue(
        level = "leakage",
        severity = "advisory",
        code = "heavy_batch_reuse",
        message = paste0("Batch `", batch_id, "` is reused across many samples."),
        node_ids = batch_id,
        edge_ids = batch_edges$edge_id[batch_edges$to == batch_id]
      )
    }
  }

  .depgraph_bind_issues(issues)
}

#' @rdname build_dependency_graph
#' @export
validate_graph <- function(graph, checks = c("ids", "references", "cardinality", "schema", "time"), error_on_fail = FALSE, levels = NULL, severities = NULL, validation_overrides = NULL) {
  .depgraph_assert(inherits(graph, "dependency_graph"), "`graph` must be a `dependency_graph`.")

  if (!missing(checks)) {
    .Deprecated(
      msg = paste0(
        "The `checks` argument of `validate_graph()` is deprecated. ",
        "Use `levels` and `severities` instead."
      ),
      package = "splitGraph"
    )
  }

  if (!is.null(validation_overrides)) {
    .depgraph_assert(
      is.list(validation_overrides) && (length(validation_overrides) == 0L ||
        (!is.null(names(validation_overrides)) &&
           all(nzchar(names(validation_overrides))))),
      "`validation_overrides` must be a named list."
    )
    graph <- .depgraph_with_overrides(graph, validation_overrides)
  }

  selected_levels <- levels
  if (is.null(selected_levels)) {
    if (missing(checks)) {
      selected_levels <- .depgraph_validation_levels
    } else {
      selected_levels <- .depgraph_old_checks_to_levels(checks)
      if (length(selected_levels) == 0L) {
        selected_levels <- .depgraph_validation_levels
      }
    }
  }
  selected_levels <- unique(as.character(selected_levels))
  .depgraph_assert(all(selected_levels %in% .depgraph_validation_levels), "`levels` contains unsupported values.")

  selected_severities <- if (is.null(severities)) .depgraph_validation_severities else unique(as.character(severities))
  .depgraph_assert(all(selected_severities %in% .depgraph_validation_severities), "`severities` contains unsupported values.")

  issue_sets <- list()
  if ("structural" %in% selected_levels) {
    issue_sets[[length(issue_sets) + 1L]] <- .validate_structural(graph)
  }
  if ("semantic" %in% selected_levels) {
    issue_sets[[length(issue_sets) + 1L]] <- .validate_semantic(graph)
  }
  if ("leakage" %in% selected_levels) {
    issue_sets[[length(issue_sets) + 1L]] <- .validate_leakage(graph)
  }

  all_issues <- .depgraph_bind_issues(issue_sets)
  issues <- all_issues
  if (nrow(issues) > 0L) {
    issues <- issues[issues$severity %in% selected_severities, , drop = FALSE]
    row.names(issues) <- NULL
  }

  all_errors <- unique(all_issues$message[all_issues$severity == "error"])
  all_warnings <- unique(all_issues$message[all_issues$severity == "warning"])
  all_advisories <- unique(all_issues$message[all_issues$severity == "advisory"])

  metrics <- list(
    n_nodes = nrow(graph$nodes$data),
    n_edges = nrow(graph$edges$data),
    n_samples = sum(graph$nodes$data$node_type == "Sample"),
    n_issues = nrow(all_issues),
    n_errors = length(all_errors),
    n_warnings = length(all_warnings),
    n_advisories = length(all_advisories)
  )

  report <- depgraph_validation_report(
    graph_name = graph$metadata$graph_name,
    issues = issues,
    metrics = metrics,
    metadata = list(
      levels = selected_levels,
      severities = selected_severities
    ),
    valid = length(all_errors) == 0L,
    errors = all_errors,
    warnings = all_warnings,
    advisories = all_advisories
  )

  if (length(all_errors) > 0L && isTRUE(error_on_fail)) {
    stop(paste(c("Graph validation failed.", report$errors), collapse = "\n"), call. = FALSE)
  }

  report
}

#' @rdname build_dependency_graph
#' @export
validate_depgraph <- function(graph, checks = c("ids", "references", "cardinality", "schema", "time"), error_on_fail = FALSE, levels = NULL, severities = NULL, validation_overrides = NULL) {
  .Deprecated(
    new = "validate_graph",
    package = "splitGraph",
    msg = "`validate_depgraph()` is deprecated. Use `validate_graph()` instead."
  )
  if (missing(checks)) {
    return(validate_graph(
      graph = graph,
      error_on_fail = error_on_fail,
      levels = levels,
      severities = severities,
      validation_overrides = validation_overrides
    ))
  }

  validate_graph(
    graph = graph,
    checks = checks,
    error_on_fail = error_on_fail,
    levels = levels,
    severities = severities,
    validation_overrides = validation_overrides
  )
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)

## ----derive-------------------------------------------------------------------
library(splitGraph)

meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5"),
  subject_id   = c("P1", "P1", "P2", "P3", "P3"),
  timepoint_id = c("T0", "T1", "T0", "T2", "T0"),
  time_index   = c(0, 1, 0, 2, 0),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "handoff-demo")

# Group so that repeated measures of the same subject never straddle a split.
constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(constraint, graph = g)

path <- tempfile(fileext = ".json")
write_split_spec(spec, path)

## ----validate-----------------------------------------------------------------
report <- validate_split_spec_json(path)
report$valid

# The R-side grouping we expect Python to reproduce:
grouping_vector(constraint)

## ----pypath-------------------------------------------------------------------
system.file("python", package = "splitGraph")

## ----conformance, eval = nzchar(Sys.which("python3")) && requireNamespace("jsonlite", quietly = TRUE)----
script   <- system.file("python", "conformance.py", package = "splitGraph")
out_path <- tempfile(fileext = ".json")

# Run the Python reader on our JSON file; it writes back what it recovered.
status <- system2(
  "python3", c("-B", shQuote(script), shQuote(path), shQuote(out_path)),
  stdout = FALSE, stderr = FALSE
)

if (status == 0 && file.exists(out_path)) {
  recovered <- jsonlite::fromJSON(out_path)

  # Grouping recovered by Python:
  print(unlist(recovered$grouping))

  # Identical to the grouping R produced?
  r_grouping <- grouping_vector(constraint)
  cat("Python matches R exactly:",
      identical(unlist(recovered$grouping)[names(r_grouping)],
                r_grouping[names(r_grouping)]), "\n")
}

---
title: "splitGraph: From Metadata to Leakage-Aware Split Design"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_document:
    toc: true
    toc_float: true
    number_sections: true
    theme: flatly
    highlight: tango
vignette: >
  %\VignetteIndexEntry{splitGraph: From Metadata to Leakage-Aware Split Design}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}

or_empty <- function(x) {
  if (is.null(x)) character() else x
}
```

## Why `splitGraph` exists

Leakage in biomedical evaluation workflows often comes from dataset structure rather
than from an obvious coding mistake. Two samples may look independent in a
model matrix while still sharing the same subject, batch, study, timepoint, or
feature provenance. If those relationships stay implicit, train/test
separation can look correct while violating the scientific separation you
actually intended.

`splitGraph` exists to make those relationships explicit before evaluation. It
turns metadata into a typed dependency graph that can be:

- validated for structural and leakage-relevant problems
- queried to inspect hidden overlap and provenance
- converted into deterministic split constraints
- translated into a stable, tool-agnostic split specification through the
  `split_spec` class and `as_split_spec()` / `validate_split_spec()` API

The package is intentionally narrow. It does not fit models, run preprocessing
pipelines, or generate resamples by itself. Its job is to represent dependency
structure clearly enough that downstream evaluation can be trustworthy.

## A realistic toy dataset

The example below includes exactly the kinds of relationships that usually
matter for leakage-aware evaluation:

- repeated subjects (`P1` and `P2`)
- reused batch (`B1`)
- one subject (`P2`) appearing across studies
- explicit time ordering with one sample missing time metadata
- a feature set derived at the full-dataset scope

```{r metadata}
meta <- data.frame(
  sample_id = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id = c("P1", "P1", "P2", "P3", "P4", "P2"),
  batch_id = c("B1", "B2", "B1", "B3", NA, "B1"),
  study_id = c("ST1", "ST1", "ST1", "ST2", "ST3", "ST2"),
  timepoint_id = c("T0", "T1", "T0", "T2", NA, "T1"),
  assay_id = c("RNAseq", "RNAseq", "RNAseq", "RNAseq", "Proteomics", "RNAseq"),
  featureset_id = c("FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_PROT", "FS_GLOBAL"),
  outcome_id = c("O_case", "O_case", "O_ctrl", "O_case", "O_ctrl", "O_ctrl"),
  stringsAsFactors = FALSE
)

meta
```

This is still a small example, but it already contains enough structure to
make naive random splitting risky.

## Fast path: `graph_from_metadata()`

When your metadata already uses the canonical column names (`sample_id`,
`subject_id`, `batch_id`, `study_id`, `timepoint_id`, `time_index`,
`assay_id`, `featureset_id`, `outcome_id` / `outcome_value`),
`graph_from_metadata()` does ingestion, typed node construction, canonical
edge construction, and optional `timepoint_precedes` derivation in a single
call:

```{r fast-path}
quick_graph <- graph_from_metadata(
  data.frame(
    sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
    subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
    batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
    timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
    time_index   = c(0, 1, 0, 1, 0, 1),
    outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl")
  ),
  graph_name = "quick_demo"
)

quick_graph
```

If you pass `outcome_value` as a numeric column (e.g. `0`/`1`), the package
will create one `Outcome` node per distinct numeric value (`outcome:0`,
`outcome:1`) and emit a warning. For most workflows you want a character
class label, so prefer `outcome_id`.

The rest of this vignette uses the explicit constructor path because it lets
us show node attributes (`time_index`, `visit_label`, `platform`,
`derivation_scope`) and edges that `graph_from_metadata()` does not build
for you on its own — specifically `featureset_generated_from_study` and
`featureset_generated_from_batch`. (`subject_has_outcome` is the one
non-default outcome edge `graph_from_metadata()` will build for you when
you pass `outcome_scope = "subject"`.) Use `graph_from_metadata()` when
the canonical columns are enough; use the explicit path when you need
custom attributes or feature-provenance edges.

## Ingest metadata and build typed nodes and edges

The first step is to standardize metadata and then turn each entity type into
canonical graph nodes. Sample-level relations become typed edges.

```{r construction}
meta <- ingest_metadata(meta, dataset_name = "VignetteDemo")

sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")

time_nodes <- create_nodes(
  data.frame(
    timepoint_id = c("T0", "T1", "T2"),
    time_index = c(0L, 1L, 2L),
    visit_label = c("baseline", "follow_up", "late_follow_up"),
    stringsAsFactors = FALSE
  ),
  type = "Timepoint",
  id_col = "timepoint_id",
  attr_cols = c("time_index", "visit_label")
)

assay_nodes <- create_nodes(
  data.frame(
    assay_id = c("RNAseq", "Proteomics"),
    modality = c("transcriptomics", "proteomics"),
    platform = c("NovaSeq", "Orbitrap"),
    stringsAsFactors = FALSE
  ),
  type = "Assay",
  id_col = "assay_id",
  attr_cols = c("modality", "platform")
)

featureset_nodes <- create_nodes(
  data.frame(
    featureset_id = c("FS_GLOBAL", "FS_PROT"),
    featureset_name = c("global_rna_signature", "proteomics_panel"),
    derivation_scope = c("per_dataset", "external"),
    feature_count = c(500L, 80L),
    stringsAsFactors = FALSE
  ),
  type = "FeatureSet",
  id_col = "featureset_id",
  attr_cols = c("featureset_name", "derivation_scope", "feature_count")
)

outcome_nodes <- create_nodes(
  data.frame(
    outcome_id = c("O_case", "O_ctrl"),
    outcome_name = c("response", "response"),
    outcome_type = c("binary", "binary"),
    observation_level = c("subject", "subject"),
    stringsAsFactors = FALSE
  ),
  type = "Outcome",
  id_col = "outcome_id",
  attr_cols = c("outcome_name", "outcome_type", "observation_level")
)

subject_edges <- create_edges(
  meta, "sample_id", "subject_id",
  "Sample", "Subject", "sample_belongs_to_subject"
)

batch_edges <- create_edges(
  meta, "sample_id", "batch_id",
  "Sample", "Batch", "sample_processed_in_batch",
  allow_missing = TRUE
)

study_edges <- create_edges(
  meta, "sample_id", "study_id",
  "Sample", "Study", "sample_from_study"
)

time_edges <- create_edges(
  meta, "sample_id", "timepoint_id",
  "Sample", "Timepoint", "sample_collected_at_timepoint",
  allow_missing = TRUE
)

assay_edges <- create_edges(
  meta, "sample_id", "assay_id",
  "Sample", "Assay", "sample_measured_by_assay"
)

featureset_edges <- create_edges(
  meta, "sample_id", "featureset_id",
  "Sample", "FeatureSet", "sample_uses_featureset"
)

outcome_edges <- create_edges(
  data.frame(
    subject_id = c("P1", "P2", "P3", "P4"),
    outcome_id = c("O_case", "O_ctrl", "O_case", "O_ctrl"),
    stringsAsFactors = FALSE
  ),
  "subject_id", "outcome_id",
  "Subject", "Outcome", "subject_has_outcome"
)

precedence_edges <- create_edges(
  data.frame(
    from_timepoint = c("T0", "T1"),
    to_timepoint = c("T1", "T2"),
    stringsAsFactors = FALSE
  ),
  "from_timepoint", "to_timepoint",
  "Timepoint", "Timepoint", "timepoint_precedes"
)

featureset_from_study <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    study_id = "ST1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "study_id",
  "FeatureSet", "Study", "featureset_generated_from_study"
)

featureset_from_batch <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    batch_id = "B1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "batch_id",
  "FeatureSet", "Batch", "featureset_generated_from_batch"
)
```

The node and edge tables are canonical and typed. The package assigns globally
unique node IDs such as `sample:S1` and `subject:P1`, so different entity
types cannot collide accidentally.

```{r construction-output}
sample_nodes
as.data.frame(sample_nodes)[, c("node_id", "node_type", "node_key", "label")]

edge_preview <- do.call(rbind, lapply(
  list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  as.data.frame
))

edge_preview[, c("from", "to", "edge_type")]
```

The node table shows the canonical sample IDs that everything else refers to.
The edge table shows the package's central design choice: dependency structure
is explicit, typed, and inspectable.

## Assemble the dependency graph

```{r graph}
graph <- build_dependency_graph(
  nodes = list(
    sample_nodes, subject_nodes, batch_nodes, study_nodes,
    time_nodes, assay_nodes, featureset_nodes, outcome_nodes
  ),
  edges = list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  graph_name = "vignette_graph",
  dataset_name = attr(meta, "dataset_name")
)

graph
summary(graph)
```

At this point the package has a single `dependency_graph` object with both
tabular and `igraph` representations behind it. The summary is useful because
it tells you exactly which entity types and relation types are present before
you derive any split rules.

### Visualize the typed structure

`plot()` renders the graph with a typed, layered layout: `Sample` on top,
peer dependencies (`Subject`, `Batch`, `Study`, `Timepoint`) in the middle
band, `Assay`/`FeatureSet` below that, and `Outcome` at the bottom. Node
colors are keyed to type and an auto-generated legend is drawn by default.

```{r plot, fig.width = 7, fig.height = 5}
plot(graph)
```

Useful options:

```{r plot-options, eval = FALSE}
plot(graph, layout = "sugiyama")         # alternative hierarchical layout
plot(graph, show_labels = FALSE)         # hide labels on dense graphs
plot(graph, legend = FALSE)              # suppress the legend
plot(graph, legend_position = "bottomright")
plot(graph, node_colors = c(Sample = "#000000"))
```

## Validate before you split

Validation is where `splitGraph` starts paying off. The graph below is
structurally valid, but it still carries leakage-relevant warnings and
advisories.

```{r validation}
validation <- validate_graph(graph)

validation
as.data.frame(validation)[, c("level", "severity", "code", "message")]
```

That output is the core value proposition of the package in one place:

- repeated subjects are surfaced explicitly
- cross-study subject overlap is surfaced explicitly
- full-dataset feature provenance is surfaced explicitly
- heavy batch reuse is surfaced explicitly

`valid = TRUE` here means the graph has no errors. It does not mean the dataset
is free of leakage risk. Warnings and advisories still matter.

The package is also intentionally strict about silent failure. If you ask for a
subset of samples and some of them do not resolve, it errors instead of
dropping them.

```{r strictness}
tryCatch(
  derive_split_constraints(graph, mode = "subject", samples = c("S1", "BAD")),
  error = function(e) e$message
)
```

That behavior is important in practice because quietly omitting samples would
change the truth of the split problem.

### When you really do need to relax a check

Some leakage-relevant rules have legitimate exceptions. For example, in
some pooled designs a single biological "sample" is intentionally paired
across two subjects. The semantic validator flags this as an error by
default, and `derive_split_constraints(mode = "subject")` refuses to
choose a subject for you — both consistent with the "no silent guessing"
stance.

To make the contrast concrete, build a tiny graph in which sample `S1`
is linked to two subjects:

```{r overrides-fixture}
multi_nodes <- graph_node_set(data.frame(
  node_id   = c("sample:S1", "subject:P1", "subject:P2"),
  node_type = c("Sample", "Subject", "Subject"),
  node_key  = c("S1", "P1", "P2"),
  label     = c("S1", "P1", "P2"),
  attrs     = I(list(list(), list(), list())),
  stringsAsFactors = FALSE
))

multi_edges <- graph_edge_set(data.frame(
  edge_id   = c("sample_belongs_to_subject:1", "sample_belongs_to_subject:2"),
  from      = c("sample:S1", "sample:S1"),
  to        = c("subject:P1", "subject:P2"),
  edge_type = c("sample_belongs_to_subject", "sample_belongs_to_subject"),
  attrs     = I(list(list(), list())),
  stringsAsFactors = FALSE
))

multi_graph <- dependency_graph(nodes = multi_nodes, edges = multi_edges)
```

Default behavior — the validator surfaces the multi-subject sample as an
error, and constraint derivation refuses to pick a subject:

```{r overrides-default}
default_report <- validate_graph(multi_graph)
default_report$valid
default_report$issues[, c("severity", "code", "message")]

tryCatch(
  derive_split_constraints(multi_graph, mode = "subject"),
  error = function(e) e$message
)
```

When the ambiguity is intended, opt in explicitly via
`validation_overrides`. The same key
(`allow_multi_subject_samples = TRUE`) is honored by both the validator
and the constraint deriver:

```{r overrides-on}
# Validator: pass.
permissive_report <- validate_graph(
  multi_graph,
  validation_overrides = list(allow_multi_subject_samples = TRUE)
)
permissive_report$valid

# Constraint derivation: pick the first listed subject and record the
# ambiguity in metadata$warnings instead of erroring.
multi_graph$metadata$validation_overrides <-
  list(allow_multi_subject_samples = TRUE)
relaxed_constraint <- derive_split_constraints(multi_graph, mode = "subject")
relaxed_constraint$sample_map[, c("sample_id", "group_id")]
relaxed_constraint$metadata$warnings
```

The override is documented under `?validate_graph`. Use it sparingly and
only when the relaxation matches your scientific intent — the message in
`metadata$warnings` is the audit trail that the choice was tolerated, not
hidden.

## Query the graph to inspect hidden structure

You can inspect local provenance, trace paths, and project direct sample
dependencies.

```{r neighbors-and-paths}
neighbors_s1 <- query_neighbors(graph, node_ids = "sample:S1", direction = "out")
neighbors_s1
as.data.frame(neighbors_s1)[, c("seed_node_id", "node_id", "node_type", "edge_type")]

subject_outcome_path <- query_shortest_paths(
  graph,
  from = "sample:S1",
  to = "outcome:O_case",
  edge_types = c("sample_belongs_to_subject", "subject_has_outcome")
)

subject_outcome_path
as.data.frame(subject_outcome_path)
```

The first query shows everything the graph knows directly about `S1`. The
second shows that `S1` reaches the subject-level outcome through its subject
node, which is exactly the kind of relationship that would stay implicit in a
plain metadata table.

```{r projected-dependencies}
shared_dependencies <- detect_shared_dependencies(
  graph,
  via = c("Subject", "Batch", "FeatureSet")
)

as.data.frame(shared_dependencies)[, c(
  "sample_id_1", "sample_id_2", "shared_node_type", "shared_node_id", "edge_type"
)]

dependency_components <- detect_dependency_components(
  graph,
  via = c("Subject", "Batch")
)

as.data.frame(dependency_components)
```

These projected queries are useful because they answer the splitting question
directly. They tell you which samples should be treated as structurally linked,
not just which metadata columns happen to match.

## Derive split constraints from the graph

`splitGraph` can derive direct constraints for subject, batch, study, and time
as well as composite constraints that combine multiple dependency sources.

```{r constraints}
subject_constraint <- derive_split_constraints(graph, mode = "subject")
batch_constraint <- derive_split_constraints(graph, mode = "batch")
study_constraint <- derive_split_constraints(graph, mode = "study")
time_constraint <- derive_split_constraints(graph, mode = "time")

strict_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "strict",
  via = c("Subject", "Batch")
)

rule_based_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "rule_based",
  priority = c("batch", "study", "subject", "time")
)

constraint_overview <- do.call(rbind, lapply(
  list(
    subject = subject_constraint,
    batch = batch_constraint,
    study = study_constraint,
    time = time_constraint,
    composite_strict = strict_constraint,
    composite_rule = rule_based_constraint
  ),
  function(x) {
    data.frame(
      strategy = x$strategy,
      groups = length(unique(x$sample_map$group_id)),
      warnings = if (is.null(x$metadata$warnings)) 0L else length(x$metadata$warnings),
      stringsAsFactors = FALSE
    )
  }
))

constraint_overview <- cbind(constraint = row.names(constraint_overview), constraint_overview)
row.names(constraint_overview) <- NULL

constraint_overview
```

That summary already shows why the package is useful: different notions of
dependency produce different splitting units.

### Batch constraints

```{r batch-constraint}
batch_constraint
as.data.frame(batch_constraint)[, c("sample_id", "group_id", "group_label", "explanation")]
```

Batch grouping keeps all `B1` samples together and preserves `S5` as an
explicit singleton because it has no batch assignment. Missing structure is not
hidden.

### Time constraints

```{r time-constraint}
time_constraint
as.data.frame(time_constraint)[, c("sample_id", "group_id", "timepoint_id", "order_rank")]
```

Time grouping adds `order_rank`, which is the field downstream tooling actually
needs for ordered evaluation. The missing timepoint on `S5` stays visible as
`NA`, so ordering is partial rather than pretended.

### Composite constraints

```{r composite-constraints}
strict_constraint
as.data.frame(strict_constraint)[, c("sample_id", "group_id", "constraint_type")]

rule_based_constraint
as.data.frame(rule_based_constraint)[, c("sample_id", "group_id", "constraint_type", "group_label")]
```

The strict composite constraint uses transitive closure: `S1`, `S2`, `S3`, and
`S6` end up in the same group because subject and batch links connect them into
one dependency component. The rule-based composite constraint is different: it
uses the highest-priority available dependency per sample, so `S5` falls back
to study-level grouping instead of becoming a composite component.

## Time ordering can come from precedence edges alone

If explicit `time_index` metadata are unavailable, `splitGraph` can still infer
time order from `timepoint_precedes` edges.

```{r precedence-only}
precedence_meta <- data.frame(
  sample_id = c("S1", "S2", "S3"),
  subject_id = c("P1", "P1", "P2"),
  study_id = c("ST1", "ST1", "ST2"),
  timepoint_id = c("T0", "T1", "T2"),
  stringsAsFactors = FALSE
)

precedence_graph <- build_dependency_graph(
  nodes = list(
    create_nodes(precedence_meta, type = "Sample", id_col = "sample_id"),
    create_nodes(precedence_meta, type = "Subject", id_col = "subject_id"),
    create_nodes(precedence_meta, type = "Study", id_col = "study_id"),
    create_nodes(
      data.frame(timepoint_id = c("T0", "T1", "T2"), stringsAsFactors = FALSE),
      type = "Timepoint",
      id_col = "timepoint_id"
    )
  ),
  edges = list(
    create_edges(
      precedence_meta, "sample_id", "subject_id",
      "Sample", "Subject", "sample_belongs_to_subject"
    ),
    create_edges(
      precedence_meta, "sample_id", "study_id",
      "Sample", "Study", "sample_from_study"
    ),
    create_edges(
      precedence_meta, "sample_id", "timepoint_id",
      "Sample", "Timepoint", "sample_collected_at_timepoint"
    ),
    create_edges(
      data.frame(
        from_timepoint = c("T0", "T1"),
        to_timepoint = c("T1", "T2"),
        stringsAsFactors = FALSE
      ),
      "from_timepoint", "to_timepoint",
      "Timepoint", "Timepoint", "timepoint_precedes"
    )
  ),
  graph_name = "precedence_only_graph"
)

precedence_time_constraint <- derive_split_constraints(precedence_graph, mode = "time")

precedence_time_constraint$metadata$time_order_source
as.data.frame(precedence_time_constraint)[, c("sample_id", "timepoint_id", "time_index", "order_rank")]
```

The important detail is that ordering is still derived, but the source is
`timepoint_precedes` rather than `time_index`.

## Translate the constraint into a split specification

The graph-derived constraint is not the end of the workflow. The main handoff
target is a canonical sample-level split specification — the `split_spec`
class. Downstream tools consume it through their own adapters, so
`split_spec` stays tool-agnostic.

```{r split-spec}
split_spec <- as_split_spec(strict_constraint, graph = graph)
split_spec

as.data.frame(split_spec)[, c(
  "sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank"
)]

split_spec_validation <- validate_split_spec(split_spec)
split_spec_validation
as.data.frame(split_spec_validation)
```

This translation step is where the package becomes operational for downstream
evaluation workflows:

- `group_id` carries the split unit
- `batch_group` and `study_group` are available for blocking
- `order_rank` is available for ordered evaluation
- the generated object is validated before handoff

## Summarize the leakage picture in one object

The final helper combines graph validation, constraint diagnostics, and
split-spec readiness into one summary object.

```{r risk-summary}
risk_summary <- summarize_leakage_risks(
  graph,
  constraint = strict_constraint,
  split_spec = split_spec
)

risk_summary
as.data.frame(risk_summary)[, c("source", "severity", "category", "message")]
```

This is a useful stopping point before model training. It gives you one place
to review whether the graph is structurally sound, whether the chosen
constraint is overly singleton-heavy, and whether the downstream split spec is
ready to use.

## Downstream handoff

`split_spec` is the tool-agnostic handoff artifact. `splitGraph` does not
know about any particular resampling package — downstream consumers provide
their own adapters so `splitGraph` stays neutral. The typical end-to-end
flow is:

1. `graph_from_metadata()` (or the explicit constructor path) → typed
   `dependency_graph`
2. `derive_split_constraints(g, mode = ...)` → `split_constraint`
3. `as_split_spec(constraint, graph = g)` → `split_spec`
4. (optional) `write_split_spec(spec, path)` → JSON, for cross-session or
   cross-language handoff
5. adapter in the downstream package → native resamples

The `sample_data` frame carried by `split_spec` exposes exactly the columns
downstream adapters consume: `sample_id` for joining against the observation
frame, `group_id` for grouped resampling, `batch_group` / `study_group` for
blocking, and `order_rank` for ordered evaluation. Adapters can be built by
any package that wants to consume a `split_spec` — for example, on top of
`rsample::group_vfold_cv()` (grouped CV keyed to `group_id`) or
`rsample::rolling_origin()` (ordered evaluation keyed to `order_rank`).

### Persisting the handoff

If the downstream consumer is in a different R session — or in a
different language entirely — write the spec (and, if useful, the graph)
to JSON. The format is documented under `?write_split_spec` and
`?write_dependency_graph`, carries a `schema_version` field, and
preserves `NA` values as JSON `null`.

```{r serialize, eval = requireNamespace("jsonlite", quietly = TRUE)}
spec_path <- tempfile(fileext = ".json")
write_split_spec(split_spec, spec_path)

# Round-trip it back into R unchanged.
spec_round_trip <- read_split_spec(spec_path)
identical(split_spec$sample_data$group_id, spec_round_trip$sample_data$group_id)

unlink(spec_path)
```

For three small worked examples of consuming a `split_spec` from
downstream code (a base-R leave-one-group-out adapter and illustrative
`rsample` adapters), see the companion vignette:

```{r cookbook-pointer, eval = FALSE}
vignette("adapter-cookbook", package = "splitGraph")
```

## Case studies

The end-to-end workflow above shows the package surface. The case studies below
show how the same graph leads to different evaluation decisions depending on
the scientific question.

### Case study 1: repeated subjects in a longitudinal cohort

Suppose the real question is whether future observations from the same subject
should be held out from training. In this setting, subject reuse and time
ordering both matter, but they solve different problems.

```{r case-study-1}
subject_groups <- grouping_vector(subject_constraint)
time_groups <- time_constraint$sample_map[, c("sample_id", "group_id", "timepoint_id", "order_rank")]

subject_groups
time_groups
```

Interpretation:

- `S1` and `S2` share subject `P1`, so subject-grouped evaluation keeps them
  together.
- `S3` and `S6` share subject `P2`, so they also stay together under a
  subject-based split.
- time grouping adds a different axis: `T0`, `T1`, and `T2` become ordered
  units with explicit `order_rank`.

If the leakage concern is repeated measurements from the same individual, use
the subject constraint. If the evaluation question is prospective prediction,
the time constraint adds the ordering information you need.

### Case study 2: a subject reused across studies

The graph intentionally includes subject `P2` in both `ST1` and `ST2`. A
study-only split would treat those studies as separate units, but the graph
shows that subject overlap breaks the intended independence.

```{r case-study-2}
cross_study_issues <- as.data.frame(validation)[
  as.data.frame(validation)$code == "subject_cross_study_overlap",
  c("severity", "code", "message")
]

p2_shared <- detect_shared_dependencies(
  graph,
  via = "Subject",
  samples = c("S3", "S6")
)

study_only_map <- study_constraint$sample_map[, c("sample_id", "group_id", "group_label")]
strict_map <- strict_constraint$sample_map[, c("sample_id", "group_id", "constraint_type")]

cross_study_issues
as.data.frame(p2_shared)
study_only_map[study_only_map$sample_id %in% c("S3", "S6"), ]
strict_map[strict_map$sample_id %in% c("S3", "S6"), ]
```

Interpretation:

- validation surfaces the cross-study subject overlap directly
- the shared-dependency query confirms that `S3` and `S6` are linked through
  the same subject
- a study-only split would place them in different groups (`ST1` versus `ST2`)
- the strict composite constraint correctly keeps them in the same dependency
  component

This is exactly the kind of failure mode `splitGraph` is designed to expose:
metadata columns suggest a legitimate study split, but graph structure shows
that the split would still leak subject information.

### Case study 3: partially observed technical metadata

Real metadata are rarely complete. Here, `S5` has no batch assignment and no
timepoint assignment. The package does not pretend those fields exist. It keeps
the sample visible and tells you how the split logic handled it.

```{r case-study-3}
batch_missing <- batch_constraint$sample_map[
  batch_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "group_label", "explanation")
]

rule_based_missing <- rule_based_constraint$sample_map[
  rule_based_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "constraint_type", "group_label", "explanation")
]

split_spec_missing <- as.data.frame(split_spec)[
  as.data.frame(split_spec)$sample_id == "S5",
  c("sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank")
]

batch_missing
rule_based_missing
split_spec_missing
```

Interpretation:

- batch-based splitting keeps `S5` as an explicit singleton because batch
  metadata are missing
- the rule-based composite strategy falls back to study-level grouping for `S5`
- the translated split specification preserves the missing batch and time
  fields as `NA` rather than silently inventing values

That behavior matters because incomplete metadata are common. `splitGraph`
stays strict about what is known, but still produces a usable, inspectable
split object.

### Case study 4: choosing a defensible split strategy

A typical practical question is not "what can the package compute?" but "which
constraint should I actually use?" The answer depends on which dependency
source is scientifically unacceptable to leak across train and test.

```{r case-study-4}
strategy_summary <- data.frame(
  constraint = c("subject", "batch", "study", "time", "composite_strict", "composite_rule"),
  groups = c(
    length(unique(subject_constraint$sample_map$group_id)),
    length(unique(batch_constraint$sample_map$group_id)),
    length(unique(study_constraint$sample_map$group_id)),
    length(unique(time_constraint$sample_map$group_id)),
    length(unique(strict_constraint$sample_map$group_id)),
    length(unique(rule_based_constraint$sample_map$group_id))
  ),
  warnings = c(
    length(or_empty(subject_constraint$metadata$warnings)),
    length(or_empty(batch_constraint$metadata$warnings)),
    length(or_empty(study_constraint$metadata$warnings)),
    length(or_empty(time_constraint$metadata$warnings)),
    length(or_empty(strict_constraint$metadata$warnings)),
    length(or_empty(rule_based_constraint$metadata$warnings))
  ),
  recommended_resampling = c(
    as_split_spec(subject_constraint, graph = graph)$recommended_resampling,
    as_split_spec(batch_constraint, graph = graph)$recommended_resampling,
    as_split_spec(study_constraint, graph = graph)$recommended_resampling,
    as_split_spec(time_constraint, graph = graph)$recommended_resampling,
    as_split_spec(strict_constraint, graph = graph)$recommended_resampling,
    as_split_spec(rule_based_constraint, graph = graph)$recommended_resampling
  ),
  stringsAsFactors = FALSE
)

strategy_summary
```

Interpretation:

- subject grouping is the right default when repeated individuals are the
  dominant leakage source
- batch grouping is appropriate when technical runs are the main contamination
  risk
- study grouping is useful for cross-study generalization only when no higher
  level dependency crosses study boundaries
- strict composite grouping is the safest choice when multiple dependency
  sources can connect samples transitively
- rule-based composite grouping is a pragmatic fallback when you want a single
  deterministic hierarchy over partially observed metadata

The package does not choose the scientific objective for you. It makes the
trade-off visible and auditable.

## When `splitGraph` is useful

`splitGraph` is a good fit when:

- sample relationships are scientifically meaningful and must influence
  evaluation
- metadata contain repeated subjects, shared batches, multiple studies, or
  temporal structure
- feature provenance or outcome level matters for leakage assessment
- you want deterministic, inspectable split constraints instead of ad hoc
  grouping code

## What `splitGraph` is not for

`splitGraph` is not:

- a general biological network analysis package
- a model training framework
- a resampling engine
- a substitute for downstream performance auditing

Its value is earlier in the workflow: it makes dependency structure explicit so
that the split design itself can be justified.

## Takeaway

If you already know your data have repeated subjects, reused batches, temporal
ordering, or shared feature provenance, then you already have a graph problem
whether you model it explicitly or not. `splitGraph` is useful because it turns
that hidden graph into an object you can validate, query, and convert into a
split design that downstream tooling can trust.

---
title: "Adapter cookbook: from split_spec to native resamples"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_document:
    toc: true
    toc_float: true
    number_sections: true
    theme: flatly
    highlight: tango
vignette: >
  %\VignetteIndexEntry{Adapter cookbook: from split_spec to native resamples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}
```

## What this vignette is for

`splitGraph` ends at a `split_spec` object. It deliberately knows nothing
about `rsample`, `tidymodels`, or any other resampling engine. The handoff
contract is the `sample_data` table inside the spec plus a few scalar
fields (`group_var`, `block_vars`, `time_var`, `ordering_required`,
`recommended_resampling`).

This cookbook shows three small, self-contained adapters that turn a
`split_spec` into something a downstream workflow can use:

1. A **base-R adapter** that returns a list of `(train, test)` row-index
   pairs — runnable here, no extra dependencies.
2. An **`rsample::group_vfold_cv()`** adapter for grouped cross-validation
   keyed to `group_id`.
3. An **`rsample::rolling_origin()`** adapter for ordered evaluation keyed
   to `order_rank`.

Adapters 2 and 3 show idiomatic glue but are not evaluated in this
vignette so that `splitGraph` does not pick up `rsample` as a build-time
dependency.

The same pattern works for any other resampling library you happen to use.

## Build a split_spec to work with

```{r build-spec}
meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
  batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
  timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
  time_index   = c(0, 1, 0, 1, 0, 1),
  outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl"),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "cookbook")
subject_constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(subject_constraint, graph = g)
spec
```

The `sample_data` table is the contract:

```{r show-sample-data}
as.data.frame(spec)[, c("sample_id", "group_id", "batch_group", "order_rank")]
```

## Adapter 1 — base R: leave-one-group-out folds

This is the simplest meaningful adapter. It groups by whatever
`split_spec$group_var` says is the split unit, and returns one held-out
group per fold.

```{r adapter-base-r}
logo_folds <- function(spec, observation_data, sample_id_col = "sample_id") {
  stopifnot(inherits(spec, "split_spec"))
  if (!sample_id_col %in% names(observation_data)) {
    stop("`observation_data` must contain a `", sample_id_col, "` column.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$group_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )
  joined$.row <- seq_len(nrow(joined))
  groups <- split(joined$.row, joined[[spec$group_var]])

  lapply(names(groups), function(g) {
    list(
      group   = g,
      train   = unlist(groups[setdiff(names(groups), g)], use.names = FALSE),
      assess  = groups[[g]]
    )
  })
}

# Pretend we have an observation frame keyed by sample_id.
obs <- data.frame(
  sample_id = meta$sample_id,
  x = rnorm(nrow(meta)),
  y = rbinom(nrow(meta), 1, 0.5)
)

folds <- logo_folds(spec, obs)
length(folds)
folds[[1]]
```

That is the entire downstream contract: take `spec`, take an observation
frame, return train/assess index lists. Anything more complicated is
specific to a resampling library.

## Adapter 2 — `rsample::group_vfold_cv()`

Grouped CV keyed to `group_id`. The downstream package would typically
ship something like this; the adapter is short enough that you can paste
it into your own analysis script.

```{r adapter-rsample-group, eval=FALSE}
spec_to_group_vfold <- function(spec, observation_data,
                                v = NULL,
                                sample_id_col = "sample_id") {
  stopifnot(inherits(spec, "split_spec"))
  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Install rsample to use this adapter.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$group_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )

  n_groups <- length(unique(joined[[spec$group_var]]))
  if (is.null(v)) v <- n_groups

  rsample::group_vfold_cv(
    data  = joined,
    group = !!spec$group_var,
    v     = v
  )
}
```

`v = NULL` (the default above) gives leave-one-group-out, which is the
right default when `splitGraph` has already grouped samples by their
deepest leakage-relevant unit (e.g. subject). Pick a smaller `v` for
k-fold-style grouped CV.

## Adapter 3 — `rsample::rolling_origin()`

When `spec$ordering_required` is `TRUE` (or `spec$time_var` is set), the
right downstream object is an ordered split rather than a grouped one.

```{r adapter-rsample-rolling, eval=FALSE}
spec_to_rolling_origin <- function(spec, observation_data,
                                   sample_id_col = "sample_id",
                                   initial = NULL,
                                   assess = 1L) {
  stopifnot(inherits(spec, "split_spec"))
  if (is.null(spec$time_var)) {
    stop("This split_spec has no `time_var`; ordered evaluation is not available.")
  }
  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Install rsample to use this adapter.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$time_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )
  ordered <- joined[order(joined[[spec$time_var]]), , drop = FALSE]

  if (is.null(initial)) initial <- max(1L, floor(nrow(ordered) * 0.6))
  rsample::rolling_origin(ordered, initial = initial, assess = assess)
}
```

The key idea: `splitGraph` puts ordering information on the spec; the
adapter is just a thin shim that consumes it.

## Going across language boundaries via JSON

If the downstream consumer is not in R, write the spec to JSON and let the
consumer (Python, Julia, a CLI) interpret it.

```{r serialize, eval = requireNamespace("jsonlite", quietly = TRUE)}
tmp <- tempfile(fileext = ".json")
write_split_spec(spec, tmp)

# Inspect the first ~30 lines so the on-disk format is visible.
cat(readLines(tmp, n = 30), sep = "\n")

# And read it back exactly.
spec2 <- read_split_spec(tmp)
identical(spec$sample_data$group_id, spec2$sample_data$group_id)

unlink(tmp)
```

The same pair exists for `dependency_graph` (`write_dependency_graph()` /
`read_dependency_graph()`). Both formats are documented under
`?write_split_spec` and `?write_dependency_graph` and include a
`schema_version` field so consumers can detect drift.

## When you need a custom adapter

The only assumptions an adapter has to honor:

- `split_spec$sample_data` is keyed by `sample_id` (character).
- `split_spec$group_var` is the column that holds the splitting unit.
- `split_spec$block_vars` are present-but-coarser blocking columns.
- `split_spec$time_var`, when non-`NULL`, defines the ordering.
- `split_spec$recommended_resampling` is a hint, not a contract — your
  adapter is free to ignore it.

That is the whole interface. As long as those five fields are honored,
anything is a valid downstream consumer.

---
title: "Cross-language handoff: R to JSON to Python to scikit-learn"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Cross-language handoff: R to JSON to Python to scikit-learn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)
```

`split_spec` is designed as an *interchange format*, not as internal plumbing
for any one downstream package. This vignette shows the full path: derive a
constraint in R, serialize it to JSON, read it in Python with the shipped
`splitspec` reference consumer, and hand the recovered grouping straight to a
scikit-learn resampler.

The Python chunks below are shown but not executed, so building the vignette
needs no Python. To keep the central claim honest rather than asserted, the
vignette *does* run the shipped Python reader through R when a `python3`
interpreter is available (see "Verify the round-trip"), and shows that the
grouping it recovers matches R's exactly.

# Derive and serialize in R

```{r derive}
library(splitGraph)

meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5"),
  subject_id   = c("P1", "P1", "P2", "P3", "P3"),
  timepoint_id = c("T0", "T1", "T0", "T2", "T0"),
  time_index   = c(0, 1, 0, 2, 0),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "handoff-demo")

# Group so that repeated measures of the same subject never straddle a split.
constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(constraint, graph = g)

path <- tempfile(fileext = ".json")
write_split_spec(spec, path)
```

The written file carries a `$schema` reference and a `schema_version`, and can
be validated against the shipped JSON Schema before it ever leaves R:

```{r validate}
report <- validate_split_spec_json(path)
report$valid

# The R-side grouping we expect Python to reproduce:
grouping_vector(constraint)
```

# Read in Python

The reference consumer lives in the installed package under `inst/python`. On
the R side its location is:

```{r pypath}
system.file("python", package = "splitGraph")
```

Point Python at that directory (or install/copy the `splitspec` package), then:

```python
import sys
# sys.path.append(<the inst/python path printed above>)
from splitspec import load_split_spec

spec = load_split_spec("split_spec.json")

spec.schema_version        # "0.2.0"
spec.constraint_mode       # "subject"
spec.recommended_resampling  # "grouped_cv"

# Grouping keyed by sample_id — identical to R's grouping_vector():
spec.grouping()
# {'S1': 'subject:P1', 'S2': 'subject:P1', 'S3': 'subject:P2',
#  'S4': 'subject:P3', 'S5': 'subject:P3'}

df = spec.to_frame()       # pandas DataFrame of sample_data
```

# Verify the round-trip

Rather than take the comment above on faith, we can run the shipped Python
reader on the exact file we just wrote and compare what it recovers to R's
`grouping_vector()`. This is what `inst/python/conformance.py` does; the chunk
below invokes it through R and only runs when a `python3` interpreter is
present, so the vignette still builds without Python.

```{r conformance, eval = nzchar(Sys.which("python3")) && requireNamespace("jsonlite", quietly = TRUE)}
script   <- system.file("python", "conformance.py", package = "splitGraph")
out_path <- tempfile(fileext = ".json")

# Run the Python reader on our JSON file; it writes back what it recovered.
status <- system2(
  "python3", c("-B", shQuote(script), shQuote(path), shQuote(out_path)),
  stdout = FALSE, stderr = FALSE
)

if (status == 0 && file.exists(out_path)) {
  recovered <- jsonlite::fromJSON(out_path)

  # Grouping recovered by Python:
  print(unlist(recovered$grouping))

  # Identical to the grouping R produced?
  r_grouping <- grouping_vector(constraint)
  cat("Python matches R exactly:",
      identical(unlist(recovered$grouping)[names(r_grouping)],
                r_grouping[names(r_grouping)]), "\n")
}
```

The same script also checks `order_rank`, and the package's test suite runs this
comparison as an automated conformance test (skipped when Python is absent, and
never on CRAN). The point is that the partition is *decided once* in R and only
*reproduced* elsewhere — the two languages cannot disagree.

# Drive scikit-learn

The grouping vector plugs directly into `GroupKFold` (or
`StratifiedGroupKFold`), guaranteeing that all samples from a subject land in
the same fold:

```python
import numpy as np
from sklearn.model_selection import GroupKFold

groups = spec.groups()          # group_id per sample, in file order
X = np.zeros((len(groups), 1))  # placeholder design matrix

for train_idx, test_idx in GroupKFold(n_splits=3).split(X, groups=groups):
    train_groups = {groups[i] for i in train_idx}
    test_groups  = {groups[i] for i in test_idx}
    assert train_groups.isdisjoint(test_groups)  # no subject leaks across
```

For an ordered evaluation (a `mode = "time"` spec), sort by `order_rank` first
and use `TimeSeriesSplit`:

```python
from sklearn.model_selection import TimeSeriesSplit

order = spec.ordered_index()    # row indices sorted by order_rank
df_ordered = spec.to_frame().iloc[order].reset_index(drop=True)

for train_idx, test_idx in TimeSeriesSplit(n_splits=3).split(df_ordered):
    ...
```

# Why this matters

The leakage-aware partition is decided *once*, in R, from explicit and validated
dependency structure — and every other language merely *reproduces* it from the
`split_spec`. Nothing about the split logic is re-implemented in Python, so the
two sides cannot drift. `split_spec` is the contract; scikit-learn (here) and
`rsample` (on the R side) are just interchangeable consumers of it. That is what
makes it an interchange format rather than internal plumbing for any one tool.

---
title: "Adapter cookbook: from split_spec to native resamples"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Adapter cookbook: from split_spec to native resamples}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}
```

## What this vignette is for

`splitGraph` ends at a `split_spec` object. It deliberately knows nothing
about `rsample`, `tidymodels`, or any other resampling engine. The handoff
contract is the `sample_data` table inside the spec plus a few scalar
fields (`group_var`, `block_vars`, `time_var`, `ordering_required`,
`recommended_resampling`), together with provenance the adapter can inspect
to choose a strategy (`constraint_mode`, `constraint_strategy`).

You do not always have to write this glue yourself. The reference downstream
consumer, [**bioLeak**](https://github.com/selcukorkmaz/bioLeak), takes a
`split_spec` directly — `bioLeak::as_leaksplits(spec, data, outcome)` builds an
executable, leakage-audited split plan from it. This cookbook is for the other
case: when you want to feed a `split_spec` into a different engine, or
understand exactly what a consumer has to honor. It shows three small,
self-contained adapters that turn a `split_spec` into something a downstream
workflow can use:

1. A **base-R adapter** that returns a list of `(train, test)` row-index
   pairs — runnable here, no extra dependencies.
2. An **`rsample::group_vfold_cv()`** adapter for grouped cross-validation
   keyed to `group_id`.
3. An **`rsample::rolling_origin()`** adapter for ordered evaluation keyed
   to `order_rank`.

Adapters 2 and 3 show idiomatic glue but are not evaluated in this
vignette so that `splitGraph` does not pick up `rsample` as a build-time
dependency.

The same pattern works for any other resampling library you happen to use.

## Build a split_spec to work with

```{r build-spec}
meta <- data.frame(
  sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
  batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
  timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
  time_index   = c(0, 1, 0, 1, 0, 1),
  outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl"),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "cookbook")
subject_constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(subject_constraint, graph = g)
spec
```

The `sample_data` table is the contract:

```{r show-sample-data}
as.data.frame(spec)[, c("sample_id", "group_id", "batch_group", "order_rank")]
```

## Adapter 1 — base R: leave-one-group-out folds

This is the simplest meaningful adapter. It groups by whatever
`split_spec$group_var` says is the split unit, and returns one held-out
group per fold.

```{r adapter-base-r}
logo_folds <- function(spec, observation_data, sample_id_col = "sample_id") {
  stopifnot(inherits(spec, "split_spec"))
  if (!sample_id_col %in% names(observation_data)) {
    stop("`observation_data` must contain a `", sample_id_col, "` column.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$group_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )
  joined$.row <- seq_len(nrow(joined))
  groups <- split(joined$.row, joined[[spec$group_var]])

  lapply(names(groups), function(g) {
    list(
      group   = g,
      train   = unlist(groups[setdiff(names(groups), g)], use.names = FALSE),
      assess  = groups[[g]]
    )
  })
}

# Pretend we have an observation frame keyed by sample_id.
obs <- data.frame(
  sample_id = meta$sample_id,
  x = rnorm(nrow(meta)),
  y = rbinom(nrow(meta), 1, 0.5)
)

folds <- logo_folds(spec, obs)
length(folds)
folds[[1]]
```

That is the entire downstream contract: take `spec`, take an observation
frame, return train/assess index lists. Anything more complicated is
specific to a resampling library.

### Honoring block variables

`group_var` is the primary split unit, but `split_spec` also advertises coarser
`block_vars` — dependency axes that should ideally not straddle a fold even when
they are not the grouping unit. They are per-sample columns aligned to
`sample_id`, so an adapter reads them exactly like `group_var`:

```{r block-vars}
spec$block_vars
head(spec$sample_data[, c("sample_id", spec$group_var, spec$block_vars)])
```

A block-aware adapter can pass these to a resampler's blocking/strata argument,
or simply audit its folds. Here we check whether any batch straddles the
train/assess boundary — a leak a subject-only split does not prevent, and
exactly what carrying `batch_group` on the spec lets a consumer catch:

```{r block-audit}
block <- spec$block_vars[[1]]
block_of <- setNames(spec$sample_data[[block]], spec$sample_data$sample_id)

do.call(rbind, lapply(folds, function(f) {
  data.frame(
    held_out_group    = f$group,
    straddling_batches = paste(
      intersect(block_of[obs$sample_id[f$train]],
                block_of[obs$sample_id[f$assess]]),
      collapse = ", "
    )
  )
}))
```

Every batch appears on both sides, because grouping by subject does not also
block by batch. Whether that matters is a scientific decision — the point is
that the spec carries enough information for the adapter to make it.

## Adapter 2 — `rsample::group_vfold_cv()`

Grouped CV keyed to `group_id`. The downstream package would typically
ship something like this; the adapter is short enough that you can paste
it into your own analysis script.

```{r adapter-rsample-group, eval=FALSE}
spec_to_group_vfold <- function(spec, observation_data,
                                v = NULL,
                                sample_id_col = "sample_id") {
  stopifnot(inherits(spec, "split_spec"))
  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Install rsample to use this adapter.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$group_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )

  n_groups <- length(unique(joined[[spec$group_var]]))
  if (is.null(v)) v <- n_groups

  rsample::group_vfold_cv(
    data  = joined,
    group = !!spec$group_var,
    v     = v
  )
}
```

`v = NULL` (the default above) gives leave-one-group-out, which is the
right default when `splitGraph` has already grouped samples by their
deepest leakage-relevant unit (e.g. subject). Pick a smaller `v` for
k-fold-style grouped CV.

## Adapter 3 — `rsample::rolling_origin()`

When `spec$ordering_required` is `TRUE` (or `spec$time_var` is set), the
right downstream object is an ordered split rather than a grouped one.

```{r adapter-rsample-rolling, eval=FALSE}
spec_to_rolling_origin <- function(spec, observation_data,
                                   sample_id_col = "sample_id",
                                   initial = NULL,
                                   assess = 1L) {
  stopifnot(inherits(spec, "split_spec"))
  if (is.null(spec$time_var)) {
    stop("This split_spec has no `time_var`; ordered evaluation is not available.")
  }
  if (!requireNamespace("rsample", quietly = TRUE)) {
    stop("Install rsample to use this adapter.")
  }

  joined <- merge(
    observation_data,
    spec$sample_data[, c("sample_id", spec$time_var)],
    by.x = sample_id_col, by.y = "sample_id", sort = FALSE
  )
  ordered <- joined[order(joined[[spec$time_var]]), , drop = FALSE]

  if (is.null(initial)) initial <- max(1L, floor(nrow(ordered) * 0.6))
  rsample::rolling_origin(ordered, initial = initial, assess = assess)
}
```

The key idea: `splitGraph` puts ordering information on the spec; the
adapter is just a thin shim that consumes it.

## Going across language boundaries via JSON

If the downstream consumer is not in R, write the spec to JSON and let the
consumer interpret it. The on-disk format is a formal, versioned contract: it
has a JSON Schema (Draft 2020-12) shipped in `inst/schema/`, each file names it
via a `$schema` key, and `validate_split_spec_json()` checks a file against it
before you consume it.

```{r serialize, eval = requireNamespace("jsonlite", quietly = TRUE)}
tmp <- tempfile(fileext = ".json")
write_split_spec(spec, tmp)

# The file opens with its $schema reference and schema_version.
cat(readLines(tmp, n = 5), sep = "\n")

# Validate the file against the shipped JSON Schema, then read it back exactly.
validate_split_spec_json(tmp)$valid
spec2 <- read_split_spec(tmp)
identical(spec$sample_data$group_id, spec2$sample_data$group_id)

unlink(tmp)
```

You do not have to write a JSON parser to consume this from Python: the package
ships a pure-Python reference reader (`inst/python/splitspec`) that recovers the
same grouping and ordering and drives scikit-learn `GroupKFold` /
`TimeSeriesSplit`. The *cross-language-handoff* vignette walks the full
R → JSON → Python → scikit-learn path:

```{r xlang-pointer, eval = FALSE}
vignette("cross-language-handoff", package = "splitGraph")
```

The same read/write pair exists for `dependency_graph`
(`write_dependency_graph()` / `read_dependency_graph()`, validated with
`validate_graph_json()`). Both formats are documented under `?write_split_spec`
and `?write_dependency_graph`. Because `schema_version` follows a documented
major-compatibility policy, a file written by an older splitGraph still loads;
`migrate_split_spec_json()` upgrades it to the current version in place.

## Letting the spec pick the adapter

The three adapters above cover different shapes of split. You do not have to
choose between them by hand: `split_spec` carries `recommended_resampling`, so a
single dispatcher can route each spec to the right one. This makes a pipeline
that handles subject, batch, time, and composite specs uniformly.

```{r dispatch}
recommend_adapter <- function(spec) {
  switch(
    spec$recommended_resampling,
    grouped_cv          = "group_vfold_cv (group = group_id)",
    blocked_cv          = "group_vfold_cv (group = group_id)",
    custom_grouped_cv   = "group_vfold_cv (group = group_id)",
    leave_one_group_out = "leave-one-group-out over group_id",
    ordered_split       = "rolling_origin (order by order_rank)",
    "group_vfold_cv (default)"
  )
}

# The subject spec recommends grouped CV; a time-mode spec recommends ordering.
recommend_adapter(spec)
time_spec <- as_split_spec(derive_split_constraints(g, mode = "time"), graph = g)
recommend_adapter(time_spec)
```

`recommended_resampling` is only a hint — your adapter is free to override it —
but it lets one entry point serve every constraint mode without inspecting the
graph.

## When you need a custom adapter

The only assumptions an adapter has to honor:

- `split_spec$sample_data` is keyed by `sample_id` (character).
- `split_spec$group_var` is the column that holds the splitting unit.
- `split_spec$block_vars` are present-but-coarser blocking columns. Depending on
  the graph these include `batch_group`, `study_group`, `site_group`,
  `region_group`, `platform_group`, and `assay_group` — an adapter can block or
  stratify on any that are present.
- `split_spec$time_var`, when non-`NULL`, defines the ordering.
- `split_spec$recommended_resampling` is a hint, not a contract — your
  adapter is free to ignore it. `constraint_mode` / `constraint_strategy` are
  available if you want to branch (e.g. treat `"time"` specs as ordered).

That is the whole interface, and it is stable: `bioLeak::as_leaksplits()`
consumes exactly these fields, and a contract test in splitGraph pins the seam
so it cannot drift. As long as those fields are honored, anything is a valid
downstream consumer.

---
title: "splitGraph: From Metadata to Leakage-Aware Split Design"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{splitGraph: From Metadata to Leakage-Aware Split Design}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE,
  eval = TRUE
)

package_root <- if (file.exists("../DESCRIPTION")) ".." else "."
if (requireNamespace("pkgload", quietly = TRUE) &&
    file.exists(file.path(package_root, "DESCRIPTION"))) {
  pkgload::load_all(package_root, export_all = FALSE, helpers = FALSE, quiet = TRUE)
} else {
  library(splitGraph)
}

or_empty <- function(x) {
  if (is.null(x)) character() else x
}
```

## Why `splitGraph` exists

Leakage in biomedical evaluation workflows often comes from dataset structure rather
than from an obvious coding mistake. Two samples may look independent in a
model matrix while still sharing the same subject, batch, study, timepoint, or
feature provenance. If those relationships stay implicit, train/test
separation can look correct while violating the scientific separation you
actually intended.

`splitGraph` exists to make those relationships explicit before evaluation. It
turns metadata into a typed dependency graph that can be:

- validated for structural and leakage-relevant problems
- queried to inspect hidden overlap and provenance
- converted into deterministic split constraints
- translated into a stable, tool-agnostic split specification through the
  `split_spec` class and `as_split_spec()` / `validate_split_spec()` API
- serialized to a schema-versioned JSON that any language can consume (an R
  session, or the shipped Python reader driving scikit-learn)

The package is intentionally narrow. It does not fit models, run preprocessing
pipelines, or generate resamples by itself. Its job is to represent dependency
structure clearly enough that downstream evaluation can be trustworthy.

## A realistic toy dataset

The example below includes exactly the kinds of relationships that usually
matter for leakage-aware evaluation:

- repeated subjects (`P1` and `P2`)
- reused batch (`B1`)
- one subject (`P2`) appearing across studies
- explicit time ordering with one sample missing time metadata
- a feature set derived at the full-dataset scope

```{r metadata}
meta <- data.frame(
  sample_id = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id = c("P1", "P1", "P2", "P3", "P4", "P2"),
  batch_id = c("B1", "B2", "B1", "B3", NA, "B1"),
  study_id = c("ST1", "ST1", "ST1", "ST2", "ST3", "ST2"),
  timepoint_id = c("T0", "T1", "T0", "T2", NA, "T1"),
  assay_id = c("RNAseq", "RNAseq", "RNAseq", "RNAseq", "Proteomics", "RNAseq"),
  featureset_id = c("FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_GLOBAL", "FS_PROT", "FS_GLOBAL"),
  outcome_id = c("O_case", "O_case", "O_ctrl", "O_case", "O_ctrl", "O_ctrl"),
  stringsAsFactors = FALSE
)

meta
```

This is still a small example, but it already contains enough structure to
make naive random splitting risky.

## Fast path: `graph_from_metadata()`

When your metadata already uses the canonical column names (`sample_id`,
`subject_id`, `batch_id`, `study_id`, `timepoint_id`, `time_index`,
`assay_id`, `featureset_id`, `site_id`, `region_id`, `platform_id`,
`outcome_id` / `outcome_value`), `graph_from_metadata()` does ingestion, typed
node construction, canonical edge construction, and optional
`timepoint_precedes` derivation in a single call. Any canonical column that is
absent is simply skipped, so you only supply the structure you have:

```{r fast-path}
quick_graph <- graph_from_metadata(
  data.frame(
    sample_id    = c("S1", "S2", "S3", "S4", "S5", "S6"),
    subject_id   = c("P1", "P1", "P2", "P2", "P3", "P3"),
    batch_id     = c("B1", "B2", "B1", "B2", "B1", "B2"),
    timepoint_id = c("T0", "T1", "T0", "T1", "T0", "T1"),
    time_index   = c(0, 1, 0, 1, 0, 1),
    outcome_id   = c("ctrl", "case", "ctrl", "case", "case", "ctrl")
  ),
  graph_name = "quick_demo"
)

quick_graph
```

If you pass `outcome_value` as a numeric column (e.g. `0`/`1`), the package
will create one `Outcome` node per distinct numeric value (`outcome:0`,
`outcome:1`) and emit a warning. For most workflows you want a character
class label, so prefer `outcome_id`.

The rest of this vignette uses the explicit constructor path because it lets
us show node attributes (`time_index`, `visit_label`, `platform`,
`derivation_scope`) and edges that `graph_from_metadata()` does not build
for you on its own — specifically `featureset_generated_from_study` and
`featureset_generated_from_batch`. (`subject_has_outcome` is the one
non-default outcome edge `graph_from_metadata()` will build for you when
you pass `outcome_scope = "subject"`.) Use `graph_from_metadata()` when
the canonical columns are enough; use the explicit path when you need
custom attributes or feature-provenance edges.

## Ingest metadata and build typed nodes and edges

The first step is to standardize metadata and then turn each entity type into
canonical graph nodes. Sample-level relations become typed edges.

```{r construction}
meta <- ingest_metadata(meta, dataset_name = "VignetteDemo")

sample_nodes <- create_nodes(meta, type = "Sample", id_col = "sample_id")
subject_nodes <- create_nodes(meta, type = "Subject", id_col = "subject_id")
batch_nodes <- create_nodes(meta, type = "Batch", id_col = "batch_id")
study_nodes <- create_nodes(meta, type = "Study", id_col = "study_id")

time_nodes <- create_nodes(
  data.frame(
    timepoint_id = c("T0", "T1", "T2"),
    time_index = c(0L, 1L, 2L),
    visit_label = c("baseline", "follow_up", "late_follow_up"),
    stringsAsFactors = FALSE
  ),
  type = "Timepoint",
  id_col = "timepoint_id",
  attr_cols = c("time_index", "visit_label")
)

assay_nodes <- create_nodes(
  data.frame(
    assay_id = c("RNAseq", "Proteomics"),
    modality = c("transcriptomics", "proteomics"),
    platform = c("NovaSeq", "Orbitrap"),
    stringsAsFactors = FALSE
  ),
  type = "Assay",
  id_col = "assay_id",
  attr_cols = c("modality", "platform")
)

featureset_nodes <- create_nodes(
  data.frame(
    featureset_id = c("FS_GLOBAL", "FS_PROT"),
    featureset_name = c("global_rna_signature", "proteomics_panel"),
    derivation_scope = c("per_dataset", "external"),
    feature_count = c(500L, 80L),
    stringsAsFactors = FALSE
  ),
  type = "FeatureSet",
  id_col = "featureset_id",
  attr_cols = c("featureset_name", "derivation_scope", "feature_count")
)

outcome_nodes <- create_nodes(
  data.frame(
    outcome_id = c("O_case", "O_ctrl"),
    outcome_name = c("response", "response"),
    outcome_type = c("binary", "binary"),
    observation_level = c("subject", "subject"),
    stringsAsFactors = FALSE
  ),
  type = "Outcome",
  id_col = "outcome_id",
  attr_cols = c("outcome_name", "outcome_type", "observation_level")
)

subject_edges <- create_edges(
  meta, "sample_id", "subject_id",
  "Sample", "Subject", "sample_belongs_to_subject"
)

batch_edges <- create_edges(
  meta, "sample_id", "batch_id",
  "Sample", "Batch", "sample_processed_in_batch",
  allow_missing = TRUE
)

study_edges <- create_edges(
  meta, "sample_id", "study_id",
  "Sample", "Study", "sample_from_study"
)

time_edges <- create_edges(
  meta, "sample_id", "timepoint_id",
  "Sample", "Timepoint", "sample_collected_at_timepoint",
  allow_missing = TRUE
)

assay_edges <- create_edges(
  meta, "sample_id", "assay_id",
  "Sample", "Assay", "sample_measured_by_assay"
)

featureset_edges <- create_edges(
  meta, "sample_id", "featureset_id",
  "Sample", "FeatureSet", "sample_uses_featureset"
)

outcome_edges <- create_edges(
  data.frame(
    subject_id = c("P1", "P2", "P3", "P4"),
    outcome_id = c("O_case", "O_ctrl", "O_case", "O_ctrl"),
    stringsAsFactors = FALSE
  ),
  "subject_id", "outcome_id",
  "Subject", "Outcome", "subject_has_outcome"
)

precedence_edges <- create_edges(
  data.frame(
    from_timepoint = c("T0", "T1"),
    to_timepoint = c("T1", "T2"),
    stringsAsFactors = FALSE
  ),
  "from_timepoint", "to_timepoint",
  "Timepoint", "Timepoint", "timepoint_precedes"
)

featureset_from_study <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    study_id = "ST1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "study_id",
  "FeatureSet", "Study", "featureset_generated_from_study"
)

featureset_from_batch <- create_edges(
  data.frame(
    featureset_id = "FS_GLOBAL",
    batch_id = "B1",
    stringsAsFactors = FALSE
  ),
  "featureset_id", "batch_id",
  "FeatureSet", "Batch", "featureset_generated_from_batch"
)
```

The node and edge tables are canonical and typed. The package assigns globally
unique node IDs such as `sample:S1` and `subject:P1`, so different entity
types cannot collide accidentally.

```{r construction-output}
sample_nodes
as.data.frame(sample_nodes)[, c("node_id", "node_type", "node_key", "label")]

edge_preview <- do.call(rbind, lapply(
  list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  as.data.frame
))

edge_preview[, c("from", "to", "edge_type")]
```

The node table shows the canonical sample IDs that everything else refers to.
The edge table shows the package's central design choice: dependency structure
is explicit, typed, and inspectable.

## Assemble the dependency graph

```{r graph}
graph <- build_dependency_graph(
  nodes = list(
    sample_nodes, subject_nodes, batch_nodes, study_nodes,
    time_nodes, assay_nodes, featureset_nodes, outcome_nodes
  ),
  edges = list(
    subject_edges, batch_edges, study_edges, time_edges,
    assay_edges, featureset_edges, outcome_edges,
    precedence_edges, featureset_from_study, featureset_from_batch
  ),
  graph_name = "vignette_graph",
  dataset_name = attr(meta, "dataset_name")
)

graph
summary(graph)
```

At this point the package has a single `dependency_graph` object with both
tabular and `igraph` representations behind it. The summary is useful because
it tells you exactly which entity types and relation types are present before
you derive any split rules.

### Visualize the typed structure

`plot()` renders the graph with a typed, layered layout: `Sample` on top,
peer dependencies (`Subject`, `Batch`, `Study`, `Timepoint`) in the middle
band, `Assay`/`FeatureSet` below that, and `Outcome` at the bottom. Node
colors are keyed to type and an auto-generated legend is drawn by default.

```{r plot, fig.width = 7, fig.height = 5}
plot(graph)
```

Useful options:

```{r plot-options, eval = FALSE}
plot(graph, layout = "sugiyama")         # alternative hierarchical layout
plot(graph, show_labels = FALSE)         # hide labels on dense graphs
plot(graph, legend = FALSE)              # suppress the legend
plot(graph, legend_position = "bottomright")
plot(graph, node_colors = c(Sample = "#000000"))
```

## Validate before you split

Validation is where `splitGraph` starts paying off. The graph below is
structurally valid, but it still carries leakage-relevant warnings and
advisories.

```{r validation}
validation <- validate_graph(graph)

validation
as.data.frame(validation)[, c("level", "severity", "code", "message")]
```

That output is the core value proposition of the package in one place:

- repeated subjects are surfaced explicitly
- cross-study subject overlap is surfaced explicitly
- full-dataset feature provenance is surfaced explicitly
- heavy batch reuse is surfaced explicitly

`valid = TRUE` here means the graph has no errors. It does not mean the dataset
is free of leakage risk. Warnings and advisories still matter.

The package is also intentionally strict about silent failure. If you ask for a
subset of samples and some of them do not resolve, it errors instead of
dropping them.

```{r strictness}
tryCatch(
  derive_split_constraints(graph, mode = "subject", samples = c("S1", "BAD")),
  error = function(e) e$message
)
```

That behavior is important in practice because quietly omitting samples would
change the truth of the split problem.

### When you really do need to relax a check

Some leakage-relevant rules have legitimate exceptions. For example, in
some pooled designs a single biological "sample" is intentionally paired
across two subjects. The semantic validator flags this as an error by
default, and `derive_split_constraints(mode = "subject")` refuses to
choose a subject for you — both consistent with the "no silent guessing"
stance.

To make the contrast concrete, build a tiny graph in which sample `S1`
is linked to two subjects:

```{r overrides-fixture}
multi_nodes <- graph_node_set(data.frame(
  node_id   = c("sample:S1", "subject:P1", "subject:P2"),
  node_type = c("Sample", "Subject", "Subject"),
  node_key  = c("S1", "P1", "P2"),
  label     = c("S1", "P1", "P2"),
  attrs     = I(list(list(), list(), list())),
  stringsAsFactors = FALSE
))

multi_edges <- graph_edge_set(data.frame(
  edge_id   = c("sample_belongs_to_subject:1", "sample_belongs_to_subject:2"),
  from      = c("sample:S1", "sample:S1"),
  to        = c("subject:P1", "subject:P2"),
  edge_type = c("sample_belongs_to_subject", "sample_belongs_to_subject"),
  attrs     = I(list(list(), list())),
  stringsAsFactors = FALSE
))

multi_graph <- dependency_graph(nodes = multi_nodes, edges = multi_edges)
```

Default behavior — the validator surfaces the multi-subject sample as an
error, and constraint derivation refuses to pick a subject:

```{r overrides-default}
default_report <- validate_graph(multi_graph)
default_report$valid
default_report$issues[, c("severity", "code", "message")]

tryCatch(
  derive_split_constraints(multi_graph, mode = "subject"),
  error = function(e) e$message
)
```

When the ambiguity is intended, opt in explicitly via
`validation_overrides`. The same key
(`allow_multi_subject_samples = TRUE`) is honored by both the validator
and the constraint deriver:

```{r overrides-on}
# Validator: pass.
permissive_report <- validate_graph(
  multi_graph,
  validation_overrides = list(allow_multi_subject_samples = TRUE)
)
permissive_report$valid

# Constraint derivation: pick the first listed subject and record the
# ambiguity in metadata$warnings instead of erroring.
multi_graph$metadata$validation_overrides <-
  list(allow_multi_subject_samples = TRUE)
relaxed_constraint <- derive_split_constraints(multi_graph, mode = "subject")
relaxed_constraint$sample_map[, c("sample_id", "group_id")]
relaxed_constraint$metadata$warnings
```

The override is documented under `?validate_graph`. Use it sparingly and
only when the relaxation matches your scientific intent — the message in
`metadata$warnings` is the audit trail that the choice was tolerated, not
hidden.

## Query the graph to inspect hidden structure

You can inspect local provenance, trace paths, and project direct sample
dependencies.

```{r neighbors-and-paths}
neighbors_s1 <- query_neighbors(graph, node_ids = "sample:S1", direction = "out")
neighbors_s1
as.data.frame(neighbors_s1)[, c("seed_node_id", "node_id", "node_type", "edge_type")]

subject_outcome_path <- query_shortest_paths(
  graph,
  from = "sample:S1",
  to = "outcome:O_case",
  edge_types = c("sample_belongs_to_subject", "subject_has_outcome")
)

subject_outcome_path
as.data.frame(subject_outcome_path)
```

The first query shows everything the graph knows directly about `S1`. The
second shows that `S1` reaches the subject-level outcome through its subject
node, which is exactly the kind of relationship that would stay implicit in a
plain metadata table.

```{r projected-dependencies}
shared_dependencies <- detect_shared_dependencies(
  graph,
  via = c("Subject", "Batch", "FeatureSet")
)

as.data.frame(shared_dependencies)[, c(
  "sample_id_1", "sample_id_2", "shared_node_type", "shared_node_id", "edge_type"
)]

dependency_components <- detect_dependency_components(
  graph,
  via = c("Subject", "Batch")
)

as.data.frame(dependency_components)
```

These projected queries are useful because they answer the splitting question
directly. They tell you which samples should be treated as structurally linked,
not just which metadata columns happen to match.

## Derive split constraints from the graph

`splitGraph` derives *direct* constraints — one grouping node per sample — for
`subject`, `batch`, `study`, `time`, `site`, `region`, `platform`, and `assay`,
plus *composite* constraints that combine several dependency sources, and
*pairwise* constraints (`relatedness`, `spatial`) built from thresholded
similarity. This section demonstrates the core four (`subject`, `batch`,
`study`, `time`) and both composite strategies; the cluster relations (`site`,
`region`, `platform`, `assay`) and the pairwise relations are covered in their
own sections below and, in more depth, in the *modeling-structure* vignette.

```{r constraints}
subject_constraint <- derive_split_constraints(graph, mode = "subject")
batch_constraint <- derive_split_constraints(graph, mode = "batch")
study_constraint <- derive_split_constraints(graph, mode = "study")
time_constraint <- derive_split_constraints(graph, mode = "time")

strict_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "strict",
  via = c("Subject", "Batch")
)

rule_based_constraint <- derive_split_constraints(
  graph,
  mode = "composite",
  strategy = "rule_based",
  priority = c("batch", "study", "subject", "time")
)

constraint_overview <- do.call(rbind, lapply(
  list(
    subject = subject_constraint,
    batch = batch_constraint,
    study = study_constraint,
    time = time_constraint,
    composite_strict = strict_constraint,
    composite_rule = rule_based_constraint
  ),
  function(x) {
    data.frame(
      strategy = x$strategy,
      groups = length(unique(x$sample_map$group_id)),
      warnings = if (is.null(x$metadata$warnings)) 0L else length(x$metadata$warnings),
      stringsAsFactors = FALSE
    )
  }
))

constraint_overview <- cbind(constraint = row.names(constraint_overview), constraint_overview)
row.names(constraint_overview) <- NULL

constraint_overview
```

That summary already shows why the package is useful: different notions of
dependency produce different splitting units.

### Batch constraints

```{r batch-constraint}
batch_constraint
as.data.frame(batch_constraint)[, c("sample_id", "group_id", "group_label", "explanation")]
```

Batch grouping keeps all `B1` samples together and preserves `S5` as an
explicit singleton because it has no batch assignment. Missing structure is not
hidden.

### Time constraints

```{r time-constraint}
time_constraint
as.data.frame(time_constraint)[, c("sample_id", "group_id", "timepoint_id", "order_rank")]
```

Time grouping adds `order_rank`, which is the field downstream tooling actually
needs for ordered evaluation. The missing timepoint on `S5` stays visible as
`NA`, so ordering is partial rather than pretended.

### Composite constraints

```{r composite-constraints}
strict_constraint
as.data.frame(strict_constraint)[, c("sample_id", "group_id", "constraint_type")]

rule_based_constraint
as.data.frame(rule_based_constraint)[, c("sample_id", "group_id", "constraint_type", "group_label")]
```

The strict composite constraint uses transitive closure: `S1`, `S2`, `S3`, and
`S6` end up in the same group because subject and batch links connect them into
one dependency component. The rule-based composite constraint is different: it
uses the highest-priority available dependency per sample, so `S5` falls back
to study-level grouping instead of becoming a composite component.

## Cluster-style relations: site, region, platform, assay

Beyond subject/batch/study/time, several other metadata columns define
*cluster-style* leakage axes — categorical groupings a model can memorize
instead of generalizing across. `splitGraph` models each as a first-class node
type with its own auto-detected column, validation rule, and constraint mode:

| Relation | Column | Edge | Mode |
|---|---|---|---|
| Collection site / center | `site_id` | `sample_collected_at_site` | `"site"` |
| Tissue / anatomical region | `region_id` | `sample_located_in_region` | `"region"` |
| Sequencing / measurement platform | `platform_id` | `sample_run_on_platform` | `"platform"` |
| Assay / modality | `assay_id` | `sample_measured_by_assay` | `"assay"` |

They all behave identically: `graph_from_metadata()` auto-detects the column,
`validate_graph()` flags samples assigned to more than one target, the mode
groups samples so no cluster straddles a split, and `as_split_spec()` carries
the assignment as a blocking annotation (`site_group`, `region_group`,
`platform_group`, `assay_group`).

### Site

In multi-site studies, the collection site is a common leakage axis: a model can
"recognize" a site rather than generalize across sites. `splitGraph` models the
site as a first-class `Site` node connected by `sample_collected_at_site` edges.
`graph_from_metadata()` auto-detects a `site_id` column, so no extra wiring is
needed.

```{r site-structure}
site_meta <- data.frame(
  sample_id  = c("S1", "S2", "S3", "S4", "S5", "S6"),
  subject_id = c("P1", "P1", "P2", "P2", "P3", "P3"),
  site_id    = c("NYC", "NYC", "BOS", "BOS", "NYC", "BOS"),
  stringsAsFactors = FALSE
)

site_graph <- graph_from_metadata(site_meta, graph_name = "multi-site")

# Group samples so that no collection site straddles a train/test split.
site_constraint <- derive_split_constraints(site_graph, mode = "site")
as.data.frame(site_constraint)[, c("sample_id", "group_id", "group_label")]
```

`mode = "site"` keeps every sample from a given site in the same group. The site
assignment is also carried into the `split_spec` as a blocking annotation
(`site_group`), so a downstream consumer can block on site even when the primary
grouping is something else (e.g. subject):

```{r site-spec}
subject_then_block_by_site <- as_split_spec(
  derive_split_constraints(site_graph, mode = "subject"),
  graph = site_graph
)
subject_then_block_by_site$block_vars
head(subject_then_block_by_site$sample_data[, c("sample_id", "group_id", "site_group")])
```

Samples assigned to more than one site are rejected rather than silently
resolved, by both `validate_graph()` and `derive_split_constraints(mode = "site")`.

### Region

The `Region` relation works the same way for a categorical tissue or anatomical
region (`region_id` column, `sample_located_in_region` edge,
`derive_split_constraints(mode = "region")`), so samples from the same region
stay together across a split.

```{r region-structure}
region_meta <- data.frame(
  sample_id = c("S1", "S2", "S3", "S4"),
  region_id = c("cortex", "cortex", "hippocampus", "hippocampus"),
  stringsAsFactors = FALSE
)
region_graph <- graph_from_metadata(region_meta, graph_name = "regions")
as.data.frame(derive_split_constraints(region_graph, mode = "region"))[
  , c("sample_id", "group_id", "group_label")
]
```

### Platform and assay

Technical measurement structure is the same story. `platform_id` (the
sequencing instrument or measurement platform) and `assay_id` (the assay or
modality) are both auto-detected, and `mode = "platform"` / `mode = "assay"`
group samples so a whole platform or assay never straddles a split — useful when
a batch effect tracks the instrument rather than the run.

```{r platform-assay}
tech_meta <- data.frame(
  sample_id   = c("S1", "S2", "S3", "S4"),
  platform_id = c("illumina", "illumina", "nanopore", "nanopore"),
  assay_id    = c("rnaseq", "rnaseq", "wgs", "wgs"),
  stringsAsFactors = FALSE
)
tech_graph <- graph_from_metadata(tech_meta, graph_name = "tech")

as.data.frame(derive_split_constraints(tech_graph, mode = "platform"))[
  , c("sample_id", "group_id", "group_label")
]
as.data.frame(derive_split_constraints(tech_graph, mode = "assay"))[
  , c("sample_id", "group_id", "group_label")
]
```

These new types are first-class in the typed layout too. Building a graph that
carries several of them shows how each gets its own colour and layer band — the
same visual vocabulary as the core types, so a mixed graph stays readable:

```{r cluster-plot, fig.width = 7, fig.height = 5}
mixed_meta <- data.frame(
  sample_id   = c("S1", "S2", "S3", "S4"),
  subject_id  = c("P1", "P1", "P2", "P2"),
  site_id     = c("NYC", "NYC", "BOS", "BOS"),
  platform_id = c("illumina", "illumina", "nanopore", "nanopore"),
  stringsAsFactors = FALSE
)
plot(graph_from_metadata(mixed_meta, graph_name = "mixed_structure"))
```

### Pairwise relations: relatedness and spatial proximity

Not every leakage source is a clean categorical group. Genetic relatedness and
spatial proximity are *pairwise and continuous*: they link individual pairs by a
similarity score. `splitGraph` models these as thresholded, undirected edges
built with `relatedness_edges_from_kinship()` and `spatial_edges_from_coords()`,
then `mode = "relatedness"` / `mode = "spatial"` form groups by transitive
closure over the surviving edges — so a chain of individually near neighbours
still lands in one group, a grouping a single column cannot express.

The example below keeps subject pairs whose kinship is at least `0.1`. `P1`–`P2`
and `P2`–`P3` clear the threshold, so those three subjects collapse into one
group by transitive closure; the unrelated `P4` stays on its own:

```{r relatedness-demo}
kin <- data.frame(
  id1     = c("P1", "P2", "P1"),
  id2     = c("P2", "P3", "P4"),
  kinship = c(0.25, 0.20, 0.02),   # P1-P4 is below the 0.1 threshold
  stringsAsFactors = FALSE
)
rel_edges <- relatedness_edges_from_kinship(kin, threshold = 0.1)

rel_meta <- data.frame(
  sample_id  = paste0("S", 1:4),
  subject_id = c("P1", "P2", "P3", "P4"),
  stringsAsFactors = FALSE
)
rel_graph <- build_dependency_graph(
  nodes = list(
    create_nodes(rel_meta, "Sample", "sample_id"),
    create_nodes(rel_meta, "Subject", "subject_id")
  ),
  edges = list(
    create_edges(rel_meta, "sample_id", "subject_id",
                 "Sample", "Subject", "sample_belongs_to_subject"),
    rel_edges
  )
)

grouping_vector(derive_split_constraints(rel_graph, mode = "relatedness"))
```

The `spatial` mode works identically over sample coordinates
(`spatial_edges_from_coords(coords, radius)`). Both relations — with the full
scikit-learn handoff — are covered in depth in the *modeling-structure*
vignette:

```{r modeling-pointer, eval = FALSE}
vignette("modeling-structure", package = "splitGraph")
```

## Time ordering can come from precedence edges alone

If explicit `time_index` metadata are unavailable, `splitGraph` can still infer
time order from `timepoint_precedes` edges.

```{r precedence-only}
precedence_meta <- data.frame(
  sample_id = c("S1", "S2", "S3"),
  subject_id = c("P1", "P1", "P2"),
  study_id = c("ST1", "ST1", "ST2"),
  timepoint_id = c("T0", "T1", "T2"),
  stringsAsFactors = FALSE
)

precedence_graph <- build_dependency_graph(
  nodes = list(
    create_nodes(precedence_meta, type = "Sample", id_col = "sample_id"),
    create_nodes(precedence_meta, type = "Subject", id_col = "subject_id"),
    create_nodes(precedence_meta, type = "Study", id_col = "study_id"),
    create_nodes(
      data.frame(timepoint_id = c("T0", "T1", "T2"), stringsAsFactors = FALSE),
      type = "Timepoint",
      id_col = "timepoint_id"
    )
  ),
  edges = list(
    create_edges(
      precedence_meta, "sample_id", "subject_id",
      "Sample", "Subject", "sample_belongs_to_subject"
    ),
    create_edges(
      precedence_meta, "sample_id", "study_id",
      "Sample", "Study", "sample_from_study"
    ),
    create_edges(
      precedence_meta, "sample_id", "timepoint_id",
      "Sample", "Timepoint", "sample_collected_at_timepoint"
    ),
    create_edges(
      data.frame(
        from_timepoint = c("T0", "T1"),
        to_timepoint = c("T1", "T2"),
        stringsAsFactors = FALSE
      ),
      "from_timepoint", "to_timepoint",
      "Timepoint", "Timepoint", "timepoint_precedes"
    )
  ),
  graph_name = "precedence_only_graph"
)

precedence_time_constraint <- derive_split_constraints(precedence_graph, mode = "time")

precedence_time_constraint$metadata$time_order_source
as.data.frame(precedence_time_constraint)[, c("sample_id", "timepoint_id", "time_index", "order_rank")]
```

The important detail is that ordering is still derived, but the source is
`timepoint_precedes` rather than `time_index`.

## Translate the constraint into a split specification

The graph-derived constraint is not the end of the workflow. The main handoff
target is a canonical sample-level split specification — the `split_spec`
class. Downstream tools consume it through their own adapters, so
`split_spec` stays tool-agnostic.

```{r split-spec}
split_spec <- as_split_spec(strict_constraint, graph = graph)
split_spec

as.data.frame(split_spec)[, c(
  "sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank"
)]

split_spec_validation <- validate_split_spec(split_spec)
split_spec_validation
as.data.frame(split_spec_validation)
```

This translation step is where the package becomes operational for downstream
evaluation workflows:

- `group_id` carries the split unit
- `batch_group` and `study_group` are available for blocking
- `order_rank` is available for ordered evaluation
- the generated object is validated before handoff

## Summarize the leakage picture in one object

The final helper combines graph validation, constraint diagnostics, and
split-spec readiness into one summary object. Crucially, when you pass the
`constraint` you chose, the summary reports a `severed` column: whether that
constraint **structurally eliminates** each leakage path (`TRUE`), leaves it
open (`FALSE`), or is not applicable (`NA`, e.g. for informational split-spec
rows).

```{r risk-summary}
risk_summary <- summarize_leakage_risks(
  graph,
  constraint = strict_constraint,
  split_spec = split_spec
)

risk_summary
as.data.frame(risk_summary)[, c("source", "severity", "category", "severed", "message")]
```

Read the `severed` column against the constraint you actually chose. Here the
strict composite constraint (`via = c("Subject", "Batch")`) severs the subject,
cross-study, and batch-reuse risks (`TRUE`) because those samples are forced
into the same group — but it does **not** address `missing_time_ordering` or
`shared_featureset_provenance` (`FALSE`), which are orthogonal to a
subject/batch grouping. That distinction is the point: the summary tells you
which of your surfaced risks your split design has actually handled, and which
still need attention (a different mode, a blocking variable, or a data fix) —
so the leakage trade-off is explicit before any model is trained.

## Downstream handoff

`split_spec` is the tool-agnostic handoff artifact. `splitGraph` does not
know about any particular resampling package — downstream consumers provide
their own adapters so `splitGraph` stays neutral. The typical end-to-end
flow is:

1. `graph_from_metadata()` (or the explicit constructor path) → typed
   `dependency_graph`
2. `derive_split_constraints(g, mode = ...)` → `split_constraint`
3. `as_split_spec(constraint, graph = g)` → `split_spec`
4. (optional) `write_split_spec(spec, path)` → JSON, for cross-session or
   cross-language handoff
5. adapter in the downstream package → native resamples

The `sample_data` frame carried by `split_spec` exposes exactly the columns
downstream adapters consume: `sample_id` for joining against the observation
frame, `group_id` for grouped resampling, `batch_group` / `study_group` for
blocking, and `order_rank` for ordered evaluation. Adapters can be built by
any package that wants to consume a `split_spec` — for example, on top of
`rsample::group_vfold_cv()` (grouped CV keyed to `group_id`) or
`rsample::rolling_origin()` (ordered evaluation keyed to `order_rank`).

### Persisting the handoff

If the downstream consumer is in a different R session — or in a different
language entirely — write the spec (and, if useful, the graph) to JSON. The
on-disk format has a formal JSON Schema (Draft 2020-12) shipped in
`inst/schema/`, and each written file references it via a `$schema` key. You can
validate a handoff file against that contract with `validate_split_spec_json()`
before consuming it, and `NA` values round-trip as JSON `null`.

```{r serialize, eval = requireNamespace("jsonlite", quietly = TRUE)}
spec_path <- tempfile(fileext = ".json")
write_split_spec(split_spec, spec_path)

# Validate the file against the shipped JSON Schema.
validate_split_spec_json(spec_path)$valid

# Round-trip it back into R unchanged.
spec_round_trip <- read_split_spec(spec_path)
identical(split_spec$sample_data$group_id, spec_round_trip$sample_data$group_id)

unlink(spec_path)
```

Because the format is a documented, versioned contract, consumers are not
limited to R. The package ships a pure-Python reference reader (`inst/python`)
that recovers the same grouping and ordering and drives scikit-learn resamplers;
the *cross-language-handoff* vignette walks the full R → JSON → Python →
scikit-learn path. For worked R adapters (a base-R leave-one-group-out adapter
and illustrative `rsample` adapters), see the *adapter-cookbook* vignette:

```{r cookbook-pointer, eval = FALSE}
vignette("adapter-cookbook", package = "splitGraph")       # R adapters
vignette("cross-language-handoff", package = "splitGraph") # Python / sklearn
```

## Case studies

The end-to-end workflow above shows the package surface. The case studies below
show how the same graph leads to different evaluation decisions depending on
the scientific question.

### Case study 1: repeated subjects in a longitudinal cohort

Suppose the real question is whether future observations from the same subject
should be held out from training. In this setting, subject reuse and time
ordering both matter, but they solve different problems.

```{r case-study-1}
subject_groups <- grouping_vector(subject_constraint)
time_groups <- time_constraint$sample_map[, c("sample_id", "group_id", "timepoint_id", "order_rank")]

subject_groups
time_groups
```

Interpretation:

- `S1` and `S2` share subject `P1`, so subject-grouped evaluation keeps them
  together.
- `S3` and `S6` share subject `P2`, so they also stay together under a
  subject-based split.
- time grouping adds a different axis: `T0`, `T1`, and `T2` become ordered
  units with explicit `order_rank`.

If the leakage concern is repeated measurements from the same individual, use
the subject constraint. If the evaluation question is prospective prediction,
the time constraint adds the ordering information you need.

### Case study 2: a subject reused across studies

The graph intentionally includes subject `P2` in both `ST1` and `ST2`. A
study-only split would treat those studies as separate units, but the graph
shows that subject overlap breaks the intended independence.

```{r case-study-2}
cross_study_issues <- as.data.frame(validation)[
  as.data.frame(validation)$code == "subject_cross_study_overlap",
  c("severity", "code", "message")
]

p2_shared <- detect_shared_dependencies(
  graph,
  via = "Subject",
  samples = c("S3", "S6")
)

study_only_map <- study_constraint$sample_map[, c("sample_id", "group_id", "group_label")]
strict_map <- strict_constraint$sample_map[, c("sample_id", "group_id", "constraint_type")]

cross_study_issues
as.data.frame(p2_shared)
study_only_map[study_only_map$sample_id %in% c("S3", "S6"), ]
strict_map[strict_map$sample_id %in% c("S3", "S6"), ]
```

Interpretation:

- validation surfaces the cross-study subject overlap directly
- the shared-dependency query confirms that `S3` and `S6` are linked through
  the same subject
- a study-only split would place them in different groups (`ST1` versus `ST2`)
- the strict composite constraint correctly keeps them in the same dependency
  component

This is exactly the kind of failure mode `splitGraph` is designed to expose:
metadata columns suggest a legitimate study split, but graph structure shows
that the split would still leak subject information.

### Case study 3: partially observed technical metadata

Real metadata are rarely complete. Here, `S5` has no batch assignment and no
timepoint assignment. The package does not pretend those fields exist. It keeps
the sample visible and tells you how the split logic handled it.

```{r case-study-3}
batch_missing <- batch_constraint$sample_map[
  batch_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "group_label", "explanation")
]

rule_based_missing <- rule_based_constraint$sample_map[
  rule_based_constraint$sample_map$sample_id == "S5",
  c("sample_id", "group_id", "constraint_type", "group_label", "explanation")
]

split_spec_missing <- as.data.frame(split_spec)[
  as.data.frame(split_spec)$sample_id == "S5",
  c("sample_id", "group_id", "batch_group", "study_group", "timepoint_id", "order_rank")
]

batch_missing
rule_based_missing
split_spec_missing
```

Interpretation:

- batch-based splitting keeps `S5` as an explicit singleton because batch
  metadata are missing
- the rule-based composite strategy falls back to study-level grouping for `S5`
- the translated split specification preserves the missing batch and time
  fields as `NA` rather than silently inventing values

That behavior matters because incomplete metadata are common. `splitGraph`
stays strict about what is known, but still produces a usable, inspectable
split object.

### Case study 4: choosing a defensible split strategy

A typical practical question is not "what can the package compute?" but "which
constraint should I actually use?" The answer depends on which dependency
source is scientifically unacceptable to leak across train and test.

```{r case-study-4}
strategy_summary <- data.frame(
  constraint = c("subject", "batch", "study", "time", "composite_strict", "composite_rule"),
  groups = c(
    length(unique(subject_constraint$sample_map$group_id)),
    length(unique(batch_constraint$sample_map$group_id)),
    length(unique(study_constraint$sample_map$group_id)),
    length(unique(time_constraint$sample_map$group_id)),
    length(unique(strict_constraint$sample_map$group_id)),
    length(unique(rule_based_constraint$sample_map$group_id))
  ),
  warnings = c(
    length(or_empty(subject_constraint$metadata$warnings)),
    length(or_empty(batch_constraint$metadata$warnings)),
    length(or_empty(study_constraint$metadata$warnings)),
    length(or_empty(time_constraint$metadata$warnings)),
    length(or_empty(strict_constraint$metadata$warnings)),
    length(or_empty(rule_based_constraint$metadata$warnings))
  ),
  recommended_resampling = c(
    as_split_spec(subject_constraint, graph = graph)$recommended_resampling,
    as_split_spec(batch_constraint, graph = graph)$recommended_resampling,
    as_split_spec(study_constraint, graph = graph)$recommended_resampling,
    as_split_spec(time_constraint, graph = graph)$recommended_resampling,
    as_split_spec(strict_constraint, graph = graph)$recommended_resampling,
    as_split_spec(rule_based_constraint, graph = graph)$recommended_resampling
  ),
  stringsAsFactors = FALSE
)

strategy_summary
```

Interpretation:

- subject grouping is the right default when repeated individuals are the
  dominant leakage source
- batch grouping is appropriate when technical runs are the main contamination
  risk
- study grouping is useful for cross-study generalization only when no higher
  level dependency crosses study boundaries
- strict composite grouping is the safest choice when multiple dependency
  sources can connect samples transitively
- rule-based composite grouping is a pragmatic fallback when you want a single
  deterministic hierarchy over partially observed metadata

The package does not choose the scientific objective for you. It makes the
trade-off visible and auditable.

## When `splitGraph` is useful

`splitGraph` is a good fit when:

- sample relationships are scientifically meaningful and must influence
  evaluation
- metadata contain repeated subjects, shared batches, multiple studies, or
  temporal structure
- feature provenance or outcome level matters for leakage assessment
- you want deterministic, inspectable split constraints instead of ad hoc
  grouping code

## What `splitGraph` is not for

`splitGraph` is not:

- a general biological network analysis package
- a model training framework
- a resampling engine
- a substitute for downstream performance auditing

Its value is earlier in the workflow: it makes dependency structure explicit so
that the split design itself can be justified.

## Takeaway

If you already know your data have repeated subjects, reused batches, temporal
ordering, or shared feature provenance, then you already have a graph problem
whether you model it explicitly or not. `splitGraph` is useful because it turns
that hidden graph into an object you can validate, query, and convert into a
split design that downstream tooling can trust.

---
title: "Modeling site, platform, relatedness, and spatial structure"
author: "Selçuk Korkmaz"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Modeling site, platform, relatedness, and spatial structure}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)
library(splitGraph)
```

Beyond the classic subject / batch / study / time relations, `splitGraph`
models several further leakage axes, in two families:

- **Cluster-style relations** — collection **site**, tissue / anatomical
  **region**, sequencing **platform**, and **assay** — are categorical
  groupings. Each is auto-detected from a metadata column and handled by its own
  constraint mode, exactly like `subject` or `batch`.
- **Pairwise relations** — genetic **relatedness** and **spatial** proximity —
  are continuous and defined *between pairs*. They are modelled as thresholded
  edges and grouped by transitive closure, a partition a single categorical
  column cannot express.

This vignette builds and groups by each, and shows how the threshold drives the
pairwise grouping.

# Cluster-style relations: site, region, platform, assay

`graph_from_metadata()` auto-detects `site_id`, `region_id`, `platform_id`, and
`assay_id` columns and builds the corresponding typed nodes and edges. Each then
has its own constraint mode. The example below uses site, platform, and assay;
`region` behaves identically (a `region_id` column and `mode = "region"`) and is
omitted only to keep the output short.

```{r cluster}
meta <- data.frame(
  sample_id   = paste0("S", 1:6),
  subject_id  = c("P1", "P1", "P2", "P2", "P3", "P3"),
  site_id     = c("NYC", "NYC", "BOS", "BOS", "NYC", "BOS"),
  platform_id = c("illumina", "illumina", "nanopore", "nanopore", "illumina", "nanopore"),
  assay_id    = c("rnaseq", "rnaseq", "rnaseq", "wgs", "wgs", "wgs"),
  stringsAsFactors = FALSE
)

g <- graph_from_metadata(meta, graph_name = "structure-demo")

grouping_vector(derive_split_constraints(g, mode = "site"))
grouping_vector(derive_split_constraints(g, mode = "platform"))
grouping_vector(derive_split_constraints(g, mode = "assay"))
```

Whatever mode is primary, every detected cluster relation is also carried into
the `split_spec` as a *blocking annotation*, so a downstream consumer can block
on site, platform, or assay even when the split unit is something else — here,
subject:

```{r block-annotations}
spec <- as_split_spec(derive_split_constraints(g, mode = "subject"), graph = g)
spec$block_vars
head(spec$sample_data[, c("sample_id", "group_id",
                          "site_group", "platform_group", "assay_group")])
```

Any of these relations can also participate in a **composite** derivation, where
several dependency sources are combined and each connected component becomes one
group:

```{r composite}
constraint <- derive_split_constraints(
  g, mode = "composite", strategy = "strict",
  via = c("Subject", "Site", "Platform")
)
grouping_vector(constraint)
```

# Pairwise relation: genetic relatedness

Some leakage is pairwise and continuous rather than a clean grouping. Genetic
relatedness is the canonical example: a kinship coefficient — typically from a
tool such as KING or PLINK — links *pairs* of subjects.
`relatedness_edges_from_kinship()` takes such a pair table, keeps pairs at or
above a threshold, and emits `subject_related_to` edges; `mode = "relatedness"`
then groups by transitive closure over those edges (so a chain of related
individuals lands in one group).

```{r relatedness}
# A kinship table over subject pairs (one sample per subject here for clarity).
# P1-P2 and P2-P3 clear the threshold and chain together; P5-P6 form a second
# related pair; P1-P4 is too weak to count.
kin <- data.frame(
  id1     = c("P1", "P2", "P1", "P5"),
  id2     = c("P2", "P3", "P4", "P6"),
  kinship = c(0.25, 0.20, 0.02, 0.30),
  stringsAsFactors = FALSE
)
rel_edges <- relatedness_edges_from_kinship(kin, threshold = 0.1)

meta_r <- data.frame(
  sample_id  = paste0("S", 1:6),
  subject_id = paste0("P", 1:6),
  stringsAsFactors = FALSE
)
samples  <- create_nodes(meta_r, "Sample", "sample_id")
subjects <- create_nodes(meta_r, "Subject", "subject_id")
belongs  <- create_edges(meta_r, "sample_id", "subject_id",
                         "Sample", "Subject", "sample_belongs_to_subject")

g_rel <- build_dependency_graph(list(samples, subjects), list(belongs, rel_edges))

rel_groups <- grouping_vector(derive_split_constraints(g_rel, mode = "relatedness"))
rel_groups
```

The grouping is a transitive closure over the `subject_related_to` edges. The
network below draws those edges between subjects, coloured by the relatedness
group each subject (and therefore its samples) lands in: the P1–P2–P3 chain
becomes one group even though P1 and P3 were never linked directly, P5–P6 form a
second, and the unrelated P4 stands alone.

```{r rel-plot, fig.width = 6.5, fig.height = 4.5}
subject_group <- setNames(rel_groups[meta_r$sample_id], meta_r$subject_id)
kept_pairs <- kin[kin$kinship >= 0.1, c("id1", "id2")]
rel_net <- igraph::graph_from_data_frame(
  kept_pairs, directed = FALSE,
  vertices = data.frame(name = meta_r$subject_id)
)

palette_rel <- c("#4C78A8", "#F58518", "#54A24B", "#B279A2")
set.seed(1)
plot(rel_net,
     vertex.color       = palette_rel[as.integer(factor(subject_group[igraph::V(rel_net)$name]))],
     vertex.size        = 34,
     vertex.label.color = "white",
     vertex.label.font  = 2,
     edge.color         = "grey60",
     edge.width         = 2,
     main               = "Relatedness clusters (kinship >= 0.1)")
```

The threshold is the key knob, and it belongs to the *edge-building* step, not
the grouping. Raising it drops weaker links: at `0.22` the P2–P3 pair (kinship
`0.20`) no longer qualifies, so that chain breaks and P3 splits into its own
group, while the stronger P5–P6 pair is untouched:

```{r rel-threshold}
rel_strict <- relatedness_edges_from_kinship(kin, threshold = 0.22)
g_rel_strict <- build_dependency_graph(list(samples, subjects), list(belongs, rel_strict))

grouping_vector(derive_split_constraints(g_rel_strict, mode = "relatedness"))
```

# Pairwise relation: spatial proximity

Spatial proximity works the same way over sample coordinates — for example spot
locations from spatial transcriptomics, positions on a tissue slide, or
geographic site coordinates. `spatial_edges_from_coords()` connects samples
within a radius (Euclidean distance over the coordinate columns), and
`mode = "spatial"` groups the resulting connected components.

```{r spatial}
# Two spatial clusters. Cluster 1 (S1-S3) is a chain: neighbouring pairs are
# within the radius, but the endpoints are not.
coords <- data.frame(
  sample_id = paste0("S", 1:6),
  x = c(0, 1, 2,  6.0, 6.9, 6.2),
  y = c(0, 1, 0,  6.0, 6.6, 5.3),
  stringsAsFactors = FALSE
)
adj_edges <- spatial_edges_from_coords(coords, radius = 1.5)

meta_s <- data.frame(
  sample_id  = paste0("S", 1:6),
  subject_id = paste0("P", 1:6),
  stringsAsFactors = FALSE
)
samples_s  <- create_nodes(meta_s, "Sample", "sample_id")
subjects_s <- create_nodes(meta_s, "Subject", "subject_id")
belongs_s  <- create_edges(meta_s, "sample_id", "subject_id",
                           "Sample", "Subject", "sample_belongs_to_subject")

g_sp <- build_dependency_graph(list(samples_s, subjects_s), list(belongs_s, adj_edges))

sp_groups <- grouping_vector(derive_split_constraints(g_sp, mode = "spatial"))
sp_groups
```

Plotting the coordinates, drawing the within-radius adjacency edges in grey, and
colouring points by the derived group makes the transitive closure concrete:
S1–S2 and S2–S3 are each within the `1.5` radius, so all three share a group
even though S1 and S3 are `2` units apart and were never linked directly. Every
sample in the second cluster is likewise reachable from the others, while the
two clusters are far enough apart to stay separate:

```{r sp-plot, fig.width = 6.5, fig.height = 5}
sp_grp <- factor(sp_groups[coords$sample_id])
row_of <- setNames(seq_len(nrow(coords)), coords$sample_id)
from_i <- row_of[sub("^sample:", "", adj_edges$data$from)]
to_i   <- row_of[sub("^sample:", "", adj_edges$data$to)]
palette_sp <- c("#4C78A8", "#F58518")

plot(coords$x, coords$y, type = "n", asp = 1, xlab = "x", ylab = "y",
     main = "Spatial groups (radius = 1.5)")
segments(coords$x[from_i], coords$y[from_i],
         coords$x[to_i],   coords$y[to_i], col = "grey60", lwd = 2)
points(coords$x, coords$y, pch = 19, cex = 3.5, col = palette_sp[as.integer(sp_grp)])
text(coords$x, coords$y, labels = coords$sample_id, col = "white", cex = 0.8, font = 2)
legend("topleft", legend = levels(sp_grp), pch = 19,
       col = palette_sp[seq_along(levels(sp_grp))], title = "Spatial group", bty = "n")
```

# Deriving on a subset is leakage-safe

Real splits are derived on a *subset* of samples — the training rows, say. For
pairwise (and composite) modes this raises a subtle question: if a sample that
*bridges* two others is left out of the subset, could those two still inherit a
shared group from the full graph? They do not. When you pass `samples =`,
grouping is recomputed within that subset, so structure that exists only through
an excluded sample never leaks across the split.

The spatial chain makes this visible. S1 and S3 shared a group only because S2
bridged them; ask for S1 and S3 alone, and they correctly fall into separate
groups:

```{r subset-scoping}
grouping_vector(
  derive_split_constraints(g_sp, mode = "spatial", samples = c("S1", "S3"))
)
```

# Thresholds are inputs, not modeling

Because the threshold (kinship cutoff, spatial radius) is applied up front in the
edge-building helpers, it is a *derivation input*, not a modeling choice:
`splitGraph` forms groups over whatever edges survive and never computes folds
itself. The resulting `split_spec` is handed to a downstream consumer for
execution, exactly as with every other mode — see the *adapter-cookbook* and
*cross-language-handoff* vignettes for that step.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints.R
\name{derive_split_constraints}
\alias{derive_split_constraints}
\alias{grouping_vector}
\title{Derive Split Constraints from Dependency Graphs}
\usage{
derive_split_constraints(
  graph,
  mode = c("subject", "batch", "study", "time", "site", "region", "platform", "assay",
    "relatedness", "spatial", "composite"),
  samples = NULL,
  strategy = c("strict", "rule_based"),
  via = NULL,
  priority = NULL,
  include_warnings = TRUE
)

grouping_vector(x)
}
\arguments{
\item{graph}{A \code{dependency_graph}.}

\item{mode}{Constraint derivation mode.}

\item{samples}{Optional sample identifiers or sample node IDs used to
restrict the returned \code{sample_map}. All requested samples must
resolve successfully.}

\item{strategy}{Composite grouping strategy. Ignored for non-composite
modes.}

\item{via}{Optional dependency sources used for composite grouping. May be
given as lower-case modes such as \code{"subject"} or node types such as
\code{"Subject"}.}

\item{priority}{Optional priority order used for
\code{strategy = "rule_based"}.}

\item{include_warnings}{Whether to retain human-readable warnings in the
returned metadata.}

\item{x}{A \code{split_constraint}.}
}
\value{
\code{derive_split_constraints()} returns a \code{split_constraint}
  whose \code{sample_map} contains grouping assignments and, for time-aware
  constraints, ordering metadata. \code{grouping_vector()} returns a named
  character vector of \code{group_id} values keyed by \code{sample_id}.
}
\description{
Convert dataset dependency structure into deterministic sample-level
grouping constraints suitable for leakage-aware evaluation design.
}
\details{
Constraint derivation rules:

\describe{
  \item{\code{mode = "subject"}}{Groups samples by the target of
  \code{sample_belongs_to_subject}. All samples linked to the same
  \code{Subject} receive the same \code{group_id}.}

  \item{\code{mode = "batch"}}{Groups samples by the target of
  \code{sample_processed_in_batch}. Samples with no batch assignment are
  retained as singleton unlinked groups and recorded in metadata warnings.}

  \item{\code{mode = "study"}}{Groups samples by the target of
  \code{sample_from_study}.}

  \item{\code{mode = "site"}}{Groups samples by the target of
  \code{sample_collected_at_site}. Samples with no site assignment are
  retained as singleton unlinked groups and recorded in metadata warnings.}

  \item{\code{mode = "region"}}{Groups samples by the target of
  \code{sample_located_in_region} (e.g. a categorical tissue or anatomical
  region). Samples with no region assignment are retained as singleton
  unlinked groups and recorded in metadata warnings.}

  \item{\code{mode = "platform"}}{Groups samples by the target of
  \code{sample_run_on_platform} (the sequencing / measurement platform or
  instrument). Samples with no platform assignment are retained as singleton
  unlinked groups and recorded in metadata warnings.}

  \item{\code{mode = "assay"}}{Groups samples by the target of
  \code{sample_measured_by_assay} (the assay / modality). Samples with no
  assay assignment are retained as singleton unlinked groups and recorded in
  metadata warnings.}

  \item{\code{mode = "relatedness"}}{Groups samples by transitive closure
  over thresholded \code{subject_related_to} edges (genetic relatedness).
  Samples that share a subject, or whose subjects are directly or indirectly
  related above threshold, land in the same connected-component group. Build
  the edges with \code{\link{relatedness_edges_from_kinship}}. Samples with
  no subject are retained as singleton groups (recorded in metadata
  warnings).}

  \item{\code{mode = "spatial"}}{Groups samples by transitive closure over
  thresholded \code{sample_adjacent_to} edges (spatial proximity). Build the
  edges with \code{\link{spatial_edges_from_coords}}. Isolated samples form
  singleton groups.}

  \item{\code{mode = "time"}}{Groups samples by the target of
  \code{sample_collected_at_timepoint}. When \code{Timepoint} nodes have
  \code{time_index} metadata, that value is used to derive
  \code{order_rank}. If \code{time_index} is unavailable, the function
  attempts to derive ordering from \code{timepoint_precedes} edges over the
  timepoint subgraph.}

  \item{\code{mode = "composite"}, \code{strategy = "strict"}}{Projects the
  selected dependency relations onto a sample graph and assigns one
  \code{group_id} per connected component. This is the transitive-closure
  interpretation of composite dependency grouping.}

  \item{\code{mode = "composite"}, \code{strategy = "rule_based"}}{Evaluates
  dependency assignments in deterministic priority order and groups each
  sample by the highest-priority available dependency source.
  Lower-priority available dependencies are retained in the explanation
  field.}
}

The returned \code{split_constraint$sample_map} always contains
\code{sample_id}, \code{sample_node_id}, \code{group_id},
\code{constraint_type}, \code{group_label}, and \code{explanation}.
Time-aware constraints also include \code{time_index}, \code{timepoint_id},
and \code{order_rank} when available.

Ambiguous direct assignments are rejected. A sample cannot be assigned to
multiple batches, studies, or timepoints when deriving direct split
constraints.
}
\examples{
meta <- data.frame(
  sample_id  = c("S1", "S2", "S3", "S4"),
  subject_id = c("P1", "P1", "P2", "P2"),
  batch_id   = c("B1", "B2", "B1", "B2")
)
g <- graph_from_metadata(meta)

constraint <- derive_split_constraints(g, mode = "subject")
grouping_vector(constraint)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwise.R
\name{pairwise_edges}
\alias{pairwise_edges}
\alias{relatedness_edges_from_kinship}
\alias{spatial_edges_from_coords}
\title{Build Pairwise Leakage Edges from Continuous Similarity}
\usage{
relatedness_edges_from_kinship(
  pairs,
  threshold,
  id1 = "id1",
  id2 = "id2",
  kinship = "kinship"
)

spatial_edges_from_coords(coords, radius, id = "sample_id", coord_cols = NULL)
}
\arguments{
\item{pairs}{A data.frame of subject pairs with two id columns and a metric
column.}

\item{threshold}{Minimum kinship value (inclusive) for a pair to be kept.}

\item{id1, id2}{Column names in \code{pairs} holding the two subject ids.}

\item{kinship}{Column name in \code{pairs} holding the kinship / relatedness
value.}

\item{coords}{A data.frame with one row per sample: a sample id column plus
the numeric coordinate columns.}

\item{radius}{Maximum distance (inclusive) for two samples to be adjacent.}

\item{id}{Column name in \code{coords} holding the sample id.}

\item{coord_cols}{Character vector of coordinate columns in \code{coords}.
Defaults to every numeric column other than \code{id}.}
}
\value{
A \code{graph_edge_set}.
}
\description{
Helpers that turn a continuous, pairwise similarity signal into the
thresholded, undirected edges consumed by
\code{derive_split_constraints(mode = "relatedness")} and
\code{derive_split_constraints(mode = "spatial")}. Only pairs that pass the
threshold become edges; the derivation modes then form groups as connected
components over those edges (transitive closure), so a chain of individually
below-radius neighbours can still land in one group.
}
\details{
\code{relatedness_edges_from_kinship()} keeps subject pairs whose kinship (or
relatedness) coefficient is \emph{at least} \code{threshold} and emits
\code{subject_related_to} edges (\code{Subject} -> \code{Subject}).

\code{spatial_edges_from_coords()} keeps sample pairs whose Euclidean
distance over the coordinate columns is \emph{at most} \code{radius} and
emits \code{sample_adjacent_to} edges (\code{Sample} -> \code{Sample}).

Both return a \code{graph_edge_set} that can be combined with the other node
and edge sets in \code{build_dependency_graph()}. The passing metric value is
carried on each edge as an attribute (\code{kinship} / \code{distance}).
}
\examples{
pairs <- data.frame(
  id1 = c("P1", "P1", "P2"),
  id2 = c("P2", "P3", "P3"),
  kinship = c(0.25, 0.02, 0.30)
)
relatedness_edges_from_kinship(pairs, threshold = 0.1)

coords <- data.frame(
  sample_id = c("S1", "S2", "S3"),
  x = c(0, 1, 9),
  y = c(0, 1, 9)
)
spatial_edges_from_coords(coords, radius = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema-validate.R
\name{validate_json}
\alias{validate_json}
\alias{validate_graph_json}
\alias{validate_split_spec_json}
\title{Validate Serialized splitGraph JSON Against the Shipped Schema}
\usage{
validate_graph_json(path)

validate_split_spec_json(path)
}
\arguments{
\item{path}{Path to a serialized \code{dependency_graph} or
\code{split_spec} JSON file.}
}
\value{
A \code{splitgraph_json_report}: a list with \code{valid} (logical),
  \code{issues} (character vector of failures), the detected
  \code{object_type}, and the schema \code{$id}.
}
\description{
Check that a JSON file written by \code{write_dependency_graph()} or
\code{write_split_spec()} conforms to the splitGraph on-disk contract. The
formal JSON Schemas (Draft 2020-12) ship in \code{inst/schema/} and are
referenced from the written JSON via the \code{$schema} key; these functions
apply a dependency-free structural check of the same invariants (required
fields, value types, node/edge-type enumerations, and referential integrity
of edge endpoints) so a handoff file can be validated without a JSON Schema
engine.
}
\examples{
if (requireNamespace("jsonlite", quietly = TRUE)) {
  meta <- data.frame(sample_id = c("S1", "S2"), subject_id = c("P1", "P2"))
  g <- graph_from_metadata(meta)
  tmp <- tempfile(fileext = ".json")
  write_dependency_graph(g, tmp)
  validate_graph_json(tmp)
  unlink(tmp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{write_dependency_graph}
\alias{write_dependency_graph}
\alias{read_dependency_graph}
\title{Serialize a Dependency Graph to JSON}
\usage{
write_dependency_graph(graph, path, pretty = TRUE)

read_dependency_graph(path)
}
\arguments{
\item{graph}{A \code{dependency_graph} produced by
\code{build_dependency_graph()} or \code{graph_from_metadata()}.}

\item{path}{Path to write to or read from.}

\item{pretty}{If \code{TRUE} (default), the JSON is indented for human
inspection. Set \code{FALSE} for a compact representation.}
}
\value{
\code{write_dependency_graph()} invisibly returns \code{path}.
  \code{read_dependency_graph()} returns a validated
  \code{dependency_graph}.
}
\description{
Write a \code{dependency_graph} to a JSON file and read it back. The on-disk
format is intentionally simple and stable: it captures the canonical node
table, the canonical edge table (each with their list-column of
attributes), the graph metadata (including \code{validation_overrides}),
and the data-model \code{schema_version}. The internal \code{igraph}
representation is not stored; it is rebuilt on read via
\code{dependency_graph()}.
}
\details{
This makes \code{split_spec}/\code{dependency_graph} objects portable
across R sessions, and across language boundaries (any consumer that can
read JSON can interpret the format).
}
\section{JSON format}{

\preformatted{
{
  "$schema": "https://.../inst/schema/dependency_graph.schema.json",
  "splitGraph_object": "dependency_graph",
  "schema_version": "0.2.0",
  "metadata": {
    "graph_name": "...",
    "dataset_name": "...",
    "created_at": "2026-04-29T10:11:12.000000+0000",
    "schema_version": "0.2.0",
    "validation_overrides": { ... }
  },
  "nodes": [
    { "node_id": "sample:S1", "node_type": "Sample",
      "node_key": "S1", "label": "S1", "attrs": { ... } },
    ...
  ],
  "edges": [
    { "edge_id": "sample_belongs_to_subject:1",
      "from": "sample:S1", "to": "subject:P1",
      "edge_type": "sample_belongs_to_subject", "attrs": { ... } },
    ...
  ]
}
}
Reading a file whose \code{schema_version} shares the installed major
version loads silently (additive-only differences); a differing major
version loads with a warning suggesting \code{migrate_dependency_graph_json()}.
The written JSON also carries a \code{$schema} reference to the formal JSON
Schema shipped in \code{inst/schema/}; validate a file against it with
\code{validate_graph_json()}.
}

\examples{
if (requireNamespace("jsonlite", quietly = TRUE)) {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2")
  )
  g <- graph_from_metadata(meta, graph_name = "demo")

  tmp <- tempfile(fileext = ".json")
  write_dependency_graph(g, tmp)
  g2 <- read_dependency_graph(tmp)
  identical(g$nodes$data$node_id, g2$nodes$data$node_id)
  unlink(tmp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitGraph-package.R
\docType{package}
\name{splitGraph-package}
\alias{splitGraph}
\alias{splitGraph-package}
\title{splitGraph: Dataset Dependency Graphs for Leakage-Aware Evaluation}
\description{
The \pkg{splitGraph} package provides typed graph objects for representing
dataset structure, sample provenance, and leakage-relevant dependencies in
biomedical evaluation workflows. It makes dataset dependency structure
explicit enough to validate, query, and convert into a stable, tool-agnostic
split specification (\code{split_spec}) for leakage-aware evaluation.
}
\section{Scope (what splitGraph does)}{

\itemize{
  \item Model dataset dependency structure as a typed graph.
  \item Validate that structure (structural, semantic, leakage-relevant).
  \item Derive deterministic split \emph{constraints} from the structure.
  \item Emit and validate the tool-agnostic \code{split_spec} interchange
    format (with a formal JSON Schema and a Python reference consumer).
}
}

\section{Non-goals (what downstream consumers own)}{

splitGraph deliberately stops at the constraint / \code{split_spec} boundary.
It does \emph{not} generate resamples or folds, perform stratified splitting,
apply purge/embargo, fit or tune models, or produce statistical leakage
evidence. Those belong to downstream consumers. The reference consumer is
\pkg{bioLeak}, whose \code{as_leaksplits()} turns a \code{split_spec} into an
executable split plan; \code{split_spec} is neutral, so other tools (an
\pkg{rsample} adapter, the shipped Python reader, etc.) can consume it
equally. See the \code{split_spec} contract in \code{?as_split_spec} and the
"Scope & relationship to bioLeak" section of the README.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/selcukorkmaz/splitGraph}
  \item Report bugs at \url{https://github.com/selcukorkmaz/splitGraph/issues}
}

}
\author{
\strong{Maintainer}: Selcuk Korkmaz \email{selcukorkmaz@gmail.com} (\href{https://orcid.org/0000-0003-4632-6850}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema-validate.R
\name{migrate_json}
\alias{migrate_json}
\alias{migrate_dependency_graph_json}
\alias{migrate_split_spec_json}
\title{Upgrade Serialized splitGraph JSON to the Current Schema Version}
\usage{
migrate_dependency_graph_json(path, out = path)

migrate_split_spec_json(path, out = path)
}
\arguments{
\item{path}{Path to the JSON file to upgrade.}

\item{out}{Path to write the upgraded file to. Defaults to \code{path}
(in-place upgrade).}
}
\value{
The output path, invisibly.
}
\description{
Read a \code{dependency_graph} or \code{split_spec} JSON file written under
an older \code{schema_version} and rewrite it at the installed version. The
round-trip fills any field introduced since the file was written with its
default (\code{NA} for missing \code{sample_data} columns), stamps the
current \code{schema_version}, and adds the \code{$schema} reference. Files
already at the current version are rewritten unchanged.
}
\examples{
if (requireNamespace("jsonlite", quietly = TRUE)) {
  meta <- data.frame(sample_id = c("S1", "S2"), subject_id = c("P1", "P2"))
  g <- graph_from_metadata(meta)
  tmp <- tempfile(fileext = ".json")
  write_dependency_graph(g, tmp)
  migrate_dependency_graph_json(tmp)
  unlink(tmp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{depgraph_validation_report}
\alias{depgraph_validation_report}
\alias{split_spec}
\alias{split_spec_validation}
\alias{leakage_risk_summary}
\title{Validation Report Object for splitGraph Graphs}
\usage{
depgraph_validation_report(
  graph_name = NULL,
  issues = NULL,
  metrics = list(),
  metadata = list(),
  valid = NULL,
  errors = NULL,
  warnings = NULL,
  advisories = NULL
)

split_spec(
  sample_data = NULL,
  group_var = "group_id",
  block_vars = character(),
  time_var = NULL,
  ordering_required = FALSE,
  constraint_mode = NULL,
  constraint_strategy = NULL,
  recommended_resampling = NULL,
  metadata = list()
)

split_spec_validation(issues = NULL, metadata = list())

leakage_risk_summary(
  overview = character(),
  diagnostics = NULL,
  validation_summary = list(),
  constraint_summary = list(),
  split_spec_summary = list(),
  metadata = list()
)
}
\arguments{
\item{graph_name}{Graph label stored on the report.}

\item{issues}{Canonical issue table. When \code{NULL}, an empty skeleton is
constructed.}

\item{metrics}{Named list of graph- and issue-level counts.}

\item{metadata}{Named list of report metadata.}

\item{valid}{Optional logical override for the overall validity flag.}

\item{errors, warnings, advisories}{Optional character vectors of
severity-specific messages.}

\item{sample_data}{Sample-level mapping table carried by a
\code{split_spec}.}

\item{group_var}{Name of the grouping column.}

\item{block_vars}{Optional blocking variable names.}

\item{time_var}{Optional ordering column name.}

\item{ordering_required}{Whether ordering is required for downstream
evaluation.}

\item{constraint_mode, constraint_strategy}{Constraint-derivation metadata.}

\item{recommended_resampling}{Optional recommended resampling routine.}

\item{overview}{Character vector of human-readable overview lines.}

\item{diagnostics}{Diagnostics data frame for leakage risks.}

\item{validation_summary, constraint_summary, split_spec_summary}{Named lists
carrying pre-computed summaries.}
}
\value{
An S3 object corresponding to the constructor that was called.
}
\description{
\code{depgraph_validation_report} is the structured return type produced by
\code{validate_graph()} and \code{validate_depgraph()}.
}
\details{
The report contains:

\itemize{
  \item \code{graph_name}: graph label when available
  \item \code{valid}: whether any \code{error}-severity issues were found
  \item \code{issues}: canonical issue table
  \item \code{summary}: counts by level, severity, and code
  \item \code{metadata}: report metadata
  \item \code{errors}, \code{warnings}, \code{advisories}: backward-compatible
    message vectors
  \item \code{metrics}: graph and issue counts
}

The canonical issue table includes the columns:
\code{issue_id}, \code{level}, \code{severity}, \code{code}, \code{message},
\code{node_ids}, \code{edge_ids}, and \code{details}.
}
\examples{
meta <- data.frame(
  sample_id  = c("S1", "S2"),
  subject_id = c("P1", "P2")
)
g <- graph_from_metadata(meta)

report <- validate_graph(g)
report$valid
summary(report)
}
\seealso{
\code{\link{validate_graph}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph-build-auto.R
\name{graph_from_metadata}
\alias{graph_from_metadata}
\title{Build a Dependency Graph Directly from a Metadata Table}
\usage{
graph_from_metadata(
  meta,
  columns = NULL,
  dataset_name = NULL,
  graph_name = NULL,
  outcome_scope = c("sample", "subject"),
  time_precedence = TRUE,
  validate = TRUE,
  validation_overrides = list()
)
}
\arguments{
\item{meta}{A \code{data.frame} containing one row per sample and optional
canonical columns: \code{sample_id} (required), \code{subject_id},
\code{batch_id}, \code{study_id}, \code{timepoint_id}, \code{time_index},
\code{assay_id}, \code{featureset_id}, \code{outcome_id}, or
\code{outcome_value}.}

\item{columns}{Optional named character vector passed to
\code{ingest_metadata()} to rename user columns to canonical names.}

\item{dataset_name, graph_name}{Optional metadata labels.}

\item{outcome_scope}{Either \code{"sample"} (default) or \code{"subject"}.
Controls whether outcome edges attach to samples or subjects.}

\item{time_precedence}{If \code{TRUE} and \code{time_index} is present,
derive \code{timepoint_precedes} edges from the ordering of
\code{time_index}.}

\item{validate}{Forwarded to \code{build_dependency_graph()}.}

\item{validation_overrides}{Forwarded to \code{build_dependency_graph()}.}
}
\value{
A validated \code{dependency_graph}.
}
\description{
One-shot convenience builder that auto-detects canonical columns in a
metadata table, creates the corresponding node and edge sets, optionally
derives timepoint ordering from \code{time_index}, and assembles a
\code{dependency_graph}. Columns that are absent or entirely missing are
silently skipped.
}
\examples{
meta <- data.frame(
  sample_id  = c("S1", "S2", "S3", "S4"),
  subject_id = c("P1", "P1", "P2", "P2"),
  batch_id   = c("B1", "B2", "B1", "B2"),
  timepoint_id = c("T1", "T2", "T1", "T2"),
  time_index = c(1, 2, 1, 2),
  outcome_id = c("ctrl", "case", "ctrl", "case")
)

g <- graph_from_metadata(meta, graph_name = "demo")
g
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{query_node_type}
\alias{query_node_type}
\alias{query_edge_type}
\alias{query_neighbors}
\alias{query_paths}
\alias{query_shortest_paths}
\alias{detect_dependency_components}
\alias{detect_shared_dependencies}
\title{Query Dependency Graph Structure}
\usage{
query_node_type(graph, node_types, ids = NULL)

query_edge_type(graph, edge_types, node_ids = NULL)

query_neighbors(
  graph,
  node_ids,
  edge_types = NULL,
  node_types = NULL,
  direction = c("out", "in", "all")
)

query_paths(
  graph,
  from,
  to,
  edge_types = NULL,
  node_types = NULL,
  mode = c("out", "in", "all"),
  max_length = NULL
)

query_shortest_paths(
  graph,
  from,
  to,
  edge_types = NULL,
  node_types = NULL,
  mode = c("out", "in", "all")
)

detect_dependency_components(
  graph,
  via = c("Subject", "Batch", "Study", "Timepoint", "Assay", "FeatureSet", "Outcome"),
  edge_types = NULL,
  min_size = 1
)

detect_shared_dependencies(
  graph,
  via = c("Subject", "Batch", "Study", "Timepoint"),
  samples = NULL
)
}
\arguments{
\item{graph}{A \code{dependency_graph}.}

\item{node_types}{Optional node types used to filter node results or allowed
path members.}

\item{ids}{Optional node identifiers used to further restrict
\code{query_node_type()}.}

\item{edge_types}{Optional edge types used to filter the traversal graph or
edge table.}

\item{node_ids, from, to}{Node identifiers to use as query seeds or endpoints.}

\item{direction, mode}{Traversal direction.}

\item{max_length}{Maximum path length (number of edges) for
\code{query_paths()}. Defaults to a documented finite cap
(\code{8}) so that \code{igraph::all_simple_paths()} cannot blow up
on dense graphs. Pass \code{Inf} to opt out and search exhaustively;
pass any non-negative integer for an explicit cap. Negative values and
non-numeric inputs are rejected.}

\item{via}{Dependency node types used for sample-level dependency detection.}

\item{min_size}{Minimum component size retained by
\code{detect_dependency_components()}.}

\item{samples}{Optional sample identifiers or sample node IDs used to
restrict direct shared-dependency detection. All requested samples must
resolve successfully.}
}
\value{
Each function returns a \code{graph_query_result}. Use
  \code{as.data.frame()} to obtain the tidy result table.
}
\description{
Query graph neighborhoods, typed nodes and edges, path structure, projected
sample dependency components, and direct shared dependencies within a
\code{dependency_graph}.
}
\details{
When a \code{samples} subset is supplied, partial matching is not allowed:
unknown sample identifiers raise an error rather than being silently
dropped.
}
\examples{
meta <- data.frame(
  sample_id  = c("S1", "S2", "S3"),
  subject_id = c("P1", "P1", "P2"),
  batch_id   = c("B1", "B2", "B1")
)
g <- graph_from_metadata(meta)

query_node_type(g, "Sample")
query_neighbors(g, "sample:S1", direction = "out")
detect_shared_dependencies(g, via = "Subject")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ingest.R
\name{create_nodes}
\alias{create_nodes}
\alias{create_edges}
\title{Create Canonical Node and Edge Tables}
\usage{
create_nodes(
  data,
  type,
  id_col,
  label_col = NULL,
  attr_cols = NULL,
  prefix = TRUE,
  dedupe = TRUE
)

create_edges(
  data,
  from_col,
  to_col,
  from_type,
  to_type,
  relation,
  attr_cols = NULL,
  allow_missing = FALSE,
  dedupe = TRUE,
  from_prefix = TRUE,
  to_prefix = TRUE
)
}
\arguments{
\item{data}{A \code{data.frame} containing entity or relationship columns.}

\item{type, from_type, to_type}{Supported node types such as \code{"Sample"}
or \code{"Subject"}.}

\item{id_col}{Column containing the source identifier for the node type.}

\item{label_col}{Optional column used for node labels.}

\item{attr_cols}{Optional columns stored in the \code{attrs} list-column.}

\item{prefix}{If \code{TRUE}, prepend typed prefixes such as \code{sample:}
to node identifiers.}

\item{dedupe}{If \code{TRUE}, collapse duplicate identifiers or duplicate
edges only when the retained definition is identical.}

\item{from_col, to_col}{Source and target identifier columns for edge
creation.}

\item{relation}{Canonical edge type.}

\item{allow_missing}{If \code{TRUE}, drop rows with missing edge endpoints
instead of erroring.}

\item{from_prefix, to_prefix}{Whether to prepend typed prefixes when
constructing the edge endpoint identifiers. Defaults preserve the
canonical prefixed-ID format.}
}
\value{
For \code{create_nodes()}, a \code{graph_node_set}. For
  \code{create_edges()}, a \code{graph_edge_set}.
}
\description{
Build canonical node and edge tables from ordinary metadata frames.
}
\details{
The package uses typed node identifiers such as \code{sample:S1} as the
canonical graph representation. If you create node sets with
\code{prefix = FALSE}, the corresponding edge endpoints must use matching
prefix settings via \code{from_prefix} and \code{to_prefix}.

When \code{dedupe = TRUE}, exact duplicate node or edge definitions are
collapsed, but conflicting definitions for the same canonical node
identifier or edge relation are rejected with an error.
}
\examples{
meta <- data.frame(
  sample_id = c("S1", "S2"),
  subject_id = c("P1", "P2")
)

samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
edges <- create_edges(
  meta,
  from_col = "sample_id",
  to_col = "subject_id",
  from_type = "Sample",
  to_type = "Subject",
  relation = "sample_belongs_to_subject"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split-spec.R
\name{as_split_spec}
\alias{as_split_spec}
\alias{validate_split_spec}
\alias{summarize_leakage_risks}
\title{Translate splitGraph Constraints into Stable Split Specifications}
\usage{
as_split_spec(constraint, graph = NULL)

validate_split_spec(x)

summarize_leakage_risks(
  graph,
  constraint = NULL,
  split_spec = NULL,
  validation = NULL
)
}
\arguments{
\item{constraint}{A \code{split_constraint}.}

\item{graph}{A \code{dependency_graph}.}

\item{x}{A \code{split_spec}.}

\item{split_spec}{An optional \code{split_spec}.}

\item{validation}{An optional \code{depgraph_validation_report}.}
}
\value{
\code{as_split_spec()} returns a \code{split_spec}.
  \code{validate_split_spec()} returns a \code{split_spec_validation}.
  \code{summarize_leakage_risks()} returns a \code{leakage_risk_summary}.
}
\description{
Translate graph-derived split constraints into a stable, inspectable
structure for sample-level grouping, blocking, and ordering, perform
preflight structural checks on that translation, and summarize structural
leakage risks.
}
\details{
The translation layer always produces canonical sample-level columns
including \code{sample_id}, \code{sample_node_id}, \code{group_id}, and
\code{primary_group}. When available, it also carries \code{batch_group},
\code{study_group}, \code{timepoint_id}, \code{time_index}, and
\code{order_rank}. Missing but relevant fields are retained as \code{NA}
columns rather than omitted.

When only a subset of samples has ordering metadata, the translated split
spec still exposes that partial ordering through \code{time_var}, but
\code{ordering_required} remains \code{FALSE}. Ordering is only marked as
required when the constraint implies complete ordering coverage.

The split-spec validator checks:
\itemize{
  \item missing required columns
  \item missing or duplicated sample identifiers
  \item missing grouping assignments
  \item singleton-only grouping structures
  \item missing ordering when ordering is required
  \item invalid or empty block variables
}

Repeated validation of the same split spec yields deterministic issue IDs
and diagnostics, which makes the returned validation object stable across
runs.

The produced \code{split_spec} is tool-agnostic. Downstream consumers are
expected to provide their own adapters to convert a \code{split_spec} into
their native split representation, so \pkg{splitGraph} has no runtime
dependency on any of them.

\code{summarize_leakage_risks()} reuses \code{validate_graph()} and
\code{split_constraint} metadata rather than duplicating downstream
evaluation logic.
}
\examples{
meta <- data.frame(
  sample_id  = c("S1", "S2", "S3", "S4"),
  subject_id = c("P1", "P1", "P2", "P2")
)
g <- graph_from_metadata(meta)

constraint <- derive_split_constraints(g, mode = "subject")
spec <- as_split_spec(constraint, graph = g)
validate_split_spec(spec)
summarize_leakage_risks(g, constraint = constraint, split_spec = spec)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph-build.R, R/validate.R
\name{build_dependency_graph}
\alias{build_dependency_graph}
\alias{build_depgraph}
\alias{as_igraph}
\alias{validate_graph}
\alias{validate_depgraph}
\title{Assemble and Validate Dependency Graphs}
\usage{
build_dependency_graph(
  nodes,
  edges,
  graph_name = NULL,
  dataset_name = NULL,
  validate = TRUE,
  validation_overrides = list()
)

build_depgraph(
  nodes,
  edges,
  graph_name = NULL,
  dataset_name = NULL,
  validate = TRUE,
  validation_overrides = list()
)

as_igraph(x)

validate_graph(
  graph,
  checks = c("ids", "references", "cardinality", "schema", "time"),
  error_on_fail = FALSE,
  levels = NULL,
  severities = NULL,
  validation_overrides = NULL
)

validate_depgraph(
  graph,
  checks = c("ids", "references", "cardinality", "schema", "time"),
  error_on_fail = FALSE,
  levels = NULL,
  severities = NULL,
  validation_overrides = NULL
)
}
\arguments{
\item{nodes, edges}{Lists of \code{graph_node_set} and \code{graph_edge_set}
objects.}

\item{graph_name, dataset_name}{Optional metadata labels.}

\item{validate}{If \code{TRUE}, run \code{validate_graph()} before returning.}

\item{validation_overrides}{Optional named list of explicit validation
exceptions. Currently supported keys:
\describe{
  \item{\code{allow_multi_subject_samples}}{If \code{TRUE}, the semantic
    validator does not flag samples linked to multiple subjects, and
    \code{derive_split_constraints(mode = "subject")} silently keeps the
    first listed subject assignment (recording the ambiguity in
    \code{metadata$warnings}). Defaults to \code{FALSE}.}
}
When passed to \code{validate_graph()} or \code{validate_depgraph()},
the override is merged into the graph's existing
\code{validation_overrides} for the duration of the call only.}

\item{x}{A \code{dependency_graph}.}

\item{graph}{A \code{dependency_graph}.}

\item{checks}{\strong{Deprecated.} Use \code{levels} and \code{severities}
instead. Retained for backward compatibility with 0.1.0 callers.}

\item{error_on_fail}{If \code{TRUE}, stop when validation errors are found
across all detected issues from the selected validation levels, even if
those errors are hidden from \code{issues} by \code{severities}.}

\item{levels}{Optional validation layers to run.}

\item{severities}{Optional severities to retain in the returned
\code{issues} table. This filter does not change whether the graph is
considered valid.}
}
\value{
For \code{build_dependency_graph()}, a \code{dependency_graph}. For
  \code{validate_graph()} and \code{validate_depgraph()}, a
  \code{depgraph_validation_report}. For \code{as_igraph()}, the underlying
  \code{igraph} object.
}
\description{
Combine canonical node and edge tables into a typed dependency graph and
perform structural, semantic, and graph-local leakage-aware validation.
}
\examples{
meta <- data.frame(
  sample_id = c("S1", "S2"),
  subject_id = c("P1", "P2")
)

samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
edges <- create_edges(
  meta,
  "sample_id",
  "subject_id",
  "Sample",
  "Subject",
  "sample_belongs_to_subject"
)

g <- build_dependency_graph(list(samples, subjects), list(edges))
validate_graph(g)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serialization.R
\name{write_split_spec}
\alias{write_split_spec}
\alias{read_split_spec}
\title{Serialize a Split Specification to JSON}
\usage{
write_split_spec(spec, path, pretty = TRUE)

read_split_spec(path)
}
\arguments{
\item{spec}{A \code{split_spec} produced by \code{as_split_spec()}.}

\item{path}{Path to write to or read from.}

\item{pretty}{If \code{TRUE} (default), the JSON is indented.}
}
\value{
\code{write_split_spec()} invisibly returns \code{path}.
  \code{read_split_spec()} returns a \code{split_spec}.
}
\description{
Write a \code{split_spec} to a JSON file and read it back. The on-disk
format captures the canonical sample-level table (\code{sample_data}) plus
all spec-level fields needed by a downstream resampling adapter
(\code{group_var}, \code{block_vars}, \code{time_var},
\code{ordering_required}, \code{constraint_mode},
\code{constraint_strategy}, \code{recommended_resampling}) and the spec
metadata.
}
\details{
\code{NA} values in \code{sample_data} are written as JSON \code{null} and
read back as \code{NA}.
}
\section{JSON format}{

\preformatted{
{
  "$schema": "https://.../inst/schema/split_spec.schema.json",
  "splitGraph_object": "split_spec",
  "schema_version": "0.2.0",
  "group_var": "group_id",
  "block_vars": ["batch_group", "study_group"],
  "time_var": "order_rank",
  "ordering_required": false,
  "constraint_mode": "subject",
  "constraint_strategy": "subject",
  "recommended_resampling": "grouped_cv",
  "metadata": { ... },
  "sample_data": [
    { "sample_id": "S1", "group_id": "subject:P1", ... },
    ...
  ]
}
}
}

\examples{
if (requireNamespace("jsonlite", quietly = TRUE)) {
  meta <- data.frame(
    sample_id  = c("S1", "S2"),
    subject_id = c("P1", "P2")
  )
  g <- graph_from_metadata(meta)
  constraint <- derive_split_constraints(g, mode = "subject")
  spec <- as_split_spec(constraint, graph = g)

  tmp <- tempfile(fileext = ".json")
  write_split_spec(spec, tmp)
  spec2 <- read_split_spec(tmp)
  identical(spec$sample_data$group_id, spec2$sample_data$group_id)
  unlink(tmp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{graph_node_set}
\alias{graph_node_set}
\alias{graph_edge_set}
\alias{dependency_graph}
\alias{new_depgraph_nodes}
\alias{new_depgraph_edges}
\alias{new_depgraph}
\alias{graph_query_result}
\alias{dependency_constraint}
\alias{split_constraint}
\title{Construct Core splitGraph S3 Objects}
\usage{
graph_node_set(
  data = NULL,
  schema_version = .depgraph_schema_version,
  source = list()
)

graph_edge_set(
  data = NULL,
  schema_version = .depgraph_schema_version,
  source = list()
)

dependency_graph(nodes, edges, graph, metadata = list(), caches = list())

new_depgraph_nodes(
  data = NULL,
  schema_version = .depgraph_schema_version,
  source = list()
)

new_depgraph_edges(
  data = NULL,
  schema_version = .depgraph_schema_version,
  source = list()
)

new_depgraph(nodes, edges, graph = NULL, metadata = list(), caches = list())

graph_query_result(
  query = "",
  params = list(),
  nodes = NULL,
  edges = NULL,
  table = NULL,
  metadata = list()
)

dependency_constraint(
  constraint_id,
  relation_types,
  sample_map,
  transitive = TRUE,
  metadata = list()
)

split_constraint(
  strategy,
  sample_map,
  recommended_downstream_args = list(),
  metadata = list()
)
}
\arguments{
\item{data}{A data frame matching the canonical schema for nodes or edges.}

\item{schema_version}{Schema version string stored on the object.}

\item{source}{Optional source metadata.}

\item{nodes, edges}{A \code{graph_node_set} and \code{graph_edge_set}.}

\item{graph}{An internal \code{igraph} object.}

\item{metadata, caches, params, recommended_downstream_args}{Named lists with
auxiliary metadata.}

\item{query}{Query label stored on a \code{graph_query_result}.}

\item{table}{Tabular query result payload.}

\item{constraint_id, relation_types, transitive}{Fields describing a
dependency constraint.}

\item{sample_map}{Sample-level mapping table for constraints.}

\item{strategy}{Split strategy identifier.}
}
\value{
An S3 object corresponding to the constructor that was called.
}
\description{
Low-level constructors for the core S3 classes used throughout
\pkg{splitGraph}.
}
\examples{
meta <- data.frame(
  sample_id = c("S1", "S2"),
  subject_id = c("P1", "P2")
)

samples <- create_nodes(meta, type = "Sample", id_col = "sample_id")
subjects <- create_nodes(meta, type = "Subject", id_col = "subject_id")
edges <- create_edges(
  meta,
  from_col = "sample_id",
  to_col = "subject_id",
  from_type = "Sample",
  to_type = "Subject",
  relation = "sample_belongs_to_subject"
)

nodes_set <- graph_node_set(rbind(samples$data, subjects$data))
edges_set <- graph_edge_set(edges$data)
nodes_set
edges_set
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ingest.R
\name{ingest_metadata}
\alias{ingest_metadata}
\title{Standardize Sample Metadata}
\usage{
ingest_metadata(data, col_map = NULL, dataset_name = NULL, strict = TRUE)
}
\arguments{
\item{data}{A sample-level \code{data.frame}.}

\item{col_map}{Optional named character vector mapping canonical names to
user-provided columns.}

\item{dataset_name}{Optional dataset label stored as an attribute on the
returned table.}

\item{strict}{If \code{TRUE}, error when required columns are missing.}
}
\value{
A standardized \code{data.frame} with canonical identifier columns
  coerced to character.
}
\description{
Normalize user-provided metadata into the canonical column contract used by
\pkg{splitGraph}.
}
\examples{
meta <- ingest_metadata(
  data.frame(sample_id = c("S1", "S2"), subject_id = c("P1", "P2"))
)
}
